from ctypes import *
from locale import currency
from xdxsmi.api.node import Node
import re
import string
import time
import subprocess
import sys
import os
# from enum import Enum

# buffer size
XDXML_DEVICE_NAME_BUFFER_SIZE                 = 64
XDXML_DEVICE_PCI_BUS_ID_BUFFER_V2_SIZE        = 16
XDXML_DEVICE_PCI_BUS_ID_BUFFER_SIZE           = 32
XDXML_THERMAL_BUFFER_SIZE                     = 5
XDXML_DRIVER_VER_BUFFER_SIZE                  = 20
XDXML_XDXML_PHY_NUM_MAX                       = 4

class _PrintableStructure(Structure):
    """
    Abstract class that produces nicer __str__ output than ctypes.Structure.
    e.g. instead of:
      >>> print str(obj)
      <class_name object at 0x7fdf82fef9e0>
    this class will print
      class_name(field_name: formatted_value, field_name: formatted_value)

    _fmt_ dictionary of <str _field_ name> -> <str format>
    e.g. class that has _field_ 'hex_value', c_uint could be formatted with
      _fmt_ = {"hex_value" : "%08X"}
    to produce nicer output.
    Default fomratting string for all fields can be set with key "<default>" like:
      _fmt_ = {"<default>" : "%d MHz"} # e.g all values are numbers in MHz.
    If not set it's assumed to be just "%s"

    Exact format of returned str from this class is subject to change in the future.
    """
    _fmt_ = {}
    def __str__(self):
        result = []
        for x in self._fields_:
            key = x[0]
            value = getattr(self, key)
            fmt = "%s"
            if key in self._fmt_:
                fmt = self._fmt_[key]
            elif "<default>" in self._fmt_:
                fmt = self._fmt_["<default>"]
            result.append(("%s: " + fmt) % (key, value))
        return self.__class__.__name__ + "(" + ','.join(result) + ")"

class nvmlPciInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('domain', c_uint64),
        ('bus', c_uint64),
        ('device', c_uint64),
        ('bdfid', c_uint64),
        ('func', c_uint64),
        ('pci_dbdf', c_char * 64),
        ('busId', c_char * XDXML_DEVICE_PCI_BUS_ID_BUFFER_SIZE),
    ]
    _fmt_ = {
            'domain'         : "0x%08X",
            'bus'            : "0x%02X",
            'device'         : "0x%02X",
            'bdfid'          : "0x%08X",
            }

## Device structures
class struct_c_xdx_device_t(Structure):
    pass # opaque handle
c_xdx_device_t = POINTER(struct_c_xdx_device_t)

#clock info structure
class xdxmlclockInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('xdx_gddr_phys_freq', c_int),
        ('xdx_gfx_freq', c_int),
        ('xdx_acpu_freq', c_int),
        ('xdx_cmcu_freq', c_int),
        ('xdx_ndf_freq', c_int),
    ]
    _fmt_ = {
            'xdx_gddr_phys_freq'      : "%d",
            'xdx_gfx_freq'            : "%d",
            'xdx_acpu_freq'           : "%d",
            'xdx_cmcu_freq'           : "%d",
            'xdx_ndf_freq'            : "%d",
            }

#power info structure
class xdxmlpowerInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('volt', c_int),
        ('current', c_int),
        ('pwr', c_int),
        ('pwr_total', c_int),
        ('power_cap', c_int),
        ('power1_cap_max', c_int),
        ('power1_cap_min', c_int),
    ]
    _fmt_ = {
            'volt'      : "%d",
            'current'            : "%d",
            'pwr'           : "%d",
            }

#temper info structure
class xdxmltemperInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('ctrl_mode', c_int),
        ('temp', c_int),
        ('temp_highest', c_int),
        ('temp_min', c_int),
        ('temp_max', c_int),
        ('temp_crit', c_int),
        ('gddr_temp', c_int),
    ]

#temper info structure
class xdxmltemperSet_t(_PrintableStructure):
    _fields_ = [
        ('ctrl_mode', c_char* 5),
        ('temp_min', c_char* 5),
        ('temp_max', c_char* 5),
        ('temp_crit', c_char* 5),
    ]

#temper type structrue
class xdxmltemperType(object):
    crtl_mode_type = 0
    temper_min_type = 1
    temper_max_type = 2
    temper_crit_type = 3

#thermal info structure
class xdxmlthermalInfo_t(_PrintableStructure):
    _fields_ = [
        ('xdx_temp_slowdown', c_int),
        ('xdx_temp_shutdown', c_int),
        ('xdx_enable_thm_slowdown', c_int),
        ('xdx_enable_thm_shutdown', c_int),
    ]

# thermal enum structure
TEMP_SLOWDOWN_TYPE = 0
TEMP_SHUTDOWN_TYPE = 1
ENABLE_THM_SLOWDOWN_TYPE = 2
ENABLE_THM_SHUTDOWN_TYPE = 3
class xdxmlthermalType(object):
    temp_slowdown_type = 0
    temp_shutdown_type = 1
    enable_thm_slowdown_type = 2
    enable_thm_shutdown_type = 3

# thermal set structure
class xdxmlthermalSet_t(_PrintableStructure):
    _fields_ = [
        ('temp_slowdown', c_char * XDXML_THERMAL_BUFFER_SIZE),
        ('temp_shutdown', c_char * XDXML_THERMAL_BUFFER_SIZE),
        ('enable_thm_slowdown', c_char * XDXML_THERMAL_BUFFER_SIZE),
        ('enable_thm_shutdown', c_char * XDXML_THERMAL_BUFFER_SIZE),
    ]

#dvfs info structure
class xdxmldvfsInfo_t(_PrintableStructure):
    _fields_ = [
        ('xdx_enable_dvfs', c_int),
    ]

## C Type mappings ##
## Enums
XDXML_FEATURE_DISABLED    = 0
XDXML_FEATURE_ENABLED     = 1

class xdxmlutilInfo_t(_PrintableStructure):
    _fields_ = [
        ('gpu',  c_int),
        ('tdm',  c_int),
        ('geom', c_int),
        ('thrd', c_int),
        ('cdm',  c_int),
    ]

class xdxmldrvInfo_t(_PrintableStructure):
    _fields_ = [
        ('DDK', c_char* XDXML_DRIVER_VER_BUFFER_SIZE),
        ('KMD', c_char* XDXML_DRIVER_VER_BUFFER_SIZE),
        ('UMD', c_char* XDXML_DRIVER_VER_BUFFER_SIZE),
    ]

class xdxmlmemInfo_t(_PrintableStructure):
    _fields_ = [
        # Moved to the new busId location below
        ('fb_free', c_ulonglong),
        ('fb_total', c_ulonglong),
        ('fb_used', c_ulonglong),
        ('sys_free', c_ulonglong),
        ('sys_total', c_ulonglong),
        ('sys_used', c_ulonglong),
    ]

#temper info structure
class xdxmlqueryInfo_t(_PrintableStructure):
    _fields_ = [
        ('display_status', c_uint),
        # Moved to the new busId location below
        ('xdx_gddr_phys_freq', c_int),
        ('xdx_gfx_freq', c_int),
        ('xdx_acpu_freq', c_int),
        ('xdx_cmcu_freq', c_int),
        ('xdx_ndf_freq', c_int),

        ('name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('vendor', c_char* 20),
        ('uuid', c_char* 10),
        ('vbios_version', c_char* 20),
        ('serial', c_char* 10),
        ('minor_number', c_int),
        ('fan_speed', c_int),
        ('fan_ctrl_mode', c_int),
        ('primary_id', c_char* 10),
        ('primary_core_id', c_char* 10),
        ('asci_type', c_char* 20),
        ('core', c_int),
        ('mmbar_type', c_char* 20),
        ('card_type', c_char* 20),
        ('fw_ver', c_char* 10),
        ('gddr_count', c_int),
        ('gddr_speed', c_ulong),
        ('gddr_size', c_int),
        ('gddr_width', c_uint),
        ('gddr_type', c_char * 10),
        ('gddr_package_model', c_uint),

        #pcie error
        ('err_corr', c_char* 10),
        ('err_fatal', c_char* 10),
        ('non_fatal', c_char* 10),

        ('ctrl_mode', c_int),
        ('temp', c_int),
        ('temp_highest', c_int),
        ('temp_min', c_int),
        ('temp_max', c_int),
        ('temp_crit', c_int),

        #thermal info
        ('xdx_temp_slowdown', c_int),
        ('xdx_temp_shutdown', c_int),
        ('xdx_enable_thm_slowdown', c_int),
        ('xdx_enable_thm_shutdown', c_int),

        ('curr_pcie_link_width', c_uint),
        ('max_pcie_link_width', c_uint),
        ('pcie_speed', c_uint),
        ('pcie_max_speed', c_uint),
        ('curr_link_gen', c_uint),
        ('max_link_gen', c_uint),
        ('device_ID', c_uint),
        ('vendor_ID', c_uint),
        ('subsystem_device_ID', c_uint),
        ('subsystem_vendor_ID', c_uint),

        #pcie event
        ('pcie_event', c_int),
        ('perst_assert', c_int),
        ('perst_de_assert', c_int),
        ('link_clk_stable', c_int),
        ('link_down', c_int),
        ('perst_assert_acpu_ack', c_int),
        ('perst_de_assert_acpu_ack', c_int),

        #pcie_info
        ('domain', c_uint64),
        ('bus', c_uint64),
        ('device', c_uint64),
        ('bdfid', c_uint64),
        ('func', c_uint64),
        ('pci_dbdf', c_char * 64),

        #memory info
        ('fb_free', c_ulonglong),
        ('fb_total', c_ulonglong),
        ('fb_used', c_ulonglong),
        ('sys_free', c_ulonglong),
        ('sys_total', c_ulonglong),
        ('sys_used', c_ulonglong),

        #GPU info wcj   
        ('bValid',c_uint64),     
        ('ui64GpuStatActive',c_uint64),
        ('ui64GpuStatBlocked',c_uint64),
        ('ui64GpuStatIdle',c_uint64),
        ('ui64GpuStatCumulative',c_uint64),

        ('aaui64DMOSStatActive_0',c_uint64 ),
        ('aaui64DMOSStatActive_1',c_uint64 ),
        ('aaui64DMOSStatActive_2',c_uint64 ),
        ('aaui64DMOSStatActive_3',c_uint64 ),
        ('aaui64DMOSStatActive_4',c_uint64 ),
        ('aaui64DMOSStatActive_5',c_uint64 ),
        ('aaui64DMOSStatActive_6',c_uint64 ),
        ('aaui64DMOSStatActive_7',c_uint64 ),
        ('aaui64DMOSStatActive_8',c_uint64 ),

        ('aaui64DMOSStatBlocked_0',c_uint64),
        ('aaui64DMOSStatBlocked_1',c_uint64),
        ('aaui64DMOSStatBlocked_2',c_uint64),
        ('aaui64DMOSStatBlocked_3',c_uint64),
        ('aaui64DMOSStatBlocked_4',c_uint64),
        ('aaui64DMOSStatBlocked_5',c_uint64),
        ('aaui64DMOSStatBlocked_6',c_uint64),
        ('aaui64DMOSStatBlocked_7',c_uint64),
        ('aaui64DMOSStatBlocked_8',c_uint64),


        ('aaui64DMOSStatIdle_0',c_uint64),
        ('aaui64DMOSStatIdle_1',c_uint64),
        ('aaui64DMOSStatIdle_2',c_uint64),
        ('aaui64DMOSStatIdle_3',c_uint64),
        ('aaui64DMOSStatIdle_4',c_uint64),
        ('aaui64DMOSStatIdle_5',c_uint64),
        ('aaui64DMOSStatIdle_6',c_uint64),
        ('aaui64DMOSStatIdle_7',c_uint64),
        ('aaui64DMOSStatIdle_8',c_uint64),

        ('aaui64DMOSStatCumulative_0',c_uint64),
        ('aaui64DMOSStatCumulative_1',c_uint64),
        ('aaui64DMOSStatCumulative_2',c_uint64),
        ('aaui64DMOSStatCumulative_3',c_uint64),
        ('aaui64DMOSStatCumulative_4',c_uint64),
        ('aaui64DMOSStatCumulative_5',c_uint64),
        ('aaui64DMOSStatCumulative_6',c_uint64),
        ('aaui64DMOSStatCumulative_7',c_uint64),
        ('aaui64DMOSStatCumulative_8',c_uint64),

        ('ui64TimeStamp',c_uint64),

        ('pronum',c_uint64),

        ('process_1',c_uint64),             
        ('pro1LMA',c_uint64),
        ('pro1Total',c_uint64),

        ('process_2',c_uint64),        
        ('pro2LMA',c_uint64),
        ('pro2Total',c_uint64),

        ('process_3',c_uint64),        
        ('pro3LMA',c_uint64),
        ('pro3Total',c_uint64),

        ('process_4',c_uint64),        
        ('pro4LMA',c_uint64),
        ('pro4Total',c_uint64),

        ('process_5',c_uint64),        
        ('pro5LMA',c_uint64),
        ('pro5Total',c_uint64),

        ('pro1NAME',c_char* 30),   
        ('pro2NAME',c_char* 30),
        ('pro3NAME',c_char* 30),
        ('pro4NAME',c_char* 30),
        ('pro5NAME',c_char* 30),
		('device_id', c_uint64),
        ('volt', c_int), 
        ('support_fan_info', c_int), 
        ('current', c_float), 
        ('pwr', c_float),
        ('pwr_total', c_float), 
        ('chip_model', c_char* 64), 
        ('support_power_info', c_int),
        ('support_temp_info', c_int),
        ('serial_number', c_char* 32),
        ('gpu_default_freq', c_uint64), 
        
        #util info
        ('util_gpu', c_int),
        ('util_tdm', c_int),
        ('util_geom', c_int),
        ('util_thrd', c_int),
        ('util_cdm', c_int),

        #gddr crc info
        ('crc_support', c_int),
        ('crc_write_cnt', c_int * XDXML_XDXML_PHY_NUM_MAX),
        ('crc_read_cnt', c_int * XDXML_XDXML_PHY_NUM_MAX),
        ('crc_write_retry_cnt', c_int * XDXML_XDXML_PHY_NUM_MAX),
        ('crc_read_retry_cnt', c_int * XDXML_XDXML_PHY_NUM_MAX),
        #init info
        ('initStaInfo', c_int),
    ]

class xdxmlstaticInfo_t(_PrintableStructure):
    _fields_ = [
        ('asic_type',  c_uint),
        ('core_num',  c_uint),
        ('gddr_count', c_uint),
        ('gddr_size',   c_uint),
        ('gddr_speed',  c_uint64),
        ('gpu_primary_id',   c_uint),
        ('vendor',   c_uint),
        ('chip_id',   c_uint),
        ('device_id',   c_uint),
        ('card_type',   c_uint),
        ('serial_number',   c_char* 32),
        ('fw_version',   c_uint),
        ('vbios_version',   c_uint),
        ('fan_ctrl_mode',   c_uint),
        ('support_fan_info',   c_uint),
        ('support_temp_info',   c_uint),
        ('support_power_info',   c_uint),
        ('gpu_default_freq',   c_uint64),
    ]

## vpu Enums
VP_JPEG_BASELINE = 12
VP_PROFILE_UNKOWN = 37
RT_IMG_FMT_UNKOWN = 15
RC_MTD_UNKOWN = 12

#struct Vpu_rt_img_fmt_info_st from "xdx_smi_api.h"
class vpu_rt_img_fmt_info_t(_PrintableStructure):
    _fields_ = [
        ('img_fmt_name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('is_valid', c_int),
        ('img_fmt_type', c_int),
    ]

#struct Vpu_enc_br_ctrl_mtd_info_st from "xdx_smi_api.h"
class vpu_enc_br_ctrl_mtd_info_t(_PrintableStructure):
    _fields_ = [
        ('br_ctrl_mth_name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('is_valid', c_int),
        ('br_ctrl_mtd_type', c_int),
    ]

#struct Vpu_enc_info_st from "xdx_smi_api.h"
class vpu_enc_info_t(_PrintableStructure):
    _fields_ = [
        ('enc_name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('is_valid', c_int),
        ('profile_type', c_int),
        ('img_fmt_list', vpu_rt_img_fmt_info_t* RT_IMG_FMT_UNKOWN),
        ('enc_br_crtl_mtd_List', vpu_enc_br_ctrl_mtd_info_t* RC_MTD_UNKOWN),
    ]

#struct Vpu_dec_info_st from "xdx_smi_api.h"
class vpu_dec_info_t(_PrintableStructure):
    _fields_ = [
        ('dec_name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('is_valid', c_int),
        ('profile_type', c_int),
        ('img_fmt_list', vpu_rt_img_fmt_info_t* RT_IMG_FMT_UNKOWN),
    ]

#struct Vpu_codec_info_st from "xdx_smi_api.h"
class vpu_codec_info_t(_PrintableStructure):
    _fields_ = [
        ('vender_name', c_char* XDXML_DEVICE_NAME_BUFFER_SIZE),
        ('is_valid', c_int),
        ('encoders', vpu_enc_info_t* VP_PROFILE_UNKOWN),
        ('decoders', vpu_dec_info_t* VP_PROFILE_UNKOWN),
    ]

def is_file_in_path(file_path, file_name):
    full_path = os.path.join(file_path, file_name)
    return os.path.isfile(full_path)

def load_xdxml_library():
    '''
    Load the library if it isn't loaded already
    '''
    global xdxml_lib
    etc_node = Node("/etc")
    curr_os = subprocess.check_output(['cat', '/etc/os-release']).decode('utf-8').lower()
    arch = subprocess.check_output('arch', shell = True).decode('utf-8').split('\n',1)[0]#when use check_output the value of arch is 'x86_64\n' but 'x86_64',thus,needed to slipt '\n'
    try:
        # use ldcache to load libxdxgpu-ml.so
        xdxml_lib = CDLL("libxdxgpu-ml.so")

        # use absolute path to load libxdxgpu-ml.so
        # if arch == "x86_64":
        #     if "id=ubuntu" in curr_os:
        #         xdxml_lib = CDLL("/opt/xdxgpu/lib/x86_64-linux-gnu/libxdxgpu-ml.so")
        #     elif "kylin" in curr_os and "server" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     elif "red hat" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     else:
        #         if(is_file_in_path("/usr/lib/x86_64-linux-gnu/xdxgpu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/usr/lib/x86_64-linux-gnu/xdxgpu/libxdxgpu-ml.so")
        #         elif(is_file_in_path("/opt/xdxgpu/lib/x86_64-linux-gnu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/opt/xdxgpu/lib/x86_64-linux-gnu/libxdxgpu-ml.so")
        #         else:
        #             print("cannot find the file libxdxgpu-ml.so")
            
        # elif arch == "aarch64":
        #     if "id=ubuntu" in curr_os:
        #         xdxml_lib = CDLL("/opt/xdxgpu/lib/aarch64-linux-gnu/libxdxgpu-ml.so")
        #     elif "kylin" in curr_os and "server" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     elif "red hat" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     else:
        #         if(is_file_in_path("/usr/lib/aarch64-linux-gnu/xdxgpu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/usr/lib/aarch64-linux-gnu/xdxgpu/libxdxgpu-ml.so")
        #         elif(is_file_in_path("/opt/xdxgpu/lib/aarch64-linux-gnu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/opt/xdxgpu/lib/aarch64-linux-gnu/libxdxgpu-ml.so")
        #         else:
        #             print("cannot find the file libxdxgpu-ml.so")
        
        # elif arch == "loongarch64":
        #     if "id=ubuntu" in curr_os:
        #         xdxml_lib = CDLL("/opt/xdxgpu/lib/loongarch64-linux-gnu/xdxgpu/libxdxgpu-ml.so")
        #     elif "kylin" in curr_os and "server" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     elif "red hat" in curr_os:
        #         xdxml_lib = CDLL("/usr/lib64/xdxgpu/libxdxgpu-ml.so")
        #     else:
        #         if(is_file_in_path("/usr/lib/loongarch64-linux-gnu/xdxgpu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/usr/lib/loongarch64-linux-gnu/xdxgpu/libxdxgpu-ml.so")
        #         elif(is_file_in_path("/opt/xdxgpu/lib/loongarch64-linux-gnu/", "libxdxgpu-ml.so")):
        #             xdxml_lib = CDLL("/opt/xdxgpu/lib/loongarch64-linux-gnu/libxdxgpu-ml.so")
        #         else:
        #             print("cannot find the file libxdxgpu-ml.so")

    except Exception as e:
        print("Failed to load xdxml library:", str(e))
        exit()


## Function access ##
_xdxmlGetFunctionPointer_cache = dict() # function pointers are cached to prevent unnecessary libLoadLock locking
def _xdxmlGetFunctionPointer(name):
    global xdxml_lib

    if name in _xdxmlGetFunctionPointer_cache:
        return _xdxmlGetFunctionPointer_cache[name]

    _xdxmlGetFunctionPointer_cache[name] = getattr(xdxml_lib, name)
    return _xdxmlGetFunctionPointer_cache[name]

def xdxml_init():
    #print("zihan_debug:enter xdxml_init")
    load_xdxml_library()

    #
    # Initialize the library
    #
    fn = _xdxmlGetFunctionPointer("xdxml_init")
    ret = fn()
    return None

def xdxml_out_of_band_init():
    load_xdxml_library()
    #
    # Initialize the library
    #
    fn = _xdxmlGetFunctionPointer("xdxml_out_of_band_init")
    ret = fn()
    return None

def xdxml_shutdown():
    fn = _xdxmlGetFunctionPointer("xdxml_shutdown")
    ret = fn()
    return None

def is_out_of_band():
    fn = _xdxmlGetFunctionPointer("is_out_of_band")
    ret = fn() 
    return ret

## Device get functions
def xdxml_device_get_count():
    c_count = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_count")
    ret = fn(byref(c_count))
    return c_count.value

def xdxml_device_get_handle_by_index(index):
    c_index = c_uint(index)
    device = c_xdx_device_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_handle_by_index")
    ret = fn(c_index, byref(device))
    return device

# a map function that handle to card index, assist detect_xdxgpus2() in smi.py
def xdxml_device_get_card_index_by_handle(handle):
    c_index = c_uint()
    device = handle
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_card_index_by_handle")
    ret = fn(device, byref(c_index))

    return c_index.value

def xdxml_device_get_name(handle):
    c_name = create_string_buffer(XDXML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_name")
    ret = fn(handle, c_name, c_uint(XDXML_DEVICE_NAME_BUFFER_SIZE))
    return c_name.value

def xdxml_device_get_pci_info(handle):
    c_info = nvmlPciInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pci_info")
    ret = fn(handle, byref(c_info))
    #_nvmlCheckReturn(ret)
    return c_info

def xdxml_device_get_clock(handle):
    clock_t = xdxmlclockInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_clock")
    ret = fn(handle, byref(clock_t))
    return clock_t  

def xdxml_device_get_curr_pcie_link_width(handle):
    curr_link_width = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_curr_pcie_link_width")
    ret = fn(handle, byref(curr_link_width))
    return curr_link_width.value

def xdxml_device_get_max_pcie_link_width(handle):
    max_link_width = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_max_pcie_link_width")
    ret = fn(handle, byref(max_link_width))
    return max_link_width.value  

def xdxml_device_get_pcie_speed(handle):
    curr_link_speed = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pcie_speed")
    ret = fn(handle, byref(curr_link_speed))
    return curr_link_speed.value

def xdxml_device_get_pcie_link_max_speed(handle):
    max_link_speed = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pcie_link_max_speed")
    ret = fn(handle, byref(max_link_speed))
    return max_link_speed.value

def xdxml_device_get_curr_pcie_link_generation(handle):
    curr_link_gen = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_curr_pcie_link_generation")
    ret = fn(handle, byref(curr_link_gen))
    return curr_link_gen.value

def xdxml_device_get_max_pcie_link_generation(handle):
    max_link_gen = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_max_pcie_link_generation")
    ret = fn(handle, byref(max_link_gen))
    return max_link_gen.value 

def xdxml_device_get_fan_speed(handle):
    fan_speed = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_speed")
    ret = fn(handle, byref(fan_speed))
    return fan_speed.value

def xdxml_device_get_min_max_fan_speed(handle):
    min = c_int()
    max = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_min_max_fan_speed")
    ret = fn(handle, byref(min), byref(max))
    fan_dict = dict(min_fan_speed = min.value, max_fan_speed = max.value)
    return fan_dict

def xdxml_device_get_fan_pwm(handle):
    fan_pwm = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_pwm")
    ret = fn(handle, byref(fan_pwm))
    return fan_pwm.value

def xdxml_device_get_pwm_min(handle):
    fan_pwm_min = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pwm_min")
    ret = fn(handle, byref(fan_pwm_min))
    return fan_pwm_min.value

def xdxml_device_get_fan_ctrl_mode(handle):
    fan_ctrl_mode = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_fan_ctrl_mode")
    ret = fn(handle, byref(fan_ctrl_mode))
    return fan_ctrl_mode.value

def xdxml_device_get_pwm_max(handle):
    fan_pwm_max = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_pwm_max")
    ret = fn(handle, byref(fan_pwm_max))
    return fan_pwm_max.value

def xdxml_device_get_power_info(handle):
    power_t = xdxmlpowerInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_power_info")
    ret = fn(handle, byref(power_t))
    return power_t

def xdxml_device_get_temper_info(handle):
    temper_t = xdxmltemperInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_temper_info")
    ret = fn(handle, byref(temper_t))
    return temper_t

def xdxml_device_get_temper(handle, temper, temp_type):
    if temp_type == "ctrl_mode":
        type = c_uint(0)
        # type = xdxmltemperType.crtl_mode_type
    elif temp_type == "temp_min":
        type = c_uint(1)
        # type = xdxmltemperType.temper_min_type
    elif temp_type == "temp_max":
        # type = xdxmltemperType.temper_max_type
        type = c_uint(2)
    elif temp_type == "temp_crit":
        type = c_uint(3)
        # type = xdxmltemperType.temper_crit_type
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_temper")
    ret = fn(handle, byref(temper), type)

def xdxml_device_get_thermal_info(handle):
    thermal_t = xdxmlthermalInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_thermal_info")
    ret = fn(handle, byref(thermal_t))
    return thermal_t

def xdxml_device_get_dvfs_info(handle):
    dvfs_t = xdxmldvfsInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_dvfs_info")
    ret = fn(handle, byref(dvfs_t))
    return dvfs_t

def xdxml_device_get_display_status(handle):
    c_mode = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_display_status")
    ret = fn(handle, byref(c_mode))
    return c_mode.value

def xdxml_device_get_mem_info(handle):
    memory_t = xdxmlmemInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_mem_info")
    ret = fn(handle, byref(memory_t))
    return memory_t

def xdxml_device_get_driver_version(handle):
    driver_ver_t = xdxmldrvInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_driver_version")
    ret = fn(handle, byref(driver_ver_t))
    return driver_ver_t

def xdxml_device_get_util_info(handle):
    util_t = xdxmlutilInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_util_info")
    ret = fn(handle, byref(util_t))
    return util_t

def xdxml_device_query(handle):
    query_t = xdxmlqueryInfo_t()
    fn =  _xdxmlGetFunctionPointer("xdxml_device_query")
    ret = fn(handle, byref(query_t))
    return query_t

def xdxml_device_query_two_parts(handle, query_t=None):
    if(query_t == None):
        query_t = xdxmlqueryInfo_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_query_two_parts")
    ret = fn(handle, byref(query_t))
    return query_t

def xdxml_device_query_static_info(handle, query_t):
    fn = _xdxmlGetFunctionPointer("xdxml_device_query_static_info")
    ret = fn(handle, byref(query_t))
    return query_t

def xdxml_device_query_dynamic_info(handle, query_t):
    fn = _xdxmlGetFunctionPointer("xdxml_device_query_dynamic_info")
    ret = fn(handle, byref(query_t))
    return query_t

def xdxml_device_query_by_out_of_band(handle):
    query_t = xdxmlqueryInfo_t()
    fn =  _xdxmlGetFunctionPointer("xdxml_device_query_by_out_of_band")
    ret = fn(handle, byref(query_t))
    return query_t

#set function
def xdxml_device_set_fan_pwm1(handle, pwm):
    c_data = c_char_p(str(pwm).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan_pwm1")
    ret = fn(handle,c_data)

def xdxml_device_set_fan_ctrl_mode(handle, fan_ctrl_mode):
    c_data = c_char_p(str(fan_ctrl_mode).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan_ctrl_mode")
    ret = fn(handle,c_data)

def xdxml_device_set_fan1_speed(handle, fan):
    c_data = c_char_p(str(fan).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_fan1_speed")
    ret = fn(handle,c_data)

def xdxml_device_set_power_warn(handle, cap):
    c_data = c_char_p(str(cap).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_power_warn")
    ret = fn(handle,c_data)

def xdxml_device_set_power_volt(handle, input):
    c_data = c_char_p(str(input).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_power_volt")
    ret = fn(handle,c_data)

def xdxml_device_set_gfx_clock(handle, gfx_clock):
    c_data = c_char_p(str(gfx_clock).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_gfx_clock")
    ret = fn(handle,c_data)

def xdxml_device_set_temper_info(handle,temper,tem_type):        
    temper_set = xdxmltemperSet_t()
    temp = str(temper).encode()
    #print("temp is " + temp.decode("utf-8"))  
    if tem_type == "temp1_crit":
        type = c_uint(3)
        temper_set.temp_crit = temp
        #print("PYTHON zihan_debug "+ str(temper_set.temp_crit))
    if tem_type == "temp1_min":
        type = c_uint(1)
        temper_set.temp_min = temp
        #print("temp_min is "+ temper_set.temp_min.decode("utf-8"))
    if tem_type == "temp1_max":
        type = c_uint(2)
        temper_set.temp_max = temp
    if tem_type == "temp1_ctrl_mode":
        type = c_uint(0)
        temper_set.ctrl_mode = temp
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_temper_info")
    ret = fn(handle, byref(temper_set), type)

def xdxml_device_set_thermal_info(handle, thermal, thm_type):
    thm_set = xdxmlthermalSet_t()
    thm = str(thermal).encode()
    if thm_type == "temp_slowdown":
        type = TEMP_SLOWDOWN_TYPE
        thm_set.temp_slowdown = thm
    elif thm_type == "temp_shutdown":
        type = TEMP_SHUTDOWN_TYPE
        thm_set.temp_shutdown = thm
    elif thm_type == "enable_slowdown":
        type = ENABLE_THM_SLOWDOWN_TYPE
        thm_set.enable_thm_slowdown = thm
    elif thm_type == "enable_shutdown":
        type = ENABLE_THM_SHUTDOWN_TYPE
        thm_set.enable_thm_shutdown = thm
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_thermal_info")
    ret = fn(handle, byref(thm_set), type)

def xdxml_device_get_thermal(handle, thermal, thm_type):
    if thm_type == "temp_slowdown":
        #type = xdxmlthermalType.temp_slowdown_type.value
        type = c_int(0)
    elif thm_type == "temp_shutdown":
        # type = xdxmlthermalType.temp_shutdown_type.value
        type = c_int(1)
    elif thm_type == "enable_slowdown":
        # type = xdxmlthermalType.enable_thm_slowdown_type
        type = c_int(2)
    elif thm_type == "enable_shutdown":
        #type = xdxmlthermalType.enable_thm_shutdown_type
        type = c_uint(3)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_thermal")
    ret = fn(handle, byref(thermal), type)

def xdxml_device_set_dvfs_enable(handle, dvfs_enable):
    c_data = c_char_p(str(dvfs_enable).encode())
    fn = _xdxmlGetFunctionPointer("xdxml_device_set_dvfs_enable")
    ret = fn(handle, c_data)

def xdxml_device_get_support_fan_info(handle):
    support_fan_info = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_support_fan_info")
    ret = fn(handle, byref(support_fan_info))
    return support_fan_info.value

def xdxml_device_get_device_model(handle):
    c_device_model_name = create_string_buffer(XDXML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_device_model")
    ret = fn(handle, c_device_model_name)
    return c_device_model_name.value

def xdxml_device_get_asic_type(handle):
    c_asic_type = c_uint()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_asic_type")
    ret = fn(handle, byref(c_asic_type))
    return c_asic_type.value

def xdxml_device_get_support_power_info(handle):
    support_power_info = c_int()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_support_power_info")
    ret = fn(handle, byref(support_power_info))
    return support_power_info.value

def xdxml_device_get_vendor(handle):
    c_vendor_name = create_string_buffer(XDXML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_vendor")
    ret = fn(handle, c_vendor_name)
    return c_vendor_name.value

# ()->bool
def xdxml_get_vpu_functions_valid():
    isValid = c_int(0)
    fn = _xdxmlGetFunctionPointer("xdxml_get_vpu_functions_valid")
    ret = fn(byref(isValid))
    return isValid.value == 1

def xdxml_device_get_vpu_info(handle):
    vpu_codec_info = vpu_codec_info_t()
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_vpu_info")
    ret = fn(handle, byref(vpu_codec_info))
    return vpu_codec_info

def xdxml_device_get_vpu_utlz(handle):
    enc_cmd_rev_fps = c_int(-1)
    enc_cmd_exc_fps = c_int(-1)
    dec_cmd_rev_fps = c_int(-1)
    dec_cmd_exc_fps = c_int(-1)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_vpu_utlz")
    ret = fn(handle, byref(enc_cmd_rev_fps), byref(enc_cmd_exc_fps), byref(dec_cmd_rev_fps), byref(dec_cmd_exc_fps))
    return enc_cmd_rev_fps.value, enc_cmd_exc_fps.value, dec_cmd_rev_fps.value, dec_cmd_exc_fps.value
    
def xdxml_device_id_to_gfx_bound(handle, device_id):
    lower_bound = c_int(0)
    upper_bound = c_int(0)
    fn = _xdxmlGetFunctionPointer("xdxml_device_id_to_gfx_bound")
    ret = fn(device_id, byref(lower_bound), byref(upper_bound))
    return lower_bound.value, upper_bound.value

def xdxml_device_get_driver_release_version(handle):
    ver = create_string_buffer(XDXML_DRIVER_VER_BUFFER_SIZE)
    ver_size = XDXML_DRIVER_VER_BUFFER_SIZE
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_driver_release_version")
    ret = fn(ver, ver_size)
    return ver.value.decode('utf-8')

def xdxml_device_get_video_decoder_encoder(handle):
    decoder = create_string_buffer(XDXML_DEVICE_NAME_BUFFER_SIZE)
    encoder = create_string_buffer(XDXML_DEVICE_NAME_BUFFER_SIZE)
    fn = _xdxmlGetFunctionPointer("xdxml_device_get_video_decoder_encoder")
    ret = fn(handle, decoder, encoder)
    return decoder.value.decode('utf-8'), encoder.value.decode('utf-8')