"""
usage:
   xdxsmi query [options]

options:
    -h, --help                    print this help message
    -f, --filename=file           redirect query output to a specified file
    -x, --xml                     generate XML file to store target GPUs info
    -i, --index=val               target gpu to read/write attributes, start from 0
    -d, --display=type            show selected info
    -l, --loop=sec                set look up interval, whose unit is second
"""

# from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.xdxgpu import XdxGpu
from xdxsmi.api.smi import XdxSmi
from get_set_attrs import setAttrs, getAttrs
import time, re

# (gpu: XdxGpu)
def getgddrCRCAttrs(gpu):
    if ((gpu.asci_type == "CHIP_PANGU_A0" and gpu.fm_ver < "v2.2.6") or
        (gpu.asci_type == "CHIP_PANGU_B0" and gpu.fm_ver < "v1.2.7")):
        gddrCRCAttrs = "\n"
        gddrCRCAttrs += """\
        GDDR CRC State (total)
            CRC Write Count         : {crc_write_cnt}
            CRC Read Count          : {crc_read_cnt}
            CRC Write Retry Count   : {crc_write_retry_cnt}
            CRC Read Retry Count    : {crc_read_retry_cnt}\n""".format(crc_write_cnt=gpu.crc_write_cnt[0],
                                                                       crc_read_cnt=gpu.crc_read_cnt[0],
                                                                       crc_write_retry_cnt=gpu.crc_write_retry_cnt[0],
                                                                       crc_read_retry_cnt=gpu.crc_read_retry_cnt[0]
                                                                      )
        return gddrCRCAttrs

    gddrCRCHeader ='\n'
    gddrCRCHeader += """\
        GDDR CRC State""".format()
    gddrCRCHeader +='\n'

    gddrCRCBody = """""".format()
    for i in range(0, int(gpu.gddr_count, 16)):
        gddrCRCBody += """\
            PHY-{i}
              CRC Write Count       : {crc_write_cnt}
              CRC Read Count        : {crc_read_cnt}
              CRC Write Retry Count : {crc_write_retry_cnt}
              CRC Read Retry Count  : {crc_read_retry_cnt}\n""".format(i=i, crc_write_cnt=gpu.crc_write_cnt[i], 
                                                                            crc_read_cnt=gpu.crc_read_cnt[i],
                                                                            crc_write_retry_cnt=gpu.crc_write_retry_cnt[i],
                                                                            crc_read_retry_cnt=gpu.crc_read_retry_cnt[i]
                                                                      )
    return gddrCRCHeader + gddrCRCBody

# (gpu: XdxGpu, args: Dict)
def printAttributes(gpu, args):
    cur_value = 'N/A'
    processInfo = """\
        Process
                GUP ID              : {vaule}
                PID                 : {vaule}
                Process Name        : {vaule}
                Memory Usage        : {vaule}""".format(vaule=cur_value)

    gpuAttrs = """\
    GPU {user_index}
        Product Name                : {name}
        Core                        : {core}
        PrimaryID                   : {primary_id}
        Primary Core ID             : {primary_core_id}
        Display Mode                : {dsp_mode}
        Accounting mode             : {acct_mode}
        GPU UUID                    : {uuid}
        Minor Number                : {minor_num}
        VBIOS Version               : {vbios}
        FW Version                  : {fm_ver}
        GDDR Count                  : {gddr_count}
        GDDR Speed                  : {gddr_speed}
        GDDR Size                   : {gddr_size}
        GDDR Bus Width              : {gddr_width}
        GDDR Effective Frequency    : {gddr_effec}
        GDDR Type                   : {gddr_type}
        GDDR Package Model          : {gddr_package_model}""".format(   user_index=gpu.user_index, name=gpu.name,
                                                                        core=gpu.core, primary_id=gpu.primary_id,
                                                                        primary_core_id=gpu.primary_core_id,
                                                                        dsp_mode=gpu.dsp_mode, acct_mode=gpu.acct_mode,
                                                                        uuid=gpu.uuid, minor_num=gpu.minor_num,
                                                                        vbios=gpu.vbios, fm_ver=gpu.fm_ver,
                                                                        gddr_count=gpu.gddr_count,
                                                                        gddr_speed=gpu.gddr_speed,
                                                                        gddr_size=gpu.gddr_size,
                                                                        gddr_width=gpu.gddr_width,
                                                                        gddr_effec=gpu.gddr_effec,
                                                                        gddr_type=gpu.gddr_type,
                                                                        gddr_package_model=gpu.gddr_package_model
                                                                    )

    gddrCRCAttrs = getgddrCRCAttrs(gpu)
    # placeholder to show the display results
    # gddrCRCAttrs = f"""\
    #     GDDR CRC State
    #         CRC Write Count         : {gpu.crc_write_cnt}
    #         CRC Read Count          : {gpu.crc_read_cnt}
    #         CRC Write Retry Count   : {gpu.crc_write_retry_cnt}
    #         CRC Read Retry Count    : {gpu.crc_read_retry_cnt}"""
    fan_spd = gpu.fan_spd
    if gpu.fan_spd=="0":
        fan_spd = "N/A"
    otherAttrs = """\
        PCIE Memory bar Type        : {pcie_mmbar_type}
        Asci Type                   : {asci_type}
        Pcie Event                  
            Reset Assert            : {perstAssert} 
            Reset DeAssert          : {perstDeAssert}
            Link Clock Stable       : {linkClkStable}
            Link Down               : {linkDown}
            Reset Assert AcpuAck    : {perstAssertAcpuAck}
            Reset DeAssert AcpuAck  : {perstDeAssertAcpuAck}
        PCI
            Bus                     : {bus}
            Device                  : {dev}
            Domain                  : {domain}
            Bus Id                  : {pci_bdf}
            Vendor ID               : {vendor_ID}
            Device ID               : {device_ID}
            Subsystem Vendor ID     : {subsystem_vendor_ID}
            Subsystem ID            : {subsystem_device_ID}
            GPU Link Info
                PCIe Generation
                    Max             : {pcie_gen_max}
                    Current         : {pcie_gen_cur}
                Link Width
                    Max             : {width_max}
                    Current         : {width_cur}
                Throughput per lane : {throughput}
        PCIE Error
            PCIE Error Corr         : {pcie_err_corr}
            PCIE Error Fatal        : {pcie_err_fatal}
            PCIE Error Non Fatal    : {pcie_err_nonfat}
        Fan Speed[RPM]              : {fan_spd}
        Performance State           : {perform_state}
        Clock Throttle Reasons      : {clk_throt_reasons}
        FB Memory Usage(MiB)
            Total                   : {fb_total}
            Used                    : {used_GPU_memory}
            Free                    : {free_GPU_memory}
        HOST Memory Usage(MiB)
            Total                   : {sys_total}
            Used                    : {sys_used}
            Free                    : {sys_free}
        ECC Mode
            Current                 : {ecc_curr}
            Pending                 : {ecc_pend}
        ECC Errors                  : {ecc_error}
        Temprature                  : {temp}
        Power Reading               : {pwr_read}
        Clocks
            Graphics                : {gfx_clk}
            GDDR                    : {gddr_clk}
        Application Clocks          : {app_clk}
        Default Application Clocks  : {def_app_clk}
        Supported Clock             : {sup_clk}
        GPU Resource Usage
            2D/TDM                  : {twoDTDMusage}
            GEOM                    : {GEOMMusage}
            3D                      : {threeDMusage}""".format( pcie_mmbar_type=gpu.pcie_mmbar_type,
                                                                asci_type=gpu.asci_type,
                                                                perstAssert=gpu.perstAssert,
                                                                perstDeAssert=gpu.perstDeAssert,
                                                                linkClkStable=gpu.linkClkStable,
                                                                linkDown=gpu.linkDown,
                                                                perstAssertAcpuAck=gpu.perstAssertAcpuAck,
                                                                perstDeAssertAcpuAck=gpu.perstDeAssertAcpuAck,
                                                                bus=gpu.bus,
                                                                dev=gpu.dev,
                                                                domain=gpu.domain,
                                                                pci_bdf=gpu.pci_bdf,
                                                                vendor_ID=gpu.vendor_ID,
                                                                device_ID=gpu.device_ID,
                                                                subsystem_vendor_ID=gpu.subsystem_vendor_ID,
                                                                subsystem_device_ID=gpu.subsystem_device_ID,
                                                                pcie_gen_max=gpu.pcie_gen_max,
                                                                pcie_gen_cur=gpu.pcie_gen_cur,
                                                                width_max=gpu.width_max,
                                                                width_cur=gpu.width_cur,
                                                                throughput=gpu.throughput,
                                                                pcie_err_corr=gpu.pcie_err_corr,
                                                                pcie_err_fatal=gpu.pcie_err_fatal,
                                                                pcie_err_nonfat=gpu.pcie_err_nonfat,
                                                                fan_spd=fan_spd,
                                                                perform_state=gpu.perform_state,
                                                                clk_throt_reasons=gpu.clk_throt_reasons,
                                                                fb_total=gpu.fb_total,
                                                                used_GPU_memory=gpu.used_GPU_memory,
                                                                free_GPU_memory=gpu.free_GPU_memory,
                                                                sys_total=gpu.sys_total,
                                                                sys_used=gpu.sys_used,
                                                                sys_free=gpu.sys_free,
                                                                ecc_curr=gpu.ecc_curr,
                                                                ecc_pend=gpu.ecc_pend,
                                                                ecc_error=gpu.ecc_error,
                                                                temp=gpu.temp,
                                                                pwr_read=gpu.pwr_read,
                                                                gfx_clk=gpu.gfx_clk,
                                                                gddr_clk=gpu.gddr_clk,
                                                                app_clk=gpu.app_clk,
                                                                def_app_clk=gpu.def_app_clk,
                                                                sup_clk=gpu.sup_clk,
                                                                twoDTDMusage=gpu.twoDTDMusage,
                                                                GEOMMusage=gpu.GEOMMusage,
                                                                threeDMusage=gpu.threeDMusage,
                                                              )

    gpuMemStr="       Process Memory Usage(MiB)"
    for i in range(1, gpu.pronum + 1):
        gpuMem =["""
                PID                     : {pid}                   
                    NAME                : {name}
                    LMA                 : {lma}
                    Total               : {total}""".format(pid=getattr(gpu,"pro{}PID".format(i)),
                                                            name=getattr(gpu,"pro{}NAME".format(i)),
                                                            lma=getattr(gpu,"pro{}LMA".format(i)),
                                                            total=getattr(gpu,"pro{}Total".format(i))
                                                           )
                        ]
        gpuMemStr += "".join(gpuMem)

    gpuAttrs += gddrCRCAttrs + otherAttrs
    for i in range(1, gpu.pronum + 1):
        getattr(gpu,"pro{}PID".format(i))

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n')
        fo.write(gpuAttrs)
        fo.write('\n')
        fo.close()
    else:
        print(gpuAttrs)
        # print(gpuMemStr)

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index = -1
    num_xdxgpus = len(xdx_smi.xdxgpus)
    try:
        if(args['--index'] == None):
            index = -1                    # default situation
        elif(args['--index'] != None and int(args['--index']) >= 0
            and int(args['--index']) < num_xdxgpus):
            index = int(args['--index'])  # assign index
        else:
            raise ValueError("No devices were found")
    except:
        print("No devices were found")
        exit()

    target_gpus = xdx_smi.xdxgpus
    if(index >= 0):
        target_gpus = [xdx_smi.xdxgpus[index]]

    timestp = time.asctime(time.localtime(time.time()))
    header = """\
==========================XDXSMI LOG==========================
    Timestamp                       : {timestp}
    Driver Version
        Driver                      : {driver_ver}                  
        DDK                         : {vers1}
        KMD                         : {vers2}
        UMD                         : {vers3}
    Attached GPUs                   : {gpuNums}""".format(  timestp=timestp,
                                                            driver_ver=xdx_smi.driver_ver,
                                                            vers1=xdx_smi.vers[1],
                                                            vers2=xdx_smi.vers[2],
                                                            vers3=xdx_smi.vers[3],
                                                            gpuNums=xdx_smi.gpuNums
                                                         )

    # show running process
    gpu = target_gpus[0]
    # filter out processes whose LMA is 0
    non_zero_lma_process = [i for i in range(1, gpu.pronum+1) if float(getattr(gpu, "pro{}LMA".format(i))) > 0]

    gpuMemStr="    top-5 Process Memory Usage(MiB)"
    for i in non_zero_lma_process:
        gpuMem =["""
            PID                         : {pid}                   
                NAME                    : {name}
                LMA                     : {lma}
                Total                   : {total}""".format(pid=getattr(gpu,"pro{}PID".format(i)),
                                                            name=getattr(gpu,"pro{}NAME".format(i)),
                                                            lma=getattr(gpu,"pro{}LMA".format(i)),
                                                            total=getattr(gpu,"pro{}Total".format(i))
                                                           )]
        gpuMemStr += "".join(gpuMem)
    gpuMemStr += "\n"


    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write(header)
        fo.close()
    else:
        print(header)
        print(gpuMemStr)

    for xdxgpu in target_gpus:
        printAttributes(xdxgpu, args)
        xdxgpu.update_dynamic_info()


if __name__ == "__main__":
    print(docopt(__doc__))
