# -*- coding: utf-8 -*-
"""
usage:
   xdxsmi query_gpu get [<ip>...] [options]

ip:
   uuid                   uuid of the gpu
   index                  index of gpu, start from 0
   temperature.gpu        current temperature of the Soc
   utilization.gpu        utilization of GPU
   utilization.memory     utilization of GPU memory
   memory.total           the total size of GPU memory
   memory.free            the free size of GPU memory
   memory.used            the size of GPU memory which has been used
   clocks.gr              gddr frequency

options:
   --format=csv           comma seperated values. value: csv, nocsv. [default: nocsv]
   --noheader=noheader    skip the first line with colum headers. value: 0, 1. [default: 0]
   --nounits=nounits      don't print units for numerical values. value: 0, 1. [default: 0]
   -h, --help             print this help message
   -f, --filename=file    redirect query output to a specified file
   -l, --loop=sec         set look up interval, whose unit is second
   -i, --index=val        target gpu to read/write attributes, start from 0      

For example:
   e.g.1 xdxsmi query_gpu get uuid index temperature.gpu utilization.gpu utilization.memory memory.total memory.free memory.used clocks.gr --format=csv --nounits=1 --noheader=1
         query all of the info with comma seperating values, no units, no header
   e.g.2 xdxsmi query_gpu get uuid temperature.gpu
         just query uuid and temperature.gpu with the default settings. (nocsv, show header, show units)
   e.g.3 xdxsmi query_gpu get
         query all of the info with the default settings. (nocsv, show header, show units)
"""

# from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *
from xdxsmi.cli.render import Render

query_unit_map = {
   "uuid":      "",
   "index":     "",
   "temperature.gpu":  "°C",
   "utilization.gpu":  "%",
   "utilization.memory": "%",
   "memory.total":       "MiB",
   "memory.free":        "MiB",
   "memory.used":        "MiB",
   "clocks.gr":          "MHz"
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
   #Collect target gpus
   
   index = -1
   num_xdxgpus = len(xdx_smi.xdxgpus)
   try:
      if(args['--index'] == None):
         index = -1                    # default situation
      elif(args['--index'] != None and int(args['--index']) >= 0
         and int(args['--index']) < num_xdxgpus):
         index = int(args['--index'])  # assign index
      else:
         raise ValueError("No devices were found")
   except:
      print("No devices were found")
      exit()
   
   target_gpus = xdx_smi.xdxgpus
   if(index >= 0):
      target_gpus = [xdx_smi.xdxgpus[index]]

   show_units = 1
   try:
      nounits = int(args['--nounits'])
      if(nounits == 1):
         show_units = 0
      elif nounits not in [0, 1]:
         raise ValueError()
   except Exception:
      print("Nounits modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      exit()

   show_header = 1
   try:
      noheader = int(args['--noheader'])
      if(noheader == 1):
         show_header = 0
      elif noheader not in [0, 1]:
         raise ValueError()
   except Exception:
      print("Noheader modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      exit()

   csv = 0
   try:
      format = args['--format']
      if(format == "csv"):
         csv = 1
      elif format not in ["csv", "nocsv"]:
         raise ValueError()
   except Exception:
      print("Format modifier is not recognized. See 'xdxsmi query_gpu -h'.")
      exit()

   if args["get"]:
      target_ip = ["uuid", "index", "temperature.gpu", "utilization.gpu",
                     "utilization.memory", "memory.total", "memory.free",
                     "memory.used", "clocks.gr"]
      # li: filter the same element between target_ip and args["<ip"]
      if args["<ip>"]:
         target_ip = [ip for ip in args["<ip>"] if ip in target_ip]

      # li: judge the input target_ip is legal
      if len(target_ip) == 0:
         element_ip = next(iter(args["<ip>"]))
         exit('''xdxsmi query_gpu get {element_ip} is not a xdxsmi query_gpu command.\
               See 'xdxsmi query_gpu --help'.'''.format(element_ip=element_ip))
      
      data = []
      header = [] # build header: target_ip name + units
      for ip in target_ip:
         element = ip
         if ip in query_unit_map and query_unit_map[ip] != "":
            element += ("[" + query_unit_map[ip] + "]")
         header.append(element)

      if show_header == 1:
         data.insert(0, header)

      # get data
      for gpu in target_gpus:
         xdev = gpu.handle
         queryInfo = xdxml_device_query_two_parts(xdev)
         
         query_gpu_dict = {
            "uuid":               "GPU-" + str(gpu.uuid),
            "index":              str(gpu.user_index),
            "temperature.gpu":    str(queryInfo.temp),
            "utilization.gpu":    str(queryInfo.util_gpu),
            "utilization.memory": str(round(gpu.fb_used*100 / gpu.fb_total)),
            "memory.total":       str(gpu.fb_total),
            "memory.free":        str(gpu.fb_free),
            "memory.used":        str(gpu.fb_total - gpu.fb_free),
            "clocks.gr":          str(int(queryInfo.gddr_speed/1e9 /8 * 1000))
         }

         row = []
         for ip in target_ip:
            append_data = query_gpu_dict[ip]
            if show_units == 1 and query_unit_map[ip] != "":
               append_data = append_data + " " + query_unit_map[ip]

            # comma seperated values
            if csv == 1 and ip != target_ip[-1]:
               append_data = append_data + ","
               
            row.append(append_data)
         
         data.append(row)
         # a part of the shown info is from xdxgpu, so we must update dynamic info
         gpu.update_dynamic_info()
      Render.render_table(data, args, upper = False)
   
   else:
      exit("%r is not a xdxsmi query_gpu command."\
            "See 'xdxsmi query_gpu --help'." % args['<command>'])

if __name__ == "__main__":
   print(docopt(__doc__))