# -*- coding: utf-8 -*-

"""
usage:
   xdxsmi thermal [options] get [<ip>...]
   xdxsmi thermal [options] set temp_slowdown <°C>
   xdxsmi thermal [options] set temp_shutdown <°C>
   xdxsmi thermal [options] set enable_shutdown <dim>

options:
   -h, --help            print this help message
   -f, --filename=file   redirect query output to a specified file
   -l, --loop=sec        set look up interval, whose unit is second
   -i, --index=val       target gpu to read/write attributes, start from 0

ip:
   temp_slowdown         The critical temperature at which the GPU starts to downclock 
   temp_shutdown         The critical temperature at which the GPU powers off
   enable_shutdown       enable to power off or not. 0: disable, 1: enable

Be cautious! This command can be used to set the thermal behaviors.
e.g. if set enable_shutdown to 1, the GPU will power off when its temperature increases to temp_shutdown.
"""

# from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *
from xdxsmi.cli.render import Render

# map from the input ip to the attr name in thm structure
thm_map = {
    "temp_slowdown" : "xdx_temp_slowdown",
    "temp_shutdown" : "xdx_temp_shutdown",
    "enable_slowdown" : "xdx_enable_thm_slowdown",
    "enable_shutdown" : "xdx_enable_thm_shutdown"
}

thm_unit_map = {
    "temp_slowdown" : "°C",
    "temp_shutdown" : "°C",
    "enable_slowdown" : "",
    "enable_shutdown" : ""
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index = -1
    num_xdxgpus = len(xdx_smi.xdxgpus)
    try:
        if(args['--index'] == None):
            index = -1                    # default situation
        elif(args['--index'] != None and int(args['--index']) >= 0
            and int(args['--index']) < num_xdxgpus):
            index = int(args['--index'])  # assign index
        else:
            raise ValueError("No devices were found")
    except:
        print("No devices were found")
        exit()
    
    target_gpus = xdx_smi.xdxgpus
    if(index >= 0):
        target_gpus = [xdx_smi.xdxgpus[index]]
       
    
    #Set value for attribute 'gfx'
    if args["set"] and args["temp_slowdown"] and args["<°C>"]:
        #Judge whether the value of temp_slowdown is valid
        try:
            newVal = str(args["<°C>"])
            for gpu in target_gpus:
                xdev = gpu.handle
                temper = xdxmltemperInfo_t()
                xdxml_device_get_temper(xdev, temper, "temp_max")

                thermal = xdxmlthermalInfo_t()
                xdxml_device_get_thermal(xdev, thermal, "temp_shutdown")
                lower = int(temper.temp_max)
                upper = int(thermal.xdx_temp_shutdown)

                if(lower < int(newVal) and int(newVal) < upper):
                    thm_type = "temp_slowdown"
                    xdxml_device_set_thermal_info(xdev, newVal, thm_type)
                else:
                    print ("[XDXSMI ERROR]: Invalid input temp_slowdown", args["<°C>"])
                    print ("[XDXSMI ERROR]: Please enter an integer between temp_max {} and temp_shutdown {}".format(lower, upper))
                    exit()
        except Exception:
            element_ip = args["<°C>"]
            exit('''xdxsmi thermal set temp_slowdown {element_ip} is not a xdxsmi thermal command.\
                See 'xdxsmi thermal --help'.'''.format(element_ip=element_ip))
    
    elif args["set"] and args["temp_shutdown"] and args["<°C>"]:
        #Judge whether the value of temp_slowdown is valid
        try:
            newVal = str(args["<°C>"])
            for gpu in target_gpus:
                xdev = gpu.handle
                thermal = xdxmlthermalInfo_t()
                xdxml_device_get_thermal(xdev, thermal, "temp_slowdown")
                lower = int(thermal.xdx_temp_slowdown)
                upper = 105
                
                if(lower < int(newVal) and int(newVal) < upper):
                    thm_type = "temp_shutdown"
                    xdxml_device_set_thermal_info(xdev, newVal, thm_type)
                else:
                    print ("[XDXSMI ERROR]: Invalid input temp_shutdown", args["<°C>"])
                    print ("[XDXSMI ERROR]: Please enter an integer between slowdown temperature {} and protective temperature {}".format(lower, upper))
                    exit()

        except Exception:
            element_ip = args["<°C>"]
            exit('''xdxsmi thermal set temp_shutdown {element_ip} is not a xdxsmi thermal command.\
                See 'xdxsmi thermal --help'.'''.format(element_ip=element_ip))

    elif args["set"] and args["enable_shutdown"] and args["<dim>"]:
        try:
            newVal = str(int(args["<dim>"]))
            if(newVal not in ["0","1"]):
                print ("[XDXSMI ERROR]: Invalid input enable_slowdown", args["<dim>"])
                print ("[XDXSMI ERROR]: Please select one of the following values and reset it:")
                print ("                0, 1")
                exit()
            for gpu in target_gpus:
                xdev = gpu.handle
                thm_type = "enable_shutdown"
                xdxml_device_set_thermal_info(xdev, newVal, thm_type)
        except Exception:
            element_ip = args["<dim>"]
            exit('''xdxsmi thermal set enable_shutdown {element_ip} is not a xdxsmi thermal command.\
                See 'xdxsmi thermal --help'.'''.format(element_ip=element_ip)) 

    #Get values for target attributes
    elif args["get"]:
        target_ip = ["temp_slowdown", "temp_shutdown", "enable_shutdown"]
        # filter the same element between target_ip and args["<ip"]
        if args["<ip>"]:
            target_ip = [ip for ip in target_ip if ip in args["<ip>"]]

        # judge the input target_ip is legal
        if len(target_ip) == 0:
            element_ip = next(iter(args["<ip>"]))
            exit('''xdxsmi thermal get {element_ip} is not a xdxsmi thermal command.\
                See 'xdxsmi thermal --help'.'''.format(element_ip=element_ip))
        
        data = []
        # i = 0
        header = ["index", "gpu", "name"] + target_ip
        data.insert(0, header)
        # instead traverse handle to traverse gpus, cause gpus include handle and user_index
        for gpu in target_gpus:
            xdev = gpu.handle 
            thmInfo = xdxml_device_get_thermal_info(xdev)
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
            for ip in target_ip:
                # unifiy the unit
                row.append(str(int(getattr(thmInfo, thm_map[ip]))) + thm_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        if sys.version_info[0] == 2:
            Render.render_table_py2(data, args)
        else:
            Render.render_table(data, args)

    else:
        # no case falls into this situation?
        print("in else")
        exit("%r is not a xdxsmi thermal command."\
             "See 'xdxsmi thermal --help'." % args['<command>'])

if __name__ == "__main__":
    print(docopt(__doc__))