/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/types.h>
#include <linux/errno.h>

#include <drm/drm_fourcc.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)
#include <linux/sched.h>
#endif
#include <drm/drm_framebuffer.h>
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 20, 0))
#include "dc8200.h"
#endif
#include "dec400.h"

#define fourcc_mod_xdx_get_tile_mode(val)                                      \
	(u8)((val)&DRM_FORMAT_MOD_XDX_DEC_TILE_MODE_MASK)

static inline void dec_set_clear(struct dm_dc *dc, u32 reg, u32 set, u32 clear)
{
	u32 value = DC_RREG32(reg);

	value &= ~clear;
	value |= set;

	DC_WREG32(reg, value);
}

static void dec400_hw_init(struct dm_dc *dc)
{
	u32 config = 0;

	config = DEC_CONTROL_RESET & (~COMPRESSION_DISABLE);
	DC_WREG32(DEC_CONTROL, config | FLUSH_ENABLE);

	config = DEC_CONTROL_EX2_RESET & (~TILE_STATUS_READ_ID_MASK) &
		 (~TILE_STATUS_READ_ID_H_MASK) & (~DISABLE_HW_DEC_FLUSH);
	DC_WREG32(DEC_CONTROL_EX2,
			     config | (TILE_STATUS_READ_ID_H << 22) |
				     TILE_STATUS_READ_ID);

	config = DEC_CONTROL_EX_RESET & (~WRITE_MISS_POLICY_MASK) &
		 (~READ_MISS_POLICY_MASK);
	DC_WREG32(DEC_CONTROL_EX,
			     config | (WRITE_MISS_POLICY1 << 19));
}

static void dec400_stream_set(struct dm_dc *dc, u32 main_base_addr,
			      u32 ts_base_addr, u8 tile_mode, u8 align_mode,
			      u8 format, u8 depth, u8 stream_id)
{
	u32 offset = stream_id << 2;

	dec_set_clear(dc, DEC_READ_CONFIG + offset,
		      (tile_mode << 25) | (align_mode << 16) | (format << 3) |
			      COMPRESSION_EN,
		      TILE_MODE_MASK | COMPRESSION_ALIGN_MODE_MASK |
			      COMPRESSION_FORMAT_MASK);

	dec_set_clear(dc, DEC_READ_EX_CONFIG + offset, (depth << 16),
		      BIT_DEPTH_MASK);

	DC_WREG32(DEC_READ_BUFFER_BASE + offset, main_base_addr);
	DC_WREG32(DEC_READ_BUFFER_END + offset, ts_base_addr - 128);
	DC_WREG32(DEC_READ_CACHE_BASE + offset, ts_base_addr);
}

static void dec400_stream_disable(struct dm_dc *dc, u8 stream_id)
{
	u32 offset = stream_id << 2;

	DC_WREG32(DEC_READ_CONFIG + offset, DEC_READ_CONFIG_RESET);
	DC_WREG32(DEC_READ_BUFFER_BASE + offset, 0xFFFFFFFF);
	DC_WREG32(DEC_READ_BUFFER_END + offset, 0xFFFFFFFF);
}

static inline bool _is_stream_changed(struct dec400l *dec400l, u8 stream_id)
{
	return dec400l->stream[stream_id].dirty;
}

static inline bool _is_stream_valid(struct dec400l *dec400l, u8 stream_id)
{
	return !!(dec400l->stream[stream_id].main_base_addr);
}

static void _enable_stream(struct dec400l *dec400l, struct dm_dc *dc,
			   u8 stream_id)
{
	if (!(dec400l->stream_status & (1 << stream_id))) {
		if (!(dec400l->stream_status))
			/* the first enabled steram */
			dec400_hw_init(dc);

		dec400l->stream_status |= 1 << stream_id;
	}
}

static void _disable_stream(struct dec400l *dec400l, struct dm_dc *dc,
			    u8 stream_id)
{
	if ((dec400l->stream_status & (1 << stream_id)))
		dec400l->stream_status &= ~(1 << stream_id);
}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 10, 0))
static u16 get_dec_tile_size(u8 tile_mode, u8 cpp)
{
	u16 multi = 0;

	switch (tile_mode) {
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_16X1:
		multi = 16;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X4:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_4X8:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_32X1:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X4_S:
		multi = 32;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X8_XMAJOR:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X8_YMAJOR:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_16X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_32X2:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X4_S:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X4_LSB:
		multi = 64;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X8:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X16:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_32X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_64X2:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X4_S:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X4_LSB:
		multi = 128;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_64X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_64X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X2:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X16:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X8:
		multi = 256;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_256X2:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X4:
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_512X1:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_128X4:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X16:
		multi = 512;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_256X4:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_64X16:
	case DRM_FORMAT_MOD_XDX_DEC_TILE_128X8:
		multi = 1024;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_512X4:
		multi = 2048;
		break;
	default:
		break;
	}

	return multi * cpp;
}
#endif

static void update_fb_format(struct dec_fb *dec_fb)
{
	struct drm_framebuffer *drm_fb = dec_fb->fb;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	u8 tile_mod = fourcc_mod_xdx_get_tile_mode(drm_fb->modifier);
#else
	u8 tile_mod = fourcc_mod_xdx_get_tile_mode(drm_fb->modifier[0]);
#endif
	u8 norm_mod = DRM_FORMAT_MOD_XDX_LINEAR;

	switch (tile_mod) {
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_32X1:
		norm_mod = DRM_FORMAT_MOD_XDX_TILE_32X1;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1:
		norm_mod = DRM_FORMAT_MOD_XDX_TILE_64X1;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1:
		norm_mod = DRM_FORMAT_MOD_XDX_TILE_128X1;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1:
		norm_mod = DRM_FORMAT_MOD_XDX_TILE_256X1;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X4:
		norm_mod = DRM_FORMAT_MOD_XDX_SUPER_TILED_XMAJOR_8X4;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_4X8:
		norm_mod = DRM_FORMAT_MOD_XDX_SUPER_TILED_YMAJOR_4X8;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_8X8_XMAJOR:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
		if (drm_fb->format->format == DRM_FORMAT_YUYV ||
		    drm_fb->format->format == DRM_FORMAT_UYVY ||
		    drm_fb->format->format == DRM_FORMAT_P010)
#else
		if (drm_fb->pixel_format == DRM_FORMAT_YUYV ||
			drm_fb->pixel_format == DRM_FORMAT_UYVY ||
			drm_fb->pixel_format == DRM_FORMAT_P010)
#endif
			norm_mod = DRM_FORMAT_MOD_XDX_TILE_8X8;
		else
			norm_mod = DRM_FORMAT_MOD_XDX_SUPER_TILED_XMAJOR;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_16X8:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
		if (drm_fb->format->format == DRM_FORMAT_NV12 ||
		    drm_fb->format->format == DRM_FORMAT_P010)
#else
		if (drm_fb->pixel_format == DRM_FORMAT_NV12 ||
			drm_fb->pixel_format == DRM_FORMAT_P010)
#endif
			norm_mod = DRM_FORMAT_MOD_XDX_TILE_8X8;
		break;
	case DRM_FORMAT_MOD_XDX_DEC_TILE_32X8:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
		if (drm_fb->format->format == DRM_FORMAT_NV12)
#else
		if (drm_fb->pixel_format == DRM_FORMAT_NV12)
#endif
			norm_mod = DRM_FORMAT_MOD_XDX_TILE_8X8;
		break;
	default:
		break;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	drm_fb->modifier = fourcc_mod_xdx_norm_code(norm_mod);
#else
	drm_fb->modifier[0] = fourcc_mod_xdx_norm_code(norm_mod);
#endif
}

static void _stream_config(struct dec_fb *dec_fb,
			   struct dec_stream *stream, u8 index)
{
	struct drm_framebuffer *drm_fb = dec_fb->fb;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 10, 0))
	u16 tile_size = 0;
	const struct drm_format_info *info = drm_fb->format;
#endif
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 0)) || (defined(REDHAT_83)))	
	u32 plane_height =
		drm_format_info_plane_height(info, drm_fb->height, index);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	u32 plane_height =
		drm_format_plane_height(drm_fb->height, drm_fb->format->format , index);
#else
	// unsed variable
//	u32 plane_height =
//		drm_format_plane_height(drm_fb->height, drm_fb->pixel_format, index);
#endif

	stream->main_base_addr = dec_fb->addr[index];
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	stream->tile_mode = fourcc_mod_xdx_get_tile_mode(drm_fb->modifier);
	if (drm_fb->modifier & DRM_FORMAT_MOD_XDX_DEC_ALIGN_32)
#else
	stream->tile_mode = fourcc_mod_xdx_get_tile_mode(drm_fb->modifier[0]);
	if (drm_fb->modifier[0] & DRM_FORMAT_MOD_XDX_DEC_ALIGN_32)
#endif
		stream->align_mode = DEC_ALIGN_32;
	else
		stream->align_mode = DEC_ALIGN_64;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	switch (drm_fb->format->format) {
#else
	switch (drm_fb->pixel_format) {
#endif
	case DRM_FORMAT_ARGB8888:
		stream->format = DEC_FORMAT_ARGB8;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_XRGB8888:
		stream->format = DEC_FORMAT_XRGB8;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_RGB565:
		stream->format = DEC_FORMAT_R5G6B5;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_ARGB1555:
		stream->format = DEC_FORMAT_A1RGB5;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_XRGB1555:
		stream->format = DEC_FORMAT_X1RGB5;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_ARGB4444:
		stream->format = DEC_FORMAT_ARGB4;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_XRGB4444:
		stream->format = DEC_FORMAT_XRGB4;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_ARGB2101010:
		stream->format = DEC_FORMAT_A2R10G10B10;
		stream->depth = DEC_DEPTH_10;
		break;
	case DRM_FORMAT_YUYV:
		stream->format = DEC_FORMAT_YUY2;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_UYVY:
		stream->format = DEC_FORMAT_UYVY;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_YVU420:
		stream->format = DEC_FORMAT_YUV_ONLY;
		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_NV12:
		WARN_ON(stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1 &&
			stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1 &&
			stream->tile_mode != DRM_FORMAT_MOD_XDX_DEC_TILE_32X8 &&
			stream->tile_mode != DRM_FORMAT_MOD_XDX_DEC_TILE_16X8);
		if (index) {
			stream->format = DEC_FORMAT_UV_MIX;
			switch (stream->tile_mode) {
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_TILE_32X8:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_TILE_32X4;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_TILE_16X8:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_TILE_16X4;
				break;
			default:
				break;
			}
		} else {
			stream->format = DEC_FORMAT_YUV_ONLY;
		}

		stream->depth = DEC_DEPTH_8;
		break;
	case DRM_FORMAT_P010:
		WARN_ON(stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1 &&
			stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1 &&
			stream->tile_mode != DRM_FORMAT_MOD_XDX_DEC_TILE_16X8 &&
			stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_TILE_8X8_XMAJOR);
		if (index) {
			stream->format = DEC_FORMAT_UV_MIX;
			switch (stream->tile_mode) {
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_32X1;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_TILE_16X8:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_TILE_16X4;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_TILE_8X8_XMAJOR:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_TILE_8X4;
				break;
			default:
				break;
			}
		} else {
			stream->format = DEC_FORMAT_YUV_ONLY;
		}

		stream->depth = DEC_DEPTH_10;
		break;
	case DRM_FORMAT_NV16:
		WARN_ON(stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1 &&
			stream->tile_mode !=
				DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1);
		if (index) {
			stream->format = DEC_FORMAT_UV_MIX;
			switch (stream->tile_mode) {
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_256X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1;
				break;
			case DRM_FORMAT_MOD_XDX_DEC_RASTER_128X1:
				stream->tile_mode =
					DRM_FORMAT_MOD_XDX_DEC_RASTER_64X1;
				break;
			default:
				break;
			}
		} else {
			stream->format = DEC_FORMAT_YUV_ONLY;
		}

		stream->depth = DEC_DEPTH_8;
		break;
	}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 10, 0))
	tile_size = get_dec_tile_size(stream->tile_mode, info->cpp[index]);
	stream->aligned_stride = ALIGN(dec_fb->stride[index], tile_size);
	stream->ts_base_addr =
		stream->main_base_addr + stream->aligned_stride * plane_height;
#endif
}

static int _dec400l_config(struct dec400l *dec400l, struct dec_fb *dec_fb,
			   u8 stream_base)
{
	struct dec_stream stream;
	u8 i, stream_id, num_planes = 0;

	if (dec_fb) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 10, 0))
		const struct drm_format_info *info = dec_fb->fb->format;

		num_planes = info->num_planes;
#else
		num_planes = drm_format_num_planes(dec_fb->fb->pixel_format);
#endif
	}

	for (i = 0; i < STREAM_COUNT; i++) {
		stream_id = stream_base + i;

		memset(&dec400l->stream[stream_id], 0,
		       sizeof(struct dec_stream));

		if (i < num_planes) {
			memset(&stream, 0, sizeof(struct dec_stream));
			_stream_config(dec_fb, &stream, i);
			memcpy(&dec400l->stream[stream_id], &stream,
			       sizeof(struct dec_stream));
		}
		dec400l->stream[stream_id].dirty = true;
	}

	if (dec_fb)
		update_fb_format(dec_fb);

	return 0;
}

int dec4001_config(struct dec400l *dec400l, struct dec_fb *dec_fb,
		  u8 stream_base)
{
	if (dec_fb && !dec_fb->fb)
		return -EINVAL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	if (dec_fb && (fourcc_mod_xdx_get_type(dec_fb->fb->modifier) !=
		       DRM_FORMAT_MOD_XDX_TYPE_COMPRESSED))
#else
	if (dec_fb && (fourcc_mod_xdx_get_type(dec_fb->fb->modifier[0]) !=
			DRM_FORMAT_MOD_XDX_TYPE_COMPRESSED))
#endif
		_dec400l_config(dec400l, NULL, stream_base);
	else
		_dec400l_config(dec400l, dec_fb, stream_base);

	return 0;
}

int dec4001_commit(struct dec400l *dec400l, struct dm_dc *dc)
{
	u8 i;

	for (i = 0; i < STREAM_TOTAL; i++) {
		if (!_is_stream_changed(dec400l, i))
			continue;

		if (_is_stream_valid(dec400l, i)) {
			_enable_stream(dec400l, dc, i);
			dec400_stream_set(dc,
					     dec400l->stream[i].main_base_addr,
					     dec400l->stream[i].ts_base_addr,
					     dec400l->stream[i].tile_mode,
					     dec400l->stream[i].align_mode,
					     dec400l->stream[i].format,
					     dec400l->stream[i].depth, i);
		} else {
			dec400_stream_disable(dc, i);
			_disable_stream(dec400l, dc, i);
		}

		dec400l->stream[i].dirty = false;
	}

	return 0;
}
