/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
#include <drm/drm_atomic_state_helper.h>
#endif
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)) || (defined(REDHAT_83)))
#include <drm/drm_probe_helper.h>
#else
#include <drm/drm_crtc_helper.h>
#endif


#include "dm_connector.h"
#include "xdx_device.h"
#include "kernel_compatibility.h"

static void xdx_dm_connector_reset(struct drm_connector *connector)
{
	struct dm_connector_state *dm_state;;

	if (connector->state) {
		dm_state = to_dm_connector_state(connector->state);
		__drm_atomic_helper_connector_destroy_state(connector->state);
		connector->state = NULL;
		kfree(dm_state);
	}

	dm_state = kzalloc(sizeof(*dm_state), GFP_KERNEL);

	if (dm_state) {
		dm_state->scaler.mode = DRM_MODE_SCALE_NONE;
		dm_state->scaler.underscan.mode = UNDERSCAN_OFF;
		__drm_atomic_helper_connector_reset(connector, &dm_state->base);
	}
}

static enum drm_connector_status
xdx_dm_connector_detect(struct drm_connector *connector, bool force)
{
	return connector_status_connected;
}

static void xdx_dm_connector_destroy(struct drm_connector *connector)
{
	struct dm_connector *dm_connector = to_dm_connector(connector);

	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
	i2c_del_adapter(&dm_connector->i2c);
	kfree(dm_connector);
}

static struct drm_connector_state *
xdx_dm_connector_atomic_duplicate_state(struct drm_connector *connector)
{
	struct dm_connector_state *state =
		to_dm_connector_state(connector->state);

	struct dm_connector_state *new_state =
		kmemdup(state, sizeof(*state), GFP_KERNEL);

	if (!new_state)
		return NULL;

	__drm_atomic_helper_connector_duplicate_state(connector,
						      &new_state->base);

	return &new_state->base;
}

static void
xdx_dm_atomic_helper_connector_destroy_state(struct drm_connector *connector,
					     struct drm_connector_state *state)
{
	struct dm_connector_state *dm_state = to_dm_connector_state(state);

	__drm_atomic_helper_connector_destroy_state(state);

	kfree(dm_state);
}

static int xdx_dm_connector_atomic_set_property(
	struct drm_connector *connector,
	struct drm_connector_state *connector_state,
	struct drm_property *property, uint64_t val)
{
	/* TODO: */
	return 0;
}

static int
xdx_dm_connector_atomic_get_property(struct drm_connector *connector,
				     const struct drm_connector_state *state,
				     struct drm_property *property,
				     uint64_t *val)
{
	/* TODO: */
	return 0;
}

static int xdx_dm_connector_late_register(struct drm_connector *connector)
{
	struct dm_connector *dm_connector = to_dm_connector(connector);
	int r;

	if ((connector->connector_type == DRM_MODE_CONNECTOR_DisplayPort) ||
	    (connector->connector_type == DRM_MODE_CONNECTOR_eDP) ||
	    (connector->connector_type == DRM_MODE_CONNECTOR_VGA)) {
		r = drm_dp_aux_register(&dm_connector->aux);
		if (r)
			return r;
	}

	return 0;
}

static void xdx_dm_connector_unregister(struct drm_connector *connector)
{
	struct dm_connector *dm_connector = to_dm_connector(connector);

	if ((connector->connector_type == DRM_MODE_CONNECTOR_DisplayPort) ||
	    (connector->connector_type == DRM_MODE_CONNECTOR_eDP) ||
	    (connector->connector_type == DRM_MODE_CONNECTOR_VGA))
		drm_dp_aux_unregister(&dm_connector->aux);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 19, 0))
#include <drm/drm_encoder.h>
/**
 * xdx_connector_for_each_possible_encoder - iterate connector's possible encoders
 * @connector: &struct drm_connector pointer
 * @encoder: &struct drm_encoder pointer used as cursor
 * @__i: int iteration cursor, for macro-internal use
 */
#define xdx_connector_for_each_possible_encoder(connector, encoder, __i) \
        for ((__i) = 0; (__i) < ARRAY_SIZE((connector)->encoder_ids) && \
                     (connector)->encoder_ids[(__i)] != 0; (__i)++) \
                for_each_if((encoder) = \
                            drm_encoder_find((connector)->dev, \
                                             (connector)->encoder_ids[(__i)])) \

#endif

static const struct drm_connector_funcs dm_connector_funcs = {
	.reset = xdx_dm_connector_reset,
	.detect = xdx_dm_connector_detect,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = xdx_dm_connector_destroy,
	.atomic_duplicate_state = xdx_dm_connector_atomic_duplicate_state,
	.atomic_destroy_state = xdx_dm_atomic_helper_connector_destroy_state,
	.atomic_set_property = xdx_dm_connector_atomic_set_property,
	.atomic_get_property = xdx_dm_connector_atomic_get_property,
	.late_register = xdx_dm_connector_late_register,
	.early_unregister = xdx_dm_connector_unregister,
};

static struct drm_encoder *connector_to_encoder(struct drm_connector *connector)
{
        struct drm_encoder *encoder;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 5, 0)) && (!defined(REDHAT_83)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0))
        unsigned int i;

        /* There is only one encoder per connector */
        drm_connector_for_each_possible_encoder (connector, encoder, i)
        return encoder;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0))
        drm_connector_for_each_possible_encoder (connector, encoder)
        return encoder;
#else//<(4,19,0)
        unsigned int i;
        xdx_connector_for_each_possible_encoder (connector, encoder, i)
        return encoder;
#endif
        return NULL;
}

static void dm_get_native_mode(struct drm_connector *connector)
{
	struct drm_encoder *encoder;
	struct dm_encoder *dm_encoder;

	encoder = connector_to_encoder(connector);

	if (encoder == NULL)
		return;

	dm_encoder = to_dm_encoder(encoder);

	dm_encoder->native_mode.clock = 0;

	if (!list_empty(&connector->probed_modes)) {
		struct drm_display_mode *preferred_mode = NULL;

		list_for_each_entry (preferred_mode, &connector->probed_modes,
				     head) {
			if (preferred_mode->type & DRM_MODE_TYPE_PREFERRED)
				dm_encoder->native_mode = *preferred_mode;

			break;
		}
	}
}

static void dm_connector_ddc_get_modes(struct drm_connector *connector,
				       struct edid *edid)
{
	struct dm_connector *dm_connector = to_dm_connector(connector);

	if (edid) {
		/* empty probed_modes */
		INIT_LIST_HEAD(&connector->probed_modes);
		dm_connector->num_modes = drm_add_edid_modes(connector, edid);

		/* sorting the probed modes before calling function
		 * dm_get_native_mode() since EDID can have
		 * more than one preferred mode. The modes that are
		 * later in the probed mode list could be of higher
		 * and preferred resolution. For example, 3840x2160
		 * resolution in base EDID preferred timing and 4096x2160
		 * preferred resolution in DID extension block later.
		 */
		drm_mode_sort(&connector->probed_modes);
		dm_get_native_mode(connector);
	} else {
		dm_connector->num_modes = 0;
	}
}

static struct drm_display_mode *
dm_create_common_mode(struct drm_encoder *encoder, char *name, int hdisplay,
		      int vdisplay)
{
	struct drm_device *ddev = encoder->dev;
	struct dm_encoder *dm_encoder = to_dm_encoder(encoder);
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *native_mode = &dm_encoder->native_mode;

	mode = drm_mode_duplicate(ddev, native_mode);

	if (mode == NULL)
		return NULL;

	mode->hdisplay = hdisplay;
	mode->vdisplay = vdisplay;
	mode->type &= ~DRM_MODE_TYPE_PREFERRED;
	strscpy(mode->name, name, DRM_DISPLAY_MODE_LEN);

	return mode;
}

static void dm_connector_add_common_modes(struct drm_encoder *encoder,
					  struct drm_connector *connector)
{
	struct dm_encoder *dm_encoder = to_dm_encoder(encoder);
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *native_mode = &dm_encoder->native_mode;
	struct dm_connector *dm_connector = to_dm_connector(connector);
	int i;
	int n;
	struct mode_size {
		char name[DRM_DISPLAY_MODE_LEN];
		int w;
		int h;
	} common_modes[] = {
		{ "640x480", 640, 480 },     { "800x600", 800, 600 },
		{ "1024x768", 1024, 768 },   { "1280x720", 1280, 720 },
		{ "1280x800", 1280, 800 },   { "1280x1024", 1280, 1024 },
		{ "1440x900", 1440, 900 },   { "1680x1050", 1680, 1050 },
		{ "1600x1200", 1600, 1200 }, { "1920x1080", 1920, 1080 },
		{ "1920x1200", 1920, 1200 }
	};

	n = ARRAY_SIZE(common_modes);

	for (i = 0; i < n; i++) {
		struct drm_display_mode *curmode = NULL;
		bool mode_existed = false;

		/* prefer to do upscaling */
		if (common_modes[i].w > native_mode->hdisplay ||
		    common_modes[i].h > native_mode->vdisplay ||
		    (common_modes[i].w == native_mode->hdisplay &&
		     common_modes[i].h == native_mode->vdisplay))
			continue;

		list_for_each_entry (curmode, &connector->probed_modes, head) {
			if (common_modes[i].w == curmode->hdisplay &&
			    common_modes[i].h == curmode->vdisplay) {
				mode_existed = true;
				break;
			}
		}

		if (mode_existed)
			continue;

		mode = dm_create_common_mode(encoder, common_modes[i].name,
					     common_modes[i].w,
					     common_modes[i].h);
		drm_mode_probed_add(connector, mode);
		dm_connector->num_modes++;
	}
}

static int xdx_dm_connector_get_modes(struct drm_connector *connector)
{
	struct dm_connector *dm_connector = to_dm_connector(connector);
	struct drm_encoder *encoder;
	struct edid *edid = dm_connector->edid;

	encoder = connector_to_encoder(connector);

	if (!drm_edid_is_valid(edid)) {
		dm_connector->num_modes =
			drm_add_modes_noedid(connector, 800, 600);
		drm_set_preferred_mode(connector, 800, 600);
	} else {
		dm_connector_ddc_get_modes(connector, edid);
		dm_connector_add_common_modes(encoder, connector);
	}

	return dm_connector->num_modes;
}

static enum drm_mode_status xdx_dm_connector_mode_valid(struct drm_connector *connector,
				   struct drm_display_mode *mode)
{
	/* TODO: */
	return MODE_OK;
}

static const struct drm_connector_helper_funcs dm_connector_helper_funcs = {
	/*
	 * If hotplugging a second bigger display in FB Con mode, bigger resolution
	 * modes will be filtered by drm_mode_validate_size(), and those modes
	 * are missing after user start lightdm. So we need to renew modes list.
	 * in get_modes call back, not just return the modes count
	 */
	.get_modes = xdx_dm_connector_get_modes,
	.mode_valid = xdx_dm_connector_mode_valid,
};

int dm_connector_init(struct xdx_dm *dm, struct dm_connector *dm_connector,
		      uint32_t id, int connector_type,
		      struct dm_encoder *dm_encoder)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	dm_connector->id = id;
	dm_connector->connector_type = connector_type;
#if 0
	ret = drm_connector_init_with_ddc(ddev, &dm_connector->base,
					  &dm_connector_funcs,
					  dm_connector->connector_type,
					  &dm_connector->i2c);
#else
	ret = drm_connector_init(ddev, &dm_connector->base,
					  &dm_connector_funcs,
					  dm_connector->connector_type);
#endif
	if (ret) {
		dev_err(xdev->dev, "connector (%d) init failed\n", id);
		dm_connector->id = (uint32_t)(-1);
		return ret;
	}

	drm_connector_helper_add(&dm_connector->base,
				 &dm_connector_helper_funcs);

	ret = drm_connector_attach_encoder(&dm_connector->base, &dm_encoder->base);
	if (ret)
		dev_err(xdev->dev, "can't attach [ENCODER:%d:%s] to [CONNECTOR:%d:%s] (err=%d)\n",
			dm_encoder->base.base.id, dm_encoder->base.name, dm_connector->base.base.id,
			dm_connector->base.name, ret);

	return ret;
}
