/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/clk.h>
#include <linux/debugfs.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/version.h>

#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_modeset_helper_vtables.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
#include <drm/drm_vblank.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0))
#include <drm/drm_irq.h>
#else
#include <drm/drmP.h>
#endif

#include <drm/xdxgpu_drm.h>

#include "xdx_device.h"

static void xdx_dm_crtc_destroy(struct drm_crtc *crtc)
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	if (dm_crtc->vblank_timer_enabled)
		hrtimer_cancel(&dm_crtc->vblank_hrtimer);

	drm_crtc_cleanup(crtc);
	dm_crtc->funcs->destroy(dm_crtc);
}

static void xdx_dm_crtc_destroy_state(struct drm_crtc_state *state)
{
	struct dm_crtc_state *dm_crtc_state = to_dm_crtc_state(state);
	struct dm_crtc *dm_ctrc = to_dm_crtc(state->crtc);

	__drm_atomic_helper_crtc_destroy_state(state);

	dm_ctrc->funcs->crtc_state_destroy(dm_crtc_state);
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0)
static void xdx_atomic_helper_crtc_reset(struct drm_crtc *crtc,
			       struct drm_crtc_state *crtc_state)
{
	//static int count = 0;
	//count++;
	//printk("%s count =%d\n",__func__,count);
	if (crtc_state)
		crtc_state->crtc = crtc;

	crtc->state = crtc_state;
}
#endif
static void xdx_dm_crtc_reset(struct drm_crtc *crtc)
{
	struct dm_crtc_state *dm_state;
	struct dm_crtc *dm_ctrc = to_dm_crtc(crtc);

	if (crtc->state) {
		xdx_dm_crtc_destroy_state(crtc->state);
		crtc->state = NULL;
	}

	dm_state = dm_ctrc->funcs->crtc_state_create(dm_ctrc);
	if (WARN_ON(!dm_state))
		return;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
	__drm_atomic_helper_crtc_reset(crtc, &dm_state->base);
#else
	xdx_atomic_helper_crtc_reset(crtc, &dm_state->base);
#endif

	dm_state->output_fmt = DRM_COLOR_FORMAT_RGB444;
	dm_state->bpc = 8;
}

static struct drm_crtc_state *
xdx_dm_crtc_atomic_duplicate_state(struct drm_crtc *crtc)
{
	struct dm_crtc *dm_ctrc = to_dm_crtc(crtc);
	struct dm_crtc_state *ori_state;
	struct dm_crtc_state *state;

	if (WARN_ON(!crtc->state))
		return NULL;

	ori_state = to_dm_crtc_state(crtc->state);

	state = dm_ctrc->funcs->crtc_duplicate_state(ori_state);
	if (!state)
		return NULL;

	__drm_atomic_helper_crtc_duplicate_state(crtc, &state->base);

	state->output_fmt = ori_state->output_fmt;
	state->bpc = ori_state->bpc;

	return &state->base;
}

static void xdx_dm_crtc_atomic_destroy_state(struct drm_crtc *crtc,
					     struct drm_crtc_state *state)
{
	struct dm_crtc *dm_ctrc = to_dm_crtc(crtc);
	struct dm_crtc_state *dm_crtc_state = to_dm_crtc_state(state);

	__drm_atomic_helper_crtc_destroy_state(state);

	dm_ctrc->funcs->crtc_state_destroy(dm_crtc_state);
}

static int xdx_dm_crtc_atomic_set_property(struct drm_crtc *crtc,
					   struct drm_crtc_state *state,
					   struct drm_property *property,
					   uint64_t val)
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);
	struct dm_crtc_state *dm_crtc_state = to_dm_crtc_state(state);

	return dm_crtc->funcs->crtc_set_property(dm_crtc, dm_crtc_state,
						 property, val);
}

static int xdx_dm_crtc_atomic_get_property(struct drm_crtc *crtc,
					   const struct drm_crtc_state *state,
					   struct drm_property *property,
					   uint64_t *val)
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);
	struct dm_crtc_state *dm_crtc_state = to_dm_crtc_state((struct drm_crtc_state *)state);

	return dm_crtc->funcs->crtc_get_property(dm_crtc, dm_crtc_state,
						 property, val);
}

#ifdef CONFIG_DEBUG_FS
static int dm_crtc_debugfs_show(struct seq_file *s, void *data)
{
	struct drm_crtc *crtc = s->private;
	struct drm_display_mode *mode = &crtc->state->adjusted_mode;

	seq_printf(s, "crtc[%u]: %s\n", crtc->base.id, crtc->name);
	seq_printf(s, "\tactive = %d\n", crtc->state->active);
	seq_printf(s, "\tsize = %dx%d\n", mode->hdisplay, mode->vdisplay);

	return 0;
}

static int dm_crtc_debugfs_open(struct inode *inode, struct file *file)
{
	return single_open(file, dm_crtc_debugfs_show, inode->i_private);
}

static const struct file_operations dm_crtc_debugfs_fops = {
	.open = dm_crtc_debugfs_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

static int dm_crtc_debugfs_init(struct drm_crtc *crtc)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	debugfs_create_file("status", 0444, crtc->debugfs_entry, crtc,
			    &dm_crtc_debugfs_fops);
#endif
	return 0;
}
#else
static int dm_crtc_debugfs_init(struct drm_crtc *crtc)
{
	return 0;
}
#endif /* CONFIG_DEBUG_FS */

static int xdx_dm_crtc_late_register(struct drm_crtc *crtc)
{
	return dm_crtc_debugfs_init(crtc);
}


static enum hrtimer_restart vblank_simulate(struct hrtimer *timer)
{
	struct dm_crtc *dm_crtc =
		container_of(timer, struct dm_crtc, vblank_hrtimer);
	struct drm_crtc *crtc = to_drm_crtc(dm_crtc);
	struct drm_device *ddev = crtc->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	u64 ret_overrun;
	bool ret;

	ret_overrun = hrtimer_forward_now(&dm_crtc->vblank_hrtimer,
					  dm_crtc->period_ns);
	if (ret_overrun != 1)
		dev_warn(xdev->dev, "%s: vblank timer overrun\n", __func__);

	ret = drm_crtc_handle_vblank(crtc);
	if (!ret)
		dev_err(xdev->dev, "failure on handling vblank");

	return HRTIMER_RESTART;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
int xdx_drm_drv_enable_vblank(struct drm_device *ddev, unsigned int pipe)
{
        struct drm_crtc* crtc = NULL;
        struct dm_crtc *dm_crtc = NULL;
        struct xdx_device *xdev = drm_to_xdev(ddev);
        struct drm_vblank_crtc *vblank = &ddev->vblank[pipe];
        int ret = 0;
        drm_for_each_crtc(crtc, ddev) {
                if (drm_crtc_index(crtc) == pipe) {
                        dm_crtc = to_dm_crtc(crtc);
                        break;
                }
        }
        if (!dm_crtc) {
                dev_err(xdev->dev, "Can't enable vblank for CRTC:%d\n", pipe);
                ret = -EINVAL;
                goto done;
        }
        if (dm_crtc->vblank_timer_enabled) {
                drm_calc_timestamping_constants(crtc, &crtc->mode);
                hrtimer_init(&dm_crtc->vblank_hrtimer, CLOCK_MONOTONIC,
                             HRTIMER_MODE_REL);
                dm_crtc->vblank_hrtimer.function = vblank_simulate;
                dm_crtc->period_ns = ktime_set(0, vblank->framedur_ns);
                hrtimer_start(&dm_crtc->vblank_hrtimer, dm_crtc->period_ns,
                              HRTIMER_MODE_REL);
        } else if (dm_crtc->funcs->enable_vblank) {
                ret = dm_crtc->funcs->enable_vblank(dm_crtc, true);
        } else {
                ret = -ENXIO;
        }

        if (ret)
                dev_err(xdev->dev, "CRTC %d: failed to enable vblank (%d)\n",
                        drm_crtc_index(crtc), ret);
        else
                dev_dbg(xdev->dev, "CRTC %d: enable vblank\n",
                         drm_crtc_index(crtc));
done:
        return ret;
}

void xdx_drm_drv_disable_vblank(struct drm_device *ddev, unsigned int pipe)
{
        struct drm_crtc* crtc = NULL;
        struct dm_crtc *dm_crtc = NULL;
        struct xdx_device *xdev = drm_to_xdev(ddev);
        int ret = 0;
        drm_for_each_crtc(crtc, ddev) {
                if (drm_crtc_index(crtc) == pipe) {
                        dm_crtc = to_dm_crtc(crtc);
                        break;
                }
        }
        if (!dm_crtc) {
                dev_err(xdev->dev, "Can't enable vblank for CRTC:%d\n", pipe);
                return;
        }
        if (dm_crtc->vblank_timer_enabled)
                hrtimer_cancel(&dm_crtc->vblank_hrtimer);
        else if (dm_crtc->funcs->enable_vblank)
                ret = dm_crtc->funcs->enable_vblank(dm_crtc, false);
        else
                ret = -ENXIO;

        if (ret)
                dev_err(xdev->dev, "CRTC %d: failed to disable vblank (%d)\n",
                        drm_crtc_index(crtc), ret);
        else
                dev_dbg(xdev->dev, "CRTC %d: disable vblank\n",
                         drm_crtc_index(crtc));
}
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
static int xdx_dm_crtc_enable_vblank(struct drm_crtc *crtc)
{
	struct drm_device *ddev = crtc->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);
	uint32_t pipe = drm_crtc_index(crtc);
	struct drm_vblank_crtc *vblank = &ddev->vblank[pipe];
	int ret = 0;

	if (dm_crtc->vblank_timer_enabled) {
		drm_calc_timestamping_constants(crtc, &crtc->mode);
		hrtimer_init(&dm_crtc->vblank_hrtimer, CLOCK_MONOTONIC,
			     HRTIMER_MODE_REL);
		dm_crtc->vblank_hrtimer.function = vblank_simulate;
		dm_crtc->period_ns = ktime_set(0, vblank->framedur_ns);
		hrtimer_start(&dm_crtc->vblank_hrtimer, dm_crtc->period_ns,
			      HRTIMER_MODE_REL);
	} else if (dm_crtc->funcs->enable_vblank) {
		ret = dm_crtc->funcs->enable_vblank(dm_crtc, true);
	} else {
		ret = -ENXIO;
	}

	if (ret)
		dev_err(xdev->dev, "CRTC %d: failed to enable vblank (%d)\n",
			drm_crtc_index(crtc), ret);
	else
		dev_dbg(xdev->dev, "CRTC %d: enable vblank\n",
			 drm_crtc_index(crtc));

	return ret;
}

static void xdx_dm_crtc_disable_vblank(struct drm_crtc *crtc)
{
	struct drm_device *ddev = crtc->dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);
	int ret = 0;

	if (dm_crtc->vblank_timer_enabled)
		hrtimer_cancel(&dm_crtc->vblank_hrtimer);
	else if (dm_crtc->funcs->enable_vblank)
		ret = dm_crtc->funcs->enable_vblank(dm_crtc, false);
	else
		ret = -ENXIO;

	if (ret)
		dev_err(xdev->dev, "CRTC %d: failed to disable vblank (%d)\n",
			drm_crtc_index(crtc), ret);
	else
		dev_dbg(xdev->dev, "CRTC %d: disable vblank\n",
			 drm_crtc_index(crtc));
}
#endif

static const struct drm_crtc_funcs xdx_dm_crtc_funcs = {
	.set_config = drm_atomic_helper_set_config,
	.destroy = xdx_dm_crtc_destroy,
	.page_flip = drm_atomic_helper_page_flip,
	.reset = xdx_dm_crtc_reset,
	.atomic_duplicate_state = xdx_dm_crtc_atomic_duplicate_state,
	.atomic_destroy_state = xdx_dm_crtc_atomic_destroy_state,
	.atomic_set_property = xdx_dm_crtc_atomic_set_property,
	.atomic_get_property = xdx_dm_crtc_atomic_get_property,
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 12, 0))
	.gamma_set = drm_atomic_helper_legacy_gamma_set,
	#endif
	.late_register = xdx_dm_crtc_late_register,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	.enable_vblank = xdx_dm_crtc_enable_vblank,
	.disable_vblank = xdx_dm_crtc_disable_vblank,
#endif
};
/*
static u8 cal_pixel_bits(u32 bus_format)
{
	u8 bpp;

	switch (bus_format) {
	case MEDIA_BUS_FMT_RGB565_1X16:
	case MEDIA_BUS_FMT_UYVY8_1X16:
		bpp = 16;
		break;
	case MEDIA_BUS_FMT_RGB666_1X18:
	case MEDIA_BUS_FMT_RGB666_1X24_CPADHI:
		bpp = 18;
		break;
	case MEDIA_BUS_FMT_UYVY10_1X20:
		bpp = 20;
		break;
	case MEDIA_BUS_FMT_BGR888_1X24:
	case MEDIA_BUS_FMT_UYYVYY8_0_5X24:
	case MEDIA_BUS_FMT_YUV8_1X24:
		bpp = 24;
		break;
	case MEDIA_BUS_FMT_RGB101010_1X30:
	case MEDIA_BUS_FMT_UYYVYY10_0_5X30:
	case MEDIA_BUS_FMT_YUV10_1X30:
		bpp = 30;
		break;
	default:
		bpp = 24;
		break;
	}

	return bpp;
}
*/

static bool dm_crtc_mode_fixup(struct drm_crtc *crtc,
			       const struct drm_display_mode *mode,
			       struct drm_display_mode *adjusted_mode)
{
	struct xdx_device *xdev = drm_to_xdev(crtc->dev);
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	if (dm_crtc->funcs->mode_fixup)
		return dm_crtc->funcs->mode_fixup(dm_crtc, mode, adjusted_mode);
	else
		dev_err(xdev->dev, "mode_fixup no implemented\n");

	return false;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
static void dm_crtc_atomic_enable(struct drm_crtc *crtc,
				  struct drm_atomic_state *old_state)
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
static void dm_crtc_atomic_enable(struct drm_crtc *crtc,
				  struct drm_crtc_state *old_state)
#else
static void dm_crtc_atomic_enable(struct drm_crtc *crtc)
#endif
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);
	struct xdx_device *xdev = drm_to_xdev(crtc->dev);

	if (dm_crtc->funcs->enable)
		dm_crtc->funcs->enable(dm_crtc);
	else
		dev_err(xdev->dev, "crtc enable no implemented\n");

	drm_crtc_vblank_on(crtc);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
static void dm_crtc_atomic_disable(struct drm_crtc *crtc,
				   struct drm_atomic_state *old_state)
#else
static void dm_crtc_atomic_disable(struct drm_crtc *crtc,
				   struct drm_crtc_state *old_state)
#endif
{
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	drm_crtc_vblank_off(crtc);

	if (dm_crtc->funcs->disable)
		dm_crtc->funcs->disable(dm_crtc);

	if (crtc->state->event && !crtc->state->active) {
		spin_lock_irq(&crtc->dev->event_lock);
		drm_crtc_send_vblank_event(crtc, crtc->state->event);
		spin_unlock_irq(&crtc->dev->event_lock);

		crtc->state->event = NULL;
	}
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
static void dm_crtc_atomic_begin(struct drm_crtc *crtc,
				 struct drm_atomic_state *old_crtc_state)
#else
static void dm_crtc_atomic_begin(struct drm_crtc *crtc,
				 struct drm_crtc_state *old_crtc_state)
#endif
{
	struct drm_property_blob *blob = crtc->state->gamma_lut;
	struct drm_color_lut *lut;
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	if (crtc->state->color_mgmt_changed) {
		if ((blob) && (blob->length)) {
			lut = (struct drm_color_lut *)blob->data;
			dm_crtc->funcs->set_gamma(dm_crtc, lut,
						  blob->length / sizeof(*lut));
			dm_crtc->funcs->enable_gamma(dm_crtc, true);
		} else {
			dm_crtc->funcs->enable_gamma(dm_crtc, false);
		}
	}
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
static void dm_crtc_atomic_flush(struct drm_crtc *crtc,
				 struct drm_atomic_state *old_crtc_state)
#else
static void dm_crtc_atomic_flush(struct drm_crtc *crtc,
				 struct drm_crtc_state *old_crtc_state)
#endif
{
	struct drm_pending_vblank_event *event = crtc->state->event;
	struct dm_crtc *dm_crtc = to_dm_crtc(crtc);

	dm_crtc->funcs->commit(dm_crtc);

	if (event) {
		spin_lock_irq(&crtc->dev->event_lock);

		if(drm_crtc_vblank_get(crtc) == 0)
			drm_crtc_arm_vblank_event(crtc, event);
		else
			drm_crtc_send_vblank_event(crtc, event);

		spin_unlock_irq(&crtc->dev->event_lock);

		crtc->state->event = NULL;
	}
}

static const struct drm_crtc_helper_funcs xdx_dm_crtc_helper_funcs = {
	.mode_fixup = dm_crtc_mode_fixup,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	.atomic_enable = dm_crtc_atomic_enable,
#else
	.enable = dm_crtc_atomic_enable,
#endif
	.atomic_disable = dm_crtc_atomic_disable,
	.atomic_begin = dm_crtc_atomic_begin,
	.atomic_flush = dm_crtc_atomic_flush,
};

int dm_universal_crtc_init(struct xdx_dm *dm, struct dm_crtc *dm_crtc,
			   struct dm_dc *dc, struct dm_crtc_funcs *funcs,
			   struct dm_plane *primary, struct dm_plane *cursor,
			   uint32_t id)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	if (!primary || !funcs)
		return -EINVAL;

	if (id >= XDX_MAX_CRTCS) {
		dev_err(xdev->dev, "wrong crtc id %d\n", id);
		return -EINVAL;
	}

	dm_crtc->id = id;
	dm_crtc->funcs = funcs;
	dm_crtc->dm = dm;
	dm_crtc->dc = dc;
	
	if (xdev->flags.virtual_vblank) {
		dm_crtc->vblank_timer_enabled = true;
		dev_dbg(xdev->dev, "dm: enable vblank timer for CRTC %d\n", id);
	}

	ret = drm_crtc_init_with_planes(ddev, &dm_crtc->base, &primary->base,
					&cursor->base, &xdx_dm_crtc_funcs,
					NULL);
	if (ret) {
		dev_err(xdev->dev, "failed to init drm crtc (%d)\n", ret);
		return ret;
	}

	drm_crtc_helper_add(&dm_crtc->base, &xdx_dm_crtc_helper_funcs);

	dc->crtcs[id] = dm_crtc;

	return 0;
}
