/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#ifndef __XDX_DMA_DRV_H__
#define __XDX_DMA_DRV_H__

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/dmaengine.h>
#include <linux/dmapool.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/of_dma.h>
#include <linux/of_device.h>
#include <linux/platform_device.h>
#include <linux/reset.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/dma-mapping.h>
#include <linux/version.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,10,0))
#include <linux/dma-map-ops.h>
#endif
#include <linux/time.h>
#include <linux/timer.h>
#include <img_types.h>
#include <linux/mempool.h>
#include "virt-dma.h"
#include <asm/cacheflush.h>
#include <linux/pci.h>
#include "./hal/vdma_hal.h"
#include "./hal/vdma_drv.h"
#include <linux/random.h>
#include <linux/smp.h>
#include <linux/delay.h>
#include "B0_HalVdma.h"
#include "xdx_irq.h"
#include "vdma_irq.h"
#include "pangu.h"
#include "HalTDMA.h"
#include "xdx_dma_dbglevel.h"

#define CMD_MAX 			 (64)
#define XDX_VDMA_ADDR_ALIGN  (64)

#define XDX_VDMA_A0_NAME "xdxVdmaA0"
#define XDX_VDMA_B0_NAME "xdxVdmaB0"
#define XDX_TDMA_B0_NAME "xdxTdmaB0"

#define DMA_CHAN_ID(_dma, _chan_num) ((((_dma)&0xf) << 8) | ((_chan_num) & 0xf))
#define DMA_DMA_NUM(_chan_id) (((_chan_id) >> 8) & 0xf)
#define DMA_CHAN_NUM(_chan_id) ((_chan_id) & 0xf)

#define XDXDMA_FLAG_VDMA (1 << 0)
#define XDXDMA_FLAG_TDMA (1 << 1)
#define XDXDMA_FLAG_A0 (1 << 2)
#define XDXDMA_FLAG_B0 (1 << 3)

#define XDXDMA_A0_VDMA_MASK (XDXDMA_FLAG_VDMA | XDXDMA_FLAG_A0)
#define XDXDMA_B0_VDMA_MASK (XDXDMA_FLAG_VDMA | XDXDMA_FLAG_B0)
#define XDXDMA_B0_TDMA_MASK (XDXDMA_FLAG_TDMA | XDXDMA_FLAG_B0)

#define DMA_TRANS_POLLINGTIMEOUT 4
#define DMA_TRANS_BUSSERR 5

struct xdx_dma_dev;

struct xdx_dma_pchan_cfg {
	int chan_id;
	int prio;
	u32 reg_offset;
};

struct xdx_dma_vchan_cfg {
	u32 flags;
	int dev_id;
	int prio;
	int num_pchan;
	const u32 *pchans;
};
struct xdx_dma_drv_cfg {
	void (*set_burst_length)(u32 *p_cfg, s8 src_burst, s8 dst_burst);
	u32 src_addr_widths;
	u32 dst_addr_widths;
	bool has_mbus_clk;
	u32 flags;
	int num_vchan;
	int desc_entry_size;
	struct xdx_dma_vchan_cfg *vchans;
	int num_pchan;
	struct xdx_dma_pchan_cfg *pchans;
	struct dma_slave_map *fn_map;
};

/*
 * Hardware channels / ports representation
 *
 * The hardware is used in several SoCs, with differing numbers
 * of channels and endpoints. This structure ties those numbers
 * to a certain compatible string.
 */
struct xdx_dma_config {
	u32 nr_max_channels;
	u32 nr_max_requests;
	u32 nr_max_vchans;
	/*
	 * In the datasheets/user manuals of newer Allwinner SoCs, a special
	 * bit (bit 2 at register 0x20) is present.
	 * It's named "DMA MCLK interface circuit auto gating bit" in the
	 * documents, and the footnote of this register says that this bit
	 * should be set up when initializing the DMA controller.
	 * Allwinner A23/A33 user manuals do not have this bit documented,
	 * however these SoCs really have and need this bit, as seen in the
	 * BSP kernel source code.
	 */
	void (*set_burst_length)(u32 *p_cfg, s8 src_burst, s8 dst_burst);
	u32 src_addr_widths;
	u32 dst_addr_widths;
	bool has_mbus_clk;
};


#define XDX_DMA_UNIT_IGNOR (-1)
typedef struct xdx_dma_desc_blk_s
{
	/*
	 * This field is not used by the DMA controller, but will be
	 * used by the CPU to go through the list (mostly for dumping
	 * or freeing it).
	 */
	struct list_head	list;
	struct list_head	next_list;
	u32					desc_total; //for debug & align to 8
	u32					desc_num;
	uint64_t			data_size;
	dma_addr_t			desc_addr;
	uint64_t			align_addr;
	struct {
		uint64_t dword;
	} __attribute__((aligned(16)))desc_entry[0];
}xdx_dma_desc_blk_t;

typedef struct xdx_dma_cmd_entry_s
{
	/*
	 * This field is not used by the DMA controller, but will be
	 * used by the CPU to go through the list (mostly for dumping
	 * or freeing it).
	 */
	struct list_head	list;
	u32					desc_num;
	dma_addr_t			desc_addr;
	struct {
		uint64_t dword[2];
	} desc_entry[0];
}xdx_dma_cmd_entry_s;

typedef struct xdx_dma_addr_seg_s {
	uint64_t offset;
	uint64_t len_unit;
	uint64_t unit_cnt;
	dma_addr_t dma_addr;
	struct xdx_dma_addr_seg_s *next;
}xdx_dma_addr_seg_t;

typedef enum _xdx_dma_xmit_type
{
	XDX_XMIT_UNKNOWN = 0,
	XDX_XMIT_COPY,
	XDX_XMIT_MEMSET,
	XDX_XMIT_POLL,
	XDX_XMIT_COPY_2D,
	XDX_XMIT_TWD_2D3D,
}xdx_dma_xmit_type_e;

typedef enum xdx_dma_addrtype_{
	XDX_DMA_ADDR_NULL,
	XDX_DMA_ADDR_DMA,
	XDX_DMA_ADDR_SGL,
}xdx_dma_addrtype_e;

typedef struct xdxdma_xmit_sg_s{
	struct scatterlist	*sgl;
	unsigned int	sg_len;
}xdxdma_xmit_sg_t;

typedef struct xdxdma_xmit_addr_s {
	xdx_dma_addrtype_e addr_type;
	union {
		xdxdma_xmit_sg_t sg_tab;
		dma_addr_t addr;
	};
	uint32_t addr_flag;
}xdxdma_xmit_addr_t;

typedef struct xdxdma_twd_2d3d_para_s {
	uint64_t width;
	uint64_t height;
	uint64_t x_stride;
	uint64_t y_stride;
	uint64_t depth;
	uint16_t len_unit;
}xdxdma_twd_2d3d_para_t;

typedef struct xdxdma_2d_para_s {
	uint64_t width;
	uint64_t height;
	uint64_t x_stride;
	uint16_t len_unit;
}xdxdma_2d_para_t;

#define XDX_DMA_PATTERN_MAX_BYTES (8*4)

typedef struct xdxdma_memset_para_s {
	uint64_t len;
	uint64_t Pattern_len;
	uint32_t Pattern[XDX_DMA_PATTERN_MAX_BYTES/4];
}xdxdma_memset_para_t;

typedef struct xdxdma_xmit_descpara_s{
	xdxdma_xmit_addr_t src;
	xdxdma_xmit_addr_t dest;
	union {
		xdxdma_twd_2d3d_para_t twiddle_2d3d;
		xdxdma_2d_para_t para_2d;
		xdxdma_memset_para_t para_memset;
		uint64_t len;
	};
}xdxdma_xmit_descpara_t;

typedef struct xdx_desc_cmd_cfg_s{
	struct list_head	list;
	union{
	VDMA_B0_CmdCfg_T	vb0_cfg;
	TDMA_CmdCfg_T		tb0_cfg;
	VDMA_CMD_Config_t	va0_cfg;
	};
}xdx_desc_cmd_cfg_t;

struct xdx_desc
{
	struct xdx_virt_dma_desc	vd;
	struct xdx_dma_dev			*sdev;
	atomic_t					cmd_cnt;
	enum dma_transfer_direction	direct;
	xdxdma_xmit_descpara_t		paras;
	xdx_dma_xmit_type_e			xmit_type;
	struct xdx_pchan 			*pchan_used;
	struct list_head			desc_src_head;
	struct list_head			desc_dest_head;
	struct list_head			cmd_cfgs_head;
};

struct xdx_cmd_reference
{
	struct xdx_desc		*desc;
	struct xdx_vchan	*vchan;
};

struct xdx_pchan
{
	struct device 		*dev;
	uint64_t			DmaCmdPoolMask;
	struct xdx_cmd_reference cmdRef[CMD_MAX];
	spinlock_t			lock;
	u32					idx;
	u16					free_cmd;
	u16					submit_cmd;
	void __iomem		*base;
	u32					vchan_refs;
	bool				inited;
};

struct xdx_vchan
{
	struct device 		*dev;
	struct xdx_virt_dma_chan	vc;
	u32					tag;
	u8					port;
	u8					irq_type;
	bool				cyclic;
	u32					pchan_cnt;
	struct xdx_pchan	**pchans;
};
struct xdx_dma_dev
{
	struct dma_device		slave;
	struct device			*dev;
	/*pdata as parent data: xdx_device*/
	void					*pdata;
	void __iomem			*base;
	void __iomem			*hostmisc_base;
	//int						irq;
	spinlock_t				lock;
	u32						mcu_id;
	int						irq_cid;
	struct xdx_irq_src		irq_src;
	struct xdx_dma_ops		*ops;
	//struct reset_control	*rstc;
	struct tasklet_struct	task;
	atomic_t				tasklet_shutdown;
	//struct list_head		pending;
	struct dma_pool			*pool;
	//struct dma_pool			*xdxpool;
	struct xdx_pchan		*pchans;
	struct xdx_vchan		*vchans;
	struct kmem_cache 		*txpool;	//txd cache pool
	struct kmem_cache 		*cmd_pool;	//cmd cfg cache pool
	u32						pendingMask;//pending vc mask
	atomic_t				cur_pID; 	//current p_channel ID
//	const struct xdx_dma_config *cfg;
	const struct xdx_dma_drv_cfg *dcfg;

//	VDMA_CMD_Config_t		*cmd_cfg;
	//volatile u32			*hostmisc_msi_addr;
	u32						num_pchans;
	u32						num_vchans;
	u32						max_request;
};

#define PVR_ALIGN(_x, _n)	(((_x)+((_n)-1U)) & ~((_n)-1U))
#define PVR_FLOOR_ALIGN(_x, _n)	((_x) & ~((_n)-1U))

struct vdma_funcs {
	void (*destroy)(struct xdx_dma_dev *dma);
	void (*reg_write)(struct xdx_dma_dev *dma, u32 reg, u32 value);
	u32 (*reg_read)(struct xdx_dma_dev *dma, u32 reg);
	void (*enable_interrupt)(struct xdx_dma_dev *dma, bool enable);
	int (*hw_init)(struct xdx_dma_dev *dma);
	void (*interrupt_handle)(struct xdx_dma_dev *dma);
};

struct xdx_dma_ops {
	int (*hw_init)(struct xdx_dma_dev*);
	int (*irq_set_en)(struct xdx_dma_dev*, u32 stat);
	int (*irq_handler)(struct xdx_dma_dev*);
	int (*parse_sgls)(struct xdx_dma_dev*, struct xdx_desc*);
	int (*package_cfg)(struct xdx_dma_dev*, struct xdx_desc*);
	int (*issue_chan)(struct xdx_vchan*, struct xdx_desc*, xdx_desc_cmd_cfg_t*);
	int (*print_reg)(struct xdx_dma_dev *);
	bool (*check_paras)(struct xdx_dma_dev *, xdx_dma_xmit_type_e, void *);
	int (*cmd_buffer_remain)(struct xdx_pchan *);
};

static inline int  xdx_dma_parse_sgls(struct xdx_dma_dev *sdev, struct xdx_desc *txd)
{
	if (sdev && sdev->ops != NULL && sdev->ops->parse_sgls)
		return sdev->ops->parse_sgls(sdev, txd);
	return -ENOTSUPP;
}

static inline int xdx_dma_package_cfg(struct xdx_dma_dev *sdev, struct xdx_desc *txd)
{
	if (sdev && sdev->ops != NULL && sdev->ops->package_cfg)
		return sdev->ops->package_cfg(sdev, txd);
	return -ENOTSUPP;
}

static inline bool  xdx_dma_check_para(
		struct xdx_dma_dev *sdev, xdx_dma_xmit_type_e xmit_type, void *para)
{
	if (sdev && sdev->ops != NULL && sdev->ops->check_paras)
		return sdev->ops->check_paras(sdev, xmit_type, para);
	return true;
}

static inline int xdx_dma_issue_chan(
		struct xdx_dma_dev *sdev, struct xdx_vchan *vchan,
		struct xdx_desc* txd, xdx_desc_cmd_cfg_t *cfg)
{
	if (sdev && sdev->ops != NULL && sdev->ops->issue_chan)
		return sdev->ops->issue_chan(vchan, txd, cfg);
	return -ENOTSUPP;
}

static inline int xdx_dma_cmd_buffer_remain(
		struct xdx_dma_dev *sdev, struct xdx_pchan *pchan)
{
	if (sdev && sdev->ops != NULL && sdev->ops->cmd_buffer_remain)
		return sdev->ops->cmd_buffer_remain(pchan);
	return -ENOTSUPP;
}

static inline struct device *xdxdma_chan2dev(struct dma_chan *chan)
{
	return &chan->dev->device;
}

xdx_dma_desc_blk_t* alloc_desc_block(struct xdx_dma_dev *sdev, dma_addr_t *p_desc_blk);
void free_desc_block(struct xdx_dma_dev *sdev, xdx_dma_desc_blk_t *desc_blk);
int xdx_dma_flush_desc_block( struct xdx_dma_dev *sdev,
		struct xdx_desc *txd, xdx_dma_desc_blk_t *desc_blk);
struct xdx_desc *to_xdx_desc(struct dma_async_tx_descriptor *tx);
struct xdx_desc* xdx_dma_alloc_txd(struct xdx_dma_dev *sdev);
void xdx_dma_free_txd(struct xdx_dma_dev *sdev, struct xdx_desc *txd);
xdx_desc_cmd_cfg_t *xdx_dma_alloc_cmd_cfg(struct xdx_dma_dev *sdev);
void xdx_dma_free_cmd_cfg(struct xdx_dma_dev *sdev, xdx_desc_cmd_cfg_t *cdm_cfg);
void xdx_dma_free_desc(struct xdx_virt_dma_desc *vd);
int xdx_dma_request_command_tag( struct xdx_vchan *vchan,
			struct xdx_dma_dev *sdev, struct xdx_pchan **ppchan);
void xdx_dma_release_command_tag(struct xdx_pchan *pchan, int cmd_tag);

static inline void *xdx_dma_get_desc_entry(struct xdx_dma_dev *sdev, xdx_dma_desc_blk_t *desc_blk, int desc_num)
{
	void *desc_entry = NULL;
	if (desc_num > desc_blk->desc_total)
		return NULL;
	desc_entry = (void*)(&desc_blk->desc_entry[0]);
	desc_entry += sdev->dcfg->desc_entry_size * desc_num;
	return desc_entry;
}

static inline struct xdx_dma_dev *to_xdx_dma_dev(struct dma_device *d)
{
	return container_of(d, struct xdx_dma_dev, slave);
}

static inline struct xdx_vchan *to_xdx_vchan(struct dma_chan *chan)
{
	return container_of(chan, struct xdx_vchan, vc.chan);
}

static inline uint32_t GetLeadOnefromU64(uint64_t u64Mask)
{
	return (63U - __builtin_clzll(u64Mask));
}

static inline uint32_t GetLeadOnefromU32(uint32_t u32Mask)
{
	return (31U - __builtin_clz(u32Mask));
}

int xdx_dma_chans_irq_set(struct xdx_dma_dev *sdev, unsigned int enable);
int xdx_dma_package_desc(struct dma_async_tx_descriptor *tx);

extern struct xdx_dma_ops vdma_a0_ops;
extern struct xdx_dma_ops vdma_b0_ops;
extern struct xdx_dma_ops tdma_b0_ops;

extern struct xdx_dma_drv_cfg xdxvdma_a0;
extern struct xdx_dma_drv_cfg xdxvdma_b0;
extern struct xdx_dma_drv_cfg xdxtdma_b0;

#endif

