/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#ifndef __XDX_SHARED_H__
#define __XDX_SHARED_H__

#include <linux/types.h>

struct xdx_device;

enum xdx_ip_block_type {
	XDX_IP_BLOCK_TYPE_COMMON,
	XDX_IP_BLOCK_TYPE_GFX,
	XDX_IP_BLOCK_TYPE_HOST,
	XDX_IP_BLOCK_TYPE_VPU,
	XDX_IP_BLOCK_TYPE_DC,
	XDX_IP_BLOCK_TYPE_ACPU,
	XDX_IP_BLOCK_TYPE_CMCU,
	XDX_IP_BLOCK_TYPE_VDMA,
	XDX_IP_BLOCK_TYPE_XDMA,
	XDX_IP_BLOCK_TYPE_CF,
	XDX_IP_BLOCK_TYPE_NDF,
	XDX_IP_BLOCK_TYPE_GDDR,
	XDX_IP_BLOCK_TYPE_TDMA,
	XDX_IP_BLOCK_TYPE_SMMU,
	XDX_IP_BLOCK_TYPE_THMSENSOR,
	XDX_IP_BLOCK_TYPE_NUM,
};

enum xdx_clockgating_state {
        XDX_CG_STATE_GATE = 0,
        XDX_CG_STATE_UNGATE,
};


enum xdx_powergating_state {
        XDX_PG_STATE_GATE = 0,
        XDX_PG_STATE_UNGATE,
};

/**
 * struct xdx_ip_funcs - general hooks for managing amdgpu IP Blocks
 * @name: Name of IP block
 * @early_init: sets up early driver state (pre sw_init),
 *              does not configure hw - Optional
 * @late_init: sets up late driver/hw state (post hw_init) - Optional
 * @sw_init: sets up driver state, does not configure hw
 * @sw_fini: tears down driver state, does not configure hw
 * @early_fini: tears down stuff before dev detached from driver
 * @hw_init: sets up the hw state
 * @hw_fini: tears down the hw state
 * @late_fini: final cleanup
 * @suspend: handles IP specific hw/sw changes for suspend
 * @resume: handles IP specific hw/sw changes for resume
 * @is_idle: returns current IP block idle status
 * @wait_for_idle: poll for idle
 *
 * These hooks provide an interface for controlling the operational state
 * of IP blocks. After acquiring a list of IP blocks for the GPU in use,
 * the driver can make chip-wide state changes by walking this list and
 * making calls to hooks from each IP block. This list is ordered to ensure
 * that the driver initializes the IP blocks in a safe sequence.
 */
struct xdx_ip_funcs {
	int (*early_init)(void *handle);
	int (*late_init)(void *handle);
	int (*sw_init)(void *handle);
	int (*sw_fini)(void *handle);
	int (*early_fini)(void *handle);
	int (*hw_init)(void *handle);
	int (*hw_fini)(void *handle);
	void (*late_fini)(void *handle);
	int (*suspend)(void *handle);
	int (*resume)(void *handle);
	bool (*is_idle)(void *handle);
	int (*wait_for_idle)(void *handle);
};

struct xdx_ip_block_version {
	char *name;
	enum xdx_ip_block_type type;
	u32 major;
	u32 minor;
	u32 rev;
	struct xdx_ip_funcs funcs;
};

struct xdx_ip_block {
	union {
		struct {
			uint32_t early_init:1;
			uint32_t sw_init:1;
			uint32_t hw_init:1;
			uint32_t late_init:1;
			uint32_t reserved:28;
		};
		uint32_t v;
	} status;
	struct xdx_ip_block_version version;
};

enum xdx_power_state {
	/* normal running state, alternative to D0 state */
	XDX_PS0,
	/*
	 * low power state, alternative to D1 state 
	 * clock gating SOC main blocks, no power gating
	 */
	XDX_PS1,
	/*
	 * sleep state, alternative to D3hot state
	 * power gating each block except AON block
	 */
	XDX_PS2,
	/*
	 * poweroff state, alternative to D3clod state
	 * poweroff entire SOC
	 */
	XDX_PS3,
};

#endif /* __XDX_SHARED_H__ */