/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/pagemap.h>
#include <linux/slab.h>
#include <linux/dma-mapping.h>
#include <linux/vmalloc.h>

#include <asm/io.h>
#include <asm/delay.h>

#include "xdx_device.h"
#include "xdx_gem.h"
#if defined(SUPPORT_XDXGPU_RPM)
#include "pm/xdx_pm.h"
#endif

#include "dc8200_mmu.h"
#if LINUX_VERSION_CODE>=KERNEL_VERSION(5, 17, 0)
#include <linux/list.h>
#include <drm/drm.h>
#include <drm/drm_device.h>
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 8, 0))
#include <drm/drm_legacy.h>
#endif
#include <linux/hash.h>
#include <linux/mm.h>
#include <linux/rculist.h>
#include <drm/drm_print.h>

struct agp_memory;
struct drm_buf_desc;
struct drm_device;
struct drm_file;
struct drm_hash_item;
struct drm_open_hash;

static int drm_ht_create(struct drm_open_hash *ht, unsigned int order)
{
	unsigned int size = 1 << order;

	ht->order = order;
	ht->table = NULL;
	if (size <= PAGE_SIZE / sizeof(*ht->table))
		ht->table = kcalloc(size, sizeof(*ht->table), GFP_KERNEL);
	else
		ht->table = vzalloc(array_size(size, sizeof(*ht->table)));
	if (!ht->table) {
		DRM_ERROR("Out of memory for hash table\n");
		return -ENOMEM;
	}
	return 0;
}
#if 0
static void drm_ht_verbose_list(struct drm_open_hash *ht, unsigned long key)
{
	struct drm_hash_item *entry;
	struct hlist_head *h_list;
	unsigned int hashed_key;
	int count = 0;

	hashed_key = hash_long(key, ht->order);
	DRM_DEBUG("Key is 0x%08lx, Hashed key is 0x%08x\n", key, hashed_key);
	h_list = &ht->table[hashed_key];
	hlist_for_each_entry(entry, h_list, head)
		DRM_DEBUG("count %d, key: 0x%08lx\n", count++, entry->key);
}

static struct hlist_node *drm_ht_find_key(struct drm_open_hash *ht,
					  unsigned long key)
{
	struct drm_hash_item *entry;
	struct hlist_head *h_list;
	unsigned int hashed_key;

	hashed_key = hash_long(key, ht->order);
	h_list = &ht->table[hashed_key];
	hlist_for_each_entry(entry, h_list, head) {
		if (entry->key == key)
			return &entry->head;
		if (entry->key > key)
			break;
	}
	return NULL;
}
#endif

static struct hlist_node *drm_ht_find_key_rcu(struct drm_open_hash *ht,
					      unsigned long key)
{
	struct drm_hash_item *entry;
	struct hlist_head *h_list;
	unsigned int hashed_key;

	hashed_key = hash_long(key, ht->order);
	h_list = &ht->table[hashed_key];
	hlist_for_each_entry_rcu(entry, h_list, head) {
		if (entry->key == key)
			return &entry->head;
		if (entry->key > key)
			break;
	}
	return NULL;
}

static int drm_ht_insert_item(struct drm_open_hash *ht, struct drm_hash_item *item)
{
	struct drm_hash_item *entry;
	struct hlist_head *h_list;
	struct hlist_node *parent;
	unsigned int hashed_key;
	unsigned long key = item->key;

	hashed_key = hash_long(key, ht->order);
	h_list = &ht->table[hashed_key];
	parent = NULL;
	hlist_for_each_entry(entry, h_list, head) {
		if (entry->key == key)
			return -EINVAL;
		if (entry->key > key)
			break;
		parent = &entry->head;
	}
	if (parent) {
		hlist_add_behind_rcu(&item->head, parent);
	} else {
		hlist_add_head_rcu(&item->head, h_list);
	}
	return 0;
}

#if 0
/*
 * Just insert an item and return any "bits" bit key that hasn't been
 * used before.
 */
static int drm_ht_just_insert_please(struct drm_open_hash *ht, struct drm_hash_item *item,
			      unsigned long seed, int bits, int shift,
			      unsigned long add)
{
	int ret;
	unsigned long mask = (1UL << bits) - 1;
	unsigned long first, unshifted_key;

	unshifted_key = hash_long(seed, bits);
	first = unshifted_key;
	do {
		item->key = (unshifted_key << shift) + add;
		ret = drm_ht_insert_item(ht, item);
		if (ret)
			unshifted_key = (unshifted_key + 1) & mask;
	} while(ret && (unshifted_key != first));

	if (ret) {
		DRM_ERROR("Available key bit space exhausted\n");
		return -EINVAL;
	}
	return 0;
}
#endif

static int drm_ht_find_item(struct drm_open_hash *ht, unsigned long key,
		     struct drm_hash_item **item)
{
	struct hlist_node *list;

	list = drm_ht_find_key_rcu(ht, key);
	if (!list)
		return -EINVAL;

	*item = hlist_entry(list, struct drm_hash_item, head);
	return 0;
}

#if 0
static int drm_ht_remove_key(struct drm_open_hash *ht, unsigned long key)
{
	struct hlist_node *list;

	list = drm_ht_find_key(ht, key);
	if (list) {
		hlist_del_init_rcu(list);
		return 0;
	}
	return -EINVAL;
}
#endif

static int drm_ht_remove_item(struct drm_open_hash *ht, struct drm_hash_item *item)
{
	hlist_del_init_rcu(&item->head);
	return 0;
}

static void drm_ht_remove(struct drm_open_hash *ht)
{
	if (ht->table) {
		kvfree(ht->table);
		ht->table = NULL;
	}
}

#endif

static inline u32 bo_gpu_page_alignment(struct xdx_bo *bo)
{
	return (bo->page_alignment << PAGE_SHIFT) / MMU_PAGE_4K_SIZE;
}

static inline u32 bo_ngpu_pages(struct xdx_bo *bo)
{
	return bo->num_pages * bo_gpu_page_alignment(bo);
}

static inline u32 _mtlb_offset(u32 address)
{
#ifdef CONFIG_MMU_1K_16M
	return (address & MMU_MTLB_MASK) >> MMU_MTLB_SHIFT;
#else
	return (address & MMU_MTLB_MASK) >> MMU_MTLB_SHIFT;
#endif
}

static inline u32 _stlb_offset(u32 address)
{
#ifdef CONFIG_MMU_1K_16M
	return (address & 0xff000000) >> MMU_STLB_16M_SHIFT;
#else
	return (address & MMU_STLB_4K_MASK) >> MMU_STLB_4K_SHIFT;
#endif
}

static inline u32 _address_to_index(struct dc8200_mmu *mmu, u32 address)
{
#ifdef CONFIG_MMU_1K_16M
	return _mtlb_offset(address);
#else
	return _mtlb_offset(address) * MMU_STLB_4K_ENTRY_NUM +
	       _stlb_offset(address);
#endif
}

static inline u32 _set_page(u32 page_address, u32 page_address_ext, bool writable)
{
	u32 entry = page_address
		    /* AddressExt */
		    | (page_address_ext << 4)
		    /* Ignore exception */
		    | (0 << 1)
		    /* Present */
		    | (1 << 0);

	if (writable) {
		/* writable */
		entry |= (1 << 2);
	}

	return entry;
}

static inline void _write_page_entry(u32 *page_entry, u32 entry_value)
{
	*page_entry = entry_value;
}

static inline u32 _read_page_entry(u32 *page_entry)
{
	return *page_entry;
}

static int _setup_process_address_space(struct xdx_device *xdev, struct dc8200_mmu *mmu)
{
	u32 *map = NULL;
	u32 free, i;
	u32 dynamic_mapping_entries, address;
	int ret = 0;
	u64 all_stlb_phyaddr;
	u32 *all_stlb_vaddr;
	u64 all_stlb_size;
#if defined(SUPPORT_XDXGPU_RPM)
	u64 offset;
	const u32 gddr_interleave_gap_scale = xdev->fw_st_info.gddr_count;
#else
	const u32 gddr_interleave_gap_scale = 1;
#endif

	dynamic_mapping_entries = MMU_MTLB_ENTRY_NUM;
	mmu->dynamic_mapping_start = 0;
#ifdef CONFIG_MMU_1K_16M
	mmu->page_table_size = dynamic_mapping_entries * MMU_STLB_16M_SIZE / 16;
#else
	mmu->page_table_size = dynamic_mapping_entries * MMU_STLB_4K_SIZE;
#endif

	mmu->page_table_entries = mmu->page_table_size / sizeof(u32);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	mmu->map_logical = kvzalloc(mmu->page_table_size, GFP_KERNEL);
#else
	mmu->map_logical = kmalloc(mmu->page_table_size, GFP_KERNEL);
	memset(mmu->map_logical, 0, mmu->page_table_size);
#endif
	if (!mmu->map_logical) {
		pr_err("Failed to alloc mmu map buffer.\n");
		return -ENOMEM;
	}

	map = mmu->map_logical;

	/* Initialize free area*/
	free = mmu->page_table_entries;

	_write_page_entry(map, (free << 8) | DC_MMU_FREE);
	_write_page_entry(map + 1, ~0U);

	mmu->heap_list = 0;
	mmu->free_nodes = false;

#ifdef CONFIG_MMU_1K_16M
	all_stlb_size = MMU_STLB_16M_SIZE * MMU_MTLB_ENTRY_NUM;
#else
	all_stlb_size = MMU_STLB_4K_SIZE * MMU_MTLB_ENTRY_NUM;
#endif
#if defined(SUPPORT_XDXGPU_RPM)
	all_stlb_size *= xdev->fw_st_info.gddr_count;
#endif

	ret = xdx_bo_create_reserved(xdev, all_stlb_size, PAGE_SIZE,
				     XDXGPU_PL_VRAM, &mmu->stlbs_bo,
				     &all_stlb_phyaddr,
				     (void **)&all_stlb_vaddr);
	if (ret) {
		dev_err(xdev->dev, "Failed to alloc all stlbs\n");
		goto failed_all_stlbs_alloc;
	}
	OSDeviceMemSetIO(all_stlb_vaddr, 0, all_stlb_size);

#if defined(SUPPORT_XDXGPU_RPM)
	offset = rpm_valid_gddr_offset(xdev, all_stlb_phyaddr);
	all_stlb_phyaddr += offset;
	all_stlb_vaddr += offset / sizeof(u32);
#endif

	for (i = 0; i < dynamic_mapping_entries; i++) {
		struct dc8200_mmu_stlb *stlb;
		struct dc8200_mmu_stlb **stlbs = (struct dc8200_mmu_stlb**)mmu->stlbs;

		stlb = kzalloc(sizeof(struct dc8200_mmu_stlb), GFP_KERNEL);
		if (!stlb) {
			ret = -ENOMEM;
			goto failed_stlb_alloc;
		}
#ifdef CONFIG_MMU_1K_16M
		stlb->size = MMU_STLB_16M_SIZE;
		stlb->physBase = all_stlb_phyaddr +
				 i * MMU_STLB_16M_SIZE *
				 gddr_interleave_gap_scale;
		stlb->logical =
			all_stlb_vaddr + i * MMU_STLB_16M_SIZE / sizeof(u32) *
			gddr_interleave_gap_scale;
#else
		stlb->size = MMU_STLB_4K_SIZE;
		stlb->physBase = all_stlb_phyaddr +
				 i * MMU_STLB_4K_SIZE *
				 gddr_interleave_gap_scale;
		stlb->logical =
			all_stlb_vaddr + i * MMU_STLB_4K_SIZE / sizeof(u32) *
			gddr_interleave_gap_scale;
#endif
		stlbs[i] = stlb;
	}

	address = (u32)all_stlb_phyaddr;

	mutex_lock(&mmu->page_table_mutex);

	for (i = mmu->dynamic_mapping_start;
	     i < mmu->dynamic_mapping_start + dynamic_mapping_entries; i++) {
		u32 mtlb_entry;

		mtlb_entry = (address & 0xffffffc0)
#ifdef CONFIG_MMU_1K_16M
			     | MMU_MTLB_16M_PAGE
#else
			     | MMU_MTLB_4K_PAGE
#endif
			     | MMU_MTLB_PRESENT;

#ifdef CONFIG_MMU_1K_16M
		address = all_stlb_phyaddr +
			  MMU_STLB_16M_SIZE *
		(i + 1 - mmu->dynamic_mapping_start) *
			  gddr_interleave_gap_scale;
#else
		address += MMU_STLB_4K_SIZE * gddr_interleave_gap_scale;
#endif
		/* Insert Slave TLB address to Master TLB entry.*/
		_write_page_entry(mmu->mtlb_logical + i, mtlb_entry);
	}

	mutex_unlock(&mmu->page_table_mutex);

	return 0;

failed_stlb_alloc:
	xdx_gem_object_put(&mmu->stlbs_bo->base);
	for (i = 0; i < dynamic_mapping_entries; i++) {
		struct dc8200_mmu_stlb **stlbs = (struct dc8200_mmu_stlb **)mmu->stlbs;

		if (stlbs[i]) {
			kfree(stlbs[i]);
			stlbs[i] = NULL;
		}
	}

failed_all_stlbs_alloc:
	kvfree(mmu->map_logical);
	return ret;
}

static void* dc8200_mmu_create(struct xdx_device *xdev)
{
	struct dc8200_mmu *mmu;
	u32 size = 0;
	int ret = 0;
#if defined(SUPPORT_XDXGPU_RPM)
	u64 offset;
#endif

	mmu = kzalloc(sizeof(*mmu), GFP_KERNEL);
	if (!mmu)
		return ERR_PTR(-ENOMEM);

	mutex_init(&mmu->page_table_mutex);

	mutex_init(&mmu->ht_mutex);
#if defined(CONFIG_SW64)
	ret = drm_ht_create(&mmu->res_ht, 14);
#else
	ret = drm_ht_create(&mmu->res_ht, 12);
#endif
	if (ret)
		goto failed_ht_create;

	mmu->mtlb_entry_num = MMU_MTLB_ENTRY_NUM;
	/* size of each MTLB entry is dwaord */
	size = PAGE_ALIGN(mmu->mtlb_entry_num * sizeof(uint32_t));
#if defined(SUPPORT_XDXGPU_RPM)
	size *= xdev->fw_st_info.gddr_count;
#endif

	/* Allocate MTLB */
	ret = xdx_bo_create_reserved(xdev, size, PAGE_SIZE, XDXGPU_PL_VRAM,
				     &mmu->mtlb_bo, &mmu->mtlb_physical,
				     (void **)&mmu->mtlb_logical);
	if (ret) {
		dev_err(xdev->dev, "Failed to alloc mtlb buffer.\n");
		goto failed_mtlb_alloc;
	}
	OSDeviceMemSetIO(mmu->mtlb_logical, 0, size);

#if defined(SUPPORT_XDXGPU_RPM)
	offset = rpm_valid_gddr_offset(xdev, mmu->mtlb_physical);
	mmu->mtlb_logical = (u32 *)((u64)mmu->mtlb_logical + offset);
	mmu->mtlb_physical += offset;
#endif

	/* each MTLB entry point to one STLB arrays */
	size = mmu->mtlb_entry_num * sizeof(struct dc8200_mmu_stlb*);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	mmu->stlbs = kvzalloc(size, GFP_KERNEL);
#else
	mmu->stlbs = kmalloc(size, GFP_KERNEL);
#endif
	if (!mmu->stlbs) {
		ret = -ENOMEM;
		goto failed_stlbs_alloc;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
	memset(mmu->stlbs, 0, size);
#endif

#if defined(SUPPORT_XDXGPU_RPM)
	mmu->mode = MMU_MODE_4K;
#else
	mmu->mode = MMU_MODE_1K;
#endif

	ret = _setup_process_address_space(xdev, mmu);
	if (ret) {
		dev_err(xdev->dev, "Failed to setup addr space\n");
		goto failed_setup;
	}

	/* Allocate safe page */
	ret = xdx_bo_create_reserved(xdev, 4096, PAGE_SIZE, XDXGPU_PL_VRAM,
				     &mmu->safe_page_bo,
				     &mmu->safe_page_physical,
				     &mmu->safe_page_logical);
	if (ret) {
		dev_err(xdev->dev, "Failed to alloc safe page.\n");
		goto failed_setup;
	}
	OSDeviceMemSetIO(mmu->safe_page_logical, 0, 4096);

	mmu->xdev = xdev;

	return mmu;

failed_setup:
	mutex_destroy(&mmu->page_table_mutex);
failed_stlbs_alloc:
	kvfree(mmu->stlbs);
	xdx_gem_object_put(&mmu->mtlb_bo->base);
failed_mtlb_alloc:
	drm_ht_remove(&mmu->res_ht);
failed_ht_create:
	mutex_destroy(&mmu->ht_mutex);
	kfree(mmu);
	return ERR_PTR(ret);
}

static int dc8200_mmu_destrory(void *mmu_ptr)
{
	struct dc8200_mmu *mmu = (struct dc8200_mmu *)mmu_ptr;
	u32 i;
	xdx_gem_object_put(&mmu->mtlb_bo->base);
	xdx_gem_object_put(&mmu->safe_page_bo->base);
	for (i = 0; i < MMU_MTLB_ENTRY_NUM; i++) {
		struct dc8200_mmu_stlb **stlbs = (struct dc8200_mmu_stlb **)mmu->stlbs;
		if (stlbs[i]) {
			kfree(stlbs[i]);
			stlbs[i] = NULL;
		}
	}
	kvfree(mmu->stlbs);
	xdx_gem_object_put(&mmu->stlbs_bo->base);


	kvfree(mmu->map_logical);

	mutex_destroy(&mmu->page_table_mutex);

	mutex_destroy(&mmu->ht_mutex);

	drm_ht_remove(&mmu->res_ht);

	kfree(mmu);

	return 0;
}

static int get_page_entry(struct dc8200_mmu *mmu, u32 address, u32 **page_table)
{
	struct xdx_device *xdev = mmu->xdev;
	struct dc8200_mmu_stlb *stlb;
	struct dc8200_mmu_stlb **stlbs = (struct dc8200_mmu_stlb **)mmu->stlbs;
	u32 mtlb_offset = _mtlb_offset(address);
#ifdef CONFIG_MMU_1K_16M
	u32 stlb_offset = 0;
#else
	u32 stlb_offset = _stlb_offset(address);
#endif

	stlb = stlbs[mtlb_offset - mmu->dynamic_mapping_start];
	if (stlb == NULL) {
		dev_err(xdev->dev, "BUG: invalid stlb,  mmu=%p stlbs=%p  mtlb_offset=0x%x %s(%d)\n",
		       mmu, stlbs, mtlb_offset, __FUNCTION__, __LINE__);
		return -ENXIO;
	}

	*page_table = &stlb->logical[stlb_offset];

	return 0;
}

static int _link(struct dc8200_mmu *mmu, u32 index, u32 node)
{
	struct xdx_device *xdev = mmu->xdev;

	if (index >= mmu->page_table_entries) {
		mmu->heap_list = node;
	} else {
		u32 *map = mmu->map_logical;

		switch (DC_ENTRY_TYPE(_read_page_entry(&map[index]))) {
		case DC_MMU_SINGLE:
			/* Previous is a single node, link to it*/
			_write_page_entry(&map[index],
					  (node << 8) | DC_MMU_SINGLE);
			break;
		case DC_MMU_FREE:
			/* Link to FREE TYPE node */
			_write_page_entry(&map[index + 1], node);
			break;
		default:
			dev_err(xdev->dev, "MMU table corrupted at index %u!", index);
			return -EINVAL;
		}
	}

	return 0;
}

static int _add_free(struct dc8200_mmu *mmu, u32 index, u32 node, u32 count)
{
	u32 *map = mmu->map_logical;

	if (count == 1) {
		/* Initialize a single page node */
		_write_page_entry(map + node, DC_SINGLE_PAGE_NODE_INITIALIZE |
						      DC_MMU_SINGLE);
	} else {
		/* Initialize the FREE node*/
		_write_page_entry(map + node, (count << 8) | DC_MMU_FREE);
		_write_page_entry(map + node + 1, ~0U);
	}

	return _link(mmu, index, node);
}

/* Collect free nodes */
static int _collect(struct dc8200_mmu *mmu)
{
	struct xdx_device *xdev = mmu->xdev;
	u32 *map = mmu->map_logical;
	u32 count = 0, start = 0, i = 0;
	u32 previous = ~0U;
	int ret = 0;

	mmu->heap_list = ~0U;
	mmu->free_nodes = false;

	/* Walk the entire page table */
	for (i = 0; i < mmu->page_table_entries; i++) {
		switch (DC_ENTRY_TYPE(_read_page_entry(&map[i]))) {
		case DC_MMU_SINGLE:
			if (count++ == 0) {
				/* Set new start node */
				start = i;
			}
			break;
		case DC_MMU_FREE:
			if (count == 0) {
				/* Set new start node */
				start = i;
			}

			count += _read_page_entry(&map[i]) >> 8;
			/* Advance the index of the page table */
			i += (_read_page_entry(&map[i]) >> 8) - 1;
			break;
		case DC_MMU_USED:
			/* Meet used node, start to collect */
			if (count > 0) {
				/* Add free node to list*/
				ret = _add_free(mmu, previous, start, count);
				if (ret)
					return ret;
				/* Reset previous unused node index */
				previous = start;
				count = 0;
			}
			break;
		default:
			dev_err(xdev->dev, "MMU page table corrupted at index %u!", i);
			return -EINVAL;
		}
	}

	/* If left node is an open node. */
	if (count > 0) {
		ret = _add_free(mmu, previous, start, count);
		if (ret)
			return ret;
	}

	return 0;
}

static int _fill_page_table(u32 *page_table, u32 page_count, u32 entry_value)
{
	u32 i;

	for (i = 0; i < page_count; i++)
		_write_page_entry(page_table + i, entry_value);

	return 0;
}

static int mmu_allocate_pages(struct dc8200_mmu *mmu, u32 page_count, u32 *address)
{
	struct xdx_device *xdev = mmu->xdev;
	bool got = false, acquired = false;
	u32 *map;
	u32 index = 0, vaddr, left;
	u32 previous = ~0U;
	u32 mtlb_offset, stlb_offset;
	int ret = 0;

	if (page_count == 0 || page_count > mmu->page_table_entries) {
		dev_err(xdev->dev, "%s has invalid arguments.\n", __FUNCTION__);
		return -EINVAL;
	}

	mutex_lock(&mmu->page_table_mutex);

	acquired = true;

	for (map = mmu->map_logical; !got;) {
		for (index = mmu->heap_list;
		     !got && (index < mmu->page_table_entries);) {
			switch (DC_ENTRY_TYPE(_read_page_entry(&map[index]))) {
			case DC_MMU_SINGLE:
				if (page_count == 1) {
					got = true;
				} else {
					/* Move to next node */
					previous = index;
					index = _read_page_entry(&map[index]) >>
						8;
				}
				break;
			case DC_MMU_FREE:
				if (page_count <=
				    (_read_page_entry(&map[index]) >> 8)) {
					got = true;
				} else {
					/* Move to next node */
					previous = index;
					index = _read_page_entry(
						&map[index + 1]);
				}
				break;
			case DC_MMU_USED:
				index++;
				break;
			default:
				/* Only link SINGLE and FREE node */
				dev_err(xdev->dev, "MMU table corrupted at index %u!",
				       index);
				ret = -EINVAL;
				goto OnError;
			}
		}

		/* If out of index */
		if (index >= mmu->page_table_entries) {
			if (mmu->free_nodes) {
				/* Collect the free node */
				ret = _collect(mmu);
				if (ret)
					goto OnError;
			} else {
				ret = -ENODATA;
				goto OnError;
			}
		}
	}

	switch (DC_ENTRY_TYPE(_read_page_entry(&map[index]))) {
	case DC_MMU_SINGLE:
		/* Unlink single node from node list */
		ret = _link(mmu, previous, _read_page_entry(&map[index]) >> 8);
		if (ret)
			goto OnError;
		break;

	case DC_MMU_FREE:
		left = (_read_page_entry(&map[index]) >> 8) - page_count;
		switch (left) {
		case 0:
			/* Unlink the entire FREE type node */
			ret = _link(mmu, previous,
				    _read_page_entry(&map[index + 1]));
			if (ret)
				goto OnError;
			break;
		case 1:
			/* Keep the map[index] as a single node,
             * mark the left as used
             */
			_write_page_entry(&map[index],
					  (_read_page_entry(&map[index + 1])
					   << 8) | DC_MMU_SINGLE);
			index++;
			break;
		default:
			/* FREE type node left */
			_write_page_entry(&map[index],
					  (left << 8) | DC_MMU_FREE);
			index += left;
			break;
		}
		break;
	default:
		/* Only link SINGLE and FREE node */
		dev_err(xdev->dev, "MMU table corrupted at index %u!", index);
		ret = -EINVAL;
		goto OnError;
	}

	/* Mark node as used */
	ret = _fill_page_table(&map[index], page_count, DC_MMU_USED);
	if (ret)
		goto OnError;

	mutex_unlock(&mmu->page_table_mutex);

#ifdef CONFIG_MMU_1K_16M
	mtlb_offset =
		index / MMU_STLB_16M_ENTRY_NUM + mmu->dynamic_mapping_start;
	stlb_offset = index % MMU_STLB_16M_ENTRY_NUM;
	vaddr = (mtlb_offset << 28) | (stlb_offset << MMU_STLB_16M_SHIFT);
#else
	mtlb_offset =
		index / MMU_STLB_4K_ENTRY_NUM + mmu->dynamic_mapping_start;
	stlb_offset = index % MMU_STLB_4K_ENTRY_NUM;
	vaddr = (mtlb_offset << MMU_MTLB_SHIFT) |
		(stlb_offset << MMU_STLB_4K_SHIFT);
#endif

	if (address != NULL) {
		*address = vaddr;
	}

	return 0;

OnError:
	if (acquired)
		mutex_unlock(&mmu->page_table_mutex);

	return ret;
}

static int mmu_free_pages(struct dc8200_mmu *mmu, u32 address, u32 page_count)
{
	u32 *node;

	if (page_count == 0)
		return -EINVAL;

	node = mmu->map_logical + _address_to_index(mmu, address);

	mutex_lock(&mmu->page_table_mutex);

	if (page_count == 1) {
		/* Mark the Single page node free */
		_write_page_entry(node, DC_SINGLE_PAGE_NODE_INITIALIZE |
						DC_MMU_SINGLE);
	} else {
		/* Mark the FREE type node free */
		_write_page_entry(node, (page_count << 8) | DC_MMU_FREE);
		_write_page_entry(node + 1, ~0U);
	}

	mmu->free_nodes = true;

	mutex_unlock(&mmu->page_table_mutex);

	return 0;
}

static int mmu_set_page(struct dc8200_mmu *mmu, u64 page_address, u32 *page_entry)
{
	u32 address_ext;
	u32 address;

	if (page_entry == NULL || (page_address & 0xFFF)) {
		return -EINVAL;
	}

	/* [31:0]. */
	address = (u32)(page_address & 0xFFFFFFFF);
	/* [39:32]. */
	address_ext = (u32)((page_address >> 32) & 0xFF);

	_write_page_entry(page_entry, _set_page(address, address_ext, true));

	return 0;
}

static int mmu_map_memory(struct dc8200_mmu *mmu, u64 physical, u32 page_count,
		u32 page_alignment, u32 *address, bool continuous, bool security)
{
	u32 virutal_address, i = 0, k = 0;
	u32 mtlb_num, mtlb_entry, mtlb_offset;
	bool allocated = false;
	int ret = 0;
#if defined(SUPPORT_XDXGPU_RPM)
	u32 gddr_cnt = mmu->xdev->fw_st_info.gddr_count;

	if (unlikely(atomic_read(&mmu->xdev->in_runpm)))
		page_count = DIV_ROUND_UP(page_count,
					  mmu->xdev->fw_st_info.gddr_count);
#endif

	ret = mmu_allocate_pages(mmu, page_count * page_alignment, &virutal_address);
	if (ret)
		goto OnError;

	*address = virutal_address;
	allocated = true;

	/*Fill mtlb security bit*/
#ifdef CONFIG_MMU_1K_16M
	mtlb_num = _mtlb_offset(virutal_address +
				page_count * MMU_PAGE_16M_SIZE - 1) -
		   _mtlb_offset(virutal_address) + 1;
#else
	mtlb_num = _mtlb_offset(virutal_address +
				page_count * page_alignment * MMU_PAGE_4K_SIZE - 1) -
		   _mtlb_offset(virutal_address) + 1;
#endif
	mtlb_offset = _mtlb_offset(virutal_address);
	mtlb_entry = mmu->mtlb_logical[mtlb_offset];

	for (i = 0; i < mtlb_num; i++) {
		mtlb_entry = mmu->mtlb_logical[mtlb_offset + i];
		if (security) {
			mtlb_entry = mtlb_entry | MMU_MTLB_SECURITY |
				     MMU_MTLB_EXCEPTION;
			_write_page_entry(&mmu->mtlb_logical[mtlb_offset + i],
					  mtlb_entry);
		} else {
			mtlb_entry = mtlb_entry & (~MMU_MTLB_SECURITY);
			_write_page_entry(&mmu->mtlb_logical[mtlb_offset + i],
					  mtlb_entry);
		}
	}

	/* Fill in page table */
	for (i = 0; i < page_count; i++) {
		u64 page_phy;
		u32 *page_entry;
		u64 *pages;

		if (continuous == true) {
#ifdef CONFIG_MMU_1K_16M
			page_phy = physical + i * MMU_PAGE_16M_SIZE;
#else
			page_phy = physical + i * page_alignment * MMU_PAGE_4K_SIZE;
#endif
		} else {
			pages = (u64 *)physical;
#if defined(SUPPORT_XDXGPU_RPM)
			if (unlikely(atomic_read(&mmu->xdev->in_runpm)))
				page_phy = pages[gddr_cnt * i];
			else
#endif
				page_phy = pages[i];
		}

		for (k = 0; k < page_alignment; k++) {
			ret = get_page_entry(mmu, virutal_address, &page_entry);
			if (ret)
				goto OnError;

			/* Write the page address to the page entry */
#ifdef CONFIG_MMU_1K_16M
			for (j = 0; j < MMU_STLB_16M_ENTRY_NUM; j++) {
				ret = mmu_set_page(mmu, page_phy, page_entry + j);
				if (ret)
					goto OnError;
			}
#else
			ret = mmu_set_page(mmu, page_phy + k * MMU_PAGE_4K_SIZE, page_entry);
			if (ret)
				goto OnError;
#endif
			/* Get next page */
#ifdef CONFIG_MMU_1K_16M
			virutal_address += MMU_PAGE_16M_SIZE;
#else
			virutal_address += MMU_PAGE_4K_SIZE;
#endif
		}
	}

	return 0;

OnError:
	if (allocated)
		mmu_free_pages(mmu, virutal_address, page_count * page_alignment);

	return ret;
}

static int mmu_unmap_memory(struct dc8200_mmu *mmu, u32 gpu_address, u32 page_count)
{
#if defined(SUPPORT_XDXGPU_RPM)
	if (unlikely(atomic_read(&mmu->xdev->in_runpm)))
		page_count = DIV_ROUND_UP(page_count,
					  mmu->xdev->fw_st_info.gddr_count);
#endif
	return mmu_free_pages(mmu, gpu_address, page_count);
}

static struct dc8200_va_node *dc8200_va_node_lookup(struct dc8200_mmu *mmu, struct xdx_bo *bo)
{
	struct dc8200_va_node *node = NULL;
	struct drm_hash_item *hash;

	if (!drm_ht_find_item(&mmu->res_ht, (unsigned long) bo, &hash))
		node = container_of(hash, typeof(*node), hash);

	return node;
}

static int dc8200_mmu_bo_map(void *handle, struct xdx_bo *bo, union xdx_address *va)
{
	struct dc8200_mmu *mmu = (struct dc8200_mmu *)handle;
	struct xdx_device *xdev = mmu->xdev;
	u64 *pages;
	int ret;
	struct dc8200_va_node *node;
	u32 page_count, page_aligment;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	union xdx_address vaddr = { 0 };
#else
	union xdx_address vaddr;
	memset(&vaddr, 0, sizeof(vaddr));
#endif

	mutex_lock(&mmu->ht_mutex);
	node = dc8200_va_node_lookup(mmu, bo);
	if (node) {
		*va = node->addr;
		atomic_inc(&node->refcnt);
		ret = 0;
		goto unlock_mutex;
	}

	pages = xdx_bo_get_pages(bo);
	if (!pages) {
		dev_err(xdev->dev, "failed to get pages\n");
		ret = -ENOMEM;
		goto unlock_mutex;
	}

	node = kzalloc(sizeof(*node), GFP_KERNEL);
	if (!node) {
		dev_err(xdev->dev, "failed to alloc node\n");
		ret = -ENOMEM;
		goto unlock_mutex;
	}
	atomic_set(&node->refcnt, 0);
	page_aligment = bo_gpu_page_alignment(bo);
	page_count = bo_ngpu_pages(bo);
#ifdef CONFIG_MMU_1K_16M
	ret = mmu_map_memory(mmu, (u64)pages, bo->num_pages,
				&vaddr.low, false, false);
#else
	ret = mmu_map_memory(mmu, (u64)pages, bo->num_pages, page_aligment,
				&vaddr.low, false, false);
#endif
	if (ret) {
		dev_err(xdev->dev, "bo %p failed to do mmu map (%d)\n", bo, ret);
		goto failed_map;
	}

	node->hash.key = (unsigned long)bo;
	ret = drm_ht_insert_item(&mmu->res_ht, &node->hash);
	if (ret) {
		dev_err(xdev->dev, "Failed to initialize a va node\n");
		goto failed_insert_node;
	}

	*va = node->addr = vaddr;
	atomic_inc(&node->refcnt);
	ret = 0;
	goto unlock_mutex;

failed_insert_node:
	mmu_unmap_memory(mmu, vaddr.low, page_count);
failed_map:
	kfree(node);
unlock_mutex:
	mutex_unlock(&mmu->ht_mutex);
	return ret;
}

static int dc8200_mmu_bo_unmap(void *handle, struct xdx_bo *bo, union xdx_address va)
{
	struct dc8200_mmu *mmu = (struct dc8200_mmu *)handle;
	struct xdx_device *xdev = mmu->xdev;
	int ret = 0;
	unsigned int node_refcnt;
	struct dc8200_va_node *node;
	u32 page_count;

	mutex_lock(&mmu->ht_mutex);
	node = dc8200_va_node_lookup(mmu, bo);
	if (!node || (node->addr.value != va.value)) {
		dev_err(xdev->dev, "bo %p no mapped or wrong va 0x%llx\n", bo,
			va.value);
		mutex_unlock(&mmu->ht_mutex);
		return -EINVAL;
	}

	node_refcnt = atomic_read(&node->refcnt);
	page_count = bo_ngpu_pages(bo);

	if (node_refcnt > 1) {
		atomic_dec(&node->refcnt);
		ret = 0;
		mutex_unlock(&mmu->ht_mutex);
	}else if (node_refcnt == 1) {
		atomic_dec(&node->refcnt);
#ifdef CONFIG_MMU_1K_16M
		ret = mmu_unmap_memory(mmu, va.low, bo->num_pages);
#else
		ret = mmu_unmap_memory(mmu, va.low, page_count);
#endif
		if (ret)
			dev_err(xdev->dev, "bo %p failed to do dc mmu unmap.\n", bo);

		drm_ht_remove_item(&mmu->res_ht, &node->hash);
		kfree(node);
		mutex_unlock(&mmu->ht_mutex);
	}

	return ret;
}

const struct xdx_mmu_funcs dc8200_mmu_funcs = {
	.create = dc8200_mmu_create,
	.destrory = dc8200_mmu_destrory,
	.bo_map = dc8200_mmu_bo_map,
	.bo_unmap = dc8200_mmu_bo_unmap,
};
