/*************************************************************************/ /*!
@File
@Title          Environment related functions
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#include <linux/version.h>
#include <asm/page.h>
#include <asm/div64.h>
#include <linux/atomic.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/io.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/hugetlb.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/genalloc.h>
#include <linux/string.h>
#include <linux/freezer.h>
#include <asm/hardirq.h>
#include <asm/tlbflush.h>
#include <linux/timer.h>
#include <linux/capability.h>
#include <linux/uaccess.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/dmaengine.h>
#include <linux/kthread.h>
#include <linux/utsname.h>
#include <linux/scatterlist.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
#include <linux/dma-map-ops.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
#include <linux/pfn_t.h>
#include <linux/pfn.h>
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)) */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
#include <linux/sched/clock.h>
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)) */
#if defined(SUPPORT_SECURE_ALLOC_KM)
#if defined(PVR_ANDROID_HAS_DMA_HEAP_FIND)
#include <linux/dma-heap.h>
#include "physmem_dmabuf.h"
#else
#include "physmem.h"
#endif
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>
#endif
#include <linux/sched.h>

#include "log2.h"
#include "osfunc.h"
#include "cache_km.h"
#include "img_defs.h"
#include "img_types.h"
#include "allocmem.h"
#include "devicemem_server_utils.h"
#include "event.h"
#include "linkage.h"
#include "pvr_uaccess.h"
#include "pvr_debug.h"
#include "pvr_bridge_k.h"
#include "pvrsrv_memallocflags.h"
#if defined(PVRSRV_ENABLE_PROCESS_STATS)
#include "process_stats.h"
#endif
#include "physmem_osmem_linux.h"
#include "dma_support.h"
#include "kernel_compatibility.h"

#include "pvrsrv_sync_server.h"
#if defined(PVRSRV_ENABLE_GPU_MEMORY_INFO)
#include "ri_server.h"
#include "pvr_ricommon.h"
#endif
#include "xdx_dma.h"
#include "dma_flags.h"
#include "vdma_drv.h"

#if defined(VIRTUAL_PLATFORM)
#define EVENT_OBJECT_TIMEOUT_US		(120000000ULL)
#else
#if defined(EMULATOR) || defined(TC_APOLLO_TCF5) || defined(SUPPORT_PANGU_EMU_GDDR)
#define EVENT_OBJECT_TIMEOUT_US		(2000000ULL)
#else
#if defined(SUPPORT_PANGU_FPGA)
#define EVENT_OBJECT_TIMEOUT_US		(1000000ULL)
#else
#define EVENT_OBJECT_TIMEOUT_US		(100000ULL)
#endif
#endif /* EMULATOR */
#endif


typedef struct {
	struct task_struct *kthread;
	PFN_THREAD pfnThread;
	void *hData;
	IMG_CHAR *pszThreadName;
	IMG_BOOL   bIsThreadRunning;
	IMG_BOOL   bIsSupportingThread;
	PFN_THREAD_DEBUG_DUMP pfnDebugDumpCB;
	DLLIST_NODE sNode;
} OSThreadData;

void OSSuspendTaskInterruptible(void)
{
	set_current_state(TASK_INTERRUPTIBLE);
	schedule();
}

static DLLIST_NODE gsThreadListHead;

static void _ThreadListAddEntry(OSThreadData *psThreadListNode)
{
	dllist_add_to_tail(&gsThreadListHead, &(psThreadListNode->sNode));
}

static void _ThreadListRemoveEntry(OSThreadData *psThreadListNode)
{
	dllist_remove_node(&(psThreadListNode->sNode));
}

static void _ThreadSetStopped(OSThreadData *psOSThreadData)
{
	psOSThreadData->bIsThreadRunning = IMG_FALSE;
}

static void _OSInitThreadList(void)
{
	dllist_init(&gsThreadListHead);
}

void OSThreadDumpInfo(DUMPDEBUG_PRINTF_FUNC* pfnDumpDebugPrintf,
                      void *pvDumpDebugFile)
{
	PDLLIST_NODE psNodeCurr, psNodeNext;

	dllist_foreach_node(&gsThreadListHead, psNodeCurr, psNodeNext)
	{
		OSThreadData *psThreadListNode;
		psThreadListNode = IMG_CONTAINER_OF(psNodeCurr, OSThreadData, sNode);

		PVR_DUMPDEBUG_LOG("  %s : %s",
				  psThreadListNode->pszThreadName,
				  (psThreadListNode->bIsThreadRunning) ? "Running" : "Stopped");

		if (psThreadListNode->pfnDebugDumpCB)
		{
			psThreadListNode->pfnDebugDumpCB(pfnDumpDebugPrintf, pvDumpDebugFile);
		}
	}
}

PVRSRV_ERROR OSPhyContigPagesAlloc(PHYS_HEAP *psPhysHeap, size_t uiSize,
							PG_HANDLE *psMemHandle, IMG_DEV_PHYADDR *psDevPAddr,
							IMG_PID uiPid)
{
	PVRSRV_DEVICE_NODE *psDevNode = PhysHeapDeviceNode(psPhysHeap);
	struct device *psDev = psDevNode->psDevConfig->pvOSDevice;
	IMG_CPU_PHYADDR sCpuPAddr;
	struct page *psPage;
	IMG_UINT32	ui32Order=0;
	gfp_t gfp_flags;
	dma_addr_t uiDmaAddr;

	PVR_ASSERT(uiSize != 0);
	/*Align the size to the page granularity */
	uiSize = PAGE_ALIGN(uiSize);

	/*Get the order to be used with the allocation */
	ui32Order = get_order(uiSize);

	gfp_flags = GFP_KERNEL;

#if !defined(PVR_LINUX_PHYSMEM_USE_HIGHMEM_ONLY)
	if (psDev)
	{
		if (*psDev->dma_mask == DMA_BIT_MASK(32))
		{
			/* Limit to 32 bit.
			 * Achieved by setting __GFP_DMA32 for 64 bit systems */
			gfp_flags |= __GFP_DMA32;
		}
		else if (*psDev->dma_mask < DMA_BIT_MASK(32))
		{
			/* Limit to whatever the size of DMA zone is. */
			gfp_flags |= __GFP_DMA;
		}
	}
#else
	PVR_UNREFERENCED_PARAMETER(psDev);
#endif

	/*allocate the pages */
	psPage = alloc_pages(gfp_flags, ui32Order);
	if (psPage == NULL)
	{
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}
	uiSize = (1 << ui32Order) * PAGE_SIZE;

	uiDmaAddr = dma_map_page(psDev, psPage, 0, uiSize, DMA_TO_DEVICE);
	if (dma_mapping_error(psDev, uiDmaAddr))
	{
		PVR_DPF((PVR_DBG_ERROR,
				"%s: Failed to dma map page: %llx, uiSize: %lu",
				__func__,
				(u64)psPage,
				uiSize));
		__free_pages(psPage, ui32Order);
		return PVRSRV_ERROR_BAD_MAPPING;
	}

	psMemHandle->uiDmaAddr = uiDmaAddr;
	psMemHandle->u.pvHandle = psPage;
	psMemHandle->uiOrder = ui32Order;
	sCpuPAddr.uiAddr = uiDmaAddr;

	/*
	 * Even when more pages are allocated as base MMU object we still need one single physical address because
	 * they are physically contiguous.
	 */
	PhysHeapCpuPAddrToDevPAddr(psPhysHeap, 1, psDevPAddr, &sCpuPAddr);

#if defined(PVRSRV_ENABLE_PROCESS_STATS)
#if !defined(PVRSRV_ENABLE_MEMORY_STATS)
	PVRSRVStatsIncrMemAllocStatAndTrack(PVRSRV_MEM_ALLOC_TYPE_ALLOC_PAGES_PT_UMA,
	                                    uiSize,
	                                    (IMG_UINT64)(uintptr_t) psPage,
	                                    uiPid);
#else
	PVRSRVStatsAddMemAllocRecord(PVRSRV_MEM_ALLOC_TYPE_ALLOC_PAGES_PT_UMA,
	                             psPage,
	                             sCpuPAddr,
	                             uiSize,
	                             uiPid
	                             DEBUG_MEMSTATS_VALUES);
#endif
#else
	PVR_UNREFERENCED_PARAMETER(uiPid);
#endif

	return PVRSRV_OK;
}

void OSPhyContigPagesFree(PHYS_HEAP *psPhysHeap, PG_HANDLE *psMemHandle)
{
	PVRSRV_DEVICE_NODE *psDevNode = PhysHeapDeviceNode(psPhysHeap);
	struct device *psDev = psDevNode->psDevConfig->pvOSDevice;
	struct page *psPage = (struct page*) psMemHandle->u.pvHandle;
	IMG_UINT32	uiSize, uiPageCount=0, ui32Order;

	PVR_UNREFERENCED_PARAMETER(psPhysHeap);

	ui32Order = psMemHandle->uiOrder;
	uiPageCount = (1 << ui32Order);
	uiSize = (uiPageCount * PAGE_SIZE);

#if defined(PVRSRV_ENABLE_PROCESS_STATS)
#if !defined(PVRSRV_ENABLE_MEMORY_STATS)
	PVRSRVStatsDecrMemAllocStatAndUntrack(PVRSRV_MEM_ALLOC_TYPE_ALLOC_PAGES_PT_UMA,
	                                      (IMG_UINT64)(uintptr_t) psPage);
#else
	PVRSRVStatsRemoveMemAllocRecord(PVRSRV_MEM_ALLOC_TYPE_ALLOC_PAGES_PT_UMA,
	                                (IMG_UINT64)(uintptr_t) psPage,
	                                OSGetCurrentClientProcessIDKM());
#endif
#endif

	dma_unmap_page(psDev, psMemHandle->uiDmaAddr, uiSize, DMA_TO_DEVICE);
	__free_pages(psPage, ui32Order);
	psMemHandle->uiOrder = 0;
}

PVRSRV_ERROR OSPhyContigPagesMap(PHYS_HEAP *psPhysHeap, PG_HANDLE *psMemHandle,
						size_t uiSize, IMG_DEV_PHYADDR *psDevPAddr,
						void **pvPtr)
{
	size_t actualSize = 1 << (PAGE_SHIFT + psMemHandle->uiOrder);
	*pvPtr = kmap((struct page*)psMemHandle->u.pvHandle);

	PVR_UNREFERENCED_PARAMETER(psDevPAddr);

	PVR_UNREFERENCED_PARAMETER(actualSize); /* If we don't take an #ifdef path */
	PVR_UNREFERENCED_PARAMETER(uiSize);
	PVR_UNREFERENCED_PARAMETER(psPhysHeap);

#if defined(PVRSRV_ENABLE_PROCESS_STATS)
#if !defined(PVRSRV_ENABLE_MEMORY_STATS)
	PVRSRVStatsIncrMemAllocStat(PVRSRV_MEM_ALLOC_TYPE_VMAP_PT_UMA, actualSize, OSGetCurrentClientProcessIDKM());
#else
	{
		IMG_CPU_PHYADDR sCpuPAddr;
		sCpuPAddr.uiAddr = 0;

		PVRSRVStatsAddMemAllocRecord(PVRSRV_MEM_ALLOC_TYPE_VMAP_PT_UMA,
									 *pvPtr,
									 sCpuPAddr,
									 actualSize,
									 OSGetCurrentClientProcessIDKM()
									 DEBUG_MEMSTATS_VALUES);
	}
#endif
#endif

	return PVRSRV_OK;
}

void OSPhyContigPagesUnmap(PHYS_HEAP *psPhysHeap, PG_HANDLE *psMemHandle, void *pvPtr)
{
#if defined(PVRSRV_ENABLE_PROCESS_STATS)
#if !defined(PVRSRV_ENABLE_MEMORY_STATS)
	/* Mapping is done a page at a time */
	PVRSRVStatsDecrMemAllocStat(PVRSRV_MEM_ALLOC_TYPE_VMAP_PT_UMA,
	                            (1 << (PAGE_SHIFT + psMemHandle->uiOrder)),
	                            OSGetCurrentClientProcessIDKM());
#else
	PVRSRVStatsRemoveMemAllocRecord(PVRSRV_MEM_ALLOC_TYPE_VMAP_PT_UMA,
	                                (IMG_UINT64)(uintptr_t)pvPtr,
	                                OSGetCurrentClientProcessIDKM());
#endif
#endif

	PVR_UNREFERENCED_PARAMETER(psPhysHeap);
	PVR_UNREFERENCED_PARAMETER(pvPtr);

	kunmap((struct page*) psMemHandle->u.pvHandle);
}

PVRSRV_ERROR OSPhyContigPagesClean(PHYS_HEAP *psPhysHeap,
		PG_HANDLE *psMemHandle,
		IMG_UINT32 uiOffset,
		IMG_UINT32 uiLength)
{
	PVRSRV_DEVICE_NODE *psDevNode = PhysHeapDeviceNode(psPhysHeap);
	struct device *psDev = psDevNode->psDevConfig->pvOSDevice;
	PVRSRV_ERROR eError = PVRSRV_OK;
	IMG_UINT32 ui32Order;

	if (uiLength == 0)
	{
		goto e0;
	}

	ui32Order = psMemHandle->uiOrder;
	if ((uiOffset + uiLength) > ((1 << ui32Order) * PAGE_SIZE))
	{
		PVR_DPF((PVR_DBG_ERROR,
				"%s: Invalid size params, uiOffset %u, uiLength %u",
				__func__,
				uiOffset,
				uiLength));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e0;
	}

	dma_sync_single_for_device(psDev, psMemHandle->uiDmaAddr,
				   (1 << ui32Order) * PAGE_SIZE, DMA_TO_DEVICE);

e0:
	return eError;
}

#if defined(__GNUC__)
#define PVRSRV_MEM_ALIGN __attribute__ ((aligned (0x8)))
#define PVRSRV_MEM_ALIGN_MASK (0x7)
#else
#error "PVRSRV Alignment macros need to be defined for this compiler"
#endif

IMG_UINT32 OSCPUCacheAttributeSize(OS_CPU_CACHE_ATTRIBUTE eCacheAttribute)
{
	IMG_UINT32 uiSize = 0;

	switch (eCacheAttribute)
	{
		case OS_CPU_CACHE_ATTRIBUTE_LINE_SIZE:
			uiSize = cache_line_size();
			break;

		default:
			PVR_DPF((PVR_DBG_ERROR, "%s: Invalid cache attribute type %d",
					__func__, (IMG_UINT32)eCacheAttribute));
			PVR_ASSERT(0);
			break;
	}

	return uiSize;
}

IMG_INT OSMemCmp(void *pvBufA, void *pvBufB, size_t uiLen)
{
	return (IMG_INT)memcmp(pvBufA, pvBufB, uiLen);
}

size_t OSStringLCat(IMG_CHAR *pszDest, const IMG_CHAR *pszSrc, size_t uDstSize)
{
	/*
	 * Let strlcat handle any truncation cases correctly.
	 * We will definitely get a NUL-terminated string set in pszDest
	 */
	size_t uSrcSize = strlcat(pszDest, pszSrc, uDstSize);

#if defined(PVR_DEBUG_STRLCPY) && defined(DEBUG)
	/* Handle truncation by dumping calling stack if debug allows */
	if (uSrcSize >= uDstSize)
	{
		PVR_DPF((PVR_DBG_WARNING,
			"%s: String truncated Src = '<%s>' %ld bytes, Dest = '%s'",
			__func__, pszSrc, (long)uDstSize, pszDest));
		OSDumpStack();
	}
#endif /* defined(PVR_DEBUG_STRLCPY) && defined(DEBUG) */

	return uSrcSize;
}

IMG_INT32 OSSNPrintf(IMG_CHAR *pStr, size_t ui32Size, const IMG_CHAR *pszFormat, ...)
{
	va_list argList;
	IMG_INT32 iCount;

	va_start(argList, pszFormat);
	iCount = vsnprintf(pStr, (size_t)ui32Size, pszFormat, argList);
	va_end(argList);

	return iCount;
}

IMG_INT32 OSVSNPrintf(IMG_CHAR *pStr, size_t ui32Size, const IMG_CHAR* pszFormat, va_list vaArgs)
{
	return vsnprintf(pStr, ui32Size, pszFormat, vaArgs);
}

size_t OSStringLength(const IMG_CHAR *pStr)
{
	return strlen(pStr);
}

size_t OSStringNLength(const IMG_CHAR *pStr, size_t uiCount)
{
	return strnlen(pStr, uiCount);
}

IMG_INT32 OSStringNCompare(const IMG_CHAR *pStr1, const IMG_CHAR *pStr2,
                          size_t uiSize)
{
#if defined(DEBUG)
	/* Double-check that we are not passing NULL parameters in. If we are we
	 * return -1 (for arg1 == NULL, arg2 != NULL)
	 * 0 (for arg1 == NULL, arg2 == NULL
	 * +1 (for arg1 != NULL, arg2 == NULL)
	 * strncmp(arg1, arg2, size) otherwise
	 */
	if (pStr1 == NULL)
	{
		if (pStr2 == NULL)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s(%p, %p, %d): Both args NULL",
				 __func__, pStr1, pStr2, (int)uiSize));
			OSDumpStack();
			return 0;	/* Both NULL */
		}
		else
		{
			PVR_DPF((PVR_DBG_ERROR, "%s(%p, %p, %d): arg1 NULL",
				 __func__, pStr1, pStr2, (int)uiSize));
			OSDumpStack();
			return -1;	/* NULL < non-NULL */
		}
	}
	else
	{
		if (pStr2 == NULL)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s(%p, %p, %d): arg2 NULL",
				 __func__, pStr1, pStr2, (int)uiSize));
			OSDumpStack();
			return +1;	/* non-NULL > NULL */
		}
		else
		{
			return strncmp(pStr1, pStr2, uiSize);
		}
	}
#else
	return strncmp(pStr1, pStr2, uiSize);
#endif
}

PVRSRV_ERROR OSStringToUINT32(const IMG_CHAR *pStr, IMG_UINT32 ui32Base,
                              IMG_UINT32 *ui32Result)
{
	if (kstrtou32(pStr, ui32Base, ui32Result) != 0)
		return PVRSRV_ERROR_CONVERSION_FAILED;

	return PVRSRV_OK;
}

IMG_UINT32 OSStringUINT32ToStr(IMG_CHAR *pszBuf, size_t uSize,
							IMG_UINT32 ui32Num)
{
	IMG_UINT32 ui32i, ui32Len = 0, ui32NumCopy = ui32Num;

	/* calculate string length required to hold the number string */
	do
	{
		ui32Len++;
		ui32NumCopy /= 10;
	} while (ui32NumCopy != 0);

	if (unlikely(ui32Len >= uSize))
	{
		/* insufficient buffer */
		return 0;
	}

	for (ui32i = 0; ui32i < ui32Len; ui32i++)
	{
		pszBuf[ui32Len - (ui32i + 1)] = '0' + ui32Num % 10;
		ui32Num = ui32Num / 10;
	}

	pszBuf[ui32Len] = '\0';
	return ui32Len;
}

#if defined(SUPPORT_NATIVE_FENCE_SYNC) || defined(SUPPORT_BUFFER_SYNC)
static struct workqueue_struct *gpFenceStatusWq;

static PVRSRV_ERROR _NativeSyncInit(void)
{
	gpFenceStatusWq = create_freezable_workqueue("xdxgpu_fence_status");
	if (!gpFenceStatusWq)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: Failed to create foreign fence status workqueue",
				 __func__));
		return PVRSRV_ERROR_INIT_FAILURE;
	}

	return PVRSRV_OK;
}

static void _NativeSyncDeinit(void)
{
	destroy_workqueue(gpFenceStatusWq);
}

struct workqueue_struct *NativeSyncGetFenceStatusWq(void)
{
	if (!gpFenceStatusWq)
	{
#if defined(DEBUG)
		PVR_ASSERT(gpFenceStatusWq);
#endif
		return NULL;
	}

	return gpFenceStatusWq;
}
#endif

PVRSRV_ERROR OSInitEnvData(void)
{
	PVRSRV_ERROR eError = PVRSRV_OK;

	LinuxInitPhysmem();

	_OSInitThreadList();

#if defined(SUPPORT_NATIVE_FENCE_SYNC) || defined(SUPPORT_BUFFER_SYNC)
	eError = _NativeSyncInit();
#endif

	return eError;
}

void OSDeInitEnvData(void)
{
#if defined(SUPPORT_NATIVE_FENCE_SYNC) || defined(SUPPORT_BUFFER_SYNC)
	_NativeSyncDeinit();
#endif

	LinuxDeinitPhysmem();
}

void OSReleaseThreadQuanta(void)
{
	schedule();
}

void OSMemoryBarrier(volatile void *hReadback)
{
	mb();

	if (hReadback)
	{
		/* Force a read-back to memory to avoid posted writes on certain buses
		 * e.g. PCI(E)
		 */
		(void) OSReadDeviceMem32(hReadback);
	}
}

void OSWriteMemoryBarrier(volatile void *hReadback)
{
	wmb();

	if (hReadback)
	{
		/* Force a read-back to memory to avoid posted writes on certain buses
		 * e.g. PCI(E)
		 */
		(void) OSReadDeviceMem32(hReadback);
	}
}

/* Not matching/aligning this API to the Clockus() API above to avoid necessary
 * multiplication/division operations in calling code.
 */
static inline IMG_UINT64 Clockns64(void)
{
	IMG_UINT64 timenow;

	/* Kernel thread preempt protection. Some architecture implementations
	 * (ARM) of sched_clock are not preempt safe when the kernel is configured
	 * as such e.g. CONFIG_PREEMPT and others.
	 */
	preempt_disable();

	/* Using sched_clock instead of ktime_get since we need a time stamp that
	 * correlates with that shown in kernel logs and trace data not one that
	 * is a bit behind. */
	timenow = sched_clock();

	preempt_enable();

	return timenow;
}

IMG_UINT64 OSClockns64(void)
{
	return Clockns64();
}

IMG_UINT64 OSClockus64(void)
{
	IMG_UINT64 timenow = Clockns64();
	IMG_UINT32 remainder;

	return OSDivide64r64(timenow, 1000, &remainder);
}

IMG_UINT32 OSClockus(void)
{
	return (IMG_UINT32) OSClockus64();
}

IMG_UINT32 OSClockms(void)
{
	IMG_UINT64 timenow = Clockns64();
	IMG_UINT32 remainder;

	return OSDivide64(timenow, 1000000, &remainder);
}

static inline IMG_UINT64 KClockns64(void)
{
	ktime_t sTime = ktime_get();

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	return sTime;
#else
	return sTime.tv64;
#endif
}

PVRSRV_ERROR OSClockMonotonicns64(IMG_UINT64 *pui64Time)
{
	*pui64Time = KClockns64();
	return PVRSRV_OK;
}

PVRSRV_ERROR OSClockMonotonicus64(IMG_UINT64 *pui64Time)
{
	IMG_UINT64 timenow = KClockns64();
	IMG_UINT32 remainder;

	*pui64Time = OSDivide64r64(timenow, 1000, &remainder);
	return PVRSRV_OK;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0))
IMG_UINT64 OSClockMonotonicRawns64(void)
{
	struct timespec64 ts;

	ktime_get_raw_ts64(&ts);
	return ts.tv_sec * 1000000000 + ts.tv_nsec;
}
#else
IMG_UINT64 OSClockMonotonicRawns64(void)
{
	struct timespec ts;

	getrawmonotonic(&ts);
	return (IMG_UINT64) ts.tv_sec * 1000000000 + ts.tv_nsec;
}
#endif

IMG_UINT64 OSClockMonotonicRawus64(void)
{
	IMG_UINT32 rem;
	return OSDivide64r64(OSClockMonotonicRawns64(), 1000, &rem);
}

/*
	OSWaitus
*/
void OSWaitus(IMG_UINT32 ui32Timeus)
{
	udelay(ui32Timeus);
}


/*
	OSSleepms
*/
void OSSleepms(IMG_UINT32 ui32Timems)
{
	msleep(ui32Timems);
}


INLINE IMG_UINT64 OSGetCurrentProcessVASpaceSize(void)
{
	return (IMG_UINT64)TASK_SIZE;
}

INLINE IMG_PID OSGetCurrentProcessID(void)
{
	if (in_interrupt())
	{
		return KERNEL_ID;
	}

	return (IMG_PID)task_tgid_nr(current);
}

INLINE IMG_PID OSGetCurrentVirtualProcessID(void)
{
	if (in_interrupt())
	{
		return KERNEL_ID;
	}

	return (IMG_PID)task_tgid_vnr(current);
}

INLINE IMG_CHAR *OSGetCurrentProcessName(void)
{
	return current->comm;
}

INLINE uintptr_t OSGetCurrentThreadID(void)
{
	if (in_interrupt())
	{
		return KERNEL_ID;
	}

	return current->pid;
}

IMG_PID OSGetCurrentClientProcessIDKM(void)
{
	return OSGetCurrentProcessID();
}

IMG_CHAR *OSGetCurrentClientProcessNameKM(void)
{
	return OSGetCurrentProcessName();
}

uintptr_t OSGetCurrentClientThreadIDKM(void)
{
	return OSGetCurrentThreadID();
}

size_t OSGetPageSize(void)
{
	return PAGE_SIZE;
}

size_t OSGetPageShift(void)
{
	return PAGE_SHIFT;
}

size_t OSGetPageMask(void)
{
	return (OSGetPageSize()-1);
}

size_t OSGetOrder(size_t uSize)
{
	return get_order(PAGE_ALIGN(uSize));
}

IMG_UINT64 OSGetRAMSize(void)
{
	struct sysinfo SI;
	si_meminfo(&SI);

	return (PAGE_SIZE * SI.totalram);
}

size_t RGXHeapDerivePageShift(void)
{
	IMG_UINT32  uiLog2PageSize = OSGetPageShift();
	IMG_BOOL bFound = IMG_FALSE;
	/* OS page shift must be at least RGX_HEAP_4KB_PAGE_SHIFT,
	 * max RGX_HEAP_2MB_PAGE_SHIFT, non-zero and a power of two*/
	if (uiLog2PageSize == 0U ||
	    (uiLog2PageSize < RGX_HEAP_4KB_PAGE_SHIFT) ||
	    (uiLog2PageSize > RGX_HEAP_2MB_PAGE_SHIFT))
	{
		PVR_DPF((PVR_DBG_ERROR,
				"%s: Provided incompatible log2 page size %u",
				__func__,
				uiLog2PageSize));
		PVR_ASSERT(0);
		return 0;
	}

	do
	{
		switch (uiLog2PageSize)
		{
			case RGX_HEAP_4KB_PAGE_SHIFT:
			case RGX_HEAP_16KB_PAGE_SHIFT:
			case RGX_HEAP_64KB_PAGE_SHIFT:
			case RGX_HEAP_256KB_PAGE_SHIFT:
			case RGX_HEAP_1MB_PAGE_SHIFT:
			case RGX_HEAP_2MB_PAGE_SHIFT:
				/* All good, RGX page size equals given page size
				 * => use it as default for heaps */
				bFound = IMG_TRUE;
				break;
			default:
				/* We have to fall back to a larger device
				 * page size than given page size because there
				 * is no exact match for any supported size. */
				uiLog2PageSize += 1U;
				break;
		}
	} while (!bFound);

	return uiLog2PageSize;
}

size_t RGXHeapDerivePageSizes(void)
{
	return 1 << RGXHeapDerivePageShift();
}

typedef struct
{
	int os_error;
	PVRSRV_ERROR pvr_error;
} error_map_t;

/* return -ve versions of POSIX errors as they are used in this form */
static const error_map_t asErrorMap[] =
{
	{-EFAULT, PVRSRV_ERROR_BRIDGE_EFAULT},
	{-EINVAL, PVRSRV_ERROR_BRIDGE_EINVAL},
	{-ENOMEM, PVRSRV_ERROR_BRIDGE_ENOMEM},
	{-ERANGE, PVRSRV_ERROR_BRIDGE_ERANGE},
	{-EPERM,  PVRSRV_ERROR_BRIDGE_EPERM},
	{-ENOTTY, PVRSRV_ERROR_BRIDGE_ENOTTY},
	{-ENOTTY, PVRSRV_ERROR_BRIDGE_CALL_FAILED},
	{-ERANGE, PVRSRV_ERROR_BRIDGE_BUFFER_TOO_SMALL},
	{-ENOMEM, PVRSRV_ERROR_OUT_OF_MEMORY},
	{-EACCES, PVRSRV_ERROR_PMR_NOT_PERMITTED},
	{-EINVAL, PVRSRV_ERROR_INVALID_PARAMS},

	{0,       PVRSRV_OK}
};

int PVRSRVToNativeError(PVRSRV_ERROR e)
{
	int os_error = -EFAULT;
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(asErrorMap); i++)
	{
		if (e == asErrorMap[i].pvr_error)
		{
			os_error = asErrorMap[i].os_error;
			break;
		}
	}
	return os_error;
}

typedef struct  _MISR_DATA_ {
	struct workqueue_struct *psWorkQueue;
	struct work_struct sMISRWork;
	const IMG_CHAR* pszName;
	PFN_MISR pfnMISR;
	void *hData;
} MISR_DATA;

/*
	MISRWrapper
*/
static void MISRWrapper(struct work_struct *data)
{
	MISR_DATA *psMISRData = container_of(data, MISR_DATA, sMISRWork);

	PVR_DPF((PVR_DBG_MESSAGE, "Waking up '%s' MISR %p", psMISRData->pszName, psMISRData));

	psMISRData->pfnMISR(psMISRData->hData);
}

/*
	OSInstallMISR
*/
PVRSRV_ERROR OSInstallMISR(IMG_HANDLE *hMISRData, PFN_MISR pfnMISR,
							void *hData, const IMG_CHAR *pszMisrName)
{
	MISR_DATA *psMISRData;
	IMG_CHAR workqueueName[50] = "xdxgpu_misr_wq";

	psMISRData = OSAllocMem(sizeof(*psMISRData));
	PVR_LOG_RETURN_IF_NOMEM(psMISRData, "psMISRData");

	psMISRData->hData = hData;
	psMISRData->pfnMISR = pfnMISR;
	psMISRData->pszName = pszMisrName;

	PVR_DPF((PVR_DBG_MESSAGE, "Installing MISR with cookie %p", psMISRData));

	if (pszMisrName != NULL) {
		strcat(workqueueName, "-");
	}
	strcat(workqueueName, pszMisrName);
	psMISRData->psWorkQueue = create_singlethread_workqueue(workqueueName);

	if (psMISRData->psWorkQueue == NULL)
	{
		PVR_DPF((PVR_DBG_ERROR, "OSInstallMISR: create_singlethreaded_workqueue failed"));
		OSFreeMem(psMISRData);
		return PVRSRV_ERROR_UNABLE_TO_CREATE_THREAD;
	}

	INIT_WORK(&psMISRData->sMISRWork, MISRWrapper);

	*hMISRData = (IMG_HANDLE) psMISRData;

	return PVRSRV_OK;
}

/*
	OSUninstallMISR
*/
PVRSRV_ERROR OSUninstallMISR(IMG_HANDLE hMISRData)
{
	MISR_DATA *psMISRData = (MISR_DATA *) hMISRData;

	PVR_DPF((PVR_DBG_MESSAGE, "Uninstalling MISR with cookie %p", psMISRData));

	destroy_workqueue(psMISRData->psWorkQueue);
	OSFreeMem(psMISRData);

	return PVRSRV_OK;
}

/*
	OSScheduleMISR
*/
PVRSRV_ERROR OSScheduleMISR(IMG_HANDLE hMISRData)
{
	MISR_DATA *psMISRData = (MISR_DATA *) hMISRData;

	/*
		Note:

		In the case of NO_HARDWARE we want the driver to be synchronous so
		that we don't have to worry about waiting for previous operations
		to complete
	*/
#if defined(NO_HARDWARE)
	psMISRData->pfnMISR(psMISRData->hData);
	return PVRSRV_OK;
#else
	{
		bool rc = queue_work(psMISRData->psWorkQueue, &psMISRData->sMISRWork);
		return rc ? PVRSRV_OK : PVRSRV_ERROR_ALREADY_EXISTS;
	}
#endif
}

/* OS specific values for thread priority */
static const IMG_INT32 ai32OSPriorityValues[OS_THREAD_LAST_PRIORITY] =
{
	  0, /* OS_THREAD_NOSET_PRIORITY */
	-20, /* OS_THREAD_HIGHEST_PRIORITY */
	-10, /* OS_THREAD_HIGH_PRIORITY */
	  0, /* OS_THREAD_NORMAL_PRIORITY */
	  9, /* OS_THREAD_LOW_PRIORITY */
	 19, /* OS_THREAD_LOWEST_PRIORITY */
};

static int OSThreadRun(void *data)
{
	OSThreadData *psOSThreadData = data;

	/* count freezable threads */
	LinuxBridgeNumActiveKernelThreadsIncrement();

	/* Returns true if the thread was frozen, should we do anything with this
	 * information? What do we return? Which one is the error case? */
	set_freezable();

	PVR_DPF((PVR_DBG_MESSAGE, "Starting Thread '%s'...", psOSThreadData->pszThreadName));

	/* Call the client's kernel thread with the client's data pointer */
	psOSThreadData->pfnThread(psOSThreadData->hData);

	if (psOSThreadData->bIsSupportingThread)
	{
		_ThreadSetStopped(psOSThreadData);
	}

	/* Wait for OSThreadDestroy() to call kthread_stop() */
	while (!kthread_freezable_should_stop(NULL))
	{
		schedule();
	}

	LinuxBridgeNumActiveKernelThreadsDecrement();

	return 0;
}

PVRSRV_ERROR OSThreadCreate(IMG_HANDLE *phThread,
                            IMG_CHAR *pszThreadName,
                            PFN_THREAD pfnThread,
                            PFN_THREAD_DEBUG_DUMP pfnDebugDumpCB,
                            IMG_BOOL bIsSupportingThread,
                            void *hData)
{
	return OSThreadCreatePriority(phThread, pszThreadName, pfnThread,
	                              pfnDebugDumpCB, bIsSupportingThread, hData,
	                              OS_THREAD_NOSET_PRIORITY);
}

PVRSRV_ERROR OSThreadCreatePriority(IMG_HANDLE *phThread,
                                    IMG_CHAR *pszThreadName,
                                    PFN_THREAD pfnThread,
                                    PFN_THREAD_DEBUG_DUMP pfnDebugDumpCB,
                                    IMG_BOOL bIsSupportingThread,
                                    void *hData,
                                    OS_THREAD_LEVEL eThreadPriority)
{
	OSThreadData *psOSThreadData;
	PVRSRV_ERROR eError;

	psOSThreadData = OSAllocZMem(sizeof(*psOSThreadData));
	PVR_LOG_GOTO_IF_NOMEM(psOSThreadData, eError, fail_alloc);

	psOSThreadData->pfnThread = pfnThread;
	psOSThreadData->hData = hData;
	psOSThreadData->kthread = kthread_run(OSThreadRun, psOSThreadData, "%s", pszThreadName);

	if (IS_ERR(psOSThreadData->kthread))
	{
		eError = PVRSRV_ERROR_OUT_OF_MEMORY;
		goto fail_kthread;
	}

	if (bIsSupportingThread)
	{
		psOSThreadData->pszThreadName = pszThreadName;
		psOSThreadData->pfnDebugDumpCB = pfnDebugDumpCB;
		psOSThreadData->bIsThreadRunning = IMG_TRUE;
		psOSThreadData->bIsSupportingThread = IMG_TRUE;

		_ThreadListAddEntry(psOSThreadData);
	}

	if (eThreadPriority != OS_THREAD_NOSET_PRIORITY &&
	    eThreadPriority < OS_THREAD_LAST_PRIORITY)
	{
		set_user_nice(psOSThreadData->kthread,
		              ai32OSPriorityValues[eThreadPriority]);
	}

	*phThread = psOSThreadData;

	return PVRSRV_OK;

fail_kthread:
	OSFreeMem(psOSThreadData);
fail_alloc:
	PVR_ASSERT(eError != PVRSRV_OK);
	return eError;
}

PVRSRV_ERROR OSThreadDestroy(IMG_HANDLE hThread)
{
	OSThreadData *psOSThreadData = hThread;
	int ret;

	/* Let the thread know we are ready for it to end and wait for it. */
	ret = kthread_stop(psOSThreadData->kthread);
	if (0 != ret)
	{
		PVR_DPF((PVR_DBG_WARNING, "kthread_stop failed(%d)", ret));
		return PVRSRV_ERROR_RETRY;
	}

	if (psOSThreadData->bIsSupportingThread)
	{
		_ThreadListRemoveEntry(psOSThreadData);
	}

	OSFreeMem(psOSThreadData);

	return PVRSRV_OK;
}

void OSPanic(void)
{
	BUG();

#if defined(__KLOCWORK__)
	/* Klocwork does not understand that BUG is terminal... */
	abort();
#endif
}

IMG_BOOL OSIsMapPhysNonContigSupported(void)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)) && !defined(CONFIG_VMAP_PFN)
	return IMG_FALSE;
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(3, 16, 0))
	return IMG_FALSE;
#else
	return IMG_TRUE;
#endif
}

void OSUnMapPhysArrayToLin(void *pvLinAddr, void *pvPrivData)
{
	if (is_vmalloc_addr(pvLinAddr))
	{
#if defined(__riscv)
		PVR_UNREFERENCED_PARAMETER(pvPrivData);
		vunmap(pvLinAddr);
#else // defined(__riscv)
#if defined(CONFIG_VMAP_PFN)
		PVR_UNREFERENCED_PARAMETER(pvPrivData);
		vunmap(pvLinAddr);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0))
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 19, 280) && LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0))
        free_vm_area(pvPrivData);
#else
		unmap_kernel_range((unsigned long) (uintptr_t) pvLinAddr,
                           get_vm_area_size(pvPrivData));
#endif
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0))
        free_vm_area(pvPrivData);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
		PVR_UNREFERENCED_PARAMETER(pvPrivData);
		vunmap(pvLinAddr);
#else
		PVR_DPF((PVR_DBG_ERROR,"%s: Cannot map into kernel, no method supported.", __func__));
		PVR_ASSERT(0);
#endif
#endif // defined(__riscv)
	}
	else
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: Given kernel address is not a vmalloc addr", __func__));
	}
}

#define PagePAToPFN(PageAddr) (PageAddr >> PAGE_SHIFT)

PVRSRV_ERROR OSMapPhysArrayToLin(IMG_CPU_PHYADDR pPagePA[],
                                 IMG_UINT32 uiPagesCount,
                                 void **ppvLinAddr,
                                 void **ppvPrivData)
{
	if (ppvLinAddr == NULL || ppvPrivData == NULL)
	{
		return PVRSRV_ERROR_INVALID_PARAMS;
	}

#if defined(__riscv)
	{
		IMG_UINT32 i;
		struct page **pages;

		pages = kmalloc(uiPagesCount * sizeof(struct page *), GFP_KERNEL);
		if (!pages) {
			return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
		}

		for (i = 0; i < uiPagesCount; i++) {
			pPagePA[i].uiAddr = PagePAToPFN(pPagePA[i].uiAddr);
			pages[i] = virt_to_page(pfn_to_phys(pPagePA[i].uiAddr) >> PAGE_SHIFT);
		}

		*ppvLinAddr = vmap(pages, uiPagesCount, VM_MAP, PAGE_KERNEL);
		kfree(pages);

		if (NULL == *ppvLinAddr) {
			return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
		}
		*ppvPrivData = NULL;
		return PVRSRV_OK;
	}
#else // defined(__riscv)
#if defined(CONFIG_VMAP_PFN)
	{
		IMG_UINT32 i;

		for (i = 0; i < uiPagesCount; i++)
		{
			pPagePA[i].uiAddr = PagePAToPFN(pPagePA[i].uiAddr);
		}

		*ppvLinAddr = vmap_pfn((unsigned long *)pPagePA,
							   (unsigned int)uiPagesCount,
							   pgprot_device(PAGE_KERNEL));
		if (NULL == *ppvLinAddr)
		{
			return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
		}
		*ppvPrivData = NULL;
		return PVRSRV_OK;
	}
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 9, 0) && LINUX_VERSION_CODE >= KERNEL_VERSION(3, 16, 0))
	{
		pte_t *pte[32], **pte_array;
		struct vm_struct *psVMA;
		PVRSRV_ERROR eError = PVRSRV_OK;
		IMG_UINT32 i = 0;

		pte_array = &pte[0];
		if (sizeof(pte) < (sizeof(pte[0]) * uiPagesCount))
		{
			pte_array = kzalloc(uiPagesCount * sizeof(*pte_array), GFP_KERNEL);
			if (NULL == pte_array)
			{
				return PVRSRV_ERROR_OUT_OF_MEMORY;
			}
		}

		psVMA = alloc_vm_area((size_t)(uiPagesCount << PAGE_SHIFT), pte_array);
		if (NULL == psVMA)
		{
			eError = PVRSRV_ERROR_FAILED_TO_ALLOC_VIRT_MEMORY;
			goto FreePTEArray;
		}

		PVR_DPF((PVR_DBG_MESSAGE, "Allocated VM: %s VMA: %p Addr: %p Size: 0x%lx count: %d", __func__,
				psVMA, psVMA->addr, psVMA->size, psVMA->nr_pages));

		for (i = 0; i < uiPagesCount; i++)
		{
			*(pte_array[i]) = pte_mkspecial(pfn_pte((unsigned long) PagePAToPFN(pPagePA[i].uiAddr),
													pgprot_device(PAGE_KERNEL)));
		}

		OSWriteMemoryBarrier(psVMA->addr);

		*ppvLinAddr = psVMA->addr;
		*ppvPrivData = psVMA;

FreePTEArray:
		if (pte_array != pte)
		{
			kfree(pte_array);
		}

		return eError;
	}
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
	{
		IMG_UINT32 i;
		struct page **pages;

		pages = kmalloc(uiPagesCount * sizeof(struct page *), GFP_KERNEL);
		if (!pages) {
			return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
		}

		for (i = 0; i < uiPagesCount; i++) {
			pPagePA[i].uiAddr = PagePAToPFN(pPagePA[i].uiAddr);
			pages[i] = virt_to_page(PFN_PHYS(pPagePA[i].uiAddr) >> PAGE_SHIFT);
		}

		*ppvLinAddr = vmap(pages, uiPagesCount, VM_MAP, PAGE_KERNEL);
		kfree(pages);

		if (NULL == *ppvLinAddr) {
			return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
		}
		*ppvPrivData = NULL;
		return PVRSRV_OK;
	}
#else
	PVR_DPF((PVR_DBG_ERROR,"%s: Cannot map into kernel, no method supported.", __func__));
	PVR_ASSERT(0);
	*ppvLinAddr = NULL;
	return PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
#endif
#endif // defined(__riscv)
}

void *
OSMapPhysToLin(IMG_CPU_PHYADDR BasePAddr,
			   size_t ui32Bytes,
			   PVRSRV_MEMALLOCFLAGS_T uiMappingFlags)
{
	void __iomem *pvLinAddr;

	if (uiMappingFlags & ~(PVRSRV_MEMALLOCFLAG_CPU_CACHE_MODE_MASK))
	{
		PVR_ASSERT(!"Found non-cpu cache mode flag when mapping to the cpu");
		return NULL;
	}

	if (! PVRSRV_VZ_MODE_IS(NATIVE))
	{
		/*
		  This is required to support DMA physheaps for GPU virtualization.
		  Unfortunately, if a region of kernel managed memory is turned into
		  a DMA buffer, conflicting mappings can come about easily on Linux
		  as the original memory is mapped by the kernel as normal cached
		  memory whilst DMA buffers are mapped mostly as uncached device or
		  cache-coherent device memory. In both cases the system will have
		  two conflicting mappings for the same memory region and will have
		  "undefined behaviour" for most processors notably ARMv6 onwards
		  and some x86 micro-architectures. As a result, perform ioremapping
		  manually for DMA physheap allocations by translating from CPU/VA
		  to BUS/PA thereby preventing the creation of conflicting mappings.
		*/
		pvLinAddr = (void __iomem *) SysDmaDevPAddrToCpuVAddr(BasePAddr.uiAddr, ui32Bytes);
		if (pvLinAddr != NULL)
		{
			return (void __force *) pvLinAddr;
		}
	}

#ifdef CONFIG_FORCE_WC_TO_UNCACHED
	if (uiMappingFlags == PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC)
		uiMappingFlags = PVRSRV_MEMALLOCFLAG_CPU_UNCACHED;
#endif

	switch (uiMappingFlags)
	{
		case PVRSRV_MEMALLOCFLAG_CPU_UNCACHED:
			pvLinAddr = (void __iomem *)ioremap(BasePAddr.uiAddr, ui32Bytes);
			break;
		case PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC:
#if defined(CONFIG_X86) || defined(CONFIG_ARM) || defined(CONFIG_ARM64)
			pvLinAddr = (void __iomem *)ioremap_wc(BasePAddr.uiAddr, ui32Bytes);
#else
			pvLinAddr = (void __iomem *)ioremap(BasePAddr.uiAddr, ui32Bytes);
#endif
			break;
		case PVRSRV_MEMALLOCFLAG_CPU_CACHED:
#if defined(CONFIG_X86) || defined(CONFIG_ARM)
			pvLinAddr = (void __iomem *)ioremap_cache(BasePAddr.uiAddr, ui32Bytes);
#else
			pvLinAddr = (void __iomem *)ioremap(BasePAddr.uiAddr, ui32Bytes);
#endif
			break;
		case PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT:
		case PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT:
			PVR_ASSERT(!"Unexpected cpu cache mode");
			pvLinAddr = NULL;
			break;
		default:
			PVR_ASSERT(!"Unsupported cpu cache mode");
			pvLinAddr = NULL;
			break;
	}

	return (void __force *) pvLinAddr;
}


IMG_BOOL
OSUnMapPhysToLin(void *pvLinAddr, size_t ui32Bytes)
{
	PVR_UNREFERENCED_PARAMETER(ui32Bytes);

	if (!PVRSRV_VZ_MODE_IS(NATIVE))
	{
		if (SysDmaCpuVAddrToDevPAddr(pvLinAddr))
		{
			return IMG_TRUE;
		}
	}

	iounmap((void __iomem *) pvLinAddr);

	return IMG_TRUE;
}

#define OS_MAX_TIMERS	8

/* Timer callback structure used by OSAddTimer */
typedef struct TIMER_CALLBACK_DATA_TAG
{
	IMG_BOOL			bInUse;
	PFN_TIMER_FUNC		pfnTimerFunc;
	void				*pvData;
	struct timer_list	sTimer;
	IMG_UINT32			ui32Delay;
	IMG_BOOL			bActive;
	struct work_struct	sWork;
}TIMER_CALLBACK_DATA;

static struct workqueue_struct *psTimerWorkQueue;

static TIMER_CALLBACK_DATA sTimers[OS_MAX_TIMERS];

static DEFINE_MUTEX(sTimerStructLock);

static void OSTimerCallbackBody(TIMER_CALLBACK_DATA *psTimerCBData)
{
	if (!psTimerCBData->bActive)
		return;

	/* call timer callback */
	psTimerCBData->pfnTimerFunc(psTimerCBData->pvData);

	/* reset timer */
	mod_timer(&psTimerCBData->sTimer, psTimerCBData->sTimer.expires + psTimerCBData->ui32Delay);
}


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
/*************************************************************************/ /*!
@Function       OSTimerCallbackWrapper
@Description    OS specific timer callback wrapper function
@Input          psTimer    Timer list structure
*/ /**************************************************************************/
static void OSTimerCallbackWrapper(struct timer_list *psTimer)
{
	TIMER_CALLBACK_DATA *psTimerCBData = from_timer(psTimerCBData, psTimer, sTimer);
#else
/*************************************************************************/ /*!
@Function       OSTimerCallbackWrapper
@Description    OS specific timer callback wrapper function
@Input          uData    Timer callback data
*/ /**************************************************************************/
static void OSTimerCallbackWrapper(uintptr_t uData)
{
	TIMER_CALLBACK_DATA *psTimerCBData = (TIMER_CALLBACK_DATA*)uData;
#endif
	int res;

	res = queue_work(psTimerWorkQueue, &psTimerCBData->sWork);
	if (res == 0)
	{
		PVR_DPF((PVR_DBG_WARNING, "OSTimerCallbackWrapper: work already queued"));
	}
}


static void OSTimerWorkQueueCallBack(struct work_struct *psWork)
{
	TIMER_CALLBACK_DATA *psTimerCBData = container_of(psWork, TIMER_CALLBACK_DATA, sWork);

	OSTimerCallbackBody(psTimerCBData);
}

IMG_HANDLE OSAddTimer(PFN_TIMER_FUNC pfnTimerFunc, void *pvData, IMG_UINT32 ui32MsTimeout)
{
	TIMER_CALLBACK_DATA *psTimerCBData;
	IMG_UINT32		ui32i;

	/* check callback */
	if (!pfnTimerFunc)
	{
		PVR_DPF((PVR_DBG_ERROR, "OSAddTimer: passed invalid callback"));
		return NULL;
	}

	/* Allocate timer callback data structure */
	mutex_lock(&sTimerStructLock);
	for (ui32i = 0; ui32i < OS_MAX_TIMERS; ui32i++)
	{
		psTimerCBData = &sTimers[ui32i];
		if (!psTimerCBData->bInUse)
		{
			psTimerCBData->bInUse = IMG_TRUE;
			break;
		}
	}
	mutex_unlock(&sTimerStructLock);
	if (ui32i >= OS_MAX_TIMERS)
	{
		PVR_DPF((PVR_DBG_ERROR, "OSAddTimer: all timers are in use"));
		return NULL;
	}

	psTimerCBData->pfnTimerFunc = pfnTimerFunc;
	psTimerCBData->pvData = pvData;
	psTimerCBData->bActive = IMG_FALSE;

	/*
		HZ = ticks per second
		ui32MsTimeout = required ms delay
		ticks = (Hz * ui32MsTimeout) / 1000
	*/
	psTimerCBData->ui32Delay = ((HZ * ui32MsTimeout) < 1000)
								?	1
								:	((HZ * ui32MsTimeout) / 1000);

	/* initialise object */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	timer_setup(&psTimerCBData->sTimer, OSTimerCallbackWrapper, 0);
#else
	init_timer(&psTimerCBData->sTimer);

	/* setup timer object */
	psTimerCBData->sTimer.function = (void *)OSTimerCallbackWrapper;
	psTimerCBData->sTimer.data = (uintptr_t)psTimerCBData;
#endif

	return (IMG_HANDLE)(uintptr_t)(ui32i + 1);
}


static inline TIMER_CALLBACK_DATA *GetTimerStructure(IMG_HANDLE hTimer)
{
	IMG_UINT32 ui32i = (IMG_UINT32)((uintptr_t)hTimer) - 1;

	PVR_ASSERT(ui32i < OS_MAX_TIMERS);

	return &sTimers[ui32i];
}

PVRSRV_ERROR OSRemoveTimer (IMG_HANDLE hTimer)
{
	TIMER_CALLBACK_DATA *psTimerCBData = GetTimerStructure(hTimer);

	PVR_ASSERT(psTimerCBData->bInUse);
	PVR_ASSERT(!psTimerCBData->bActive);

	/* free timer callback data struct */
	psTimerCBData->bInUse = IMG_FALSE;

	return PVRSRV_OK;
}

PVRSRV_ERROR OSEnableTimer (IMG_HANDLE hTimer)
{
	TIMER_CALLBACK_DATA *psTimerCBData = GetTimerStructure(hTimer);

	PVR_ASSERT(psTimerCBData->bInUse);
	PVR_ASSERT(!psTimerCBData->bActive);

	/* Start timer arming */
	psTimerCBData->bActive = IMG_TRUE;

	/* set the expire time */
	psTimerCBData->sTimer.expires = psTimerCBData->ui32Delay + jiffies;

	/* Add the timer to the list */
	add_timer(&psTimerCBData->sTimer);

	return PVRSRV_OK;
}


PVRSRV_ERROR OSDisableTimer (IMG_HANDLE hTimer)
{
	TIMER_CALLBACK_DATA *psTimerCBData = GetTimerStructure(hTimer);

	PVR_ASSERT(psTimerCBData->bInUse);
	PVR_ASSERT(psTimerCBData->bActive);

	/* Stop timer from arming */
	psTimerCBData->bActive = IMG_FALSE;
	smp_mb();

	flush_workqueue(psTimerWorkQueue);

	/* remove timer */
	del_timer_sync(&psTimerCBData->sTimer);

	/*
	 * This second flush is to catch the case where the timer ran
	 * before we managed to delete it, in which case, it will have
	 * queued more work for the workqueue. Since the bActive flag
	 * has been cleared, this second flush won't result in the
	 * timer being rearmed.
	 */
	flush_workqueue(psTimerWorkQueue);

	return PVRSRV_OK;
}


PVRSRV_ERROR OSEventObjectCreate(const IMG_CHAR *pszName, IMG_HANDLE *hEventObject)
{
	PVR_UNREFERENCED_PARAMETER(pszName);

	PVR_LOG_RETURN_IF_INVALID_PARAM(hEventObject, "hEventObject");

	return LinuxEventObjectListCreate(pszName, hEventObject);
}


PVRSRV_ERROR OSEventObjectDestroy(IMG_HANDLE hEventObject)
{
	PVR_LOG_RETURN_IF_INVALID_PARAM(hEventObject, "hEventObject");

	return LinuxEventObjectListDestroy(hEventObject);
}

#define _FREEZABLE IMG_TRUE
#define _NON_FREEZABLE IMG_FALSE

/*
 * EventObjectWaitTimeout()
 */
static PVRSRV_ERROR EventObjectWaitTimeout(IMG_HANDLE hOSEventKM,
										   IMG_UINT64 uiTimeoutus)
{
	PVRSRV_ERROR eError;

	if (hOSEventKM && uiTimeoutus > 0)
	{
		eError = LinuxEventObjectWait(hOSEventKM, uiTimeoutus, _NON_FREEZABLE);
	}
	else
	{
		PVR_DPF((PVR_DBG_ERROR, "OSEventObjectWait: invalid arguments %p, %lld", hOSEventKM, uiTimeoutus));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
	}

	return eError;
}

PVRSRV_ERROR OSEventObjectWaitTimeout(IMG_HANDLE hOSEventKM, IMG_UINT64 uiTimeoutus)
{
	return EventObjectWaitTimeout(hOSEventKM, uiTimeoutus);
}

PVRSRV_ERROR OSEventObjectWait(IMG_HANDLE hOSEventKM)
{
	return OSEventObjectWaitTimeout(hOSEventKM, EVENT_OBJECT_TIMEOUT_US);
}

PVRSRV_ERROR OSEventObjectWaitKernel(IMG_HANDLE hOSEventKM,
                                     IMG_UINT64 uiTimeoutus)
{
	PVRSRV_ERROR eError;

#if defined(PVRSRV_SERVER_THREADS_INDEFINITE_SLEEP)
	if (hOSEventKM)
	{
		if (uiTimeoutus > 0)
			eError = LinuxEventObjectWait(hOSEventKM, uiTimeoutus,
			                              _FREEZABLE);
		else
			eError = LinuxEventObjectWaitUntilSignalled(hOSEventKM);
	}
#else /* defined(PVRSRV_SERVER_THREADS_INDEFINITE_SLEEP) */
	if (hOSEventKM && uiTimeoutus > 0)
	{
		eError = LinuxEventObjectWait(hOSEventKM, uiTimeoutus,
		                              _FREEZABLE);
	}
#endif /* defined(PVRSRV_SERVER_THREADS_INDEFINITE_SLEEP) */
	else
	{
		PVR_DPF((PVR_DBG_ERROR, "OSEventObjectWaitKernel: invalid arguments %p",
		        hOSEventKM));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
	}

	return eError;
}

void OSEventObjectDumpDebugInfo(IMG_HANDLE hOSEventKM)
{
	LinuxEventObjectDumpDebugInfo(hOSEventKM);
}

PVRSRV_ERROR OSEventObjectOpen(IMG_HANDLE hEventObject, IMG_HANDLE *phOSEvent)
{
	PVRSRV_ERROR eError;

	PVR_LOG_RETURN_IF_INVALID_PARAM(phOSEvent, "phOSEvent");
	PVR_LOG_GOTO_IF_INVALID_PARAM(hEventObject, eError, error);

	eError = LinuxEventObjectAdd(hEventObject, phOSEvent);
	PVR_LOG_GOTO_IF_ERROR(eError, "LinuxEventObjectAdd", error);

	return PVRSRV_OK;

error:
	*phOSEvent = NULL;
	return eError;
}

PVRSRV_ERROR OSEventObjectClose(IMG_HANDLE hOSEventKM)
{
	PVR_LOG_RETURN_IF_INVALID_PARAM(hOSEventKM, "hOSEventKM");

	return LinuxEventObjectDelete(hOSEventKM);
}

PVRSRV_ERROR OSEventObjectSignal(IMG_HANDLE hEventObject)
{
	PVR_LOG_RETURN_IF_INVALID_PARAM(hEventObject, "hEventObject");

	return LinuxEventObjectSignal(hEventObject);
}

PVRSRV_ERROR OSCopyToUser(void *pvProcess,
						  void __user *pvDest,
						  const void *pvSrc,
						  size_t ui32Bytes)
{
	PVR_UNREFERENCED_PARAMETER(pvProcess);

	if (pvr_copy_to_user(pvDest, pvSrc, ui32Bytes)==0)
		return PVRSRV_OK;
	else
		return PVRSRV_ERROR_FAILED_TO_COPY_VIRT_MEMORY;
}

PVRSRV_ERROR OSCopyFromUser(void *pvProcess,
							void *pvDest,
							const void __user *pvSrc,
							size_t ui32Bytes)
{
	PVR_UNREFERENCED_PARAMETER(pvProcess);

	if (likely(pvr_copy_from_user(pvDest, pvSrc, ui32Bytes)==0))
		return PVRSRV_OK;
	else
		return PVRSRV_ERROR_FAILED_TO_COPY_VIRT_MEMORY;
}

IMG_UINT64 OSDivide64r64(IMG_UINT64 ui64Divident, IMG_UINT32 ui32Divisor, IMG_UINT32 *pui32Remainder)
{
	*pui32Remainder = do_div(ui64Divident, ui32Divisor);

	return ui64Divident;
}

IMG_UINT32 OSDivide64(IMG_UINT64 ui64Divident, IMG_UINT32 ui32Divisor, IMG_UINT32 *pui32Remainder)
{
	*pui32Remainder = do_div(ui64Divident, ui32Divisor);

	return (IMG_UINT32) ui64Divident;
}

#if defined(SUPPORT_DMA_TRANSFER)
static PVRSRV_ERROR OSDmaEnvInit(void);
#endif

/* One time osfunc initialisation */
PVRSRV_ERROR PVROSFuncInit(void)
{
	{
		PVR_ASSERT(!psTimerWorkQueue);

		psTimerWorkQueue = create_freezable_workqueue("xdxgpu_timer");
		if (psTimerWorkQueue == NULL)
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: couldn't create timer workqueue",
					 __func__));
			return PVRSRV_ERROR_UNABLE_TO_CREATE_THREAD;
		}
	}

	{
		IMG_UINT32 ui32i;

		for (ui32i = 0; ui32i < OS_MAX_TIMERS; ui32i++)
		{
			TIMER_CALLBACK_DATA *psTimerCBData = &sTimers[ui32i];

			INIT_WORK(&psTimerCBData->sWork, OSTimerWorkQueueCallBack);
		}
	}
#if defined(SUPPORT_DMA_TRANSFER)
	OSDmaEnvInit();
#endif
	return PVRSRV_OK;
}

/*
 * Osfunc deinitialisation.
 * Note that PVROSFuncInit may not have been called
 */
void PVROSFuncDeInit(void)
{
	if (psTimerWorkQueue != NULL)
	{
		destroy_workqueue(psTimerWorkQueue);
		psTimerWorkQueue = NULL;
	}
}

void OSDumpStack(void)
{
	dump_stack();
}

PVRSRV_ERROR OSChangeSparseMemCPUAddrMap(void **psPageArray,
                                         IMG_UINT64 sCpuVAddrBase,
                                         IMG_CPU_PHYADDR sCpuPAHeapBase,
                                         IMG_UINT32 ui32AllocPageCount,
                                         IMG_UINT32 *pai32AllocIndices,
                                         IMG_UINT32 ui32FreePageCount,
                                         IMG_UINT32 *pai32FreeIndices,
                                         IMG_BOOL bIsLMA)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
	pfn_t sPFN;
#else
	IMG_UINT64 uiPFN;
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)) */

	PVRSRV_ERROR eError;

	struct mm_struct *psMM = current->mm;
	struct vm_area_struct *psVMA = NULL;
	struct address_space *psMapping = NULL;
	struct page *psPage = NULL;

	IMG_UINT64 uiCPUVirtAddr = 0;
	IMG_UINT32 ui32Loop = 0;
	IMG_UINT32 ui32PageSize = OSGetPageSize();
	IMG_BOOL bMixedMap = IMG_FALSE;

	/*
	 * Acquire the lock before manipulating the VMA
	 * In this case only mmap_sem lock would suffice as the pages associated with this VMA
	 * are never meant to be swapped out.
	 *
	 * In the future, in case the pages are marked as swapped, page_table_lock needs
	 * to be acquired in conjunction with this to disable page swapping.
	 */

	/* Find the Virtual Memory Area associated with the user base address */
	psVMA = find_vma(psMM, (uintptr_t)sCpuVAddrBase);
	if (NULL == psVMA)
	{
		eError = PVRSRV_ERROR_PMR_NO_CPU_MAP_FOUND;
		return eError;
	}

	/* Acquire the memory sem */
	mmap_write_lock(psMM);

	psMapping = psVMA->vm_file->f_mapping;

	/* Set the page offset to the correct value as this is disturbed in MMAP_PMR func */
	psVMA->vm_pgoff = (psVMA->vm_start >>  PAGE_SHIFT);

	/* Delete the entries for the pages that got freed */
	if (ui32FreePageCount && (pai32FreeIndices != NULL))
	{
		for (ui32Loop = 0; ui32Loop < ui32FreePageCount; ui32Loop++)
		{
			uiCPUVirtAddr = (uintptr_t)(sCpuVAddrBase + (pai32FreeIndices[ui32Loop] * ui32PageSize));

			unmap_mapping_range(psMapping, uiCPUVirtAddr, ui32PageSize, 1);

#ifndef PVRSRV_UNMAP_ON_SPARSE_CHANGE
			/*
			 * Still need to map pages in case remap flag is set.
			 * That is not done until the remap case succeeds
			 */
#endif
		}
		eError = PVRSRV_OK;
	}

	if ((psVMA->vm_flags & VM_MIXEDMAP) || bIsLMA)
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
		psVMA->vm_flags |= VM_MIXEDMAP;
#else
		vm_flags_set(psVMA, VM_MIXEDMAP);
#endif
		bMixedMap = IMG_TRUE;
	}
	else
	{
		if (ui32AllocPageCount && (NULL != pai32AllocIndices))
		{
			for (ui32Loop = 0; ui32Loop < ui32AllocPageCount; ui32Loop++)
			{

				psPage = (struct page *)psPageArray[pai32AllocIndices[ui32Loop]];
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
				sPFN = page_to_pfn_t(psPage);

				if (!pfn_t_valid(sPFN) || page_count(pfn_t_to_page(sPFN)) == 0)
#else
				uiPFN = page_to_pfn(psPage);

				if (!pfn_valid(uiPFN) || (page_count(pfn_to_page(uiPFN)) == 0))
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)) */
				{
					bMixedMap = IMG_TRUE;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
					psVMA->vm_flags |= VM_MIXEDMAP;
#else
					vm_flags_set(psVMA, VM_MIXEDMAP);
#endif
					break;
				}
			}
		}
	}

	/* Map the pages that got allocated */
	if (ui32AllocPageCount && (NULL != pai32AllocIndices))
	{
		for (ui32Loop = 0; ui32Loop < ui32AllocPageCount; ui32Loop++)
		{
			int err;

			uiCPUVirtAddr = (uintptr_t)(sCpuVAddrBase + (pai32AllocIndices[ui32Loop] * ui32PageSize));
			unmap_mapping_range(psMapping, uiCPUVirtAddr, ui32PageSize, 1);

			if (bIsLMA)
			{
				phys_addr_t uiAddr = sCpuPAHeapBase.uiAddr +
				                     ((IMG_DEV_PHYADDR *)psPageArray)[pai32AllocIndices[ui32Loop]].uiAddr;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
				sPFN = phys_to_pfn_t(uiAddr, 0);
				psPage = pfn_t_to_page(sPFN);
#else
				uiPFN = uiAddr >> PAGE_SHIFT;
				psPage = pfn_to_page(uiPFN);
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)) */
			}
			else
			{
				psPage = (struct page *)psPageArray[pai32AllocIndices[ui32Loop]];
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
				sPFN = page_to_pfn_t(psPage);
#else
				uiPFN = page_to_pfn(psPage);
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0)) */
			}

			if (bMixedMap)
			{
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)) || (defined(REDHAT_83)))
				vm_fault_t vmf;

				vmf = vmf_insert_mixed(psVMA, uiCPUVirtAddr, sPFN);
				if (vmf & VM_FAULT_ERROR)
				{
					err = vm_fault_to_errno(vmf, 0);
				}
				else
				{
					err = 0;
				}
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 0))
				err = vm_insert_mixed(psVMA, uiCPUVirtAddr, sPFN);
#else
				err = vm_insert_mixed(psVMA, uiCPUVirtAddr, uiPFN);
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) */
			}
			else
			{
				err = vm_insert_page(psVMA, uiCPUVirtAddr, psPage);
			}

			if (err)
			{
				PVR_DPF((PVR_DBG_MESSAGE, "Remap failure error code: %d", err));
				eError = PVRSRV_ERROR_PMR_CPU_PAGE_MAP_FAILED;
				goto eFailed;
			}
		}
	}

	eError = PVRSRV_OK;
eFailed:
	mmap_write_unlock(psMM);

	return eError;
}

/*************************************************************************/ /*!
@Function       OSDebugSignalPID
@Description    Sends a SIGTRAP signal to a specific PID in user mode for
                debugging purposes. The user mode process can register a handler
                against this signal.
                This is necessary to support the Rogue debugger. If the Rogue
                debugger is not used then this function may be implemented as
                a stub.
@Input          ui32PID    The PID for the signal.
@Return         PVRSRV_OK on success, a failure code otherwise.
*/ /**************************************************************************/
PVRSRV_ERROR OSDebugSignalPID(IMG_UINT32 ui32PID)
{
	int err;
	struct pid *psPID;

	psPID = find_vpid(ui32PID);
	if (psPID == NULL)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: Failed to get PID struct.", __func__));
		return PVRSRV_ERROR_NOT_FOUND;
	}

	err = kill_pid(psPID, SIGTRAP, 0);
	if (err != 0)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: Signal Failure %d", __func__, err));
		return PVRSRV_ERROR_SIGNAL_FAILED;
	}

	return PVRSRV_OK;
}

/*************************************************************************/ /*!
@Function       OSIsKernelThread
@Description    This API determines if the current running thread is a kernel
                thread (i.e. one not associated with any userland process,
                typically an MISR handler.)
@Return         IMG_TRUE if it is a kernel thread, otherwise IMG_FALSE.
*/ /**************************************************************************/
IMG_BOOL OSIsKernelThread(void)
{
	/*
	 * Kernel threads have a NULL memory descriptor.
	 *
	 * See https://www.kernel.org/doc/Documentation/vm/active_mm.txt
	 */
	return current->mm == NULL;
}

void OSDumpVersionInfo(DUMPDEBUG_PRINTF_FUNC *pfnDumpDebugPrintf,
					   void *pvDumpDebugFile)
{
	PVR_DUMPDEBUG_LOG("OS kernel info: %s %s %s %s",
					utsname()->sysname,
					utsname()->release,
					utsname()->version,
					utsname()->machine);
}

#if defined(SUPPORT_DMA_TRANSFER)
typedef struct _OSDMADescPages_ {
	struct page **pages;
	IMG_UINT32 uiDescNumPages;
	IMG_UINT32 uiSgCnt;
	struct sg_table *psSg;
	IMG_UINT64 uiOffset;
	IMG_UINT64 ulMapSize;
	IMG_BOOL bIsFormUser;
}OSDMADescPages;

typedef struct _OSDMADesc_
{
	struct list_head List;
	struct list_head PList;
	struct device *psDmaDevice;
	IMG_UINT64 uiSize;
	IMG_UINT uDmaCookie;
	void *pvDmaDataParent;
	struct dma_chan *pDmaChan;
	OSDMADescPages tPages;
	OSDMADescPages tPages_dest;
	struct dma_async_tx_descriptor* psDescriptors;
} OSDMADesc;

typedef struct _OSDMAEntry_
{
	struct list_head List;
	struct list_head lhDescHead;
	void *pvHandleParent;
	IMG_UINT32 uiNumPages;
	IMG_UINT64 uiSize;
	IMG_UINT uiNumDescriptor;
	IMG_UINT aDescDone;
	/* Sparse PMR transfer information */
	IMG_BOOL bIsSparse;
	IMG_BOOL bTest;
//	OSDMADesc DTxDesc;
} OSDMAEntry;

typedef struct _OSDMATranferHandle_
{
	IMG_BOOL bSucceed;
	IMG_BOOL bIsFromUser;
	IMG_BOOL bAdvanceTimeline;
	IMG_UINT uiEntryMax;
	IMG_UINT uiRefCount;
	IMG_UINT uiCount;
	enum dma_transfer_direction eDirection;

	PVRSRV_DEVICE_NODE *psDevNode;
	PFN_SERVER_CLEANUP pfnServerCleanup;
	void* pvServerCleanupData;
	struct work_struct sCleanWork;
	struct list_head lReadyListHead;
	struct list_head lSubmitListHead;
	struct list_head lDoneListHead;
	struct list_head lDoneCleanupHead;
	struct list_head lhEntryHead;
	spinlock_t DoneListLock;
	atomic_t aPrepCount;
	atomic_t aDoneCount;
	atomic_t bAllPrepFlag;
	struct completion start_cleanup;
	OSDMAEntry sDMAEntry[0];
} OSDMATranferHandle;

static struct workqueue_struct *psDmaCleanUpWorkQueue;
#if defined(SUPPORT_DMA_TRANSFER)
static PVRSRV_ERROR OSDmaEnvInit(void)
{
	psDmaCleanUpWorkQueue = create_workqueue("xdxgpu_dma");
	if (IS_ERR_OR_NULL(psDmaCleanUpWorkQueue)) {
		PVR_DPF((PVR_DBG_ERROR, "Creat DMA queue failed!(%ld)", PTR_ERR(psDmaCleanUpWorkQueue)));
		psDmaCleanUpWorkQueue = NULL;
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}
	return PVRSRV_OK;
}
#endif

#define XDXDMA_SG_SEGMENT_MAX  (PAGE_SIZE / 32)
//#define XDXDMA_SG_SEGMENT_MAX  (0)

static int OSDmaRequestHostDmaAddr(
	mempool_t *pMemPool,
	OSDMADesc *pDmaTxDesc,
	void *puiAddress,
	IMG_UINT64 ulMaxSize,
	OSDMADescPages *tPages,
	bool bIsFormUser,
	bool bSegReq,
	enum dma_transfer_direction direction
)
{
	int iRet = 0;
	IMG_UINT64 offset = ((IMG_UINT64)puiAddress) & (PAGE_SIZE - 1);
	unsigned int num_pages = (ulMaxSize + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	int num_pinned_pages;
	IMG_UINT64 ulMapedSize;
	enum dma_data_direction datadir = DMA_NONE;
	unsigned int gup_flags = 0;

	if (!pDmaTxDesc || !tPages || !pDmaTxDesc->psDmaDevice)
		return -EINVAL;
	tPages->psSg = mempool_alloc(pMemPool, GFP_NOWAIT);
	if (!tPages->psSg)
		return -ENOMEM;

	if (direction == DMA_DEV_TO_MEM  || direction == DMA_MEM_TO_MEM) {
		gup_flags = FOLL_WRITE;
		datadir = DMA_FROM_DEVICE;
	} else if (direction == DMA_MEM_TO_DEV) {
	    datadir = DMA_TO_DEVICE;
	} else {
	    datadir = DMA_BIDIRECTIONAL;
	}

#if XDXDMA_SG_SEGMENT_MAX
	if (bSegReq)
		num_pages = MIN(num_pages, XDXDMA_SG_SEGMENT_MAX);
#endif
	ulMapedSize = (num_pages << PAGE_SHIFT) - offset;
	ulMapedSize = MIN(ulMaxSize, ulMapedSize);
	tPages->pages = OSAllocZMem(num_pages * sizeof(struct page *));
	if (!tPages->pages) {
		iRet = -ENOMEM;
		goto e0;
	}
	if (!ulMapedSize) {
		dev_err(pDmaTxDesc->psDmaDevice,
				"[%s]ERR:puiAddress:%lx ulMaxSize:%llx ulMapedSize:%llx offset:%llx, num_pages:%x",
				__func__, (unsigned long)puiAddress, ulMaxSize, ulMapedSize, offset, num_pages);
		iRet = -ENOMEM;
		goto e1;
	}

	tPages->ulMapSize = ulMapedSize;
	tPages->bIsFormUser = bIsFormUser;
	tPages->uiOffset = offset;
	if (bIsFormUser)
	{
		unsigned long tmp_addr;
		int iRequestNmu = num_pages;
		IMG_INT32 iRetry = 0;

		num_pinned_pages = 0;
		tmp_addr = (unsigned long)puiAddress;
		do {
			iRet = get_user_pages_fast(tmp_addr, (int)iRequestNmu, gup_flags,
						&tPages->pages[num_pinned_pages]);
			if (iRet < 0) {
				if (current->mm != NULL)
				{
					dev_err(pDmaTxDesc->psDmaDevice,
						"[%d-%s] - Pin User page[Saddr:%lx-%lx, num_pinned_pages:+%d] failed.%d",
						current->pid, current->comm, (unsigned long)puiAddress, tmp_addr, num_pinned_pages, iRet);
				} else {
					dev_err(pDmaTxDesc->psDmaDevice, "Pin User page[Saddr:%lx-%lx, num_pinned_pages:+%d] failed.%d",
						(unsigned long)puiAddress, tmp_addr, num_pinned_pages, iRet);
				}
			}else {
				num_pinned_pages += iRet;
				tmp_addr += (unsigned long) PAGE_SIZE * iRet;
				iRequestNmu -= iRet;
			}
		}  while (iRetry++ < 5 && iRequestNmu > 0);
	} else {
		IMG_UINT32 i;
		void* tmp_dma_addr;
		tmp_dma_addr = (void*)puiAddress;
		if (is_vmalloc_addr(puiAddress)) {
			for(i = 0;i < num_pages; i++) {
				tPages->pages[i] = vmalloc_to_page(tmp_dma_addr);
				tmp_dma_addr += PAGE_SIZE;
			}
		} else {
			for(i = 0;i < num_pages; i++) {
				tPages->pages[i] = virt_to_page((IMG_UINT64*)tmp_dma_addr);
				tmp_dma_addr += PAGE_SIZE;
			}
		}
		num_pinned_pages = num_pages;
	}
	tPages->uiDescNumPages = num_pinned_pages;
	if (!num_pinned_pages || num_pinned_pages != num_pages) {
		dev_err(pDmaTxDesc->psDmaDevice, "Get Pages failed:(%d/%u)",
			num_pinned_pages, num_pages);
		iRet = -ENOMEM;
		goto e1;
	}

	iRet = sg_alloc_table_from_pages(tPages->psSg, tPages->pages,
				num_pinned_pages, tPages->uiOffset, tPages->ulMapSize, GFP_KERNEL);
	if (iRet < 0) {
		dev_err(pDmaTxDesc->psDmaDevice, "sg_alloc_table_from_pages failed.%d", iRet);
		goto e2;
	}
	iRet = dma_map_sg(pDmaTxDesc->psDmaDevice, tPages->psSg->sgl, tPages->psSg->orig_nents, datadir);
	if (!iRet) {
		dev_err(pDmaTxDesc->psDmaDevice, "%s: Error mapping SG list", __func__);
		iRet = -EINVAL;
		goto e3;
	}

	if (direction == DMA_MEM_TO_MEM || direction == DMA_MEM_TO_DEV) {
		dma_sync_sg_for_cpu(pDmaTxDesc->psDmaDevice,
						tPages->psSg->sgl,
						tPages->psSg->nents,
						DMA_TO_DEVICE);
	}

	tPages->uiSgCnt = iRet;
	return iRet;
e3:
	dma_unmap_sg(pDmaTxDesc->psDmaDevice, tPages->psSg->sgl, tPages->psSg->nents, datadir);
e2:
	sg_free_table(tPages->psSg);
	if (bIsFormUser)
	{
		IMG_UINT32 i;
		/* Unpin pages */
		for (i  = 0; i < tPages->uiDescNumPages; i++)
		{
			put_page(tPages->pages[i]);
		}
	}
e1:
	OSFreeMem(tPages->pages);
	tPages->pages = NULL;
e0:
	mempool_free(tPages->psSg, pMemPool);
	tPages->psSg = NULL;
	return iRet;
}

static int OSDmaReleaseHostDmaAddr(
	mempool_t *pMemPool,
	OSDMADesc *pDmaTxDesc,
	OSDMADescPages *tPages,
	enum dma_transfer_direction direction
)
{
	enum dma_data_direction datadir = DMA_NONE;

	if (!tPages->psSg || !tPages->pages || !tPages->ulMapSize)
		return -EINVAL;

	if (direction == DMA_DEV_TO_MEM || direction == DMA_MEM_TO_MEM) {
		datadir = DMA_FROM_DEVICE;
	} else if (direction == DMA_MEM_TO_DEV) {
	    datadir = DMA_TO_DEVICE;
	} else {
	    datadir = DMA_BIDIRECTIONAL;
	}

	dma_unmap_sg(pDmaTxDesc->psDmaDevice, tPages->psSg->sgl, tPages->psSg->orig_nents, datadir);
	sg_free_table(tPages->psSg);
	mempool_free(tPages->psSg, pMemPool);
	tPages->psSg = NULL;
	if (tPages->bIsFormUser)
	{
		IMG_UINT32 i;
		for (i = 0; i < tPages->uiDescNumPages; i++)
		{
			if (direction == DMA_DEV_TO_MEM || direction == DMA_MEM_TO_MEM)
				set_page_dirty_lock(tPages->pages[i]);
			put_page(tPages->pages[i]);
		}
	}
	OSFreeMem(tPages->pages);
	tPages->pages = NULL;
	return 0;
}

static void OSDmaTransferCleanupDone(OSDMATranferHandle *psHandle)
{
	if (atomic_read(&psHandle->bAllPrepFlag) == IMG_TRUE &&
		(atomic_read(&psHandle->aDoneCount) == atomic_read(&psHandle->aPrepCount)))
	{
		complete(&psHandle->start_cleanup);
	}
	return;
}

static void OSDmaDescRelease(
			OSDMATranferHandle *psOSDMAHandle,
			OSDMADesc *psTxDesc, IMG_BOOL bFlush)
{
	OSDmaReleaseHostDmaAddr(psOSDMAHandle->psDevNode->sg_pool, psTxDesc,
				&psTxDesc->tPages, psOSDMAHandle->eDirection);
	OSDmaReleaseHostDmaAddr(psOSDMAHandle->psDevNode->sg_pool, psTxDesc,
				&psTxDesc->tPages_dest, psOSDMAHandle->eDirection);
	return;
}

static void OSDmaTransferCleanupWork(struct work_struct *work)
{
	OSDMATranferHandle *psOSDMAHandle =
			container_of(work, OSDMATranferHandle, sCleanWork);
	LIST_HEAD(sHead);
	OSDMADesc *psTxDesc, *psTempDesc;

	spin_lock_bh(&psOSDMAHandle->DoneListLock);
	list_splice_tail_init(&psOSDMAHandle->lDoneListHead, &sHead);
	spin_unlock_bh(&psOSDMAHandle->DoneListLock);
	list_for_each_entry_safe(psTxDesc, psTempDesc, &sHead, List)
	{
		OSDmaDescRelease(psOSDMAHandle, psTxDesc, IMG_TRUE);
		list_move_tail(&psTxDesc->List, &psOSDMAHandle->lDoneCleanupHead);
//		atomic_inc(&psOSDMAHandle->aDoneCount);
	}
//	OSDmaTransferCleanupDone(psOSDMAHandle);
}

static void OSDmaTxDescDoneCb(void *pvsTxDesc, const struct dmaengine_result *dmaResult)
{
	OSDMADesc *psTxDesc = (OSDMADesc*)pvsTxDesc;
	OSDMAEntry *psOSDMAEntry =
			(OSDMAEntry *)psTxDesc->pvDmaDataParent;
	OSDMATranferHandle *psOSDMAHandle =
			(OSDMATranferHandle *)psOSDMAEntry->pvHandleParent;
	unsigned long ulIrqFlags;

	if (dmaResult->result != DMA_TRANS_NOERROR)
		PVR_DPF((PVR_DBG_ERROR, "dmaengine_result.result : %d", dmaResult->result));
//	spin_lock(&psOSDMAHandle->DoneListLock);
	spin_lock_irqsave(&psOSDMAHandle->DoneListLock, ulIrqFlags);
	psTxDesc->psDescriptors = NULL;
	list_move_tail(&psTxDesc->List, &psOSDMAHandle->lDoneListHead);
//	spin_unlock(&psOSDMAHandle->DoneListLock);
	spin_unlock_irqrestore(&psOSDMAHandle->DoneListLock, ulIrqFlags);
	atomic_inc(&psOSDMAHandle->aDoneCount);
	OSDmaTransferCleanupDone(psOSDMAHandle);
//	if (psDmaCleanUpWorkQueue)
//		queue_work(psDmaCleanUpWorkQueue, &psOSDMAHandle->sCleanWork);
//	else
//		schedule_work(&psOSDMAHandle->sCleanWork);
	return;
}

static PVRSRV_ERROR OSDmaTriggerTransfer(OSDMATranferHandle *psOSDMAHandle, struct dma_chan *pvChan)
{
	OSDMADesc *TxDesc, *TempDesc;
	LIST_HEAD(sHead);
	int ret;

	if (list_empty(&psOSDMAHandle->lReadyListHead))
		return PVRSRV_OK;

	list_splice_tail_init(&psOSDMAHandle->lReadyListHead, &sHead);
	list_for_each_entry_safe(TxDesc, TempDesc, &sHead, List)
	{
		spin_lock_bh(&psOSDMAHandle->DoneListLock);
		list_move_tail(&TxDesc->List, &psOSDMAHandle->lSubmitListHead);
		spin_unlock_bh(&psOSDMAHandle->DoneListLock);
		TxDesc->uDmaCookie = dmaengine_submit(TxDesc->psDescriptors);
		ret = dma_submit_error(TxDesc->uDmaCookie);
		if (ret)
		{
			PVR_DPF((PVR_DBG_ERROR, "Dma Sumbumit Error.%d", ret));
			OSDmaDescRelease(psOSDMAHandle, TxDesc, IMG_FALSE);
			return PVRSRV_ERROR_NOT_IMPLEMENTED;
		}
	}
	dma_async_issue_pending(pvChan);
	return PVRSRV_OK;
}

static PVRSRV_ERROR OSDmaCancelTransfer(OSDMATranferHandle *psOSDMAHandle)
{
	OSDMADesc *TxDesc;
	LIST_HEAD(sHead);

	spin_lock_irq(&psOSDMAHandle->DoneListLock);
	list_splice_tail(&psOSDMAHandle->lReadyListHead, &sHead);
	list_splice_tail(&psOSDMAHandle->lSubmitListHead, &sHead);
	list_for_each_entry(TxDesc, &sHead, List) {
		TxDesc->psDescriptors->callback = NULL;
		TxDesc->psDescriptors->callback_param = NULL;
	}
	spin_unlock_irq(&psOSDMAHandle->DoneListLock);
	return PVRSRV_OK;
}
static PVRSRV_ERROR OSDmaDescEnQueueReady(
		OSDMATranferHandle *psOSDMAHandle,
		OSDMADesc *psTxDesc)
{
	PVRSRV_ERROR eError = PVRSRV_OK;

	if (!psOSDMAHandle || !psTxDesc)
		return PVRSRV_ERROR_INVALID_PARAMS;
//	spin_lock(&psOSDMAHandle->DoneListLock);
	list_add_tail(&psTxDesc->List , &psOSDMAHandle->lReadyListHead);
//	spin_unlock(&psOSDMAHandle->DoneListLock);
	atomic_inc(&psOSDMAHandle->aPrepCount);
	eError = OSDmaTriggerTransfer(psOSDMAHandle, psTxDesc->pDmaChan);
	return eError;
}

#if defined(SUPPORT_VALIDATION) && defined(PVRSRV_DEBUG_DMA)
static void
DMADumpPhysicalAddresses(struct page **ppsHostMemPages,
						 IMG_UINT32 uiNumPages,
						 IMG_DMA_ADDR *sDmaAddr,
						 IMG_UINT64 ui64Offset)
{
	IMG_CPU_PHYADDR sPagePhysAddr;
	IMG_UINT32 uiIdx;

	PVR_DPF((PVR_DBG_MESSAGE, "DMA Transfer Address Dump"));
	PVR_DPF((PVR_DBG_MESSAGE, "Hostmem phys addresses:"));

	for (uiIdx = 0; uiIdx < uiNumPages; uiIdx++)
	{
		sPagePhysAddr.uiAddr = page_to_phys(ppsHostMemPages[uiIdx]);
		if (uiIdx == 0)
		{
			sPagePhysAddr.uiAddr += ui64Offset;
			PVR_DPF((PVR_DBG_MESSAGE, "\tHost mem start at 0x%llX", sPagePhysAddr.uiAddr));
		}
		else
		{
			PVR_DPF((PVR_DBG_MESSAGE, "\tHost Mem Page %d at 0x%llX", uiIdx, sPagePhysAddr.uiAddr));
		}
	}
	PVR_DPF((PVR_DBG_MESSAGE, "Devmem CPU phys address: 0x%llX",
			sDmaAddr->uiAddr));
}
#endif

void OSDmaForceCleanup(void *pvChan, void *pvOSData)
{
	OSDMATranferHandle *psOSDMAHandle = (OSDMATranferHandle *)pvOSData;
	IMG_UINT ui32Retries;

	pr_err("OSDmaForceCleanup ...\n");
	psOSDMAHandle->bSucceed = IMG_FALSE;
	psOSDMAHandle->bAdvanceTimeline = IMG_TRUE;

	/* Need to wait for outstanding DMA Engine ops before advancing the
	   user-supplied timeline in case of error. dmaengine_terminate_sync
	   cannot be called from within atomic context, so cannot invoke it
	   from inside the cleanup kernel thread. */
	cancel_work_sync(&psOSDMAHandle->sCleanWork);
	for (ui32Retries = 0; ui32Retries < DMA_ERROR_SYNC_RETRIES; ui32Retries++)
	{
		if (dmaengine_terminate_sync(pvChan) == 0)
		{
			break;
		}
	}
	if (ui32Retries == DMA_ERROR_SYNC_RETRIES)
	{
		/* We cannot guarantee all outstanding DMAs were terminated
		 * so we let the UM fence time out as a fallback mechanism */
		psOSDMAHandle->bAdvanceTimeline = IMG_FALSE;
	}

	complete(&psOSDMAHandle->start_cleanup);

}

static PVRSRV_ERROR OSDmaCleanTransfer(
		OSDMATranferHandle *psOSDMAHandle,
		OSDMAEntry *psOSDMAEntry)
{
#if 0
	return PVRSRV_OK;

#else
	OSDMADesc *psDesc, *pDescTemp;

	list_for_each_entry_safe(psDesc, pDescTemp, &psOSDMAEntry->lhDescHead, PList)
	{
		list_del(&psDesc->PList);
		OSFreeMem(psDesc);
	}
	return PVRSRV_OK;
#endif
}

static PVRSRV_ERROR OSDmaCleanHandle(OSDMATranferHandle *psOSDMAHandle)
{
	OSDMAEntry *pEntry, *pTempEntry;
	OSDMADesc *psDesc, *pstDesc;
	int iErrorCount = 0;
	LIST_HEAD(sHead);
	LIST_HEAD(sClHead);

	cancel_work_sync(&psOSDMAHandle->sCleanWork);
	spin_lock_bh(&psOSDMAHandle->DoneListLock);
	list_splice_tail_init(&psOSDMAHandle->lSubmitListHead, &sHead);
	list_for_each_entry_safe(psDesc, pstDesc, &sHead, List)
	{
		psDesc->psDescriptors->callback = NULL;
		psDesc->psDescriptors->callback_result = NULL;
	}
	list_splice_tail_init(&psOSDMAHandle->lDoneListHead, &sHead);
	spin_unlock_bh(&psOSDMAHandle->DoneListLock);
//	spin_unlock_irqrestore(&psOSDMAHandle->DoneListLock, ulIrqFlags);

	list_splice_tail_init(&psOSDMAHandle->lReadyListHead, &sHead);
	if (!list_empty(&sHead))
		PVR_DPF((PVR_DBG_ERROR, "[%s]Dma TxDescriptor exception Release (%d/%d)",
			__func__, atomic_read(&psOSDMAHandle->aDoneCount),
			atomic_read(&psOSDMAHandle->aPrepCount)));
	list_for_each_entry_safe(psDesc, pstDesc, &sHead, List)
	{
		OSDmaDescRelease(psOSDMAHandle, psDesc, IMG_FALSE);
		iErrorCount++;
	}
	list_for_each_entry_safe(pEntry, pTempEntry, &psOSDMAHandle->lhEntryHead, List)
	{
		list_del(&pEntry->List);
		OSDmaCleanTransfer(psOSDMAHandle, pEntry);
		OSFreeMem(pEntry);
	}
	return PVRSRV_OK;
}

static IMG_UINT32 OSDmaGetContinuesPagesCnt(
		IMG_DMA_ADDR *psDmaAddrs,
		IMG_BOOL *pDmaPageVailds,
		IMG_UINT32 ui32DmaPages,
		IMG_UINT64 ulDmaPageSize,
		IMG_BOOL *pbSkiped)
{
	IMG_INT32 i;
	IMG_UINT32 ui32ContinuesCnt = 1;

	if (ui32DmaPages < 1)
		return 0;

	*pbSkiped = IMG_FALSE;
	for (i = 0; i < ui32DmaPages; i++)
	{
		if (!pDmaPageVailds[i])
		{
			*pbSkiped = IMG_TRUE;
			return 0;
		}
		if ((i + 1) == ui32DmaPages || !pDmaPageVailds[i + 1]  ||
				(psDmaAddrs[i].uiAddr + ulDmaPageSize) != psDmaAddrs[i + 1].uiAddr)
		{
			return ui32ContinuesCnt;
		}
		ui32ContinuesCnt++;
	}
	return ui32ContinuesCnt;
}


#ifdef SUPPORT_DMA_TRANSFER_B0
static PVRSRV_ERROR OSDMAPrepTwiddle2D3D(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		DMA_TWD_2D3D_PARA *pt3DParas,
		IMG_UINT32 uiFlags,
		enum dma_transfer_direction direction,
		OSDMAEntry *psOSDMAEntry
)
{
	struct device *dma_dev = ((PVRSRV_RGXDEV_INFO	*)(psDevNode->pvDevice))->dma_dev;
	struct dma_async_tx_descriptor *psDesc;
	OSDMADesc *pDmaTxDesc;
	PVRSRV_ERROR eError = PVRSRV_OK;
	int iRet;
	bool bIsFromUser = !!(uiFlags | DMA_FLAG_FROM_USER);

	pDmaTxDesc = OSAllocZMem(sizeof(OSDMADesc));
	if (!pDmaTxDesc)
	{
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}
	pDmaTxDesc->psDmaDevice = dma_dev;
	pDmaTxDesc->pDmaChan = pvChan;
	pDmaTxDesc->pvDmaDataParent = psOSDMAEntry;
	switch (direction) {
		case DMA_MEM_TO_DEV:
			iRet = OSDmaRequestHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc,
				(void *)pt3DParas->SrcAddr, pt3DParas->Size,
				&pDmaTxDesc->tPages, bIsFromUser, 0, direction);
			if (iRet <= 0)
			{
				PVR_LOG_GOTO_WITH_ERROR("OSDmaRequestHostDmaAddr", eError,
							PVRSRV_ERROR_BAD_MAPPING, e0);
			}
			if (iRet != 1)
			{
				// the dma addr must be contig while 2D/3D copy
				PVR_DPF((PVR_DBG_ERROR,
					"Src host addr is not continuous, maybe not enable IOMMU."));
				eError = PVRSRV_ERROR_INVALID_PARAMS;
				goto e1;
			}
			pt3DParas->SrcAddr = sg_dma_address(pDmaTxDesc->tPages.psSg->sgl);
			pt3DParas->SrcAddrFlag = DMA_ADDR_FLAG_HOST_PORT;
			pDmaTxDesc->uiSize = pDmaTxDesc->tPages.ulMapSize;
			break;
		case DMA_DEV_TO_DEV:
			pDmaTxDesc->uiSize = pt3DParas->Size;
			break;
		case DMA_DEV_TO_MEM:
		case DMA_MEM_TO_MEM:
			/*Not support M2M.*/
		default:
			PVR_DPF((PVR_DBG_ERROR, "[%s]%d:invalid direction\n", __func__, __LINE__));
			PVR_GOTO_WITH_ERROR(eError, PVRSRV_ERROR_INVALID_PARAMS, e0);
			break;
	}

	psDesc = xdx_prep_twd_2d3d_memcpy(pvChan, pt3DParas, direction, 0);
	if (IS_ERR_OR_NULL(psDesc))
	{
		PVR_DPF((PVR_DBG_ERROR,
			"xdx_prep_twd_2d3d_memcpy failed. %ld.", PTR_ERR(psDesc)));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e1;
	}
	psDesc->callback_param = pDmaTxDesc;
	psDesc->callback_result = OSDmaTxDescDoneCb;
	pDmaTxDesc->psDescriptors = psDesc;
	list_add_tail(&pDmaTxDesc->PList, &psOSDMAEntry->lhDescHead);
	OSDmaDescEnQueueReady(psOSDMAEntry->pvHandleParent, pDmaTxDesc);
	return PVRSRV_OK;
e1:
	OSDmaDescRelease(psOSDMAEntry->pvHandleParent, pDmaTxDesc, 0);
e0:
	OSFreeMem(pDmaTxDesc);
	PVR_DPF((PVR_DBG_ERROR, "[%s]%d:Error:%d.\n", __func__, __LINE__, eError));
	return eError;
}

static PVRSRV_ERROR OSDMAPrep2DCopy(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		DMA_XMIT_2D_PARA *pt2DParas,
		enum dma_transfer_direction direction,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry)
{
	struct device *dma_dev = ((PVRSRV_RGXDEV_INFO	*)(psDevNode->pvDevice))->dma_dev;
	struct dma_async_tx_descriptor *psDesc;
	OSDMADesc *pDmaTxDesc;
	PVRSRV_ERROR eError = PVRSRV_OK;
	int iRet;
	bool bIsFromUser = !!(uiFlags | DMA_FLAG_FROM_USER);

	pDmaTxDesc = OSAllocZMem(sizeof(OSDMADesc));
	if (!pDmaTxDesc)
	{
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}
	pDmaTxDesc->psDmaDevice = dma_dev;
	pDmaTxDesc->pDmaChan = pvChan;
	pDmaTxDesc->pvDmaDataParent = psOSDMAEntry;
	switch (direction) {
		case DMA_MEM_TO_DEV:
			iRet = OSDmaRequestHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc,
				(void *)pt2DParas->SrcAddr, pt2DParas->Size,
				&pDmaTxDesc->tPages, bIsFromUser, 0, direction);
			if (iRet <= 0)
			{
				PVR_LOG_GOTO_WITH_ERROR("OSDmaRequestHostDmaAddr", eError,
							PVRSRV_ERROR_BAD_MAPPING, e0);
			}
			if (iRet != 1)
			{
				// the dma addr must be contig while 2D/3D copy
				PVR_DPF((PVR_DBG_ERROR,
					"Dest host addr is not continuous, maybe not enable IOMMU."));
				eError = PVRSRV_ERROR_INVALID_PARAMS;
				goto e1;
			}
			pt2DParas->SrcAddr = sg_dma_address(pDmaTxDesc->tPages.psSg->sgl);
			pt2DParas->SrcAddrFlag = DMA_ADDR_FLAG_HOST_PORT;
			pDmaTxDesc->uiSize = pDmaTxDesc->tPages.ulMapSize;
			break;
		case DMA_DEV_TO_DEV:
			pDmaTxDesc->uiSize = pt2DParas->Size;
			break;
		case DMA_DEV_TO_MEM:
		case DMA_MEM_TO_MEM:
		default:
			PVR_DPF((PVR_DBG_ERROR, "[%s]%d:invalid direction\n", __func__, __LINE__));
			PVR_GOTO_WITH_ERROR(eError, PVRSRV_ERROR_INVALID_PARAMS, e0);
			break;
	}

	psDesc = xdx_prep_2d_memcpy(pvChan, pt2DParas, direction, 0);
	if (IS_ERR_OR_NULL(psDesc))
	{
		PVR_DPF((PVR_DBG_ERROR,
			"xdx_prep_2d_memcpy failed. %ld.", PTR_ERR(psDesc)));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e1;
	}
	psDesc->callback_param = pDmaTxDesc;
	psDesc->callback_result = OSDmaTxDescDoneCb;
	pDmaTxDesc->psDescriptors = psDesc;
	list_add(&pDmaTxDesc->PList, &psOSDMAEntry->lhDescHead);
	OSDmaDescEnQueueReady(psOSDMAEntry->pvHandleParent, pDmaTxDesc);
	return PVRSRV_OK;
e1:
	OSDmaDescRelease(psOSDMAEntry->pvHandleParent, pDmaTxDesc, 0);
e0:
	OSFreeMem(pDmaTxDesc);
	PVR_DPF((PVR_DBG_ERROR, "[%s]%d:Error:%d.\n", __func__, __LINE__, eError));
	return eError;
}
#endif

/*************************************************************************/
/*!
@Function       OSDmaPrepareTransferSegment
@Description    This API is used to ask OS to perform a DMA transfer operation
@Return  <= 0 failed.
		 >  0  prepare success scatter list count.
*/
/**************************************************************************/
static PVRSRV_ERROR OSDmaPrepCpySgDevHostSeg(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		IMG_DMA_ADDR *psDmaAddr,
		IMG_UINT64 *puiAddress,
		IMG_UINT64 uiMaxSize,
		enum dma_transfer_direction direction,
		OSDMAEntry *psOSDMAEntry,
		IMG_UINT32 uiFlags,
		IMG_UINT64 *uiRetSize)
{
	OSDMADesc *pDmaTxDesc = NULL;
	struct device *dma_dev = ((PVRSRV_RGXDEV_INFO	*)(psDevNode->pvDevice))->dma_dev;
	struct dma_async_tx_descriptor *psDesc;
	IMG_UINT64 ulMapedSize;
	int iRet;
	PVRSRV_ERROR eError;

	pDmaTxDesc = OSAllocZMem(sizeof(OSDMADesc));
//	pDmaTxDesc = &psOSDMAEntry->DTxDesc;
	PVR_LOG_GOTO_IF_NOMEM(pDmaTxDesc, eError, e0);
	pDmaTxDesc->psDmaDevice = dma_dev;
	pDmaTxDesc->pDmaChan = pvChan;
	pDmaTxDesc->pvDmaDataParent = psOSDMAEntry;
	iRet = OSDmaRequestHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc,
				puiAddress, uiMaxSize, &pDmaTxDesc->tPages,
				!!(uiFlags & DMA_FLAG_FROM_USER), 1, direction);
	if (iRet < 0) {
		PVR_DPF((PVR_DBG_ERROR, "%s: OSDmaRequestHostDmaAddr failed(%d)", __func__, iRet));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e0;
	}
	ulMapedSize = pDmaTxDesc->tPages.ulMapSize;
	pDmaTxDesc->uiSize = ulMapedSize;
	psDesc = dmaengine_prep_slave_sg(pvChan, pDmaTxDesc->tPages.psSg->sgl,
				pDmaTxDesc->tPages.uiSgCnt, direction, 0);
	if (!psDesc)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: dmaengine_prep_slave_sg failed", __func__));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e1;
	}

	psDesc->callback_param = pDmaTxDesc;
	psDesc->callback_result = OSDmaTxDescDoneCb;
	psDesc->phys = psDmaAddr->uiAddr; //set dma phyiscal address here!!!
	pDmaTxDesc->psDescriptors = psDesc;
	list_add(&pDmaTxDesc->PList, &psOSDMAEntry->lhDescHead);
	if (psOSDMAEntry->bTest) {
		PVR_DPF((PVR_DBG_ERROR, "[%s]%d: prepare HA:%pK DA:%llx Size:%llx success\n",
			__func__, __LINE__, puiAddress, psDmaAddr->uiAddr, ulMapedSize));
	}
	OSDmaDescEnQueueReady(psOSDMAEntry->pvHandleParent, pDmaTxDesc);
	*uiRetSize = ulMapedSize;
	return PVRSRV_OK;
e1:
	OSDmaReleaseHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc,
				&pDmaTxDesc->tPages, direction);
e0:
	*uiRetSize = 0;
	if (pDmaTxDesc)
		OSFreeMem(pDmaTxDesc);
	PVR_DPF((PVR_DBG_ERROR, "[%s]%d:Error:%d.\n", __func__, __LINE__, eError));
	return eError;
}

static PVRSRV_ERROR OSDmaPrepareCpyDevAndHost(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		enum dma_transfer_direction direction,
		tXDXDmaDualAddrs  *tpDualAddrs,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	IMG_UINT64 ulRemainSize, ulDmaAddrContinusSize;
	tXDXDmaAddrHost *tpHostAddr;
	tXDXDmaAddrDev *tpDevAddr;
	IMG_UINT64 ulRetSize = 0;
	IMG_UINT64 ulDevRemainSize = 0;
	IMG_UINT64 ulDevRequestSize = 0;
	IMG_BOOL bSkipFlag = 0;
	IMG_UINT32 ui32DmaPagesIdx = 0;
	IMG_DMA_ADDR sDmaAddrs;
	IMG_UINT64 ulDmaPageSize = OSGetPageSize();
	IMG_UINT64 ulHostAddr;
	IMG_UINT32 uiDmaOffse;

	if (!tpDualAddrs|| !psOSDMAEntry)
		return PVRSRV_ERROR_INVALID_PARAMS;
	if (direction != DMA_MEM_TO_DEV && direction != DMA_DEV_TO_MEM)
		return PVRSRV_ERROR_INVALID_PARAMS;
	tpHostAddr = &tpDualAddrs->HostAddr1;
	tpDevAddr = &tpDualAddrs->DevAddrs2;
	if (tpDevAddr->ui32DmaPages == 0)
		return PVRSRV_ERROR_INVALID_PARAMS;

	ulHostAddr = tpHostAddr->puiAddress;
	uiDmaOffse = tpDevAddr->uiDmaOffse;
	ulRemainSize = tpHostAddr->uiSize;

	while (ulRemainSize > 0) {
		if (tpDevAddr->ui32DmaPages <= 1 || !tpDevAddr->pDmaPageVailds)
		{
			/*
			if the GPU address is continuous, submit only one sg list packeg.
			*/
			ulDevRequestSize = ulRemainSize;
			sDmaAddrs.uiAddr = tpDevAddr->psDmaAddrs[0].uiAddr + (tpHostAddr->uiSize - ulRemainSize);
			uiDmaOffse = 0;
		}
		else
		{
			/*Get continuous GPU address. then packge the sg list.*/
			ulDmaAddrContinusSize = OSDmaGetContinuesPagesCnt(
						&tpDevAddr->psDmaAddrs[ui32DmaPagesIdx], &tpDevAddr->pDmaPageVailds[ui32DmaPagesIdx],
						tpDevAddr->ui32DmaPages - ui32DmaPagesIdx, ulDmaPageSize, &bSkipFlag);
			ulDevRequestSize = (ulDmaPageSize * ulDmaAddrContinusSize) - (uiDmaOffse & (PAGE_SIZE -1));
			ulDevRequestSize = MIN(ulDevRequestSize, ulRemainSize);
			sDmaAddrs.uiAddr = tpDevAddr->psDmaAddrs[ui32DmaPagesIdx].uiAddr;

			uiDmaOffse = 0;
			if (bSkipFlag == IMG_TRUE)
			{
				ui32DmaPagesIdx++;
				ulRemainSize -= ulDevRequestSize;
				ulHostAddr += ulDevRequestSize;
				PVR_DPF((PVR_DBG_VERBOSE, "[%s]Skip++,ulHostAddr:%llx ulRemainSize:%llx",
						__func__, ulHostAddr, ulRemainSize));
				continue;
			}
			if (ulDmaAddrContinusSize == 0)
			{
				PVR_DPF((PVR_DBG_ERROR,
						"[%s]%d:DmaAddrs:%llx, DmaOffse:%x, DevRequestSize:%llx,DmaPagesIdx:%x,DmaPages:%x",
						__func__, __LINE__, tpDevAddr->psDmaAddrs[0].uiAddr, tpDevAddr->uiDmaOffse,
						ulDevRequestSize, ui32DmaPagesIdx, tpDevAddr->ui32DmaPages));
				return PVRSRV_ERROR_INVALID_PARAMS;
			}
			ui32DmaPagesIdx += ulDmaAddrContinusSize;
		}
//		PVR_DPF((PVR_DBG_ERROR,
//				"[%s]%d:ulHostAddr:%lx sDmaAddrs:%llx, DevRequestSize:%llx,DmaPagesId:%x/%x",
//				__func__, __LINE__, ulHostAddr, sDmaAddrs.uiAddr,
//				ulDevRequestSize, ui32DmaPagesIdx, tpDevAddr->ui32DmaPages));
		ulDevRemainSize = ulDevRequestSize;
		while (ulDevRemainSize > 0)
		{
			eError = OSDmaPrepCpySgDevHostSeg(psDevNode,
				pvChan, &sDmaAddrs, (IMG_UINT64 *)ulHostAddr, ulDevRemainSize,
				direction, psOSDMAEntry, uiFlags, &ulRetSize);
			if (eError != PVRSRV_OK)
			{
				PVR_DPF((PVR_DBG_ERROR, "[%s]%d:prepare filed.%d", __func__, __LINE__, eError));
				goto e0;
			}
			ulDevRemainSize -= ulRetSize;
			ulHostAddr += ulRetSize;
			sDmaAddrs.uiAddr += ulRetSize;
		}
//		eError = OSDmaTriggerTransfer((OSDMATranferHandle *)(psOSDMAEntry->pvHandleParent), pvChan);
		ulRemainSize -= ulDevRequestSize;
	}
	return PVRSRV_OK;
e0:
	return eError;
}

static PVRSRV_ERROR OSDmaPrepareCpyDev2Dev(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		enum dma_transfer_direction direction,
		tXDXDmaDualAddrs  *tpDualAddrs,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	tXDXDmaAddrDev *tpSrcAddr, *tpDecAddr;
	IMG_DMA_ADDR sDecDmaAddrs, sSrcDmaAddrs;
	OSDMADesc *pDmaTxDesc = NULL;
	struct device *dma_dev = ((PVRSRV_RGXDEV_INFO	*)(psDevNode->pvDevice))->dma_dev;
	struct dma_async_tx_descriptor *psDesc;

	if (!tpDualAddrs|| !psOSDMAEntry)
		return PVRSRV_ERROR_INVALID_PARAMS;
	if (direction != DMA_DEV_TO_DEV)
		return PVRSRV_ERROR_INVALID_PARAMS;

	tpSrcAddr = &tpDualAddrs->DevAddrs1;
	tpDecAddr = &tpDualAddrs->DevAddrs2;
	if (tpSrcAddr->ui32DmaPages == 0 || tpDecAddr->ui32DmaPages == 0)
		return PVRSRV_ERROR_INVALID_PARAMS;

	sDecDmaAddrs.uiAddr = tpDecAddr->psDmaAddrs[0].uiAddr;
	sSrcDmaAddrs.uiAddr = tpSrcAddr->psDmaAddrs[0].uiAddr;
	pDmaTxDesc = OSAllocZMem(sizeof(OSDMADesc));
	PVR_LOG_GOTO_IF_NOMEM(pDmaTxDesc, eError, e0);
	pDmaTxDesc->psDmaDevice = dma_dev;
	pDmaTxDesc->pDmaChan = pvChan;
	pDmaTxDesc->pvDmaDataParent = psOSDMAEntry;
	pDmaTxDesc->uiSize = tpSrcAddr->uiSize;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
	psDesc = dmaengine_prep_dma_memcpy(pvChan, (dma_addr_t)(sDecDmaAddrs.uiAddr | XDX_DMA_ADDR_GPU),
		(dma_addr_t)(sSrcDmaAddrs.uiAddr | XDX_DMA_ADDR_GPU), tpSrcAddr->uiSize, 0);
#else
	psDesc = xdx_dmaengine_prep_dma_memcpy(pvChan, (dma_addr_t)(sDecDmaAddrs.uiAddr | XDX_DMA_ADDR_GPU),
                (dma_addr_t)(sSrcDmaAddrs.uiAddr | XDX_DMA_ADDR_GPU), tpSrcAddr->uiSize, 0);
#endif
	if (!psDesc)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: dmaengine_prep_dma_memcpy failed", __func__));
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e0;
	}

	psDesc->callback_param = pDmaTxDesc;
	psDesc->callback_result = OSDmaTxDescDoneCb;
	psDesc->phys = sDecDmaAddrs.uiAddr; //set dma phyiscal address here!!!
	pDmaTxDesc->psDescriptors = psDesc;
	list_add(&pDmaTxDesc->PList, &psOSDMAEntry->lhDescHead);
	if (psOSDMAEntry->bTest) {
		PVR_DPF((PVR_DBG_ERROR, "[%s]%d: prepare HA:%llx DA:%llx Size:%llx success\n",
			__func__, __LINE__, sDecDmaAddrs.uiAddr, sSrcDmaAddrs.uiAddr, tpSrcAddr->uiSize));
	}
	OSDmaDescEnQueueReady(psOSDMAEntry->pvHandleParent, pDmaTxDesc);
	return PVRSRV_OK;
e0:
	if (pDmaTxDesc)
		OSFreeMem(pDmaTxDesc);
	PVR_DPF((PVR_DBG_ERROR, "[%s]%d:Error:%d.\n", __func__, __LINE__, eError));
	return eError;
}

static PVRSRV_ERROR OSDmaPrepareCpyHost2Host(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		enum dma_transfer_direction direction,
		tXDXDmaDualAddrs  *tpDualAddrs,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry
)
{
	//:TODO
	PVR_DPF((PVR_DBG_ERROR, "[%s]DMA Unsupport Type(Host2Host)", __func__));
	return PVRSRV_ERROR_NOT_SUPPORTED;
}

#define XDX_DMA_MEMSET_MAX_LEN (8 << 20)  //The maximum support is 32M,which is transmitted in 8M units here
static PVRSRV_ERROR OSDMAPrepMemset(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		tXDXDDmaMamemset  *ptMemsetparas,
		enum dma_transfer_direction direction,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	tXDXDmaAddrDev *tpDstAddr = NULL;
	IMG_DMA_ADDR sDstDmaAddrs;
	OSDMADesc *pDmaTxDesc = NULL;
	struct device *dma_dev = ((PVRSRV_RGXDEV_INFO	*)(psDevNode->pvDevice))->dma_dev;
	struct dma_async_tx_descriptor *psDesc;
	IMG_UINT64 remain_len = ptMemsetparas->uiSize, copy_size = 0, offset = 0, iRet = 0;
	sDstDmaAddrs.uiAddr = 0;

	if (!ptMemsetparas|| !psOSDMAEntry)
		return PVRSRV_ERROR_INVALID_PARAMS;

	if (direction == DMA_MEM_TO_DEV) {
		tpDstAddr = &ptMemsetparas->DevAddrs1;
		if (tpDstAddr->ui32DmaPages == 0)
			return PVRSRV_ERROR_INVALID_PARAMS;
	}

	while (remain_len > 0) {
		copy_size = MIN(remain_len, XDX_DMA_MEMSET_MAX_LEN);
		pDmaTxDesc = OSAllocZMem(sizeof(OSDMADesc));
		PVR_LOG_GOTO_IF_NOMEM(pDmaTxDesc, eError, e0);
		pDmaTxDesc->psDmaDevice = dma_dev;
		pDmaTxDesc->pDmaChan = pvChan;
		pDmaTxDesc->pvDmaDataParent = psOSDMAEntry;
		pDmaTxDesc->uiSize = copy_size;

		if (direction == DMA_MEM_TO_DEV) {
			sDstDmaAddrs.uiAddr = tpDstAddr->psDmaAddrs->uiAddr + offset;
			psDesc = dmaengine_prep_dma_memset(pvChan, (dma_addr_t)(sDstDmaAddrs.uiAddr | XDX_DMA_ADDR_GPU),
				ptMemsetparas->value, copy_size, 0);
			if (!psDesc) {
				PVR_DPF((PVR_DBG_ERROR, "%s: dmaengine_prep_dma_memset failed", __func__));
				eError = PVRSRV_ERROR_INVALID_PARAMS;
				goto e0;
			}
		} else {
			iRet = OSDmaRequestHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc,
				(void *)(ptMemsetparas->HostAddr1.puiAddress + offset), copy_size, &pDmaTxDesc->tPages,
				!!(uiFlags & DMA_FLAG_FROM_USER), 1, direction);
			if (iRet < 0) {
				PVR_DPF((PVR_DBG_ERROR, "%s: OSDmaRequestHostDmaAddr failed(%lld)", __func__, iRet));
				eError = PVRSRV_ERROR_INVALID_PARAMS;
				goto e0;
			}
			copy_size = pDmaTxDesc->tPages.ulMapSize;
			pDmaTxDesc->uiSize = copy_size;

			psDesc = xdx_dmaengine_prep_dma_memset_sg(pvChan, pDmaTxDesc->tPages.psSg->sgl,
					pDmaTxDesc->tPages.uiSgCnt, ptMemsetparas->value, 0);
			if (!psDesc) {
				PVR_DPF((PVR_DBG_ERROR, "%s: dmaengine_prep_dma_memset_sg failed", __func__));
				eError = PVRSRV_ERROR_INVALID_PARAMS;
				goto e1;
			}
		}

		psDesc->callback_param = pDmaTxDesc;
		psDesc->callback_result = OSDmaTxDescDoneCb;
		if (direction == DMA_MEM_TO_DEV)
			psDesc->phys = sDstDmaAddrs.uiAddr; //set dma phyiscal address here!!!
		pDmaTxDesc->psDescriptors = psDesc;
		list_add(&pDmaTxDesc->PList, &psOSDMAEntry->lhDescHead);
		if (psOSDMAEntry->bTest && direction == DMA_MEM_TO_DEV) {
			PVR_DPF((PVR_DBG_ERROR, "[%s]%d: prepare DA:%llx Size:%llx success\n",
				__func__, __LINE__, sDstDmaAddrs.uiAddr, tpDstAddr->uiSize));
		}
		OSDmaDescEnQueueReady(psOSDMAEntry->pvHandleParent, pDmaTxDesc);

		remain_len -= copy_size;
		offset += copy_size;
	}
	return PVRSRV_OK;
e1:
	OSDmaReleaseHostDmaAddr(psDevNode->sg_pool, pDmaTxDesc, &pDmaTxDesc->tPages, direction);
e0:
	if (pDmaTxDesc)
		OSFreeMem(pDmaTxDesc);
	PVR_DPF((PVR_DBG_ERROR, "[%s]%d:Error:%d.\n", __func__, __LINE__, eError));
	return eError;
}

static PVRSRV_ERROR OSDmaPrepareTransferCpy(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		tXDXDmaDualAddrs *tpDualAddrs,
		enum dma_transfer_direction direction,
		IMG_UINT32 uiFlags,
		OSDMAEntry *psOSDMAEntry)
{
	PVRSRV_ERROR eError = PVRSRV_OK;

	switch (direction) {
		case DMA_MEM_TO_DEV:
		case DMA_DEV_TO_MEM:
			eError = OSDmaPrepareCpyDevAndHost(psDevNode,
					pvChan, direction, tpDualAddrs, uiFlags, psOSDMAEntry);
		break;
		case DMA_DEV_TO_DEV:
			eError = OSDmaPrepareCpyDev2Dev(psDevNode,
					pvChan, direction, tpDualAddrs, uiFlags, psOSDMAEntry);
		break;
		case DMA_MEM_TO_MEM:
			eError = OSDmaPrepareCpyHost2Host(psDevNode,
					pvChan, direction, tpDualAddrs, uiFlags, psOSDMAEntry);
		break;
		default:
			PVR_DPF((PVR_DBG_ERROR, "DMA unknown direction(%d)", direction));
			return PVRSRV_ERROR_NOT_SUPPORTED;
	};
	return eError;
}


PVRSRV_ERROR OSDmaWaitDone(void *pvData)
{
	OSDMATranferHandle *psOSDMAHandle = (OSDMATranferHandle*)pvData;

#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_VERBOSE, "OSDmaWaitDone thread waiting (%p) on completion", pvData));
#endif
	PVR_DPF((PVR_DBG_VERBOSE,"OSDmaWaitDone wait start_cleanup .!.\n"));

	if (atomic_read(&psOSDMAHandle->bAllPrepFlag) != IMG_TRUE)
		PVR_DPF((PVR_DBG_ERROR, "OSDmaWaitDone not all ready!"));
	if (//psOSDMAHandle->bAllPrepFlag == IMG_TRUE &&
		atomic_read(&psOSDMAHandle->aDoneCount) != atomic_read(&psOSDMAHandle->aPrepCount)) {
#if 1 //defined(DEBUG)
//	{
		long res;
		res = wait_for_completion_timeout(
				&psOSDMAHandle->start_cleanup, msecs_to_jiffies(3000));
		if (res <= 0)
		{
			if (current->mm != NULL)
			{
				PVR_DPF((PVR_DBG_ERROR, "[%d-%s]: [%s]complete timeout Done:%d  Submit:%d terminate all(%ld).\n",
					current->pid, current->comm, __func__, atomic_read(&psOSDMAHandle->aDoneCount),
					atomic_read(&psOSDMAHandle->aPrepCount), res));
			}
			else
			{
				PVR_DPF((PVR_DBG_ERROR, "[%s]complete timeout Done:%d  Submit:%d terminate all(%ld).\n",
					__func__, atomic_read(&psOSDMAHandle->aDoneCount), atomic_read(&psOSDMAHandle->aPrepCount), res));
			}
//			dmaengine_terminate_sync(psOSDMAHandle->psDevNode->hDmaTxChan);
			xdx_vdma_print_regs(psOSDMAHandle->psDevNode);
			OSDmaCancelTransfer(psOSDMAHandle);
			flush_work(&psOSDMAHandle->sCleanWork);
		}
	}
#else
	wait_for_completion(&psOSDMAHandle->start_cleanup);
#endif
//	}
//	cancel_work_sync(&psOSDMAHandle->sCleanWork);
	OSDmaTransferCleanupWork(&psOSDMAHandle->sCleanWork);
	OSDmaCleanHandle(psOSDMAHandle);
	psOSDMAHandle->pfnServerCleanup(psOSDMAHandle->pvServerCleanupData,
								psOSDMAHandle->bAdvanceTimeline);
//	memset(psOSDMAHandle, 0x66, sizeof(OSDMATranferHandle));
	OSFreeMem(psOSDMAHandle);
	return 0;
}

PVRSRV_ERROR OSDmaSubmitTransfer(void *pvOSData, void *pvChan)
{
	OSDMATranferHandle *psOSDMAHandle = (OSDMATranferHandle*)pvOSData;
	PVRSRV_ERROR eError = PVRSRV_OK;

	eError = OSDmaTriggerTransfer(psOSDMAHandle, (struct dma_chan *)pvChan);
	atomic_set(&psOSDMAHandle->bAllPrepFlag, IMG_TRUE);
	return eError;
}

/*************************************************************************/ /*!
@Function       OSDmaTransfer
@Description    This API is used to ask OS to perform a DMA transfer operation
@Return
*/ /**************************************************************************/
PVRSRV_ERROR OSDmaPrepareTransfer(
		PVRSRV_DEVICE_NODE *psDevNode,
		void* pvChan,
		eOSDMAXmitTypeE eType,
		void *vpParas,
		IMG_UINT64 uiSize,
		IMG_UINT32 uiFlags,
		void* pvOSData)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	OSDMATranferHandle* psOSDMAHandle = (OSDMATranferHandle *)pvOSData;
	enum dma_transfer_direction direction;
	OSDMAEntry *psOSDMAEntry;

	if (uiFlags & DMA_FLAG_MEM_TO_DEV)
		direction = DMA_MEM_TO_DEV;
	else if (uiFlags & DMA_FLAG_DEV_TO_DEV)
		direction = DMA_DEV_TO_DEV;
	else
		direction = DMA_DEV_TO_MEM;

	psOSDMAEntry = OSAllocZMem(sizeof(OSDMAEntry));
	PVR_LOG_GOTO_IF_NOMEM(psOSDMAEntry, eError, e0);
	INIT_LIST_HEAD(&psOSDMAEntry->lhDescHead);
	psOSDMAHandle->bIsFromUser =
			(uiFlags & DMA_FLAG_FROM_USER) ? IMG_TRUE : IMG_FALSE;
	psOSDMAEntry->pvHandleParent = psOSDMAHandle;
	psOSDMAEntry->uiSize = uiSize;
	psOSDMAHandle->eDirection = direction;
	switch (eType)
	{
		case XdxDmaTypeCPY:
			eError = OSDmaPrepareTransferCpy(psDevNode, pvChan,
				(tXDXDmaDualAddrs *)vpParas, direction, uiFlags, psOSDMAEntry);
			PVR_LOG_GOTO_IF_ERROR(eError, "OSDmaPrepareTransferCpy", e0);
			break;
#ifdef SUPPORT_DMA_TRANSFER_B0
		case XdxDmaTypeTWD2D3D:
			eError = OSDMAPrepTwiddle2D3D(psDevNode, pvChan,
				(DMA_TWD_2D3D_PARA *)vpParas, uiFlags, direction, psOSDMAEntry);
			PVR_LOG_GOTO_IF_ERROR(eError, "OSDMAPrep3DXmit", e0);
			break;
		case XdxDmaType2D:
			eError = OSDMAPrep2DCopy(psDevNode, pvChan,
				(DMA_XMIT_2D_PARA *)vpParas, direction, uiFlags, psOSDMAEntry);
			PVR_LOG_GOTO_IF_ERROR(eError, "OSDMAPrep2DXmit", e0);
			break;
#endif
		case XdxDmaTypeMemset:
			eError = OSDMAPrepMemset(psDevNode, pvChan,
				(tXDXDDmaMamemset *)vpParas, direction, uiFlags, psOSDMAEntry);
			PVR_LOG_GOTO_IF_ERROR(eError, "OSDMAPrepMemset", e0);
			break;
		case XdxDmaTypePoll:
		case XdxDmaTypeCylic:
		default:
			eError = PVRSRV_ERROR_NOT_SUPPORTED;
			break;
	}
	psOSDMAHandle->uiCount++;
	list_add_tail(&psOSDMAEntry->List, &psOSDMAHandle->lhEntryHead);
	return eError;
e0:
	if (psOSDMAEntry)
		OSFreeMem(psOSDMAEntry);
	return eError;
}

void OSDmaFreeData(void *psOSCleanup)
{
	OSDMATranferHandle *psOSDMAHandle = (OSDMATranferHandle *)psOSCleanup;

	if (!psOSDMAHandle)
		return;
	OSDmaCleanHandle(psOSDMAHandle);
	if (psOSDMAHandle->pfnServerCleanup)
		psOSDMAHandle->pfnServerCleanup(
					psOSDMAHandle->pvServerCleanupData,
					psOSDMAHandle->bAdvanceTimeline);
	OSFreeMem(psOSDMAHandle);
	return;
}

PVRSRV_ERROR OSDmaAllocData(PVRSRV_DEVICE_NODE *psDevNode, IMG_UINT32 uiNumDMA,
						IMG_HANDLE pvServerCleanupParam, PFN_SERVER_CLEANUP pfnServerCleanup,
						void **pvOutData)
{
	PVRSRV_ERROR eError;
	OSDMATranferHandle *psOSDMAHandle;
	IMG_UINT32 ui32BufSize = sizeof(struct _OSDMATranferHandle_);

	psOSDMAHandle = OSAllocZMem(ui32BufSize + (sizeof(OSDMAEntry) * uiNumDMA) );
	PVR_LOG_GOTO_IF_NOMEM(psOSDMAHandle, eError, e0);
	psOSDMAHandle->psDevNode = psDevNode;
	init_completion(&psOSDMAHandle->start_cleanup);
	psOSDMAHandle->pvServerCleanupData = pvServerCleanupParam;
	psOSDMAHandle->pfnServerCleanup = pfnServerCleanup;
	psOSDMAHandle->uiCount = 0;
	INIT_LIST_HEAD(&psOSDMAHandle->lReadyListHead);
	INIT_LIST_HEAD(&psOSDMAHandle->lSubmitListHead);
	INIT_LIST_HEAD(&psOSDMAHandle->lDoneListHead);
	INIT_LIST_HEAD(&psOSDMAHandle->lDoneCleanupHead);
	INIT_LIST_HEAD(&psOSDMAHandle->lhEntryHead);
	atomic_set(&psOSDMAHandle->bAllPrepFlag, IMG_FALSE);
	INIT_WORK(&psOSDMAHandle->sCleanWork, OSDmaTransferCleanupWork);
	spin_lock_init(&psOSDMAHandle->DoneListLock);
	atomic_set(&psOSDMAHandle->aDoneCount, 0);
	atomic_set(&psOSDMAHandle->aPrepCount, 0);
	psOSDMAHandle->uiEntryMax = uiNumDMA;
	*pvOutData = psOSDMAHandle;
	return PVRSRV_OK;
e0:
	return eError;
}

#endif /* SUPPORT_DMA_TRANSFER */

