/*************************************************************************/ /*!
@File           pvr_dvfs.c
@Title          System level interface for DVFS
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#ifndef PVR_DVFS_DEVICE_H
#define PVR_DVFS_DEVICE_H

#include "opaque_types.h"
#include "pvrsrv_error.h"

#ifdef XDXGPU
/* Devfreq events */
#define DEVFREQ_GOV_START			0x1
#define DEVFREQ_GOV_STOP			0x2
#define DEVFREQ_GOV_UPDATE_INTERVAL		0x3
#define DEVFREQ_GOV_SUSPEND			0x4
#define DEVFREQ_GOV_RESUME			0x5

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
struct devfreq_governor {
	struct list_head node;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))
    const char name[16];
    const u64 attrs;
    const u64 flags;
#else
    const char name[DEVFREQ_NAME_LEN];
	const unsigned int immutable;
#endif
	int (*get_target_freq)(struct devfreq *this, unsigned long *freq);
	int (*event_handler)(struct devfreq *devfreq,
				unsigned int event, void *data);
};
#endif

int xdx_dvfs_init(PPVRSRV_DEVICE_NODE psDeviceNode);
void xdx_dvfs_deinit(PPVRSRV_DEVICE_NODE psDeviceNode);
#endif /* XDXGPU */

PVRSRV_ERROR InitDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode);

void DeinitDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode);

PVRSRV_ERROR RegisterDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode);

void UnregisterDVFSDevice(PPVRSRV_DEVICE_NODE psDeviceNode);

PVRSRV_ERROR SuspendDVFS(PPVRSRV_DEVICE_NODE psDeviceNode);

#ifdef XDXGPU
PVRSRV_ERROR ResumeDVFS(PPVRSRV_DEVICE_NODE psDeviceNode);

PVRSRV_ERROR XdxDVFSGovInit(void);

PVRSRV_ERROR XdxDVFSGovDeinit(void);

int xdx_dvfs_early_set_freq(PPVRSRV_DEVICE_NODE psDeviceNode, u32 ui32Freq);

int xdx_dvfs_early_set_volt(PPVRSRV_DEVICE_NODE psDeviceNode, u32 ui32Volt);

void xdx_dvfs_set_opp_availability_above_floor(PVRSRV_DEVICE_NODE *psDeviceNode, bool enable_opp);

const static int xdx_opp_group0_vgid0_volt = 820000;
const static int xdx_opp_group1_vgid0_volt = 940000;
const static int xdx_opp_group2_vgid0_volt = 1020000;

#ifndef XDX_DVFS_FULL_OPP_TABLE
const static IMG_OPP xdx_opp_group0[] = {
	{ 720000, 500000000 },
	{ 720000, 700000000 },
	{ 740000, 800000000 },
	{ 840000, 1000000000 },
};

const static IMG_OPP xdx_opp_group1[] = {
	{ 720000, 500000000 },
	{ 720000, 700000000 },
	{ 740000, 800000000 },
	{ 820000, 1000000000 },
	{ 960000, 1200000000 },
};

const static IMG_OPP xdx_opp_group2[] = {
	{ 720000, 500000000 },
	{ 720000, 700000000 },
	{ 740000, 800000000 },
	{ 820000, 1000000000 },
	{ 940000, 1200000000 },
	{ 1040000, 1300000000 },
};

/* Table used for the second type of XDX 160 device */
const static IMG_OPP xdx_opp_group2_extra[] = {
	{ 720000, 500000000 },
	{ 720000, 700000000 },
	{ 740000, 800000000 },
	{ 820000, 1000000000 },
	{ 940000, 1200000000 },
};

#else /* XDX_DVFS_FULL_OPP_TABLE */
/* FIXME: Shall involve  vgid in OPP table */
/* Will replace with full OPP table with adequate test */
const static IMG_OPP xdx_opp_group0[] = {
	{ 720000, 500000000 }, { 720000, 525000000 }, { 720000, 550000000 },
	{ 720000, 575000000 }, { 720000, 600000000 }, { 720000, 625000000 },
	{ 720000, 650000000 }, { 720000, 675000000 }, { 720000, 700000000 },
	{ 720000, 725000000 }, { 720000, 750000000 }, { 720000, 775000000 },
	{ 740000, 800000000 }, { 740000, 825000000 }, { 760000, 850000000 },
	{ 760000, 875000000 }, { 780000, 900000000 }, { 800000, 925000000 },
	{ 820000, 950000000 }, { 820000, 975000000 }, { 840000, 1000000000 },
};

const static IMG_OPP xdx_opp_group1[] = {
	{ 720000, 500000000 },	{ 720000, 525000000 },	{ 720000, 550000000 },
	{ 720000, 575000000 },	{ 720000, 600000000 },	{ 720000, 625000000 },
	{ 720000, 650000000 },	{ 720000, 675000000 },	{ 720000, 700000000 },
	{ 720000, 725000000 },	{ 720000, 750000000 },	{ 720000, 775000000 },
	{ 740000, 800000000 },	{ 740000, 825000000 },	{ 760000, 850000000 },
	{ 760000, 875000000 },	{ 780000, 900000000 },	{ 800000, 925000000 },
	{ 820000, 950000000 },	{ 820000, 975000000 },	{ 820000, 1000000000 },
	{ 840000, 1025000000 }, { 860000, 1050000000 }, { 880000, 1075000000 },
	{ 900000, 1100000000 }, { 900000, 1125000000 }, { 920000, 1150000000 },
	{ 940000, 1175000000 }, { 960000, 1200000000 },
};

const static IMG_OPP xdx_opp_group2[] = {
	{ 720000, 500000000 },	 { 720000, 525000000 },	  { 720000, 550000000 },
	{ 720000, 575000000 },	 { 720000, 600000000 },	  { 720000, 625000000 },
	{ 720000, 650000000 },	 { 720000, 675000000 },	  { 720000, 700000000 },
	{ 720000, 725000000 },	 { 720000, 750000000 },	  { 720000, 775000000 },
	{ 740000, 800000000 },	 { 740000, 825000000 },	  { 760000, 850000000 },
	{ 760000, 875000000 },	 { 780000, 900000000 },	  { 800000, 925000000 },
	{ 820000, 950000000 },	 { 820000, 975000000 },	  { 820000, 1000000000 },
	{ 840000, 1025000000 },	 { 860000, 1050000000 },  { 880000, 1075000000 },
	{ 900000, 1100000000 },	 { 900000, 1125000000 },  { 920000, 1150000000 },
	{ 940000, 1175000000 },	 { 940000, 1200000000 },  { 980000, 1225000000 },
	{ 1000000, 1250000000 }, { 1020000, 1275000000 }, { 1040000, 1300000000 },
};
#endif /* XDX_DVFS_FULL_OPP_TABLE */

#endif /* XDXGPU */

#endif /* PVR_DVFS_DEVICE_H */
