/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "pvrversion.h"
//#include "services_kernel_client.h"
#include "syscommon.h"
#include "rgx_fwif_shared.h"
#include "pvr_debug.h"
#include "pvrsrv_memallocflags.h"
#include "pvrsrv_memallocflags_internal.h"
#include "physmem.h"
#include "physheap.h"
#include "pmr.h"

#include "xdx_gem.h"

#include "pvr_device.h"
#include "pvr_dmabuf.h"
#include "pvr_gem.h"

static uint64_t placement_to_mem_alloc_flag(enum xdxgpu_gem_placement placement)
{
	uint64_t pvr_flags = 0;

	switch (placement) {
	case XDXGPU_PL_VRAM:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_PHYS_HEAP_HINT(GPU_LOCAL);
		break;
	case XDXGPU_PL_PRIV:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_PHYS_HEAP_HINT(GPU_PRIVATE);
		break;
	case XDXGPU_PL_TT:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_PHYS_HEAP_HINT(CPU_LOCAL);
		break;
	case XDXGPU_PL_SYSTEM:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_PHYS_HEAP_HINT(CPU_LOCAL);
		break;
	}

	return pvr_flags;
}

static uint64_t gpu_cache_to_mem_alloc_flag(enum xdxgpu_gem_gpu_cache cache)
{
	uint64_t pvr_flags = 0;

	switch(cache)
	{
	case XDXGPU_GEM_GPU_UNCACHED_WC:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC;
		break;
	case XDXGPU_GEM_GPU_UNCACHED:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_UNCACHED;
		break;
	case XDXGPU_GEM_GPU_CACHE_COHERENT:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT;
		break;
	case XDXGPU_GEM_GPU_CACHE_INCOHERENT:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT;
		break;
	}

	return pvr_flags;
}

static uint64_t cpu_cache_to_mem_alloc_flag(enum xdxgpu_gem_cpu_cache cache)
{
	uint64_t pvr_flags = 0;

	switch(cache)
	{
	case XDXGPU_GEM_CPU_UNCACHED_WC:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC;
		break;
	case XDXGPU_GEM_CPU_UNCACHED:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_UNCACHED;
		break;
	case XDXGPU_GEM_CPU_CACHE_COHERENT:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT;
		break;
	case XDXGPU_GEM_CPU_CACHE_INCOHERENT:
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT;
		break;
	}

	return pvr_flags;
}

static uint64_t mapping_to_mem_alloc_flag(uint32_t mapping)
{
	uint64_t pvr_flags = 0;

	if (mapping & XDXGPU_GEM_MAPPING_GPU_READABLE)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_READABLE;

	if (mapping & XDXGPU_GEM_MAPPING_GPU_WRITEABLE)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_WRITEABLE;

	if (mapping & XDXGPU_GEM_MAPPING_GPU_READ_PERMITTED)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_READ_PERMITTED;

	if (mapping & XDXGPU_GEM_MAPPING_GPU_WRITE_PERMITTED)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_GPU_WRITE_PERMITTED;

	if (mapping & XDXGPU_GEM_MAPPING_CPU_READABLE)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_READABLE;

	if (mapping & XDXGPU_GEM_MAPPING_CPU_WRITEABLE)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_WRITEABLE;

	if (mapping & XDXGPU_GEM_MAPPING_CPU_READ_PERMITTED)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_READ_PERMITTED;

	if (mapping & XDXGPU_GEM_MAPPING_CPU_WRITE_PERMITTED)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_CPU_WRITE_PERMITTED;

	return pvr_flags;
}

int pvr_bo_create(void *pvrdev, void *pvr_fpriv,
		  struct xdx_bo_create_params *bp, void **ppsPMR,
		  uint64_t *puiPMRFlags)
{
	PVRSRV_DEVICE_NODE *psDeviceNode = pvrdev;
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDeviceNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	PVRSRV_CONNECTION_PRIV *psConnectionPriv = pvr_fpriv;
	CONNECTION_DATA *psConnection =
		psConnectionPriv ? psConnectionPriv->pvConnectionData : NULL;
	uint64_t pvr_flags = 0;
	PVRSRV_ERROR eError;
	PMR *psPMR;
	bool contiguous = false;
	uint32_t i = 0;
	uint32_t *pui32MappingTable = NULL;
	uint32_t ui32NumChunks = 0;

	pvr_flags |= placement_to_mem_alloc_flag(bp->pro.placement);
	pvr_flags |= gpu_cache_to_mem_alloc_flag(bp->pro.gpu_cache);
	pvr_flags |= cpu_cache_to_mem_alloc_flag(bp->pro.cpu_cache);
	pvr_flags |= mapping_to_mem_alloc_flag(bp->pro.mapping);

	if (bp->pro.flags & XDXGPU_GEM_ALLOC_CLEARED)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_ZERO_ON_ALLOC;

	if (bp->pro.flags & XDXGPU_GEM_ALLOC_CONTIGUOUS)
		contiguous = true;

	if (bp->pro.flags & XDXGPU_GEM_ALLOC_DC)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_DC_ON_RESERVE;

	if (bp->pro.flags & XDXGPU_GEM_ALLOC_VPU)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_VPU;

	/* deferred allocation */
	pvr_flags |= PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC;

	if (bp->type == xdx_bo_type_kernel)
		pvr_flags |= PVRSRV_MEMALLOCFLAG_KERNEL_CPU_MAPPABLE;

	ui32NumChunks = contiguous ? 1 : (bp->size / bp->alignment);
	pui32MappingTable = (uint32_t *)OSAllocMem(ui32NumChunks * sizeof(*pui32MappingTable));
	if (pui32MappingTable == NULL) {
		dev_err(dev, "%s: failed to allocate OS memory\n", __func__);
		return -EINVAL;
	}

	for (i = 0; i < ui32NumChunks; ++i) {
		pui32MappingTable[i] = i;
	}

	eError = PhysmemNewRamBackedPMR(psConnection,
					psDeviceNode,
					bp->size,
					ui32NumChunks,
					ui32NumChunks,
					pui32MappingTable,
					get_order(bp->alignment) + PAGE_SHIFT,
					pvr_flags,
					sizeof("GEM"),
					"GEM",
					OSGetCurrentClientProcessIDKM(),
					&psPMR,
					PDUMP_NONE,
					puiPMRFlags);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "%s: failed to create PMR\n", __func__);
		OSFreeMem(pui32MappingTable);
		return -EINVAL;
	}

	*ppsPMR = psPMR;
	OSFreeMem(pui32MappingTable);

	return 0;
}

struct dma_buf *pvr_export_dmabuf(void *pvrdev, void *psPMR, struct drm_gem_object *gobj, int flags)
{
	PVRSRV_DEVICE_NODE *psDevNode = pvrdev;
	PVRSRV_ERROR eError;
	struct dma_buf *psDmaBuf;

	eError = PhysmemExportPMRToDmaBuf(psDevNode, psPMR, gobj, flags, &psDmaBuf);
	if (eError != PVRSRV_OK)
		return ERR_PTR(-ENOMEM);

	return psDmaBuf;
}

int pvr_query_pmr_info(void *psPMRPtr, struct xdx_bo_property *pro,
		       uint32_t *psize)
{
	PMR *psPMR = psPMRPtr;
	PMR_FLAGS_T uiPMRFlags;
	IMG_DEVMEM_SIZE_T uiLogicalSize;
	PVRSRV_PHYS_HEAP ePhysHeap;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	uint64_t uiGPUCacheMode, uiCPUCacheMode;
	PHYS_HEAP *psPhysHeap;
	PHYS_HEAP_TYPE heap_type;
	PVRSRV_ERROR status;

	PMR_LogicalSize(psPMR, &uiLogicalSize);

	uiPMRFlags = PMR_Flags(psPMR);

	ePhysHeap = (PVRSRV_PHYS_HEAP)PVRSRV_GET_PHYS_HEAP_HINT(uiPMRFlags);

	status = PhysHeapAcquireByDevPhysHeap(ePhysHeap, psDevNode, &psPhysHeap);
	if (status != PVRSRV_OK) {
		dev_err(dev, "Wrong PhysHeap %d, status %d\n", ePhysHeap, status);
		return -EINVAL;
	}

	heap_type = PhysHeapGetType(psPhysHeap);
	if (heap_type == PHYS_HEAP_TYPE_UMA)
		pro->placement = XDXGPU_PL_TT;
	else if (heap_type == PHYS_HEAP_TYPE_LMA &&
		 ePhysHeap == PVRSRV_PHYS_HEAP_GPU_PRIVATE)
		pro->placement = XDXGPU_PL_PRIV;
	else if (heap_type == PHYS_HEAP_TYPE_LMA)
		pro->placement = XDXGPU_PL_VRAM;
	else {
		pro->placement = XDXGPU_PL_VRAM;
		dev_err(dev, "%s: Wrong PhysHeap %d, eType 0x%x\n", __func__,
			ePhysHeap, heap_type);
	}

	PhysHeapRelease(psPhysHeap);

	uiCPUCacheMode = PVRSRV_CPU_CACHE_MODE(uiPMRFlags);
	switch (uiCPUCacheMode) {
	case PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC:
		pro->cpu_cache = XDXGPU_GEM_CPU_UNCACHED_WC;
		break;
	case PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT:
		pro->cpu_cache = XDXGPU_GEM_CPU_CACHE_COHERENT;
		break;
	case PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT:
		pro->cpu_cache = XDXGPU_GEM_CPU_CACHE_INCOHERENT;
		break;
	case PVRSRV_MEMALLOCFLAG_CPU_UNCACHED:
		pro->cpu_cache = XDXGPU_GEM_CPU_UNCACHED;
		break;
	default:
		dev_err(dev, "%s: unhandled cpu cache mode: 0x%llx\n", __func__,
			uiCPUCacheMode);
		break;
	}

	uiGPUCacheMode = PVRSRV_GPU_CACHE_MODE(uiPMRFlags);
	switch (uiGPUCacheMode) {
	case PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC:
		pro->gpu_cache = XDXGPU_GEM_GPU_UNCACHED_WC;
		break;
	case PVRSRV_MEMALLOCFLAG_GPU_UNCACHED:
		pro->gpu_cache = XDXGPU_GEM_GPU_UNCACHED;
		break;
	case PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT:
		pro->gpu_cache = XDXGPU_GEM_GPU_CACHE_COHERENT;
		break;
	case PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT:
		pro->gpu_cache = XDXGPU_GEM_GPU_CACHE_INCOHERENT;
		break;
	default:
		dev_err(dev, "%s: unhandled gpu cache mode: 0x%llx\n", __func__,
			uiGPUCacheMode);
		break;
	}

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_GPU_READABLE)
		pro->mapping |= XDXGPU_GEM_MAPPING_GPU_READABLE;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_GPU_WRITEABLE)
		pro->mapping |= XDXGPU_GEM_MAPPING_GPU_WRITEABLE;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_GPU_READ_PERMITTED)
		pro->mapping |= XDXGPU_GEM_MAPPING_GPU_READ_PERMITTED;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_GPU_WRITE_PERMITTED)
		pro->mapping |= XDXGPU_GEM_MAPPING_GPU_WRITE_PERMITTED;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_CPU_READABLE)
		pro->mapping |= XDXGPU_GEM_MAPPING_CPU_READABLE;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_CPU_WRITEABLE)
		pro->mapping |= XDXGPU_GEM_MAPPING_CPU_WRITEABLE;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_CPU_READ_PERMITTED)
		pro->mapping |= XDXGPU_GEM_MAPPING_CPU_READ_PERMITTED;

	if (uiPMRFlags & PVRSRV_MEMALLOCFLAG_CPU_WRITE_PERMITTED)
		pro->mapping |= XDXGPU_GEM_MAPPING_CPU_WRITE_PERMITTED;

	pro->pmr_flags = uiPMRFlags;
	pro->pmr_flags &= (PVRSRV_MEMALLOCFLAG_CPU_READABLE |
			  PVRSRV_MEMALLOCFLAG_CPU_WRITEABLE |
			  PVRSRV_MEMALLOCFLAG_CPU_CACHE_MODE_MASK |
			  PVRSRV_MEMALLOCFLAGS_GPU_MMUFLAGSMASK |
			  PVRSRV_PHYS_HEAP_HINT_MASK);

	*psize = uiLogicalSize;

	return 0;
}

int pvr_import_dmabuf(void *pvrdev, struct dma_buf *dmabuf, void **ppsPMRPtr, struct drm_gem_object **gobj)
{
	PVRSRV_DEVICE_NODE *psDevNode = pvrdev;
	PVRSRV_ERROR eError;

	eError = PhysmemImportDmaBufToPMR(psDevNode, dmabuf, (PMR **)ppsPMRPtr, gobj);
	if (eError != PVRSRV_OK)
		return -ENOMEM;

	return 0;
}

static void *xdx_kvmalloc(size_t size, gfp_t flags)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
        return kvmalloc(size, GFP_KERNEL);
#else
	if (size <= KMALLOC_MAX_SIZE) {
		void *ptr = kmalloc(size, flags | __GFP_NOWARN);
		if (ptr)
			return ptr;
	}

	return __vmalloc(size, flags | __GFP_HIGHMEM, PAGE_KERNEL);
#endif
}


int pvr_pmr_get_pages(void *hPMR, uint32_t ui32Log2PageSize, uint64_t **pages, uint32_t *num_pages)
{
	PVRSRV_ERROR eError;
	PMR *psPMR = (PMR *)hPMR;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	IMG_DEV_PHYADDR *psDevPhyAddr;
	IMG_DEVMEM_SIZE_T uiLogicalSize;
	IMG_BOOL *pbValid;
	uint32_t ui32SizeInPages;
	int ret = 0;
	uint32_t i;

	eError = PMRLockSysPhysAddresses(psPMR);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "%s: failed to lock sys phys address\n", __func__);
		ret = -EBUSY;
		goto pro_end;
	}

	PMR_LogicalSize(psPMR, &uiLogicalSize);

	ui32SizeInPages = uiLogicalSize >> ui32Log2PageSize;

	psDevPhyAddr = OSAllocZMem(ui32SizeInPages * sizeof(IMG_DEV_PHYADDR));
	if (psDevPhyAddr == NULL) {
		ret = -ENOMEM;
		goto err_no_mem;
	}

	pbValid = OSAllocZMem(ui32SizeInPages * sizeof(IMG_BOOL));
	if (pbValid == NULL) {
		OSFreeMem(psDevPhyAddr);
		ret = -ENOMEM;
		goto err_no_mem;
	}

	eError = PMR_DevPhysAddr(psPMR, ui32Log2PageSize, ui32SizeInPages, 0,
				 psDevPhyAddr, pbValid);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "%s: failed to get device physical address\n",
			__func__);
		ret = -EBUSY;
		goto err_phy_addr;
	}

	*pages = xdx_kvmalloc(sizeof(uint64_t) * ui32SizeInPages, GFP_KERNEL);
	if (!*pages) {
		ret = -ENOMEM;
		goto err_phy_addr;
	}

	for (i = 0; i < ui32SizeInPages; i++)
		(*pages)[i] = psDevPhyAddr[i].uiAddr;

	*num_pages = ui32SizeInPages;
	ret = 0;

err_phy_addr:
	OSFreeMem(psDevPhyAddr);
	OSFreeMem(pbValid);
err_no_mem:
	if (ret != 0)
		PMRUnlockSysPhysAddresses(psPMR);
pro_end:
	return ret;
}

void pvr_pmr_put_pages(void *hPMR, uint64_t *pages)
{
	PMR *psPMR = (PMR *)hPMR;

	PMRUnlockSysPhysAddresses(psPMR);
	kvfree(pages);
}

int pvr_pmr_mmap(void *hPMR, struct vm_area_struct *vma)
{
	PMR *psPMR = (PMR *)hPMR;
	PVRSRV_ERROR eError;
	PVRSRV_MEMALLOCFLAGS_T uiProtFlags =
	    (BITMASK_HAS(vma->vm_flags, VM_READ) ? PVRSRV_MEMALLOCFLAG_CPU_READABLE : 0) |
	    (BITMASK_HAS(vma->vm_flags, VM_WRITE) ? PVRSRV_MEMALLOCFLAG_CPU_WRITEABLE : 0);

	eError = PMRMMapPMR(psPMR, vma, uiProtFlags);
	if (eError != PVRSRV_OK) {
		pr_err("%s: failed to mmap, ret %d\n", __func__, eError);
		return -ENOENT;
	}

	return 0;
}

int pvr_pmr_vmmap(void *hPMR, void **vaddr, void **priv)
{
	PMR *psPMR = (PMR *)hPMR;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	PVRSRV_ERROR eError;
	size_t uiMappedSize;
	IMG_DEVMEM_SIZE_T uiLogicalSize;

	PMR_LogicalSize(psPMR, &uiLogicalSize);

	eError = PMRAcquireKernelMappingData(psPMR, 0, uiLogicalSize, vaddr,
					     &uiMappedSize, priv);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "%s: failed to vmmap psPMR %p eError(%d)\n", __func__, psPMR, eError);
		return -ENOENT;
	}

	return 0;
}

void pvr_pmr_vunmap(void *hPMR, void *vaddr, void *priv)
{
	PMR *psPMR = (PMR *)hPMR;

	PMRReleaseKernelMappingData(psPMR, priv);
}

int pvr_pmr_export_handle(void *pvr_fpriv, void *hPMR, uint32_t *pvr_handle)
{
	PVRSRV_CONNECTION_PRIV *psConnectionPriv = pvr_fpriv;
	CONNECTION_DATA *psConnection =
		psConnectionPriv ? psConnectionPriv->pvConnectionData : NULL;
	PMR *psPMR = (PMR *)hPMR;
	PVRSRV_ERROR eError;
	IMG_HANDLE hPMRPtr;

	*pvr_handle = 0;

	/* Lock over handle creation. */
	LockHandle(psConnection->psHandleBase);
	eError = PVRSRVAllocHandleUnlocked(psConnection->psHandleBase,
				&hPMRPtr, (void *)psPMR,
				PVRSRV_HANDLE_TYPE_PHYSMEM_PMR,
				PVRSRV_HANDLE_ALLOC_FLAG_MULTI,
				NULL);
	if(unlikely(eError != PVRSRV_OK)) {
		UnlockHandle(psConnection->psHandleBase);
		return -ENOENT;
	}
	UnlockHandle(psConnection->psHandleBase);

	*pvr_handle = (unsigned long)hPMRPtr;
	return 0;
}

int pvr_pmr_lookup(void *pvr_fpriv, uint32_t pvr_handle, void **ppsPMRPtr)
{
	PVRSRV_CONNECTION_PRIV *psConnectionPriv = pvr_fpriv;
	CONNECTION_DATA *psConnection =
		psConnectionPriv ? psConnectionPriv->pvConnectionData : NULL;
	PVRSRV_ERROR eError;
	uint64_t tm = (uint64_t)pvr_handle;
	uint64_t *hPMRPtr = (uint64_t *)tm;

	*ppsPMRPtr = NULL;
	/* Lock over handle lookup. */
	LockHandle(psConnection->psHandleBase);
	eError = PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
				ppsPMRPtr,
				hPMRPtr, PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);
	if(unlikely(eError != PVRSRV_OK)) {
		UnlockHandle(psConnection->psHandleBase);
		return -ENOENT;
	}

	pvr_pmr_get(*ppsPMRPtr);

	PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase, hPMRPtr, PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);

	UnlockHandle(psConnection->psHandleBase);

	return 0;
}

void pvr_pmr_get(void *psPMR)
{
	PMRRefPMR((PMR *)psPMR);
}

void pvr_pmr_put(void *psPMR)
{
	PMRUnrefPMR((PMR *)psPMR);
}

void pvr_pmr_set_reservation(void *psPMR, struct dma_resv *resv)
{
	PMRSetDmaReservation((PMR *)psPMR, resv);
}

int pvr_pmr_get_page_prot(void *hPMR, struct vm_area_struct *vma)
{
	PMR *psPMR = (PMR *)hPMR;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	PVRSRV_ERROR eError = PVRSRV_OK;

	eError = OSGetPMRPageProt(psPMR, vma);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "Get wrong page prot, eError %d\n", eError);
		return VM_FAULT_SIGSEGV;
	}

	return 0;
}

int pvr_pmr_get_cpu_paddr(void *hPMR, uint64_t offset, unsigned long *cpu_paddr)
{
	PMR *psPMR = (PMR *)hPMR;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	IMG_BOOL bValid;
	PVRSRV_ERROR eError = PVRSRV_OK;
	IMG_CPU_PHYADDR cpuAddr;

	eError = PMR_CpuPhysAddr(psPMR, PAGE_SHIFT, 1, offset, &cpuAddr,
				 &bValid);
	if (eError != PVRSRV_OK) {
		dev_err(dev, "Get wrong cpu PAddr, eError %d\n", eError);
		return VM_FAULT_OOM;
	}

	*cpu_paddr = cpuAddr.uiAddr;

	return 0;
}

/* copy valid bytes from sparse PMR to kernel vaddr */
int sparse_pmr_copy(struct xdx_bo *bo, void *vaddr)
{
	PVRSRV_ERROR eError, eError1;
	IMG_UINT32 i = 0;
	PMR *psPMR = (PMR *)bo->psPMR;
	PPVRSRV_DEVICE_NODE psDevNode = PMR_DeviceNode(psPMR);
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;
	struct device *dev = psDevConfig->pvOSDevice;
	IMG_BOOL *pbValid = NULL;
	IMG_DEV_PHYADDR *apsDevPAddr;
	IMG_UINT32 ui32NumOfPages = bo->num_pages;
	size_t uiMappedSize, uiPageSize;
	IMG_HANDLE hPrivData = NULL;
	void *pvReadBuffer = NULL;
	void *pvKernAddr = vaddr;

	uiPageSize = PAGE_SIZE;

	/* Allocate OS memory for PMR page list */
	apsDevPAddr = OSAllocMem(ui32NumOfPages * sizeof(IMG_DEV_PHYADDR));
	PVR_LOG_GOTO_IF_NOMEM(apsDevPAddr, eError, ErrorFreePMRPageListMem);

	/* Allocate OS memory for PMR page state */
	pbValid = OSAllocMem(ui32NumOfPages * sizeof(IMG_BOOL));
	PVR_LOG_GOTO_IF_NOMEM(pbValid, eError, ErrorFreePMRPageStateMem);
	OSCachedMemSet(pbValid, 0, ui32NumOfPages * sizeof(IMG_BOOL));

	/*
	 * Check whether allocated PMR can be locked
	 * and obtain physical addresses of underlying memory pages.
	 */
	eError = PMRLockSysPhysAddresses(psPMR);
	if (eError != PVRSRV_OK) {
		PVR_DPF((PVR_DBG_ERROR, "Failed to lock PMR"));
		return eError;
	}

	/* Get the Device physical addresses of the pages */
	eError = PMR_DevPhysAddr(psPMR,
				 OSGetPageShift(), ui32NumOfPages,
				 0, apsDevPAddr, pbValid);
	if (eError != PVRSRV_OK) {
		PVR_DPF((PVR_DBG_ERROR,
		"Failed to map PMR pages into device physical addresses"));
		goto ErrorUnlockPhysAddresses;
	}

	/* Check whether physical page is OS PAGE_SIZE aligned */
	for (i = 0; i < ui32NumOfPages; i++) {
		if (pbValid[i]) {
			if ((apsDevPAddr[i].uiAddr & OSGetPageMask()) != 0) {
				PVR_DPF((PVR_DBG_ERROR,
				"Physical memory of PMR is not page aligned"));
				eError = PVRSRV_ERROR_MEMORY_TEST_FAILED;
				goto ErrorUnlockPhysAddresses;
			}
		}
	}

	/* Acquire kernel virtual address of each physical page and read
	 * from it and check where contents are intact.
	 */
	for (i = 0; i < ui32NumOfPages; i++) {
		if (pbValid[i]) {
			eError =
			PMRAcquireSparseKernelMappingData(psPMR,
							  (i * uiPageSize),
							  uiPageSize,
							  &pvReadBuffer,
							  &uiMappedSize,
							  &hPrivData);
			if (eError != PVRSRV_OK) {
				PVR_DPF((PVR_DBG_ERROR,
				"Failed to Acquire Kernel Mapping of PMR"));
				goto ErrorUnlockPhysAddresses;
			}
			OSCachedMemCopy(pvKernAddr,
					pvReadBuffer, uiMappedSize);
			pvKernAddr += uiMappedSize;

			eError = PMRReleaseKernelMappingData(psPMR, hPrivData);
			PVR_LOG_IF_ERROR(eError,
					 "PMRReleaseKernelMappingData");
		}
	}

ErrorUnlockPhysAddresses:
	eError1 = PMRUnlockSysPhysAddresses(psPMR);
	if (eError1 != PVRSRV_OK) {
		eError = (eError == PVRSRV_OK) ? eError1 : eError;
		PVR_DPF((PVR_DBG_ERROR, "Failed to unlock PMR"));
	}

ErrorFreePMRPageStateMem:
	OSFreeMem(pbValid);
ErrorFreePMRPageListMem:
	OSFreeMem(apsDevPAddr);

	if (eError != PVRSRV_OK) {
		dev_err(dev, "%s: failed to copy psPMR %p eError(%d)\n",
			__func__, psPMR, eError);
		return -ENOENT;
	}

	return 0;
}
