/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include "allocmem.h"
#include "dma_km.h"
#include "xdx_device.h"
#include "pangu.h"

/* bridge headers */
#include "pvr_device.h"

/* PVR Headers */
#include "pvrsrv_device.h"
#include "syscommon.h"
#include "rgx_fwif_shared.h"
#include "rgxdevice.h"
#include "pvrsrv.h"

#include "osfunc.h"
#include "xdx_irq.h"

/* Macros used for debug mem snapshot */
#define ENABLE_PCIE_WRITE_FLUSH         /* Ensure posted write are completed */
#ifdef SUPPORT_DMA_TRANSFER
static IMG_BOOL gbSnapshotDmaWr = IMG_FALSE; /* Use DMA to accelerate H2D (disabled now, under testing)*/
static IMG_BOOL gbSnapshotDmaRd = IMG_TRUE;  /* Use DMA to accelerate D2H */
#else
static IMG_BOOL gbSnapshotDmaWr = IMG_FALSE;
static IMG_BOOL gbSnapshotDmaRd = IMG_FALSE;
#endif

struct PVRSRV_SNAPSHOT_HEAP_INFO_TAG
{
        PHYS_HEAP_USAGE_FLAGS   ui32Flags;
        PVRSRV_PHYS_HEAP        ePhysHeapIdx;
};
typedef struct PVRSRV_SNAPSHOT_HEAP_INFO_TAG PVRSRV_SNAPSHOT_HEAP_INFO;

#ifdef XDXGPU
#define PVRSRV_SNAPSHOT_HEAP_MAX     XDX_SNAPSHOT_HEAP_MAX
#else
#define PVRSRV_SNAPSHOT_HEAP_MAX     4 /* GPU local,private,mmu,dc */
#endif
static PVRSRV_SNAPSHOT_HEAP_INFO gasSnapshotHeap[PVRSRV_SNAPSHOT_HEAP_MAX] = {{0 , 0}, {0, 0}, {0, 0}, {0, 0}};
static IMG_UINT8 gui8SnapshotHeapCount = 0;

static void _pvr_error_notify(IMG_HANDLE hSysData,
			      PVRSRV_ROBUSTNESS_NOTIFY_DATA *psErrorData)
{
	struct xdx_device *xdev = (struct xdx_device *)hSysData;
	uint32_t level;

	switch (psErrorData->eResetReason) {
	case RGX_CONTEXT_RESET_REASON_NONE:
	case RGX_CONTEXT_RESET_REASON_GUILTY_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_GUILTY_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_HARD_CONTEXT_SWITCH:
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_OK:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_OK:
		level = PVR_DBG_MESSAGE;
		break;
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_HWR:
	case RGX_CONTEXT_RESET_REASON_FW_EXEC_ERR:
		level = PVR_DBG_WARNING;
		break;
	case RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_ERR:
	case RGX_CONTEXT_RESET_REASON_FW_WATCHDOG:
	case RGX_CONTEXT_RESET_REASON_FW_PAGEFAULT:
	case RGX_CONTEXT_RESET_REASON_HOST_WDG_FW_ERR:
	default:
		level = PVR_DBG_ERROR;
		break;
	}

	if (psErrorData->pid > 0) {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "PID %d experienced error: %d\n",
				psErrorData->pid, psErrorData->eResetReason);
	} else {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "Device experienced error: %d\n",
				psErrorData->eResetReason);
	}

	if ((psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM) ||
	    (psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM)) {
		dev_err(xdev->dev, "ExtJobRef 0x%x, DM %d",
			psErrorData->uErrData.sChecksumErrData.ui32ExtJobRef,
			psErrorData->uErrData.sChecksumErrData.eDM);
	}
}

#if defined(SUPPORT_DMA_TRANSFER)
static void Cleanup(void* pvCleanupData, IMG_BOOL bAdvanceTimeline)
{
	/* nothing to do */
}
#endif

/* #define _DBG printk */
#define _DBG(...)

static void _ResetSnapshotBuffer(struct xdx_device *psXdxDev)
{
        int i = 0;
        PHYS_HEAP_ITERATOR **ppsIter = NULL;
        IMG_BYTE **ppbBuf = NULL;

        for (i = 0; i < gui8SnapshotHeapCount; ++i)
        {
                ppsIter = (PHYS_HEAP_ITERATOR**)&psXdxDev->gfx.pvr_heap_iter[i];
                if (*ppsIter)
                {
                        LMA_HeapIteratorDestroy(*ppsIter);
                        *ppsIter = NULL;
                }

                ppbBuf = (IMG_BYTE **)&psXdxDev->gfx.pvr_pwr_buffer[i];
                if (*ppbBuf)
                {
                        OSFreeMem(*ppbBuf);
                        *ppbBuf = NULL;
                }
        }
        return;
}

static PVRSRV_ERROR _PrePower(PVRSRV_DEVICE_NODE *psDevNode,
                              PHYS_HEAP_USAGE_FLAGS ui32Flags,
                              PVRSRV_PHYS_HEAP ePhysHeapIdx,
                              PHYS_HEAP_ITERATOR **ppsIter,
                              IMG_BYTE **ppbBuf)
{
	IMG_UINT64 uiHeapTotalSize, uiHeapUsedSize, uiHeapFreeSize;
	IMG_UINT64 uiSize = 0, uiOffset = 0;
	IMG_DEV_PHYADDR sDevPAddr = {0};
	PVRSRV_ERROR eError;

	eError = LMA_HeapIteratorCreate(psDevNode, ui32Flags, ppsIter);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorCreate", return_error);

	eError = LMA_HeapIteratorGetHeapStats(*ppsIter, &uiHeapTotalSize,
	                                      &uiHeapUsedSize);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorGetHeapStats",
	                      return_error);
	uiHeapFreeSize = uiHeapTotalSize - uiHeapUsedSize;

	_DBG("(%s()) suspend snapshot heap stats: flag=0x%x heap=0x%x total=0x%" IMG_UINT64_FMTSPECx ", "
	     "used=0x%" IMG_UINT64_FMTSPECx ", free=0x%" IMG_UINT64_FMTSPECx,
	     __func__, ui32Flags, ePhysHeapIdx,
             uiHeapTotalSize, uiHeapUsedSize, uiHeapFreeSize);

    	*ppbBuf = OSAllocMem(uiHeapUsedSize);
	PVR_LOG_GOTO_IF_NOMEM(*ppbBuf, eError, destroy_iterator);

	while (LMA_HeapIteratorNext(*ppsIter, &sDevPAddr, &uiSize))
	{
#if defined(SUPPORT_DMA_TRANSFER)
		int i;
		void* pvOSData = NULL;
		tXDXDmaDualAddrs tDualAddrs;
		DMADevAddrType tDevAddr;
#endif
		void *pvCpuVAddr = NULL;
		IMG_CPU_PHYADDR sCpuPAddr = {0};

		if (uiOffset + uiSize > uiHeapUsedSize)
		{
			PVR_DPF((PVR_DBG_ERROR, "uiOffset = %" IMG_UINT64_FMTSPECx ", "
			        "uiSize = %" IMG_UINT64_FMTSPECx, uiOffset, uiSize));

			PVR_LOG_GOTO_WITH_ERROR("LMA_HeapIteratorNext", eError,
			                        PVRSRV_ERROR_INVALID_OFFSET,
			                        free_buffer);
		}

                if (gbSnapshotDmaRd == IMG_FALSE)
                {
                        PhysHeapDevPAddrToCpuPAddr(
                                   psDevNode->apsPhysHeap[ePhysHeapIdx],
                                   1,
                                   &sCpuPAddr,
                                   &sDevPAddr);
                        pvCpuVAddr = OSMapPhysToLin(sCpuPAddr, uiSize,
                                                    PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC);
                }
		_DBG("(%s()) suspend snapshot iterator: dev_paddr=%px, "
		     "size=0x%05" IMG_UINT64_FMTSPECx, __func__,
		     (void *) sDevPAddr.uiAddr, uiSize);

                if (gbSnapshotDmaRd == IMG_FALSE)
                {
                        /* copy memory */
                        memcpy(*ppbBuf + uiOffset, pvCpuVAddr, uiSize);
                }
#if defined(SUPPORT_DMA_TRANSFER)
                else
                {
                        /* TODO
                         * Performance is enough for now, s3 test less then 700ms,
                         * WIll use dma combine to accelerate transfer
                        * */
                        eError = OSDmaAllocData(psDevNode, 1, NULL, Cleanup, &pvOSData);

                        memset(&tDevAddr, 0, sizeof(DMADevAddrType));
                        tDevAddr.ui32AddrNums = uiSize / PAGE_SIZE;
                        tDevAddr.psDmaAddrVals = OSAllocZMem(tDevAddr.ui32AddrNums * (sizeof(IMG_DMA_ADDR) + sizeof(IMG_BOOL)));
                        tDevAddr.pbValids = (IMG_BOOL*)(&tDevAddr.psDmaAddrVals[tDevAddr.ui32AddrNums]);
                        for (i = 0; i < tDevAddr.ui32AddrNums; i++) {
                                tDevAddr.psDmaAddrVals[i].uiAddr = sDevPAddr.uiAddr + i * PAGE_SIZE;
                                tDevAddr.pbValids[i] = 1;
                        }

			memset(&tDualAddrs, 0, sizeof(tXDXDmaDualAddrs));
                        tDualAddrs.HostAddr1.puiAddress = (u64)(*ppbBuf + uiOffset);
                        tDualAddrs.HostAddr1.uiSize = uiSize;
                        tDualAddrs.DevAddrs2.pDmaPageVailds = tDevAddr.pbValids;
                        tDualAddrs.DevAddrs2.psDmaAddrs = (IMG_DMA_ADDR*)tDevAddr.psDmaAddrVals;
                        tDualAddrs.DevAddrs2.ui32DmaPages = tDevAddr.ui32AddrNums;
                        tDualAddrs.DevAddrs2.uiDmaOffse = 0;
                        eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTxChan,
                                                      XdxDmaTypeCPY, &tDualAddrs, uiSize, 0x2, pvOSData);

                        OSDmaSubmitTransfer(pvOSData, psDevNode->hDmaTxChan);
                        eError = OSDmaWaitDone(pvOSData);

                        OSFreeMem(tDevAddr.psDmaAddrVals);
                }
#endif

                if (gbSnapshotDmaRd == IMG_FALSE)
                {
                        /* and now poison it */
#ifdef ENABLE_PCIE_WRITE_FLUSH
                        OSCachedMemSetWMB(pvCpuVAddr, 0x9b, uiSize);
#else
                        memset(pvCpuVAddr, 0x9b, uiSize);
#endif
                }

		uiOffset += uiSize;

                if (gbSnapshotDmaRd == IMG_FALSE)
                {
		        OSUnMapPhysToLin(pvCpuVAddr, uiSize);
                }
	}

	return PVRSRV_OK;

free_buffer:
	OSFreeMem(*ppbBuf);
        *ppbBuf = NULL;
destroy_iterator:
	LMA_HeapIteratorDestroy(*ppsIter);
        *ppsIter = NULL;
return_error:
	return eError;
}

static PVRSRV_ERROR PrePower(IMG_HANDLE hSysData,
                             PVRSRV_SYS_POWER_STATE eNewPowerState,
                             PVRSRV_SYS_POWER_STATE eCurrentPowerState,
                             PVRSRV_POWER_FLAGS ePwrFlags)
{
   	struct xdx_device *psXdxDev = (struct xdx_device *)hSysData;
	PVRSRV_ERROR eError = PVRSRV_OK;
        int i = 0;

	_DBG("(%s()) suspend snapshot state: current=%u, new=%u; flags: 0x%08x", __func__,
	     eCurrentPowerState,
	     eNewPowerState, ePwrFlags);

	/* The transition might be both from ON or OFF states to OFF state so check
	 * only for the *new* state. Also this is only valid for suspend requests. */
	if (eNewPowerState != PVRSRV_SYS_POWER_STATE_OFF ||
	    !BITMASK_HAS(ePwrFlags, PVRSRV_POWER_FLAGS_SUSPEND_REQ))
	{
		return PVRSRV_OK;
	}

        if (!gui8SnapshotHeapCount)
        {
                dev_info(psXdxDev->dev, "init snapshot heap info: max=%d\n", PVRSRV_SNAPSHOT_HEAP_MAX);

                gasSnapshotHeap[gui8SnapshotHeapCount].ui32Flags = PHYS_HEAP_USAGE_GPU_LOCAL;
                gasSnapshotHeap[gui8SnapshotHeapCount].ePhysHeapIdx = PVRSRV_PHYS_HEAP_GPU_LOCAL;
                gui8SnapshotHeapCount++;
                dev_info(psXdxDev->dev, "init snapshot heap info: gpu local\n");

                if (psXdxDev->real_vram_size > psXdxDev->aper_size)
                {
#if defined(SUPPORT_DMA_TRANSFER)
                        gbSnapshotDmaWr = IMG_TRUE;
                        gbSnapshotDmaRd = IMG_TRUE;
#endif
                        if (gbSnapshotDmaWr == IMG_FALSE)
                        {
                                dev_err(psXdxDev->dev, "DMA write not defined for invisible heap!!\n");
                                PVR_ASSERT(0);
                        }
                        if (gbSnapshotDmaRd == IMG_FALSE)
                        {
                                dev_err(psXdxDev->dev, "DMA read not defined for invisible heap!!\n");
                                PVR_ASSERT(0);
                        }

                        gasSnapshotHeap[gui8SnapshotHeapCount].ui32Flags = PHYS_HEAP_USAGE_GPU_PRIVATE;
                        gasSnapshotHeap[gui8SnapshotHeapCount].ePhysHeapIdx = PVRSRV_PHYS_HEAP_GPU_PRIVATE;
                        gui8SnapshotHeapCount++;
                        dev_info(psXdxDev->dev, "init snapshot heap info: gpu private\n");
                }

#ifdef PVRSRV_PHYS_HEAP_MMU
                gasSnapshotHeap[gui8SnapshotHeapCount].ui32Flags = PHYS_HEAP_USAGE_MMU;
                gasSnapshotHeap[gui8SnapshotHeapCount].ePhysHeapIdx = PVRSRV_PHYS_HEAP_MMU;
                gui8SnapshotHeapCount++;
                dev_info(psXdxDev->dev, "init snapshot heap info: gpu mmu\n");
#endif
#ifdef PVRSRV_PHYS_HEAP_DC
                gasSnapshotHeap[gui8SnapshotHeapCount].ui32Flags = PHYS_HEAP_USAGE_DC;
                gasSnapshotHeap[gui8SnapshotHeapCount].ePhysHeapIdx = PVRSRV_PHYS_HEAP_DC;
                gui8SnapshotHeapCount++;
                dev_info(psXdxDev->dev, "init snapshot heap info: gpu dc\n");
#endif
        }
        PVR_ASSERT(gui8SnapshotHeapCount);
        dev_info(psXdxDev->dev, "snapshot save heapCnt=%d dmaRd=%d\n",
                 gui8SnapshotHeapCount, gbSnapshotDmaRd);

        PVR_ASSERT(ARRAY_SIZE(psXdxDev->gfx.pvr_heap_iter) < gui8SnapshotHeapCount);
        PVR_ASSERT(ARRAY_SIZE(psXdxDev->gfx.pvr_pwr_buffer) != ARRAY_SIZE(psXdxDev->gfx.pvr_heap_iter));

        for (i = 0; i < gui8SnapshotHeapCount; ++i)
        {
                eError = _PrePower(psXdxDev->gfx.pvrdev,
                                   gasSnapshotHeap[i].ui32Flags,
                                   gasSnapshotHeap[i].ePhysHeapIdx,
                                   (PHYS_HEAP_ITERATOR**)&psXdxDev->gfx.pvr_heap_iter[i],
                                   (IMG_BYTE **)&psXdxDev->gfx.pvr_pwr_buffer[i]);
                if (eError != PVRSRV_OK)
                {
                        dev_err(psXdxDev->dev, "snapshot save failed on heap=%d!!\n",
                                gasSnapshotHeap[i].ePhysHeapIdx);
                        goto reset_buffer;
                }

        }
        return PVRSRV_OK;

reset_buffer:
        _ResetSnapshotBuffer(psXdxDev);
        return eError;
}

static PVRSRV_ERROR _PostPower(PVRSRV_DEVICE_NODE *psDevNode,
                               PHYS_HEAP_USAGE_FLAGS ui32Flags,
                               PVRSRV_PHYS_HEAP ePhysHeapIdx,
                               PHYS_HEAP_ITERATOR **ppsIter,
                               IMG_BYTE **ppbBuf)
{
	IMG_DEV_PHYADDR sDevPAddr = {0};
	IMG_UINT64 uiSize = 0, uiOffset = 0;
	PVRSRV_ERROR eError;

	_DBG("(%s()) resume snapshot heap stats: flag=0x%x heap=0x%x",
	     __func__, ui32Flags, ePhysHeapIdx);

	eError = LMA_HeapIteratorReset(*ppsIter);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorReset", free_buffer);

	while (LMA_HeapIteratorNext(*ppsIter, &sDevPAddr, &uiSize))
	{
#if defined(SUPPORT_DMA_TRANSFER)
		int i;
		void* pvOSData = NULL;
		tXDXDmaDualAddrs tDualAddrs;
		DMADevAddrType tDevAddr;
#endif
		void *pvCpuVAddr = NULL;
		IMG_CPU_PHYADDR sCpuPAddr = {0};

                if (gbSnapshotDmaWr == IMG_FALSE)
                {
                        PhysHeapDevPAddrToCpuPAddr(psDevNode->apsPhysHeap[ePhysHeapIdx],
                                   1,
                                   &sCpuPAddr,
                                   &sDevPAddr);
                        pvCpuVAddr = OSMapPhysToLin(sCpuPAddr, uiSize,
                                               PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC);
                }
		_DBG("(%s()) resume snapshot iterator: dev_paddr=%px, "
		     "size=0x%05" IMG_UINT64_FMTSPECx, __func__,
		     (void *) sDevPAddr.uiAddr, uiSize);

                if (gbSnapshotDmaWr == IMG_FALSE)
                {
#ifdef ENABLE_PCIE_WRITE_FLUSH
                        OSCachedMemCopyWMB(pvCpuVAddr, *ppbBuf + uiOffset, uiSize);
#else
                        memcpy(pvCpuVAddr, *ppbBuf + uiOffset, uiSize);
#endif
                }
#if defined(SUPPORT_DMA_TRANSFER)
                else
                {
                        eError = OSDmaAllocData(psDevNode, 1, NULL, Cleanup, &pvOSData);

                        memset(&tDevAddr, 0, sizeof(DMADevAddrType));
                        tDevAddr.ui32AddrNums = uiSize / PAGE_SIZE;
                        tDevAddr.psDmaAddrVals = OSAllocZMem(tDevAddr.ui32AddrNums * (sizeof(IMG_DMA_ADDR) + sizeof(IMG_BOOL)));
                        tDevAddr.pbValids = (IMG_BOOL*)(&tDevAddr.psDmaAddrVals[tDevAddr.ui32AddrNums]);
                        for (i = 0; i < tDevAddr.ui32AddrNums; i++) {
                                tDevAddr.psDmaAddrVals[i].uiAddr = sDevPAddr.uiAddr + i * PAGE_SIZE;
                                tDevAddr.pbValids[i] = 1;
                        }

			memset(&tDualAddrs, 0, sizeof(tXDXDmaDualAddrs));
                        tDualAddrs.HostAddr1.puiAddress = (u64)(*ppbBuf + uiOffset);
                        tDualAddrs.HostAddr1.uiSize = uiSize;
                        tDualAddrs.DevAddrs2.pDmaPageVailds = tDevAddr.pbValids;
                        tDualAddrs.DevAddrs2.psDmaAddrs = (IMG_DMA_ADDR*)tDevAddr.psDmaAddrVals;
                        tDualAddrs.DevAddrs2.ui32DmaPages = tDevAddr.ui32AddrNums;
                        tDualAddrs.DevAddrs2.uiDmaOffse = 0;
                        eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTxChan,
                                                      XdxDmaTypeCPY, &tDualAddrs, uiSize,
                                                      0x1, pvOSData);

                        OSDmaSubmitTransfer(pvOSData, psDevNode->hDmaTxChan);
                        eError = OSDmaWaitDone(pvOSData);

                        OSFreeMem(tDevAddr.psDmaAddrVals);
                }
#endif

		uiOffset += uiSize;

                if (gbSnapshotDmaWr == IMG_FALSE)
                {
		        OSUnMapPhysToLin(pvCpuVAddr, uiSize);
                }
	}

	LMA_HeapIteratorDestroy(*ppsIter);
        *ppsIter = NULL;

	OSFreeMem(*ppbBuf);
	*ppbBuf = NULL;

	return PVRSRV_OK;

free_buffer:
	OSFreeMem(*ppbBuf);
	*ppbBuf = NULL;

	return eError;
}

static PVRSRV_ERROR PostPower(IMG_HANDLE hSysData,
                              PVRSRV_SYS_POWER_STATE eNewPowerState,
                              PVRSRV_SYS_POWER_STATE eCurrentPowerState,
                              PVRSRV_POWER_FLAGS ePwrFlags)
{
   	struct xdx_device *psXdxDev = (struct xdx_device *)hSysData;
	PVRSRV_ERROR eError = PVRSRV_OK;
        int i;

	_DBG("(%s()) resume snapshot state: current=%u, new=%u; flags=0x%08x", __func__,
	     eCurrentPowerState,
	     eNewPowerState, ePwrFlags);

	/* The transition might be both to ON or OFF states from OFF state so check
	 * only for the *current* state. Also this is only valid for resume requests. */
	if (eCurrentPowerState != PVRSRV_SYS_POWER_STATE_OFF ||
	    !BITMASK_HAS(ePwrFlags, PVRSRV_POWER_FLAGS_RESUME_REQ))
	{
		return PVRSRV_OK;
	}

#if !defined(SUPPORT_DONT_SNOOP) && !defined(__arm64__) && !defined(__aarch64__) && !defined(__loongarch__)
	{
		uint32_t host_checkin;

		host_checkin = xdx_mmio_readl(psXdxDev, PANGU_REG_HOST_MISC_BASE + mmHOST_MISC_CHECKIN);
		host_checkin &= ~HOST_MISC_CHECKIN_DF_HOST_NO_SNOOP_ENABLE_MASK;
		xdx_mmio_writel(psXdxDev, PANGU_REG_HOST_MISC_BASE + mmHOST_MISC_CHECKIN, host_checkin);
	}
#endif
        PVR_ASSERT(gui8SnapshotHeapCount);
        dev_info(psXdxDev->dev, "snapshot restore heapCnt=%d dmaWr=%d\n",
                 gui8SnapshotHeapCount, gbSnapshotDmaWr);

        PVR_ASSERT(ARRAY_SIZE(psXdxDev->gfx.pvr_heap_iter) < gui8SnapshotHeapCount);
        PVR_ASSERT(ARRAY_SIZE(psXdxDev->gfx.pvr_pwr_buffer) != ARRAY_SIZE(psXdxDev->gfx.pvr_heap_iter));

        for (i = 0; i < gui8SnapshotHeapCount; ++i)
        {
                PVR_ASSERT(psXdxDev->gfx.pvr_heap_iter[i]);
                PVR_ASSERT(psXdxDev->gfx.pvr_pwr_buffer[i]);

                eError = _PostPower(psXdxDev->gfx.pvrdev,
                                    gasSnapshotHeap[i].ui32Flags,
                                    gasSnapshotHeap[i].ePhysHeapIdx,
                                    (PHYS_HEAP_ITERATOR**)&psXdxDev->gfx.pvr_heap_iter[i],
                                    (IMG_BYTE **)&psXdxDev->gfx.pvr_pwr_buffer[i]);
                if (eError != PVRSRV_OK)
                {
                        dev_err(psXdxDev->dev, "snapshot restore failed on heap=%d!!\n",
                                gasSnapshotHeap[i].ePhysHeapIdx);
                        goto reset_buffer;
                }
        }
        return PVRSRV_OK;

reset_buffer:
        _ResetSnapshotBuffer(psXdxDev);
        return eError;
}

PVRSRV_ERROR SysDevInit(void *pvOSDevice, PVRSRV_DEVICE_CONFIG **ppsDevConfig)
{
	struct device *dev = pvOSDevice;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	PVRSRV_DEVICE_CONFIG *psDevConfig;
	RGX_DATA *psRGXData;
	RGX_TIMING_INFORMATION *psRGXTimingInfo;
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	uint32_t alloc_size = sizeof(PVRSRV_DEVICE_CONFIG) +
			      sizeof(*psRGXData) + sizeof(*psRGXTimingInfo) +
			      sizeof(PHYS_HEAP_CONFIG) * HEAP_NUMS;

#ifdef CONFIG_FW_MAIN_ON_SYSMEM
	alloc_size += sizeof(PHYS_HEAP_CONFIG);
#endif

	psDevConfig = kmalloc(alloc_size, GFP_KERNEL);
	if (!psDevConfig)
		return PVRSRV_ERROR_OUT_OF_MEMORY;

	memset(psDevConfig, 0, alloc_size);

	psRGXData = (RGX_DATA *)((uint8_t *)psDevConfig + sizeof(*psDevConfig));
	psRGXTimingInfo = (RGX_TIMING_INFORMATION *)((uint8_t *)psRGXData +
						     sizeof(*psRGXData));
	pasPhysHeaps = (PHYS_HEAP_CONFIG *)((uint8_t *)psRGXTimingInfo +
					    sizeof(*psRGXTimingInfo));

	/* Setup RGX specific timing data */
	psRGXTimingInfo->ui32CoreClockSpeed = xdev->pm.current_gfxclk;
	psRGXTimingInfo->bEnableActivePM = IMG_FALSE;
	psRGXTimingInfo->bEnableRDPowIsland = IMG_FALSE;
	psRGXTimingInfo->ui32ActivePMLatencyms = 10;

	psRGXData->psRGXTimingInfo = psRGXTimingInfo;

	/* Setup DevConfig */
	psDevConfig->pvOSDevice = pvOSDevice;
	psDevConfig->hDevData = psRGXData;
	psDevConfig->pszName = "xdxgpu";
	psDevConfig->ui32IRQ = XDX_IRQ_CLIENTID_GPU;

	/* Setup heaps */
	pvr_init_phys_heaps(xdev, pasPhysHeaps,
			    &psDevConfig->ui32PhysHeapCount);
	psDevConfig->pasPhysHeaps = pasPhysHeaps;
	psDevConfig->eDefaultHeap = PVRSRV_PHYS_HEAP_GPU_LOCAL;

	psDevConfig->sRegsCpuPBase.uiAddr = xdev->rmmio_base + xdev->hw_info.gfx_regs.base;
	psDevConfig->ui32RegsSize = xdev->hw_info.gfx_regs.size;

#if defined(SUPPORT_DMA_TRANSFER)
	psDevConfig->pszDmaTxChanName = "xdx-dma";
	psDevConfig->pszDmaRxChanName = "xdx-dma";
	psDevConfig->bHasDma = true;
	/* Following two values are expressed in number of bytes */
	psDevConfig->ui32DmaTransferUnit = 8;
	psDevConfig->ui32DmaAlignment = 32;
#endif

	psDevConfig->eCacheSnoopingMode = PVRSRV_DEVICE_SNOOP_NONE;
	psDevConfig->bHasNonMappableLocalMemory = IMG_FALSE;
	psDevConfig->bHasFBCDCVersion31 = IMG_FALSE;

	/* Set psDevConfig->pfnSysDevErrorNotify callback */
	psDevConfig->pfnSysDevErrorNotify = _pvr_error_notify;

	/* power management on HW system */
	psDevConfig->pfnPrePowerState = PrePower;
	psDevConfig->pfnPostPowerState = PostPower;

	/* clock frequency */
	psDevConfig->pfnClockFreqGet = NULL;

	psDevConfig->hSysData = xdev;

	*ppsDevConfig = psDevConfig;

	return PVRSRV_OK;
}

void SysDevDeInit(PVRSRV_DEVICE_CONFIG *psDevConfig)
{

	kfree(psDevConfig);
}

struct pvr_irq {
	struct xdx_irq_src base;
	PFN_LISR pfnLISR;
	void *pvData;
	struct xdx_device *xdev;
};

static inline struct pvr_irq *xdx_to_pvr(struct xdx_irq_src *src)
{
	return container_of(src, struct pvr_irq, base);
}

static int pvr_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
		       unsigned type, enum xdx_interrupt_state state)
{
	uint32_t offset = PANGU_REG_HOST_MISC_BASE + mmMSI_MSIX_HW_ENABLE_PF0;
	uint32_t val = xdx_mmio_readl(xdev, offset);
	uint32_t all_gpu_core_mask;

	if (xdev->asic_type == CHIP_PANGU_A0) {
		all_gpu_core_mask =
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE0)) |
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE1)) |
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE2)) |
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE3));
	} else {
		all_gpu_core_mask =
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE0)) |
			BIT(xdev->soc_funcs.get_msix_idx(MSIX_GPU_CORE1));
	}

	if(state == XDX_IRQ_STATE_ENABLE){
		val = val | all_gpu_core_mask;
	}else{
		val = val & (~all_gpu_core_mask);
	}
	xdx_mmio_writel(xdev, offset, val);
	return 0;
}

/* Modified from SampleIRQCount */
static inline IMG_BOOL XDXCompareIRQCount(const volatile IMG_UINT32 * paui32Input,
		volatile IMG_UINT32 * paui32Output)
{
	IMG_UINT32 ui32TID;
	IMG_BOOL bReturnVal = IMG_FALSE;

	for (ui32TID = 0; ui32TID < RGXFW_THREAD_NUM; ui32TID++) {
		if (paui32Output[ui32TID] != paui32Input[ui32TID]) {
			bReturnVal = IMG_TRUE;
			break;
		}
	}
	return bReturnVal;
}

static int pvr_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
			   struct xdx_irq_entry *entry)
{
	struct pvr_irq *irq = xdx_to_pvr(src);

	if (entry->poll_enable) {
		PVRSRV_DEVICE_NODE *psDeviceNode = irq->pvData;
		PVRSRV_RGXDEV_INFO *psDevInfo = psDeviceNode->pvDevice;
		const RGXFWIF_OSDATA *psFwOsData = psDevInfo->psRGXFWIfFwOsData;

		if (!XDXCompareIRQCount(psFwOsData->aui32InterruptCount,
				psDevInfo->aui32SampleIRQCount)) {
			/* return when no GPU irq raised */
			return -EIO;
		}
	}

	if (!xdev->irq.msi_enabled) {
		/* clear hardware irq status */
		xdx_mmio_writeb(xdev, PANGU_REG_DOORBELL_126, 0);
	}

	if (irq->pfnLISR(irq->pvData))
		return 0;

	return -EIO;
}

static struct xdx_irq_src_funcs pvr_src_funcs = {
	.set = pvr_irq_set,
	.process = pvr_irq_process,
};

PVRSRV_ERROR SysInstallDeviceLISR(IMG_HANDLE hSysData, IMG_UINT32 ui32IRQ,
				  const IMG_CHAR *pszName, PFN_LISR pfnLISR,
				  void *pvData, IMG_HANDLE *phLISRData)
{
	struct xdx_device *xdev = (struct xdx_device *)hSysData;
	struct pvr_irq *irq;

	irq = kmalloc(sizeof(struct pvr_irq), GFP_KERNEL);
	if (!irq)
		return PVRSRV_ERROR_OUT_OF_MEMORY;

	irq->base.id = XDX_IRQ_CLIENTID_GPU;
	irq->base.funcs = &pvr_src_funcs;
	irq->pfnLISR = pfnLISR;
	irq->pvData = pvData;
	irq->xdev = xdev;

	xdx_irq_register_src(xdev, &irq->base);
	xdev->gfx.irq_src = &irq->base;

	*phLISRData = (IMG_HANDLE)irq;
	return PVRSRV_OK;
}

PVRSRV_ERROR SysUninstallDeviceLISR(IMG_HANDLE hLISRData)
{
	struct pvr_irq *irq = (struct pvr_irq *)hLISRData;
	struct xdx_device *xdev = irq->xdev;
	int ret;

	ret = xdx_irq_unregister_src(xdev, &irq->base);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
		return PVRSRV_ERROR_NOT_READY;
	}

	kfree(irq);
	xdev->gfx.irq_src = NULL;

	return PVRSRV_OK;
}

PVRSRV_ERROR SysDebugInfo(PVRSRV_DEVICE_CONFIG *psDevConfig,
			  DUMPDEBUG_PRINTF_FUNC *pfnDumpDebugPrintf,
			  void *pvDumpDebugFile)
{
	PVR_UNREFERENCED_PARAMETER(psDevConfig);
	PVR_UNREFERENCED_PARAMETER(pfnDumpDebugPrintf);
	PVR_UNREFERENCED_PARAMETER(pvDumpDebugFile);

	return PVRSRV_OK;
}
