/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/device.h>
#include "img_defs.h"
#include "pvrsrv_error.h"
#include "pvr_debug.h"
#include "dllist.h"
#include "osfunc.h"
#include "allocmem.h"
#include "physmem_lma.h"

#if defined(SUPPORT_RGX)
# include "rgx_memallocflags.h"
# include "rgxmmudefs_km.h"
#endif

#include "pvr_notifier.h"
#include "pvrsrv.h"
#include "htbuffer.h"
#include "pvr_ricommon.h"

#if defined(SUPPORT_GPUVIRT_VALIDATION)
#include "physmem_lma.h"
#endif

#include "smmu_platform.h"
#include "smmu_common.h"
#include "mmu_common.h"
#include "xdx_smmu_drv.h"
#include "xdx_device.h"
#include "pmr.h"
#include "devicemem_server_utils.h"
#include "xdx_gfx_convert.h"

#define INVALID_SMMU_PAGE_ADDR ~((IMG_UINT64)0x0)

struct smmu_mem_mapping_s
{
	/*! OS/system Handle for this allocation */
	PG_HANDLE				sMemHandle;
	/*! CPU virtual address of this allocation */
	void					*pvCpuVAddr;
	/*! Device physical address of this allocation */
	IMG_DEV_PHYADDR			dev_phy_addr;
	/*! Size of this allocation */
	size_t					uiSize;
	/*! Number of current mappings of this allocation */
	uint32_t				uiCpuVAddrRefCount;
	/*! Node for the defer free list */
	DLLIST_NODE				sMMUMappingItem;
	struct xdx_smmu_s	*xsmmu;
};

struct _PMR_SMMU_INFO_TAG
{
	struct smmu_context_s		*smmu_ctx;
	IMG_UINT64					ui64LogicSize;
	IMG_UINT64					ui64MapedOffset;
	IMG_UINT64					ui64SMMUVirAddr;
	IMG_BOOL 					bSparseAlloc;
	IMG_BOOL 					bPhysContig;
	DLLIST_NODE					lMapList;
	struct list_head			ctx_list;
};

struct xdx_smmu_env_s {
	PHYS_HEAP	*psPageHeap;
	RA_ARENA	*psPageMemRA;
	IMG_BOOL	bReady;
	IMG_UINT32	ui32OSid;
	POS_LOCK	pLock;
	struct xdx_smmu_s *xsmmu;
	struct DLLIST_NODE_TAG sTmpMMUMappingHead;
};

struct xdx_smmu_ctx_env_s {
	RA_ARENA *psVaddrRA;
	IMG_UINT64 Vaddr_base;
	IMG_UINT64 Vaddr_size;
	struct smmu_context_s *smmu_ctx;
};

static PVRSRV_ERROR _SMMU_PhysMem_RAImportAlloc(RA_PERARENA_HANDLE hArenaHandle,
		RA_LENGTH_T uiSize,
		RA_FLAGS_T uiflags,
		RA_LENGTH_T uBaseAlignment,
		const IMG_CHAR *pszAnnotation,
		RA_BASE_T *puiBase,
		RA_LENGTH_T *puiActualSize,
		RA_PERISPAN_HANDLE *phPriv)
{
	struct xdx_smmu_s *xsmmu = (struct xdx_smmu_s *)hArenaHandle;
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)xsmmu->env_data;
	struct smmu_mem_mapping_s *psMapping;
	PVRSRV_ERROR eError;
	uint32_t uiPid = 0;

	PVR_UNREFERENCED_PARAMETER(pszAnnotation);
	PVR_UNREFERENCED_PARAMETER(uiflags);
	PVR_UNREFERENCED_PARAMETER(uBaseAlignment);

	psMapping = OSAllocMem(sizeof(struct smmu_mem_mapping_s));
	PVR_GOTO_IF_NOMEM(psMapping, eError, e0);

#if defined(SUPPORT_GPUVIRT_VALIDATION)
	/*
	 * Store the OSid in the PG_HANDLE.uiOSid field for use by the
	 * pfnDevPxFree() routine.
	 */
	psMapping->sMemHandle.uiOSid = smmu_env->ui32OSid;
	eError = PhysHeapPagesAllocGPV(smmu_env->psPageHeap,
			TRUNCATE_64BITS_TO_SIZE_T(uiSize),
			&psMapping->sMemHandle,
			&psMapping->dev_phy_addr,
			smmu_env->ui32OSid,
			uiPid);
#else
	eError = PhysHeapPagesAlloc(smmu_env->psPageHeap,
			TRUNCATE_64BITS_TO_SIZE_T(uiSize),
			&psMapping->sMemHandle,
			&psMapping->dev_phy_addr,
			uiPid);
#endif
	PVR_GOTO_IF_ERROR(eError, e1);

	psMapping->xsmmu = xsmmu;
	psMapping->uiSize = TRUNCATE_64BITS_TO_SIZE_T(uiSize);
	psMapping->uiCpuVAddrRefCount = 0;
	*phPriv = (RA_PERISPAN_HANDLE) psMapping;
	/* Note: This assumes this memory never gets paged out */
	*puiBase = (RA_BASE_T)psMapping->dev_phy_addr.uiAddr;
	*puiActualSize = uiSize;
	SMMU_DBG(xsmmu->xdx_dev->dev,
			"psMapping:%llx Alloc:%llx",
			(IMG_UINT64)psMapping, psMapping->dev_phy_addr.uiAddr);
	return PVRSRV_OK;
e1:
	OSFreeMem(psMapping);
e0:
	return eError;
}

static void _SMMU_PhysMem_RAImportFree(RA_PERARENA_HANDLE hArenaHandle,
		RA_BASE_T uiBase, RA_PERISPAN_HANDLE hPriv)
{
	struct smmu_mem_mapping_s *psMapping = (struct smmu_mem_mapping_s *)hPriv;
	struct xdx_smmu_s *xsmmu = (struct xdx_smmu_s *)hArenaHandle;
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)xsmmu->env_data;
	PVR_UNREFERENCED_PARAMETER(uiBase);

	/* Check we have dropped all CPU mappings */
	PVR_ASSERT(psMapping->uiCpuVAddrRefCount == 0);

	/* Add mapping to defer free list */
	psMapping->xsmmu = NULL;
	OSLockAcquire(smmu_env->pLock);
	dllist_add_to_tail(&smmu_env->sTmpMMUMappingHead, &psMapping->sMMUMappingItem);
	SMMU_DBG(xsmmu->xdx_dev->dev,
			"psMapping:%llx Free ADD:%llx",
			(IMG_UINT64)psMapping, psMapping->dev_phy_addr.uiAddr);
	OSLockRelease(smmu_env->pLock);
}

void xdx_smmu_cleanup_smmu_env(struct xdx_smmu_s *xsmmu)
{
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)xsmmu->env_data;
	PDLLIST_NODE psNode, psNextNode;
	DLLIST_NODE psNewNode;

	OSLockAcquire(smmu_env->pLock);
	dllist_replace_head(&smmu_env->sTmpMMUMappingHead, &psNewNode);
	OSLockRelease(smmu_env->pLock);

	dllist_foreach_node(&psNewNode, psNode, psNextNode)
	{
		struct smmu_mem_mapping_s *psMapping =
				IMG_CONTAINER_OF(psNode, struct smmu_mem_mapping_s, sMMUMappingItem);
		PhysHeapPagesFree(smmu_env->psPageHeap, &psMapping->sMemHandle);
		SMMU_DBG(xsmmu->xdx_dev->dev,
				"psMapping:%llx, PhysHeapPagesFree:%llx ",
				(IMG_UINT64)psMapping, psMapping->dev_phy_addr.uiAddr);
		dllist_remove_node(psNode);
		OSFreeMem(psMapping);
	}
}

int xdx_smmu_free_phys_mem(
		struct smmu_context_s *smmu_ctx, struct smmu_mem_desc_s *psMemDesc)
{
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)smmu_ctx->xsmmu->env_data;

	PVR_ASSERT(psMemDesc->bValid);
	if (--psMemDesc->psMapping->uiCpuVAddrRefCount == 0)
	{
		SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"PhysHeapPagesUnMap:%llx ", psMemDesc->dev_phy_addr.uiAddr);
		PhysHeapPagesUnMap(smmu_env->psPageHeap,
		                   &psMemDesc->psMapping->sMemHandle,
		                   psMemDesc->psMapping->pvCpuVAddr);
	}
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"pvCpuVAddr:%pK RA:%llx Free",
		psMemDesc->pvCpuVAddr, psMemDesc->dev_phy_addr.uiAddr);
	psMemDesc->pvCpuVAddr = NULL;
	RA_Free(smmu_env->psPageMemRA, (RA_BASE_T)psMemDesc->dev_phy_addr.uiAddr);
	psMemDesc->bValid = IMG_FALSE;
	return 0;
}

int xdx_smmu_alloc_phys_mem(struct smmu_context_s *smmu_ctx,
	struct smmu_mem_desc_s *psMemDesc, size_t uiBytes, size_t uiAlignment)
{
	PVRSRV_ERROR eError;
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)smmu_ctx->xsmmu->env_data;
	RA_BASE_T uiPhysAddr;

	eError = RA_Alloc(smmu_env->psPageMemRA,
			uiBytes,
			RA_NO_IMPORT_MULTIPLIER,
			0, /* flags */
			uiAlignment,
			"",
			&uiPhysAddr,
			NULL,
			(RA_PERISPAN_HANDLE *)&psMemDesc->psMapping);

	PVR_LOG_RETURN_IF_ERROR(eError, "RA_Alloc");

	psMemDesc->bValid = IMG_TRUE;
	psMemDesc->pvCpuVAddr = NULL;
	psMemDesc->dev_phy_addr.uiAddr = (uint64_t) uiPhysAddr;

	if (psMemDesc->psMapping->uiCpuVAddrRefCount == 0)
	{
		eError = PhysHeapPagesMap(smmu_env->psPageHeap,
					&psMemDesc->psMapping->sMemHandle,
					psMemDesc->psMapping->uiSize,
					&psMemDesc->psMapping->dev_phy_addr,
					&psMemDesc->psMapping->pvCpuVAddr);
		if (eError != PVRSRV_OK)
		{
			RA_Free(smmu_env->psPageMemRA, psMemDesc->dev_phy_addr.uiAddr);
			return -ENOMEM;
		}
		SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"PhysPagesMap:%llx -> pvCpuVAddr:%pK",
			psMemDesc->dev_phy_addr.uiAddr,
			psMemDesc->psMapping->pvCpuVAddr);
	}

	psMemDesc->psMapping->uiCpuVAddrRefCount++;
	psMemDesc->uiOffset =
		(psMemDesc->dev_phy_addr.uiAddr - psMemDesc->psMapping->dev_phy_addr.uiAddr);
	psMemDesc->pvCpuVAddr =
		(IMG_UINT8 *)psMemDesc->psMapping->pvCpuVAddr + psMemDesc->uiOffset;
	psMemDesc->uiSize = uiBytes;
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"Alloc pvCpuVAddr:%pK RA:%llx alloc",
		psMemDesc->pvCpuVAddr, psMemDesc->dev_phy_addr.uiAddr);
	PVR_ASSERT(psMemDesc->pvCpuVAddr != NULL);
	return 0;
}

int xdx_smmu_copy_pxe_entry(void *page_entry, void *entry_data, int size)
{
	OSCachedMemCopyWMB(page_entry, entry_data, size);
	return 0;
}

int xdx_smmu_phypage_clean(
	struct smmu_context_s *smmu_ctx, struct smmu_mem_desc_s *sMemDesc,
	const struct smmu_pxe_config_s *pxe_cfg, IMG_UINT32 uiOffset, IMG_UINT32 uiLength)
{
	PVRSRV_ERROR eError;
	struct xdx_smmu_env_s *smmu_env = (struct xdx_smmu_env_s *)smmu_ctx->xsmmu->env_data;

	eError = PhysHeapPagesClean(smmu_env->psPageHeap, &sMemDesc->psMapping->sMemHandle,
			uiOffset * pxe_cfg->uiBytesPerEntry + sMemDesc->uiOffset,
			uiLength * pxe_cfg->uiBytesPerEntry);
	if (eError != PVRSRV_OK)
		return -EIO;
	return 0;
}

/* Allocate the requested device virtual address region
 * from the heap */
static PVRSRV_ERROR _XDXSMMUVAAllocAndMap(
		struct smmu_context_s *smmu_ctx,
		PMR *psPMR, IMG_UINT64 *u64SMMUVAddr)
{
	struct xdx_smmu_env_s *smmu_env;
	struct xdx_smmu_ctx_env_s *ctx_env;
	PDLLIST_NODE plListHead = NULL;
	DLLIST_NODE *psThis, *psNext;
	PVRSRV_ERROR eError;
	PVRSRV_ERROR eErrorLock;
	int ret;
	IMG_DEVMEM_SIZE_T uiAllocationActualSize = 0;
	IMG_DEV_VIRTADDR psDevVAddr;
	IMG_UINT32 ui32NumDevPages;
	IMG_DEVMEM_SIZE_T ui32AllocedSize;
	IMG_DEVMEM_SIZE_T ui32PadSize;
	struct _PMR_SMMU_INFO_TAG *PmrSMMUInfo = NULL;
	IMG_DEV_PHYADDR asDevPAddr[PMR_MAX_TRANSLATION_STACK_ALLOC];
	IMG_BOOL abValid[PMR_MAX_TRANSLATION_STACK_ALLOC];
	IMG_DEV_PHYADDR *psDevPAddr = NULL;
	IMG_BOOL *pbValid = NULL;

	if (smmu_ctx == NULL)
		return PVRSRV_ERROR_NOT_SUPPORTED;
	if (psPMR == NULL || u64SMMUVAddr == NULL)
		return PVRSRV_ERROR_INVALID_PARAMS;
	smmu_env = (struct xdx_smmu_env_s *)smmu_ctx->xsmmu->env_data;
	ctx_env = (struct xdx_smmu_ctx_env_s *)smmu_ctx->ctx_env_data;

	OSLockAcquire(smmu_env->pLock);
	plListHead = (PDLLIST_NODE)PMRGetSmmuMapList(psPMR);
	if (!plListHead)
	{
		OSLockRelease(smmu_env->pLock);
		return PVRSRV_ERROR_NOT_SUPPORTED;
	}
	dllist_foreach_node(plListHead, psThis, psNext)
	{
		struct _PMR_SMMU_INFO_TAG *_PmrSMMUInfo;

		_PmrSMMUInfo = IMG_CONTAINER_OF(psThis, struct _PMR_SMMU_INFO_TAG, lMapList);
		if (_PmrSMMUInfo->smmu_ctx == smmu_ctx)
		{
			PmrSMMUInfo = _PmrSMMUInfo;
			break;
		}
	}
	OSLockRelease(smmu_env->pLock);

	if (!PmrSMMUInfo)
	{
		PmrSMMUInfo = OSAllocZMem(sizeof(struct _PMR_SMMU_INFO_TAG));
		PVR_GOTO_IF_NOMEM(PmrSMMUInfo, eError, e0);
		PmrSMMUInfo->ui64SMMUVirAddr = INVALID_SMMU_PAGE_ADDR;
		PmrSMMUInfo->smmu_ctx = smmu_ctx;
		PmrSMMUInfo->ui64MapedOffset = 0;
	}
	ui32AllocedSize = RGXGetGDDRHeapAllocedSize(PMRGetFlavourData(psPMR));
	if (PmrSMMUInfo->ui64MapedOffset && ui32AllocedSize != PmrSMMUInfo->ui64MapedOffset)
	{
		SMMU_DBG(xdx_smmu_dev(smmu_ctx),
			"ui64MapedOffset = %llx  ui32AllocedSize:%llx",
			PmrSMMUInfo->ui64MapedOffset ,ui32AllocedSize);
	}
	else if(INVALID_SMMU_PAGE_ADDR != PmrSMMUInfo->ui64SMMUVirAddr && !PMR_IsSparse(psPMR))
	{
		goto success_out;
	}
	else if (INVALID_SMMU_PAGE_ADDR == PmrSMMUInfo->ui64SMMUVirAddr)
	{
		PMR_PhysicalSize(psPMR, &PmrSMMUInfo->ui64LogicSize);
		eError = RA_Alloc(ctx_env->psVaddrRA,
				PmrSMMUInfo->ui64LogicSize,
				RA_NO_IMPORT_MULTIPLIER,
				0, /* flags: this RA doesn't use flags*/
				PAGE_SIZE,
				"SVM_Virtual_Alloc",
				&(PmrSMMUInfo->ui64SMMUVirAddr),
				&uiAllocationActualSize,
				NULL);/* don't care about per-import priv data */
		PVR_LOG_GOTO_IF_ERROR(eError, "RA_Alloc", e0);
	}

	if(PmrSMMUInfo->ui64MapedOffset != ui32AllocedSize)
	{
		psDevVAddr.uiAddr = PmrSMMUInfo->ui64SMMUVirAddr + PmrSMMUInfo->ui64MapedOffset;
		SMMU_DBG(xdx_smmu_dev(smmu_ctx), "ui64SMMUVirAddr = %llx "
			"psDevVAddr = %llx ui32AllocedSize = %llx",
			(uint64_t)PmrSMMUInfo->ui64SMMUVirAddr, psDevVAddr.uiAddr,ui32AllocedSize);
//		ui32AllocedSize = ui32AllocedSize - PmrSMMUInfo->ui64MapedOffset;
		ui32PadSize = ui32AllocedSize - PmrSMMUInfo->ui64MapedOffset;

		ret = xdx_smmu_alloc(smmu_ctx,
						ui32PadSize,
						NULL,
						0, /* IMG_UINT32 uiProtFlags */
						0, /* alignment is n/a since we supply devvaddr */
						psDevVAddr.uiAddr,
						RGX_HEAP_4KB_PAGE_SHIFT);
		if (ret < 0)
		{
			eError = PVRSRV_ERROR_NOT_IMPLEMENTED;
			SMMU_ERR(xdx_smmu_dev(smmu_ctx), "xdx_smmu_alloc() failed (%d)", ret);
			goto e1;
		}

		// /* number of pages (device pages) that allocation spans */
		ui32NumDevPages = 0xffffffffU & (((ui32AllocedSize - 1) >> RGX_HEAP_4KB_PAGE_SHIFT) + 1);

		PmrSMMUInfo->bPhysContig = PMR_IsPhysContig(PMRGetFlavourData(psPMR));
		PmrSMMUInfo->bSparseAlloc = PMR_IsSparse(psPMR);
		if (PmrSMMUInfo->bPhysContig != IMG_TRUE)
		{
			if (ui32NumDevPages > PMR_MAX_TRANSLATION_STACK_ALLOC)
			{
				psDevPAddr = OSAllocMem(ui32NumDevPages * sizeof(IMG_DEV_PHYADDR));
				PVR_LOG_GOTO_IF_NOMEM(psDevPAddr, eError, e2);
				pbValid = OSAllocMem(ui32NumDevPages * sizeof(IMG_BOOL));
				if (pbValid == NULL)
				{
					/* Should allocation fail, clean-up here before exit */
					OSFreeMem(psDevPAddr);
					PVR_LOG_GOTO_WITH_ERROR("pbValid", eError, PVRSRV_ERROR_OUT_OF_MEMORY, e2);
				}
			}
			else
			{
				psDevPAddr = asDevPAddr;
				pbValid = abValid;
			}
			eError = PMRLockSysPhysAddresses(psPMR);
			PVR_GOTO_IF_ERROR(eError, e3);
			eError = PMR_DevPhysAddr(psPMR,
				RGX_HEAP_4KB_PAGE_SHIFT,
				ui32NumDevPages,
				0,
				psDevPAddr,
				pbValid);
			eErrorLock = PMRUnlockSysPhysAddresses(psPMR);
			PVR_GOTO_IF_ERROR(eError, e3);
			PVR_GOTO_IF_ERROR(eErrorLock, e3);
		}
		else
		{
			psDevPAddr = asDevPAddr;
			pbValid = abValid;
			eError = PMRLockSysPhysAddresses(psPMR);
			PVR_GOTO_IF_ERROR(eError, e2);
			eError = PMR_DevPhysAddr(psPMR,
				RGX_HEAP_4KB_PAGE_SHIFT,
				1,
				((IMG_DEVMEM_OFFSET_T)PmrSMMUInfo->ui64MapedOffset << RGX_HEAP_4KB_PAGE_SHIFT),
				psDevPAddr,
				pbValid);
			eErrorLock = PMRUnlockSysPhysAddresses(psPMR);
			PVR_GOTO_IF_ERROR(eError, e2);
			PVR_GOTO_IF_ERROR(eErrorLock, e2);
		}
		if(PmrSMMUInfo->bSparseAlloc)
		{
			if (PmrSMMUInfo->ui64MapedOffset)
			{
				smmu_unmap_pages(smmu_ctx, PmrSMMUInfo->ui64SMMUVirAddr, NULL,
					(IMG_DEVMEM_OFFSET_T)PmrSMMUInfo->ui64MapedOffset << RGX_HEAP_4KB_PAGE_SHIFT);
			}
			ret = smmu_map_pages(smmu_ctx,
								PmrSMMUInfo->ui64SMMUVirAddr,
								0,
								(uint64_t *)psDevPAddr,
								pbValid,
								NULL,
								ui32NumDevPages);
			if (ret < 0)
			{
				eError = PVRSRV_ERROR_NOT_IMPLEMENTED;
				SMMU_ERR(xdx_smmu_dev(smmu_ctx), "smmu_map_pages() failed (%d)", ret);
				goto e3;
			}
		}
		else if(PmrSMMUInfo->bPhysContig == IMG_TRUE)
		{
			ret = smmu_map_pages_linear(smmu_ctx,
								psDevVAddr.uiAddr,
								psDevPAddr[0].uiAddr,
								ui32NumDevPages);
			if (ret < 0)
			{
				eError = PVRSRV_ERROR_NOT_IMPLEMENTED;
				SMMU_ERR(xdx_smmu_dev(smmu_ctx),
					"smmu_map_pages_linear() failed (%d)", ret);
				goto e3;
			}
		}
		else
		{
			ret = smmu_map_pages_fast(smmu_ctx,
								psDevVAddr.uiAddr,
								PmrSMMUInfo->ui64MapedOffset,
								(uint64_t *)psDevPAddr,
								ui32NumDevPages);
			if (ret < 0)
			{
				eError = PVRSRV_ERROR_NOT_IMPLEMENTED;
				SMMU_ERR(xdx_smmu_dev(smmu_ctx), "smmu_map_pages_fast() failed (%d)", ret);
				goto e3;
			}
		}
		SMMU_DBG(xdx_smmu_dev(smmu_ctx), "SMMU MAP:%llx!", psDevVAddr.uiAddr);
		xdx_smmu_group_sync_tlb(xdx_smmu_ctx_group(smmu_ctx));
		PmrSMMUInfo->ui64MapedOffset += ui32PadSize;
	}
	OSLockAcquire(smmu_env->pLock);
	dllist_add_to_tail((PDLLIST_NODE)PMRGetSmmuMapList(psPMR), &PmrSMMUInfo->lMapList);
	OSLockRelease(smmu_env->pLock);
	mutex_lock(&smmu_ctx->lock);
	list_add_tail(&PmrSMMUInfo->ctx_list, &smmu_ctx->maped_head);
	mutex_unlock(&smmu_ctx->lock);

	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"PmrSMMUInfo:%llx SMMUVAddr=  %llx + %llx (%llx) psPMR = %llx",
		(uint64_t)PmrSMMUInfo,
		PmrSMMUInfo->ui64SMMUVirAddr, PmrSMMUInfo->ui64LogicSize,
		uiAllocationActualSize, (uint64_t)psPMR);
success_out:
	*u64SMMUVAddr = PmrSMMUInfo->ui64SMMUVirAddr;
	return PVRSRV_OK;
e3:
	if (pbValid && pbValid != abValid)
	{
		OSFreeMem(psDevPAddr);
		OSFreeMem(pbValid);
	}
e2:
	xdx_smmu_free(smmu_ctx, psDevVAddr.uiAddr, ui32AllocedSize, RGX_HEAP_4KB_PAGE_SHIFT);
e1:
	RA_Free(ctx_env->psVaddrRA, PmrSMMUInfo->ui64SMMUVirAddr);
e0:
	PmrSMMUInfo->ui64SMMUVirAddr = INVALID_SMMU_PAGE_ADDR;
	if (PmrSMMUInfo)
		OSFreeMem(PmrSMMUInfo);
	return eError;
}

PVRSRV_ERROR XdxSMMUComAllocAndMap(
			struct xdx_device *xdev,
			PMR *psPMR, IMG_UINT64 *u64SMMUVAddr)
{
	struct smmu_context_s *smmu_ctx;

	if (!xdx_smmu_valid(xdev_to_smmu(xdev)))
		return PVRSRV_ERROR_NOT_READY;
	smmu_ctx = xdev_to_smmu_comctx(xdev);
	if (!smmu_ctx)
		return PVRSRV_ERROR_NOT_SUPPORTED;
	return _XDXSMMUVAAllocAndMap(smmu_ctx, psPMR, u64SMMUVAddr);
}

PVRSRV_ERROR XdxSMMUAllocAndMap(
			struct smmu_context_s *smmu_ctx,
			PMR *psPMR, IMG_UINT64 *u64SMMUVAddr)
{
	if (!smmu_ctx)
		return PVRSRV_ERROR_NOT_SUPPORTED;
	if (!xdx_smmu_valid(xdev_to_smmu(smmu_ctx->xsmmu->xdx_dev)))
		return PVRSRV_ERROR_NOT_READY;
	return _XDXSMMUVAAllocAndMap(smmu_ctx, psPMR, u64SMMUVAddr);
}

PVRSRV_ERROR XdxSMMUAllocAndMapGsid(
			struct xdx_device *xdev, IMG_UINT32 Gsid,
			PMR *psPMR, IMG_UINT64 *u64SMMUVAddr)
{
	struct smmu_context_s *smmu_ctx;

	if (!xdx_smmu_valid(xdev_to_smmu(xdev)))
		return PVRSRV_ERROR_NOT_READY;
	smmu_ctx = xdx_smmu_get_ctx_sid(xdev, Gsid);
	if (!smmu_ctx)
		return PVRSRV_ERROR_NOT_SUPPORTED;

	return _XDXSMMUVAAllocAndMap(smmu_ctx, psPMR, u64SMMUVAddr);
}

static void _XDXSMMUFreeAndUnMap(struct _PMR_SMMU_INFO_TAG *PmrSMMUInfo)
{
	IMG_DEV_VIRTADDR sSMMUVMBase;
	IMG_UINT32 uiMapedLength;
	struct smmu_context_s *smmu_ctx;
	struct xdx_smmu_ctx_env_s *ctx_env;

	if (PmrSMMUInfo == NULL)
		return ;
	smmu_ctx = PmrSMMUInfo->smmu_ctx;
	mutex_lock(&smmu_ctx->lock);
	list_del(&PmrSMMUInfo->ctx_list);
	mutex_unlock(&smmu_ctx->lock);
	dllist_remove_node(&PmrSMMUInfo->lMapList);
	uiMapedLength = PmrSMMUInfo->ui64MapedOffset;
	sSMMUVMBase.uiAddr = PmrSMMUInfo->ui64SMMUVirAddr;
	ctx_env = (struct xdx_smmu_ctx_env_s *)smmu_ctx->ctx_env_data;

//	PVR_DPF((PVR_DBG_MESSAGE,"%s SMMUVMBase.uiAddr = 0x%llx\n",__func__,sSMMUVMBase.uiAddr));

//	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
//		"PmrSMMUInfo:%llx RA_Free:%llx!",
///		(IMG_UINT64)PmrSMMUInfo, sSMMUVMBase.uiAddr);
	if(PmrSMMUInfo->bSparseAlloc)
	{
		smmu_unmap_pages(smmu_ctx,
					sSMMUVMBase.uiAddr,
					NULL,
					uiMapedLength >> RGX_HEAP_4KB_PAGE_SHIFT);
	}
	else
	{
		smmu_unmap_pages_fast(smmu_ctx,
					sSMMUVMBase.uiAddr,
					uiMapedLength >> RGX_HEAP_4KB_PAGE_SHIFT);
	}

	xdx_smmu_free(smmu_ctx,
				sSMMUVMBase.uiAddr,
				uiMapedLength,
				RGX_HEAP_4KB_PAGE_SHIFT);

	RA_Free(ctx_env->psVaddrRA, sSMMUVMBase.uiAddr);
	SMMU_DBG(xdx_smmu_dev(smmu_ctx),
		"PmrSMMUInfo:%llx RA_Free:%llx!",
		(IMG_UINT64)PmrSMMUInfo, sSMMUVMBase.uiAddr);
	xdx_smmu_group_sync_tlb(xdx_smmu_ctx_group(smmu_ctx));
	PmrSMMUInfo->ui64SMMUVirAddr = INVALID_SMMU_PAGE_ADDR;
	OSFreeMem(PmrSMMUInfo);
}

void XDXSMMUFreeAndUnMap(PMR *psPMR)
{
	PDLLIST_NODE plListHead = NULL;
	DLLIST_NODE *psThis, *psNext;
	DLLIST_NODE lNewListHead;
	struct _PMR_SMMU_INFO_TAG *PmrSMMUInfo = NULL;
	struct xdx_device *xdev = gfx_to_xdev(PMR_DeviceNode(psPMR));
	struct xdx_smmu_s *xsmmu;
	struct xdx_smmu_env_s *smmu_env;

	if (!xdx_smmu_is_inited(xdev_to_smmu(xdev)))
		return;
	xsmmu = xdev_to_smmu_dev(xdev);
	smmu_env = (struct xdx_smmu_env_s *)xsmmu->env_data;
	OSLockAcquire(smmu_env->pLock);
	plListHead = (PDLLIST_NODE)PMRGetSmmuMapList(psPMR);
	if (!plListHead || dllist_is_empty(plListHead))
	{
		OSLockRelease(smmu_env->pLock);
		return;
	}
	dllist_replace_head(plListHead, &lNewListHead);
	OSLockRelease(smmu_env->pLock);

	dllist_foreach_node(&lNewListHead, psThis, psNext)
	{
		PmrSMMUInfo = IMG_CONTAINER_OF(psThis, struct _PMR_SMMU_INFO_TAG, lMapList);
		_XDXSMMUFreeAndUnMap(PmrSMMUInfo);
	}
}

int xdx_smmu_ctx_unmap_all(struct smmu_context_s *smmu_ctx)
{
	struct _PMR_SMMU_INFO_TAG *PmrSMMUInfo, *_PmrSMMUInfo;

	list_for_each_entry_safe(PmrSMMUInfo, _PmrSMMUInfo, &smmu_ctx->list, ctx_list)
	{
		_XDXSMMUFreeAndUnMap(PmrSMMUInfo);
	}
	return 0;
}

int xdx_smmu_plat_env_init(struct xdx_device *xdev, struct xdx_smmu_s *xsmmu)
{
	struct xdx_smmu_env_s *smmu_env;
	PVRSRV_DEVICE_NODE *psPVRDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	int ret;
	PVRSRV_ERROR eError;

	smmu_env = devm_kcalloc(xdev->dev, 1, sizeof(struct xdx_smmu_env_s), GFP_KERNEL);
	if (!smmu_env)
		return -ENOMEM;
	smmu_env->bReady = IMG_FALSE;
	if (!psPVRDeviceNode->psMMUPhysHeap) {
		pr_err("[%s] SMMU heap not ready.", __func__);
		ret =  -ENODEV;
		goto errout;
	}
	smmu_env->psPageHeap = psPVRDeviceNode->psMMUPhysHeap;
	smmu_env->psPageMemRA = RA_Create("XdxSmmuPageTables",
							PhysHeapGetPageShift(smmu_env->psPageHeap),
							RA_LOCKCLASS_1,
							_SMMU_PhysMem_RAImportAlloc,
							_SMMU_PhysMem_RAImportFree,
							xsmmu, /* priv */
							RA_POLICY_DEFAULT);
	if (!smmu_env->psPageMemRA) {
		pr_err("[%s] SMMU heap Creat RA failed.", __func__);
		ret = -ENODEV;
		goto errout;
	}
	eError = OSLockCreate(&smmu_env->pLock);
	if (eError != PVRSRV_OK) {
		ret = -EINVAL;
		goto errout;
	}
	smmu_env->bReady = IMG_TRUE;
	smmu_env->xsmmu = xsmmu;
	xsmmu->env_data = smmu_env;
	dllist_init(&smmu_env->sTmpMMUMappingHead);
	return 0;
errout:
	xsmmu->env_data = NULL;
	devm_kfree(xdev->dev, smmu_env);
	return ret;
}

int xdx_smmu_plat_env_deinit(struct xdx_device *xdev, struct xdx_smmu_s *xsmmu)
{
	struct xdx_smmu_env_s *smmu_env;

	smmu_env = (struct xdx_smmu_env_s *)xsmmu->env_data;

	smmu_env->bReady = IMG_FALSE;
	xsmmu->env_data = NULL;
	RA_Delete(smmu_env->psPageMemRA);
	smmu_env->psPageHeap = NULL;
	OSLockDestroy(smmu_env->pLock);
	devm_kfree(xdev->dev, smmu_env);
	return 0;
}

int xdx_smmu_ctx_env_init(struct xdx_device *xdev, struct smmu_context_s *smmu_ctx,
		IMG_UINT64 Vaddr_base ,IMG_UINT64 Vaddr_size)
{
	struct xdx_smmu_ctx_env_s *ctx_env;
	IMG_CHAR RaName[32];
	int ret;

	ctx_env = devm_kcalloc(xdev->dev, 1, sizeof(struct xdx_smmu_ctx_env_s), GFP_KERNEL);
	if (!ctx_env)
		return -ENOMEM;
	snprintf(RaName, 32, "smmu-group:%d",smmu_ctx->group->grp_id);
	ctx_env->psVaddrRA = RA_Create_With_Span(RaName,
			OSGetPageShift(), 0, Vaddr_base, Vaddr_size, RA_POLICY_DEFAULT);
	if (ctx_env->psVaddrRA == NULL) {
		ret = -ENODEV;
		goto errout;
	}
	ctx_env->smmu_ctx = smmu_ctx;
	smmu_ctx->ctx_env_data = ctx_env;
	return 0;
errout:
	smmu_ctx->ctx_env_data = NULL;
	devm_kfree(xdev->dev, ctx_env);
	return ret;
}

int xdx_smmu_ctx_env_destroy(struct smmu_context_s *smmu_ctx)
{
	struct xdx_smmu_ctx_env_s *ctx_env;

	ctx_env = smmu_ctx->ctx_env_data;
	xdx_smmu_cleanup_smmu_env(smmu_ctx->xsmmu);
	smmu_ctx->ctx_env_data = NULL;
	RA_Delete(ctx_env->psVaddrRA);
	devm_kfree( smmu_ctx->xsmmu->xdx_dev->dev, ctx_env);
	return 0;
}
