/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdx_device.h"
#include "xdx_thm.h"
#include "xdx_pm.h"
#include "device.h"
#include "pvrsrv.h"
#if defined(SUPPORT_LINUX_DVFS)
#include "pvr_dvfs_device.h"
#endif
#include "xdx_gfx_convert.h"
#include <linux/reboot.h>

static int xdx_thm_suspend(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL) && defined(XDXGPU)
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 8, 18))
	PVRSRV_DEVICE_NODE	*psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	IMG_DVFS_DEVICE		*psDVFSDevice = &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	int err;
#endif
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	cancel_delayed_work_sync(&xdev->thm.work);

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL) && defined(XDXGPU)
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 8, 18))
	if (xdx_dvfs_enabled(xdev->dev)) {
		err = thermal_zone_device_disable(psDVFSDevice->psThermalZoneDevice);
		if (err) {
			dev_err(xdev->dev, "Failed to disable thermal zone\n");
			return err;
		}
		dev_info(xdev->dev, "Succeed to disable thermal zone\n");
	}
#endif
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	return 0;
}

static int xdx_thm_resume(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL) && defined(XDXGPU)
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 8, 18))
	PVRSRV_DEVICE_NODE	*psDeviceNode 	= (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	IMG_DVFS_DEVICE		*psDVFSDevice 	= &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	int err;
	if (xdx_dvfs_enabled(xdev->dev)) {
		err = thermal_zone_device_enable(psDVFSDevice->psThermalZoneDevice);
		if (err) {
			dev_err(xdev->dev, "Failed to enable thermal zone\n");
			return err;
		}
		dev_info(xdev->dev, "Succeed to enable thermal zone\n");
	}
#endif
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	queue_delayed_work(xdev->thm.workq, &xdev->thm.work, msecs_to_jiffies(TIMEOUT_MS));

	return 0;
}

static int xdx_thm_early_init(void *handle)
{
	return 0;
}

static int xdx_thm_early_fini(void *handle)
{
	return 0;
}

static int xdx_thm_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	xdev->thm.workq = create_freezable_workqueue("thermal workq");
	INIT_DELAYED_WORK(&xdev->thm.work, xdx_thermal_monitor);

	return 0;
}

static int xdx_thm_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	cancel_delayed_work_sync(&xdev->thm.work);
	destroy_workqueue(xdev->thm.workq);

	return 0;
}

static int xdx_thm_hw_init(void *handle)
{
	return 0;
}

static int xdx_thm_hw_fini(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version thermal_block_version = {
	.name = "thm",
	.type = XDX_IP_BLOCK_TYPE_THMSENSOR,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.suspend = xdx_thm_suspend,
		.resume = xdx_thm_resume,
		.early_init = xdx_thm_early_init,
		.early_fini = xdx_thm_early_fini,
		.sw_init = xdx_thm_sw_init,
		.sw_fini = xdx_thm_sw_fini,
		.hw_init = xdx_thm_hw_init,
		.hw_fini = xdx_thm_hw_fini,
	},
};


#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL)
static int xdx_thmzd_bind(struct thermal_zone_device *thermal,
			struct thermal_cooling_device *cdev)
{
	int err;
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	IMG_DVFS_DEVICE		*psDVFSDevice 	= &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);

	if (cdev != psDVFSDevice->psDevfreqCoolingDevice) {
		return 0;
	}

	err = thermal_zone_bind_cooling_device(thermal, 0, cdev,
				THERMAL_NO_LIMIT, THERMAL_NO_LIMIT, THERMAL_WEIGHT_DEFAULT);

	if (err < 0) {
		dev_err(xdev->dev, "Failed to bind cooling device to trip 0\n");
		return err;
	}

	return 0;
}

static int xdx_thmzd_unbind(struct thermal_zone_device *thermal,
			struct thermal_cooling_device *cdev)
{
	int err;
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	IMG_DVFS_DEVICE		*psDVFSDevice 	= &psDeviceNode->psDevConfig->sDVFS.sDVFSDevice;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);

	if (cdev != psDVFSDevice->psDevfreqCoolingDevice)
		return 0;

	err = thermal_zone_unbind_cooling_device(thermal, 0, cdev);

	if (err < 0) {
		dev_err(xdev->dev, "Failed to bind cooling device to trip 0\n");
		return err;
	}

	return 0;
}

static int xdx_thmzd_get_temp(struct thermal_zone_device *thermal, int *t)
{
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);
	int gddr_temp = get_gddr_temp(xdev);
	int soc_temp = get_gpu_temp(xdev);

	*t = (gddr_temp > soc_temp) ? gddr_temp : soc_temp;

	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0))
static int xdx_thmzd_get_trip_type(struct thermal_zone_device *thermal, int trip,
				 enum thermal_trip_type *trip_type)
{
	if (trip == 0)
		*trip_type = THERMAL_TRIP_ACTIVE;
	else if (trip == 1)
		*trip_type = THERMAL_TRIP_CRITICAL;
	else
		return -EINVAL;

	return 0;
}

static int xdx_thmzd_get_trip_hyst(struct thermal_zone_device *thermal, int trip,
				 int *trip_hyst)
{
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);

	if (trip == 0)
		*trip_hyst = xdev->thm.hyst;
	else if (trip == 1)
		*trip_hyst = 0;
	else
		return -EINVAL;

	return 0;
}

static int xdx_thmzd_get_trip_temp(struct thermal_zone_device *thermal, int trip,
				 int *temp)
{
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);

	if (trip == 0)
		*temp = xdev->thm.slowdown_temp;
	else if (trip == 1)
		*temp = xdev->thm.shutdown_temp;
	else
		return -EINVAL;

	return 0;
}
#endif

static int xdx_thmzd_get_crit_temp(struct thermal_zone_device *thermal,
				 int *temp)
{
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);

	*temp = xdev->thm.shutdown_temp;

	return 0;
}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 206))
static void xdx_thmzd_critical(struct thermal_zone_device *thermal)
{
	PVRSRV_DEVICE_NODE 	*psDeviceNode 	= thermal->devdata;
	struct xdx_device 	*xdev 			= gfx_to_xdev(psDeviceNode);
	int gddr_temp 						= get_gddr_temp(xdev);
	int soc_temp 						= get_gpu_temp(xdev);
	int temp							= (gddr_temp > soc_temp) ? gddr_temp : soc_temp;

	dev_emerg(xdev->dev, "%s: XDXGPU Temp Too High %dC\n", thermal->type, temp);
	if (xdev->thm.enable_shutdown) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 13, 19))
		hw_protection_shutdown("XDXGPU Thermal Protection", 1000);
#else
		orderly_poweroff(true);
#endif
	}
}
#endif

static struct thermal_zone_device_ops xdx_thmzd_dev_ops = {
	.bind = xdx_thmzd_bind,
	.unbind = xdx_thmzd_unbind,
	.get_temp = xdx_thmzd_get_temp,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0))
	.get_trip_type = xdx_thmzd_get_trip_type,
	.get_trip_hyst = xdx_thmzd_get_trip_hyst,
	.get_trip_temp = xdx_thmzd_get_trip_temp,
#endif
	.get_crit_temp = xdx_thmzd_get_crit_temp,
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 206))
	.critical = xdx_thmzd_critical,
#endif
};

static struct thermal_zone_params xdx_thmzd_zone_params = {
	.governor_name = "step_wise",
};

int RegisterThermalZoneDevice(struct device *dev,
								 IMG_DVFS_DEVICE *psDVFSDevice,
								 struct devfreq_cooling_power *powerOps)
{
	struct device_node *of_node;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	int err = 0;
	PVRSRV_VZ_RET_IF_MODE(GUEST, err);

	if (!psDVFSDevice)
	{
		return -EINVAL;
	}

	if (!powerOps)
		dev_info(dev, "Enabling cooling without registering powerOps\n");

	of_node = of_node_get(dev->of_node);

	psDVFSDevice->psDevfreqCoolingDevice = of_devfreq_cooling_register_power(
		of_node, psDVFSDevice->psDevFreq, powerOps);

	if (IS_ERR(psDVFSDevice->psDevfreqCoolingDevice))
	{
		err = PTR_ERR(psDVFSDevice->psDevfreqCoolingDevice);
		dev_err(dev, "Failed to register as devfreq cooling device %d", err);
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0))
	psDVFSDevice->psThermalZoneDevice =
				thermal_zone_device_register("xdx_thmzd", 2, 0, psDeviceNode,
				&xdx_thmzd_dev_ops, &xdx_thmzd_zone_params, 0, 1000);
#else
	psDVFSDevice->psThermalZoneDevice = thermal_tripless_zone_device_register("xdx_thmzd",
							psDeviceNode, &xdx_thmzd_dev_ops, &xdx_thmzd_zone_params);
#endif
	if (IS_ERR(psDVFSDevice->psThermalZoneDevice))
		return -ENODEV;

	#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 8, 18))
		err = thermal_zone_device_enable(psDVFSDevice->psThermalZoneDevice);
		if (err) {
			dev_err(dev, "Failed to enable thermal zone\n");
			thermal_zone_device_unregister(psDVFSDevice->psThermalZoneDevice);
			return err;
		}
	#endif

	dev_info(dev, "Register thermal_zone%d\n", psDVFSDevice->psThermalZoneDevice->id);

	of_node_put(of_node);

	return 0;
}

void UnregisterThermalZoneDevice(struct device *dev, IMG_DVFS_DEVICE *psDVFSDevice)
{
	int id = -1;

	if (!IS_ERR_OR_NULL(psDVFSDevice->psThermalZoneDevice)) {
		id = psDVFSDevice->psThermalZoneDevice->id;
		thermal_zone_device_unregister(psDVFSDevice->psThermalZoneDevice);
		psDVFSDevice->psThermalZoneDevice = NULL;
	}

	if (!IS_ERR_OR_NULL(psDVFSDevice->psDevfreqCoolingDevice)) {
		devfreq_cooling_unregister(psDVFSDevice->psDevfreqCoolingDevice);
		psDVFSDevice->psDevfreqCoolingDevice = NULL;
	}

	if (id != -1)
		dev_info(dev, "Unregister thermal_zone%d\n", id);
}
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

void xdx_thermal_monitor(struct work_struct *wk) {
	struct delayed_work *delayed_wk 	= (struct delayed_work *) wk;
	struct xdx_thermal 	*thm 			= container_of(delayed_wk, struct xdx_thermal, work);
	struct xdx_device 	*xdev 			= thm->xdev;
	int gddr_temp 						= get_gddr_temp(xdev);
	int soc_temp 						= get_gpu_temp(xdev);
	u32 gpu_freq						= get_gpu_freq(xdev);
	u32 gpu_volt 						= get_gpu_volt(xdev);
#if defined(SUPPORT_LINUX_DVFS) && defined(XDXGPU)
	PVRSRV_DEVICE_NODE 	*psDeviceNode	= xdev->gfx.pvrdev;
	IMG_DVFS_DEVICE_CFG	*psDVFSDeviceCfg= &psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg;
	const IMG_OPP		*pasOPPTable 	= psDVFSDeviceCfg->pasOPPTable;
	const int			oppSize         = psDVFSDeviceCfg->ui32OPPTableSize;
#endif

#if defined(SUPPORT_LINUX_DVFS) && defined(XDXGPU)
	if (thm->dvfs_enabled != xdx_dvfs_enabled(xdev->dev)) {
		thm->dvfs_enabled = !thm->dvfs_enabled;
		thm->slowdown_status = false;
	}
#endif

#if defined(SUPPORT_LINUX_DVFS) && defined(CONFIG_DEVFREQ_THERMAL) && defined(XDXGPU)
	if (xdx_dvfs_enabled(xdev->dev)) {
		queue_delayed_work(thm->workq, &thm->work, msecs_to_jiffies(TIMEOUT_MS));
		return;
	}
#endif /* SUPPORT_LINUX_DVFS && CONFIG_DEVFREQ_THERMAL */

	thm->curr_temp = (gddr_temp > soc_temp) ? gddr_temp : soc_temp;
//	dev_dbg(xdev->dev, "THM GPU Temp %d\n", thm->curr_temp);

	if (thm->curr_temp >= thm->shutdown_temp) {
		if (thm->enable_shutdown) {
			dev_warn(xdev->dev, "XDXGPU Thermal Protection, GPU Temp %d, Freq %u, Volt %u\n", \
				thm->curr_temp, gpu_freq, gpu_volt);
			orderly_poweroff(true);
		}
	}
	else if (thm->curr_temp >= thm->slowdown_temp && thm->slowdown_status) {
		dev_dbg(xdev->dev, "THM GPU Temp keeps high even if downclock!\n");
	}
	else if (thm->curr_temp >= thm->slowdown_temp && !thm->slowdown_status) {
		thm->slowdown_status = true;
#if defined(SUPPORT_LINUX_DVFS) && defined(XDXGPU)
		if (xdx_dvfs_enabled(xdev->dev)) {
			dev_dbg(xdev->dev, "DVFS Begins Protecting GPU, GPU Temp %d, Temp Hyst %d, Curr Freq %u, Curr Volt %u, Tgt Freq %u, Tgt Volt %u\n", \
				thm->curr_temp, xdev->thm.hyst, gpu_freq, gpu_volt, pasOPPTable[0].ui32Freq, pasOPPTable[0].ui32Volt);
			xdx_dvfs_set_opp_availability_above_floor(psDeviceNode, false);
		}
		else
#endif
		{
			dev_dbg(xdev->dev, "THM Begins Protecting GPU, GPU Temp %d, Temp Hyst %d, Curr Freq %u, Curr Volt %u, Tgt Freq %u, Tgt Volt %u\n", \
				thm->curr_temp, xdev->thm.hyst, gpu_freq, gpu_volt, MIN_WORK_CLOCK_HZ, (xdev->asic_type == CHIP_PANGU_B0) ? MIN_WORK_VOLT_MV : gpu_volt);
			xdx_dvfs_set_freq(xdev, MIN_WORK_CLOCK_HZ);
			if (xdev->asic_type == CHIP_PANGU_B0)
				xdx_power_set(xdev, cH02CPowerDBCmdDVFS, cH02CPowerDVFSSubCmdSetVolt, 1, MIN_WORK_VOLT_MV);
		}
	}
	else if (thm->curr_temp <= (thm->slowdown_temp - thm->hyst) && thm->slowdown_status){
		thm->slowdown_status = false;
#if defined(SUPPORT_LINUX_DVFS) && defined(XDXGPU)
		if (xdx_dvfs_enabled(xdev->dev)) {
			dev_dbg(xdev->dev, "DVFS Ends Protecting GPU, GPU Temp %d, Temp Hyst %d, Curr Freq %u, Curr Volt %u, Tgt Freq %u, Tgt Volt %u\n", \
				thm->curr_temp, xdev->thm.hyst, gpu_freq, gpu_volt, pasOPPTable[oppSize - 1].ui32Freq, pasOPPTable[oppSize - 1].ui32Volt);
			xdx_dvfs_set_opp_availability_above_floor(psDeviceNode, true);
		}
		else
#endif
		{
			dev_dbg(xdev->dev, "THM Ends Protecting GPU, GPU Temp %d, Temp Hyst %d, Curr Freq %u, Curr Volt %u, Tgt Freq %u, Tgt Volt %u\n", \
				thm->curr_temp, xdev->thm.hyst, gpu_freq, gpu_volt, thm->gfx_default, (xdev->asic_type == CHIP_PANGU_B0) ? thm->volt_default : gpu_volt);
			if (xdev->asic_type == CHIP_PANGU_B0)
				xdx_power_set(xdev, cH02CPowerDBCmdDVFS, cH02CPowerDVFSSubCmdSetVolt, 1, thm->volt_default);
			xdx_dvfs_set_freq(xdev, thm->gfx_default);
		}
	}

	queue_delayed_work(thm->workq, &thm->work, msecs_to_jiffies(TIMEOUT_MS));
}

int xdx_thm_init(struct xdx_device *xdev)
{
	xdev->thm.xdev = xdev;
	xdev->thm.enable_shutdown = true;
#if defined(SUPPORT_LINUX_DVFS) && defined(XDXGPU)
	xdev->thm.dvfs_enabled = false;
#endif
	xdev->thm.curr_temp = get_gpu_temp(xdev);
	xdev->thm.slowdown_temp = TEMP_SLOWDOWN;
	xdev->thm.hyst = (xdev->fw_st_info.device_id == 0x11E0) ? TEMP_HYSTERESIS + 3 : TEMP_HYSTERESIS;
	xdev->thm.shutdown_temp = TEMP_SHUTDOWN;
	xdev->thm.gfx_default = xdev->fw_st_info.default_gpu_freq;
	xdev->thm.volt_default = xdev->fw_st_info.gpu_core0_volt ? xdev->fw_st_info.gpu_core0_volt : get_gpu_volt(xdev);
	xdev->thm.slowdown_status = false;

	dev_info(xdev->dev, "slowdown_temp %d  hyst %d  device_id %x\n", xdev->thm.slowdown_temp, xdev->thm.hyst, xdev->fw_st_info.device_id);
	queue_delayed_work(xdev->thm.workq, &xdev->thm.work, 0);

    return 0;
}

int xdx_thm_fini(struct xdx_device *xdev)
{
	xdev->thm.workq = NULL;

    return 0;
}
