/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#ifndef _SMMU_COMMON_H_
#define _SMMU_COMMON_H_

#include <linux/types.h>
#include <linux/mutex.h>
#include "xdx_smmu.h"
#include "xdx_device.h"
#include "xdx_smmu_drv.h"

/* services/server/include/ */
/*!
 * The level of the MMU
 */
enum smmu_level_e {
	SMMU_LEVEL_0 = 0,	/* Level 0 = Page */
	SMMU_LEVEL_1,
	SMMU_LEVEL_2,
	SMMU_LEVEL_3,
	SMMU_LEVEL_4,
	SMMU_LEVEL_LAST
};
#define SMMU_MAX_LEVEL 4
/*
 *	P(C/D/T) Entry Config:
 *	MSB-----------------------------------------------LSB
 *	| PT Addr:   | variable PT ctrl | protection flags: |
 *	| bits c+v   | b bits           | a bits            |
 *	-----------------------------------------------------
 *	where v is the variable page table modifier and is optional
 */
/*!
 *	Generic MMU entry description. This is used to describe PC, PD and PT entries.
 */
struct smmu_pxe_config_s {
	uint8_t		uiBytesPerEntry; /*! Size of an entry in bytes */
	uint8_t		uiAddrLog2Align; /*! Physical address Log 2 alignment */
	uint64_t	uiVarCtrlMask;   /*! Variable control mask */
	uint8_t		uiVarCtrlShift;  /*! Variable control shift */
	uint64_t	uiProtMask;      /*! Protection flags mask */
	uint64_t	uiValidEnMask;   /*! Entry valid bit mask */
	uint8_t		uiValidEnShift;  /*! Entry valid bit shift */
	/*! Page catalogue index mask */
	uint64_t	uiIndexMask;
	/*! Page catalogue index shift */
	uint8_t		uiIndexShift;
	uint32_t	uiNumEntries;
};

struct smmu_mem_mapping_s;

struct smmu_mem_desc_s {
	/*! Does this MMU object have physical backing */
	bool				bValid;
	/*! Device Physical address of physical backing */
	IMG_DEV_PHYADDR			dev_phy_addr;
	/*! CPU virtual address of physical backing */
	void					*pvCpuVAddr;
	/*! Mapping data for this MMU object */
	struct smmu_mem_mapping_s		*psMapping;
	/*! Memdesc offset into the psMapping */
	uint32_t uiOffset;
	/*! Size of the Memdesc */
	uint32_t uiSize;
};

struct smmu_page_config_s {
	const struct smmu_pxe_config_s *pxe_cfgs[SMMU_LEVEL_LAST];
	uint32_t	top_level;
	uint32_t	page_size;
	uint64_t	page_mask;
	uint32_t	page_shift;
	uint64_t	offset_in_page;
};

struct smmu_grp_para_s {
	uint32_t gsid_cnt;
	uint32_t *gsid_tab;
	uint32_t flags;
	uint64_t vaddr_rang_start;
	uint64_t vaddr_rang_size;
};

struct smmu_lvlx_info_s {
	uint32_t ui32NumOfEntries;
	uint32_t ui32RefCount;
	struct smmu_mem_desc_s sMemDesc;
	struct smmu_lvlx_info_s *next_level_info[0];
};

struct smmu_context_s {
	struct xdx_smmu_s *xsmmu;
	const struct smmu_page_config_s *page_cfg;
	struct smmu_group_s *group;
	struct smmu_lvlx_info_s *base_lvl_info;
	void *ctx_env_data;
	struct mutex lock;
	struct list_head maped_head;
	struct list_head list;
};

int xdx_smmu_ctx_creat(struct xdx_device *xdev,
			const struct smmu_grp_para_s *grp_para,
			const struct smmu_page_config_s *pcfg);

void xdx_smmu_ctx_destroy(struct smmu_context_s *smmu_ctx);

int xdx_smmu_alloc(struct smmu_context_s *smmu_ctx,
		uint64_t uSize, uint64_t *puActualSize, uint32_t uiProtflags,
		uint64_t uDevVAddrAlignment, uint64_t pdev_vaddr, uint32_t uiLog2PageSize);

void xdx_smmu_free(struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr,
		uint64_t uiSize, uint32_t uiLog2DataPageSize);

uint32_t smmu_unmap_pages_fast(struct smmu_context_s *smmu_ctx,
			uint64_t dev_vaddr_base, uint32_t pages_cnt);

int smmu_map_pages_linear(struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr_base,
			uint64_t dev_phy_addr, uint32_t pages_cnt);

int smmu_map_pages_fast(struct smmu_context_s *smmu_ctx,
			uint64_t dev_vaddr_base, uint32_t page_offset,
			uint64_t dev_phy_addr[], uint32_t pages_cnt);

void smmu_unmap_pages(struct smmu_context_s *smmu_ctx, uint64_t dev_vaddr_base,
	uint32_t map_indices[], uint32_t pages_cnt);

int smmu_map_pages(struct smmu_context_s *smmu_ctx,
			uint64_t dev_vaddr_base, uint32_t page_offset,
			uint64_t dev_phy_addr[], bool valid[],
			uint32_t map_indices[], uint32_t pages_cnt);

struct smmu_context_s *xdx_smmu_get_comgrp(struct xdx_device *xdev);
struct smmu_context_s *xdx_smmu_get_ctx_sid(struct xdx_device *xdev, uint32_t gsid);

static inline struct device *xdx_smmu_dev(struct smmu_context_s *smmu_ctx)
{
	return smmu_ctx->xsmmu->xdx_dev->dev;
}

static inline struct smmu_group_s *xdx_smmu_ctx_group(struct smmu_context_s *smmu_ctx)
{
	return smmu_ctx->group;
}

static inline struct smmu_context_s *xdev_to_smmu_comctx(struct xdx_device *xdev)
{
	struct xdx_smmu *smmu = xdev_to_smmu(xdev);

	if (list_empty(&smmu->ctxs_head))
		return NULL;

	return list_first_entry(&smmu->ctxs_head, struct smmu_context_s, list);
}

static inline void xdev_add_smmu_ctx(struct xdx_device *xdev, struct smmu_context_s *ctx)
{
	struct xdx_smmu *smmu = xdev_to_smmu(xdev);

	if (!ctx)
		return;
	list_add_tail(&ctx->list, &smmu->ctxs_head);
}
#endif /* #ifdef MMU_COMMON_H */
