/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/module.h>
#include <linux/platform_device.h>

#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
#include <drm/drm_drv.h>
#endif
#include <drm/drm_fb_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)) || (defined(REDHAT_83))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif
#include <drm/drm_gem.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
#include <drm/drm_gem_framebuffer_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
#include <drm/drm_ioctl.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)) || (defined(REDHAT_83))
#include <drm/drm_probe_helper.h>
#else
#include <drm/drm_crtc_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 13, 0))
#include <drm/drm_vblank.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0))
#include <drm/drm_irq.h>
#else
#include <drm/drmP.h>
#endif

#include "xdx_vkms.h"
#include "xdx_gem.h"
#include "xdx_device.h"

bool enable_cursor;

static void vkms_atomic_commit_tail(struct drm_atomic_state *old_state)
{
	struct drm_device *dev = old_state->dev;
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 11, 0))
	struct drm_crtc *crtc;
	struct drm_crtc_state *old_crtc_state;
	int i;
#endif

	drm_atomic_helper_commit_modeset_disables(dev, old_state);

	drm_atomic_helper_commit_planes(dev, old_state, 0);

	drm_atomic_helper_commit_modeset_enables(dev, old_state);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0))
	drm_atomic_helper_fake_vblank(old_state);
#endif

	drm_atomic_helper_commit_hw_done(old_state);

	drm_atomic_helper_wait_for_vblanks(dev, old_state);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(4, 11, 0))
	for_each_old_crtc_in_state (old_state, crtc, old_crtc_state, i) {
		struct vkms_crtc_state *vkms_state =
			to_vkms_crtc_state(old_crtc_state);

		flush_work(&vkms_state->composer_work);
	}
#endif

	drm_atomic_helper_cleanup_planes(dev, old_state);
}

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0))
static void xdx_framebuffer_destroy(struct drm_framebuffer *fb)
{
        struct drm_xdxgpu_framebuffer *xdx_fb = to_xdx_fb(fb);
        int i;

        for (i=0; i<4; i++)
                drm_gem_object_unreference_unlocked(xdx_fb->obj[i]);

        drm_framebuffer_cleanup(fb);
        kfree(xdx_fb);
}

static int xdx_framebuffer_create_handle(struct drm_framebuffer *fb,
                                                  struct drm_file *file_priv,
                                                  unsigned int *handle)
{
        struct drm_xdxgpu_framebuffer *xdx_fb = to_xdx_fb(fb);

        return drm_gem_handle_create(file_priv, xdx_fb->obj[0], handle);
}

static const struct drm_framebuffer_funcs xdx_fb_funcs = {
        .destroy = xdx_framebuffer_destroy,
        .create_handle = xdx_framebuffer_create_handle,
};

int xdx_framebuffer_init(struct drm_device *dev,
                        struct drm_xdxgpu_framebuffer *xdx_fb,
                        const struct drm_mode_fb_cmd2 *mode_cmd,
                        int num_planes,
                        struct drm_gem_object **obj)
{
        int ret, i;
        for (i=0; i<num_planes; i++)
                xdx_fb->obj[i] = obj[i];

        drm_helper_mode_fill_fb_struct(&xdx_fb->base, mode_cmd);

        ret = drm_framebuffer_init(dev, &xdx_fb->base, &xdx_fb_funcs);
        if (ret) {
                return ret;
        }
        return 0;
}

struct drm_framebuffer *
xdx_drm_gem_fb_create(struct drm_device *dev,
                               struct drm_file *file_priv,
                               const struct drm_mode_fb_cmd2 *mode_cmd)
{
        struct drm_gem_object *obj[4];
        struct drm_xdxgpu_framebuffer *xdx_fb;
        int ret, i;
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 11, 0))
        const struct xdx_drm_format_info *info;
        info = xdx_drm_get_format_info(dev, mode_cmd);
#else
        const struct drm_format_info *info;
        info = drm_get_format_info(dev, mode_cmd);
#endif
        if (!info)
                return ERR_PTR(-EINVAL);

        for (i=0; i<info->num_planes; i++) {
/*
 *              unsigned int width = mode_cmd->width / (i ? info->hsub : 1);
 *              unsigned int height = mode_cmd->height / (i ? info->vsub : 1);
                unsigned int min_size;
 */
                obj[i] = drm_gem_object_lookup(file_priv, mode_cmd->handles[i]);
                if (obj[i] ==  NULL) {
                        dev_err(&dev->pdev->dev, "No GEM object associated to handle 0x%08X, "
                                "can't create framebuffer\n", mode_cmd->handles[i]);
                        return ERR_PTR(-ENOENT);
                }

                /* Handle is imported dma-buf, so cannot be migrated to VRAM for scanout */
                if (obj[i]->import_attach) {
                        dev_err(&dev->pdev->dev, "Cannot create framebuffer from imported dma_buf\n");
                        return ERR_PTR(-EINVAL);
                }
/*
 *              min_size = (height - 1) * mode_cmd->pitches[i]
 *                       + drm_format_info_min_pitch(info, i, width)
 *                       + mode_cmd->offsets[i];
 *
 *              if (objs[i]->size < min_size) {
 *                      drm_gem_object_put_unlocked(objs[i]);
 *                      ret = -EINVAL;
 *                      goto err_gem_object_put;
 *              }
 */
        }

        xdx_fb = kzalloc(sizeof(*xdx_fb), GFP_KERNEL);
        if (xdx_fb == NULL) {
                return ERR_PTR(-ENOMEM);
        }

        ret = xdx_framebuffer_init(dev, xdx_fb, mode_cmd, i, obj);
        if (ret) {
                kfree(xdx_fb);
                return ERR_PTR(ret);
        }

        return &xdx_fb->base;
}
#endif

static const struct drm_mode_config_funcs vkms_mode_funcs = {
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0))
	.fb_create = xdx_drm_gem_fb_create,
#else
	.fb_create = drm_gem_fb_create,
#endif
	.atomic_check = drm_atomic_helper_check,
	.atomic_commit = drm_atomic_helper_commit,
};

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 10, 0))
static struct drm_mode_config_helper_funcs vkms_mode_config_helpers = {
        .atomic_commit_tail = vkms_atomic_commit_tail,
};
#else
static const struct drm_mode_config_helper_funcs vkms_mode_config_helpers = {
	.atomic_commit_tail = vkms_atomic_commit_tail,
};
#endif

static int vkms_modeset_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);

	drm_mode_config_init(ddev);
	ddev->mode_config.funcs = &vkms_mode_funcs;
	ddev->mode_config.min_width = XRES_MIN;
	ddev->mode_config.min_height = YRES_MIN;
	ddev->mode_config.max_width = XRES_MAX;
	ddev->mode_config.max_height = YRES_MAX;
	ddev->mode_config.preferred_depth = 24;
	ddev->mode_config.helper_private = &vkms_mode_config_helpers;

	return xdx_vkms_output_init(xdev, 0);
}

static int vkms_early_init(void *handle)
{
	return 0;
}

static int vkms_late_init(void *handle)
{
	return 0;
}

static int vkms_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 15, 0))
	ddev->irq_enabled = true;
#endif
	ret = drm_vblank_init(ddev, 1);
	if (ret) {
		dev_err(xdev->dev, "Failed to vblank\n");
		return ret;
	}

	ret = vkms_modeset_init(xdev);

	return ret;
}

static int vkms_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	struct vkms_device *vkmsdev = xdev_to_vkmsdev(xdev);
	struct drm_device *ddev = xdev_to_drm(xdev);

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 11, 0))
	xdx_drm_atomic_helper_shutdown(ddev);
#else
	drm_atomic_helper_shutdown(ddev);
#endif
	drm_mode_config_cleanup(ddev);
	destroy_workqueue(vkmsdev->output.composer_workq);

	return 0;
}

static int vkms_early_fini(void *handle)
{
	return 0;
}

static int vkms_hw_init(void *handle)
{
	return 0;
}

static int vkms_hw_fini(void *handle)
{
	return 0;
}
/*
static void vkms_late_fini(void *handle)
{

}
*/
static int vkms_suspend(void *handle)
{
	return 0;
}

static int vkms_resume(void *handle)
{
	return 0;
}

static bool vkms_is_idle(void *handle)
{
	return 0;
}

static int vkms_wait_for_idle(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version vkms_ip_block_version =
{
	.name = "vkms",
	.type = XDX_IP_BLOCK_TYPE_DC,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.early_init = vkms_early_init,
		.late_init = vkms_late_init,
		.sw_init = vkms_sw_init,
		.sw_fini = vkms_sw_fini,
		.early_fini = vkms_early_fini,
		.hw_init = vkms_hw_init,
		.hw_fini = vkms_hw_fini,
		.suspend = vkms_suspend,
		.resume = vkms_resume,
		.is_idle = vkms_is_idle,
		.wait_for_idle = vkms_wait_for_idle,		
	},
};
