/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#ifndef _XDX_VKMS_H_
#define _XDX_VKMS_H_

#include <linux/hrtimer.h>
#include <linux/debugfs.h>

#include <drm/drm.h>
#include <drm/drm_gem.h>
#include <drm/drm_encoder.h>
#if ((LINUX_VERSION_CODE > KERNEL_VERSION(6, 0, 0)))
#include <drm/drm_framebuffer.h>
#endif
#include "xdx_shared.h"

#define XRES_MIN 20
#define YRES_MIN 20

#define XRES_DEF 1920
#define YRES_DEF 1080

#define XRES_MAX 8192
#define YRES_MAX 8192

extern bool enable_cursor;
struct xdx_device;

struct vkms_composer {
	struct drm_framebuffer fb;
	struct drm_rect src, dst;
	unsigned int offset;
	unsigned int pitch;
	unsigned int cpp;
};

/**
 * vkms_plane_state - Driver specific plane state
 * @base: base plane state
 * @composer: data required for composing computation
 */
struct vkms_plane_state {
	struct drm_plane_state base;
	struct vkms_composer *composer;
	struct dentry *dump_debugfs;
	struct debugfs_blob_wrapper dump_blob;
};

/**
 * vkms_crtc_state - Driver specific CRTC state
 * @base: base CRTC state
 * @composer_work: work struct to compose and add CRC entries
 * @n_frame_start: start frame number for computed CRC
 * @n_frame_end: end frame number for computed CRC
 */
struct vkms_crtc_state {
	struct drm_crtc_state base;
	struct work_struct composer_work;

	int num_active_planes;
	/* stack of active planes for crc computation, should be in z order */
	struct vkms_plane_state **active_planes;

	/* below three are protected by vkms_output.composer_lock */
	bool crc_pending;
	u64 frame_start;
	u64 frame_end;
};

struct vkms_output {
	struct drm_crtc crtc;
	struct drm_encoder encoder;
	struct drm_connector connector;
	struct hrtimer vblank_hrtimer;
	ktime_t period_ns;
	struct drm_pending_vblank_event *event;
	/* ordered wq for composer_work */
	struct workqueue_struct *composer_workq;
	/* protects concurrent access to composer */
	spinlock_t lock;

	/* protected by @lock */
	bool composer_enabled;
	struct vkms_crtc_state *composer_state;

	spinlock_t composer_lock;
};

struct vkms_device {
	struct xdx_ip_block base;
	struct vkms_output output;
	uint32_t num_crtc;
};

#define vkms_output_to_vkmsdev(target)                                         \
	container_of(target, struct vkms_device, output)

#define drm_crtc_to_vkms_output(target)                                        \
	container_of(target, struct vkms_output, crtc)

#define drm_encoder_to_vkms_output(target)                                     \
	container_of(target, struct vkms_output, encoder)

#define to_vkms_crtc_state(target)                                             \
	container_of(target, struct vkms_crtc_state, base)

#define to_vkms_plane_state(target)                                            \
	container_of(target, struct vkms_plane_state, base)

/* CRTC */
int xdx_vkms_crtc_init(struct drm_device *dev, struct drm_crtc *crtc,
		       struct drm_plane *primary, struct drm_plane *cursor);

bool xdx_vkms_get_vblank_timestamp(struct drm_device *dev, unsigned int pipe,
				   int *max_error, ktime_t *vblank_time,
				   bool in_vblank_irq);

int xdx_vkms_output_init(struct xdx_device *xdev, int index);

struct drm_plane *xdx_vkms_plane_init(struct xdx_device *xdev,
				      enum drm_plane_type type, int index);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)) || (defined(REDHAT_83))
/* CRC Support */
int xdx_vkms_verify_crc_source(struct drm_crtc *crtc, const char *src_name,
			       size_t *values_cnt);
const char *const *xdx_vkms_get_crc_sources(struct drm_crtc *crtc,
					size_t *count);
int xdx_vkms_set_crc_source(struct drm_crtc *crtc, const char *src_name);
#else
int xdx_vkms_set_crc_source(struct drm_crtc *crtc, const char *src_name, size_t *value_cnt);
#endif

/* Composer Support */
void xdx_vkms_composer_worker(struct work_struct *work);

extern const struct xdx_ip_block_version vkms_ip_block_version;

#endif /* _XDX_VKMS_H_ */
