/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <linux/module.h>
/* Our header */
#include <linux/of_reserved_mem.h>
#include <linux/of_irq.h>

#include "hantro.h"
#include "hantro_vcmd.h"
#include "hantro_device.h"
#include "hantro_priv.h"

static vcmd_slice_str * slice_list;
static int vcmd_slice_num;

vcmd_slice_str *get_vcmd_slice_head(void)
{
	return slice_list;
}

void set_vcmd_slice_listnull(void)
{
	slice_list = NULL;
}

vcmd_core_str *get_dev_of_core(vcmd_dev_str *vcmd_dev, u32 id)
{
	vcmd_core_str *vcmd_core;

	vcmd_core = vcmd_dev->vcmd_core;
	while (vcmd_core) {
		if (vcmd_core->core_id == id)
			break;
		vcmd_core = vcmd_core->core_next;
	}
	if (!vcmd_core)
		XDX_DEV_ERROR("ERROR %s,%d\n", __func__, __LINE__);
	return vcmd_core;
}

vcmd_slice_str *get_slice_by_sliceidx(u32 sliceidx)
{
	vcmd_slice_str *slice_node;
	slice_node = slice_list;
	while (slice_node) {
		if (slice_node->sliceidx == sliceidx)
			break;
		slice_node = slice_node->slice_next;
	}
	if (!slice_node)
		XDX_DEV_ERROR("ERROR %s,%d\n", __func__, __LINE__);

	return slice_node;
}

vcmd_dev_str *get_dev_by_sliceidx(u32 sliceidx, u32 group_codec)
{
	vcmd_slice_str *slice_node;
	vcmd_dev_str *codec_dev;

	slice_node = get_slice_by_sliceidx(sliceidx);
	if (!slice_node) {
		XDX_DEV_ERROR("ERROR %s,%d\n", __func__, __LINE__);
		return NULL;
	}
	if (group_codec == HANTRO_DECODER)
		codec_dev = &slice_node->dec_vcmd;
	else
		codec_dev = &slice_node->enc_vcmd;

	return codec_dev;
}

u32 get_vcmd_slice_config(u32 sliceidx)
{
	vcmd_slice_str *slice_node;

	slice_node = get_slice_by_sliceidx(sliceidx);
	if (!slice_node) {
		XDX_DEV_ERROR("ERROR %s,%d\n", __func__, __LINE__);
		return 0;
	}
	return slice_node->config;
}

void add_vcmd_slice_config(u32 sliceidx, u32 config)
{
	vcmd_slice_str *slice_node;

	slice_node = get_slice_by_sliceidx(sliceidx);
	if (!slice_node) {
		XDX_DEV_ERROR("ERROR %s,%d\n", __func__, __LINE__);
		return;
	}
	slice_node->config |= config;
}

int get_vcmd_slice_num(void)
{
	return vcmd_slice_num;
}

void add_vcmd_slice(struct device *dev, vcmd_slice_str *slice_node)
{
	u32 num = 0;
	vcmd_slice_str *head;

	head = slice_list;

	slice_node->dec_vcmd.dev = dev;
	slice_node->enc_vcmd.dev = dev;
	slice_node->dec_vcmd.type = HANTRO_DECODER;
	slice_node->enc_vcmd.type = HANTRO_ENCODER;
	vcmd_slice_num++;

	if (!head) {
		slice_node->sliceidx = 0;
		slice_list = slice_node;
		return;
	}
	while (head->slice_next) {
		head = head->slice_next;
		num++;
	}
	slice_node->sliceidx = num + 1;
	head->slice_next = slice_node;
}

void add_vcmd_core(vcmd_slice_str *slice_node, vcmd_core_str *vcmd_core)
{
	vcmd_core_str **head, *temp_head;

	if (vcmd_core->type == HANTRO_CORE_DEC ||
	    vcmd_core->type == HANTRO_CORE_DECJPG) {

		head = &slice_node->dec_vcmd.vcmd_core;
		vcmd_core->core_id = slice_node->dec_vcmd.subsys_num;
		vcmd_core->parent_dev = &slice_node->dec_vcmd;
		slice_node->dec_vcmd.subsys_num++;
		slice_node->config |= CONFIG_HWDEC;
	} else if (vcmd_core->type == HANTRO_CORE_ENC ||
		   vcmd_core->type == HANTRO_CORE_IM ||
		   vcmd_core->type == HANTRO_CORE_ENCJPG ){

		head = &slice_node->enc_vcmd.vcmd_core;
		vcmd_core->core_id = slice_node->enc_vcmd.subsys_num;
		vcmd_core->parent_dev = &slice_node->enc_vcmd;
		slice_node->enc_vcmd.subsys_num++;
		slice_node->config |= CONFIG_HWENC;
	} else {
		XDX_DEV_ERROR("unkonwn vcmd_core->type %d\n", vcmd_core->type);
		return;
	}

	if (!(*head)) {
		*head = vcmd_core;
		return;
	}
	temp_head = *head;

	while (temp_head->core_next)
		temp_head = temp_head->core_next;
	temp_head->core_next = vcmd_core;
}

#ifdef USE_DTB_PROBE

static void get_node_info(struct device_node *ofnode, u64 *addr, u64 *size,
			  u32 *irq)
{
	int i, na, ns;
	struct fwnode_handle *fwnode;
	u32 reg_u32[4];
	u64 ioaddress, iosize;
	struct resource r;
	int endian = of_device_is_big_endian(ofnode);

	fwnode = &ofnode->fwnode;
	na = of_n_addr_cells(ofnode);
	ns = of_n_size_cells(ofnode);
	if (na > 2 || ns > 2) {
		XDX_DEV_ERROR("cell size too big");
		return;
	}

	fwnode_property_read_u32_array(fwnode, "reg", reg_u32, na + ns);
	if (na == 2) {
		if (!endian) {
			ioaddress = reg_u32[0];
			ioaddress <<= 32;
			ioaddress |= reg_u32[1];
		} else {
			ioaddress = reg_u32[1];
			ioaddress <<= 32;
			ioaddress |= reg_u32[0];
		}
	} else {
		ioaddress = reg_u32[0];
	}
	if (ns == 2) {
		if (!endian) {
			iosize = reg_u32[na];
			iosize <<= 32;
			iosize |= reg_u32[na + 1];
		} else {
			iosize = reg_u32[na + 1];
			iosize <<= 32;
			iosize |= reg_u32[na];
		}
	} else {
		iosize = reg_u32[na];
	}
	*addr = ioaddress;
	*size = iosize;
	XDX_DEV_DEBUG("node regio =%llx:%llx", ioaddress, iosize);

	for (i = 0; i < 4; i++) {
		*(irq + i) = -1;
		if (of_irq_to_resource(ofnode, i, &r) > 0) {
			int irq_val = of_irq_get(ofnode, i);

			XDX_DEV_DEBUG("irq %d:%s = %lld:%d", i, r.name, r.start,
				irq_val);
			if (irq_val > 0)
				*(irq + i) = irq_val;
		}
	}
}

static void get_submodule(struct vcmd_config *vcmd_cfg_p,
			  struct subsys_addr *temp_subsys_addr,
			  struct device_node *ofnode)
{
	int type;
	int irq[4];
	u64 ioaddress, iosize;

	get_node_info(ofnode, &ioaddress, &iosize, &irq[0]);
	type = getnodetype(ofnode->name);

	switch (type) {
	case HANTRO_CORE_VCMD: {
		vcmd_cfg_p->vcmd_base_addr = ioaddress;
		vcmd_cfg_p->vcmd_iosize = iosize;
		vcmd_cfg_p->vcmd_irq = irq[0];
		break;
	}
	case HANTRO_CORE_DEC400: {
		temp_subsys_addr->submodule_dec400_addr = ioaddress;
		break;
	}
	case HANTRO_CORE_AXIFE: {
		if (temp_subsys_addr->submodule_axife_addr[0] == 0xFFFF)
			temp_subsys_addr->submodule_axife_addr[0] = ioaddress;
		else
			temp_subsys_addr->submodule_axife_addr[1] = ioaddress;
		break;
	}
	case HANTRO_CORE_CACHE: {
		temp_subsys_addr->submodule_L2Cache_addr = ioaddress;
		break;
	}
	case HANTRO_CORE_MMU: {
		temp_subsys_addr->submodule_MMU_addr = ioaddress;
		break;
	}
	default:
		break;
	}
}

vcmd_slice_str *hantro_vcmd_analyze_subnode(struct device *dev,
					    struct device_node *slice)
{
	struct device_node *child_level_0, *child_level_1;
	vcmd_slice_str *cur_slice;
	vcmd_core_str *vcmd_core;
	struct vcmd_config *vcmd_cfg_p;
	struct subsys_addr temp_subsys_addr;
	int irq[4];
	u64 ioaddress, iosize;

	cur_slice = kzalloc(sizeof(vcmd_slice_str), GFP_KERNEL);
	add_vcmd_slice(dev, cur_slice);

	for_each_child_of_node(slice, child_level_0) {
		vcmd_core = kzalloc(sizeof(vcmd_dev_str), GFP_KERNEL);
		vcmd_cfg_p = &vcmd_core->vcmd_core_cfg;

		vcmd_core->type = getnodetype(child_level_0->name);
		switch (vcmd_core->type) {
		case HANTRO_CORE_DEC:
			vcmd_cfg_p->sub_module_type = VCMD_TYPE_DECODER;
			break;
		case HANTRO_CORE_ENC:
			vcmd_cfg_p->sub_module_type = VCMD_TYPE_ENCODER;
			break;
		case HANTRO_CORE_IM:
			vcmd_cfg_p->sub_module_type = VCMD_TYPE_CUTREE;
			break;
		case HANTRO_CORE_DECJPG:
			vcmd_cfg_p->sub_module_type = VCMD_TYPE_JPEG_DECODER;
			break;
		case HANTRO_CORE_ENCJPG:
			vcmd_cfg_p->sub_module_type = VCMD_TYPE_JPEG_ENCODER;
			break;
		default:
			XDX_DEV_ERROR("Error, DTB codec name is not recognized\n");
			break;
		}
		vcmd_cfg_p->submodule_main_addr = 0xffff;
		vcmd_cfg_p->submodule_L2Cache_addr = 0xffff;
		vcmd_cfg_p->submodule_dec400_addr = 0xffff;
		vcmd_cfg_p->submodule_axife_addr[0] = 0xffff;
		vcmd_cfg_p->submodule_axife_addr[1] = 0xffff;
		vcmd_cfg_p->submodule_MMU_addr = 0xffff;

		get_node_info(child_level_0, &ioaddress, &iosize, &irq[0]);
		temp_subsys_addr.submodule_main_addr = ioaddress;
		add_vcmd_core(cur_slice, vcmd_core);
		for_each_child_of_node(child_level_0, child_level_1) {
			get_submodule(vcmd_cfg_p, &temp_subsys_addr,
				      child_level_1);
		}
		if (temp_subsys_addr.submodule_main_addr != 0xffff) {
			if (temp_subsys_addr.submodule_main_addr >
			    vcmd_cfg_p->vcmd_base_addr) {
				vcmd_cfg_p->submodule_main_addr =
					temp_subsys_addr.submodule_main_addr -
					vcmd_cfg_p->vcmd_base_addr;
			} else {
				XDX_DEV_ERROR("Error, DTB submodule_main_addr is not correct\n");
			}
		}

		if (temp_subsys_addr.submodule_dec400_addr != 0xffff) {
			if (temp_subsys_addr.submodule_dec400_addr >
			    vcmd_cfg_p->vcmd_base_addr) {
				vcmd_cfg_p->submodule_dec400_addr =
					temp_subsys_addr.submodule_dec400_addr -
					vcmd_cfg_p->vcmd_base_addr;
				cur_slice->config |= CONFIG_DEC400;
			} else {
				XDX_DEV_ERROR("Error, DTB submodule_dec400_addr is not correct\n");
			}
		}

		if (temp_subsys_addr.submodule_L2Cache_addr != 0xffff) {
			if (temp_subsys_addr.submodule_L2Cache_addr >
			    vcmd_cfg_p->vcmd_base_addr) {
				vcmd_cfg_p->submodule_L2Cache_addr =
					temp_subsys_addr.submodule_L2Cache_addr -
					vcmd_cfg_p->vcmd_base_addr;
				cur_slice->config |= CONFIG_L2CACHE;
			} else {
				XDX_DEV_ERROR("Error, DTB submodule_L2Cache_addr is not correct\n");
			}
		}

		if (temp_subsys_addr.submodule_MMU_addr != 0xffff) {
			if (temp_subsys_addr.submodule_MMU_addr >
			    vcmd_cfg_p->vcmd_base_addr) {
				vcmd_cfg_p->submodule_MMU_addr =
					temp_subsys_addr.submodule_MMU_addr -
					vcmd_cfg_p->vcmd_base_addr;
			} else {
				XDX_DEV_ERROR("Error, DTB submodule_L2Cache_addr is not correct\n");
			}
		}
	}
	return cur_slice;
}
#endif
