/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdx_acpu.h"
#include "xdx_device.h"
#include "rgxdevice.h"
#include "xdxacpu_irq.h"
#include "rgxfwutils.h"


static inline uint64_t acpu_doorbell_rreg64(uint64_t reg_base, uint32_t off)
{
	uint64_t data;
	rmb();
	data = OSReadHWReg64(reg_base, off);
	mb();
	return data;
}

static inline void acpu_doorbell_wreg64(uint64_t reg_base, uint32_t off, uint64_t value)
{
	wmb();
	OSWriteHWReg64(reg_base, off, value);
	mb();
}

/* mmio register read/write functions */
uint64_t cmcu_doorbell_readq(struct xdx_acpu *acpu, uint32_t offset)
{
	u64 data;
	wmb();
	data = OSReadHWReg64(((void __iomem *)acpu->rmmio), offset);
	mb();
	return data;
}

void cmcu_doorbell_writeq(struct xdx_acpu *acpu, uint32_t offset, uint64_t v)
{
	wmb();
	OSWriteHWReg64(((void __iomem *)acpu->rmmio), offset, v);
	mb();
}

void
acpu_doorbell_submit(struct xdx_acpu *acpu, uint32_t dbNum,
		   uint64_t ui64Value)
{
	acpudb_pool_t *pool = acpu->acpu_db_pool;
	mutex_lock(&pool->pool_mutex);
	acpu_doorbell_wreg64((uint64_t)acpu->rmmio, DOORBELL_OFFSET(dbNum),
			     ui64Value);
	WARN_ON(BIT_ISSET(pool->db_submit, dbNum));
	BIT_SET(pool->db_submit, dbNum);
	mutex_unlock(&pool->pool_mutex);
}

int xdx_acpu_fini(struct xdx_device *xdev)
{
	int err = xdxacpu_isr_uninstall(xdev);
	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to uninstall xdx acpu isr !!!",
			__func__);
		return err;
	}
	acpu_doorbell_pool_destroy(xdev->acpu.acpu_db_pool);
	iounmap(xdev->acpu.rmmio);

	return 0;
}

static int xdx_acpu_suspend(void *handle)
{
#ifdef SUPPORT_DMA_TRANSFER
	struct xdx_device *xdev = (struct xdx_device *)handle;
	int r;

	r = acpu_wait_all_db_done(xdev->acpu.acpu_db_pool, 100);
	if (r)
		dev_err(xdev->dev,
			"wait acpu doorbell clear timeout (%d).\n", r);

	return r;
#else
        return 0;
#endif
}

static int xdx_acpu_resume(void *handle)
{
#ifdef SUPPORT_DMA_TRANSFER
	struct xdx_device *xdev = (struct xdx_device *)handle;
	PVRSRV_DEVICE_NODE *psDeviceNode = xdev->gfx.pvrdev;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDeviceNode->pvDevice;
	PVRSRV_ERROR eError;
	IMG_DEV_PHYADDR acpuPhysAddr;
	IMG_BOOL bValid;

	eError = RGXGetPhyAddr(psDevInfo->psTinyCopyMemDesc->psImport->hPMR,
			       &acpuPhysAddr,
			       psDevInfo->psTinyCopyMemDesc->uiOffset,
			       OSGetPageShift(),
			       1,
			       &bValid);
	if (eError != PVRSRV_OK)
	{
		dev_err(xdev->dev,
			"failed to get tiny copy kccb addr (%d).\n", eError);
		return eError;
	}
	psDevInfo->psAcpuFWIfOsInit->psTinyCopyKCCB = acpuPhysAddr;
#endif
	return 0;
}

static int xdx_acpu_early_init(void *handle)
{
	return 0;
}

static int xdx_acpu_early_fini(void *handle)
{
	return 0;
}

static int xdx_acpu_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	u32 cap, ctrl;
	ddkhandshake_cmd_type ddkhandshake_cmd;
	u64 fwcap;
	struct xdx_acpu *acpu = &xdev->acpu;
	int ret, err;

	dev_info(xdev->dev, "acpu init start\n");
	acpu->rmmio_size = XGP_CMCU_DOORBELLS_SIZE;
	acpu->rmmio = ioremap(xdev->rmmio_base + XGP_CMCU_DOORBELLS_BASE, acpu->rmmio_size);
	if (acpu->rmmio == NULL) {
		dev_err(xdev->dev, "acpu failed to map mmio registers\n");
		ret = -ENODEV;
		goto err1;
	}
	err = xdxacpu_isr_install(xdev);
	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to install xdx acpu irq !!!",
			__func__);
		ret = -ENODEV;
		goto err2;
	}
	ret =
	    acpu_doorbell_pool_create(MAX_DOORBELL_POOL_SIZE, (resource_size_t)acpu->rmmio,
				   &(acpu->acpu_db_pool));
	if (PVRSRV_OK != ret) {
		PVR_DPF((PVR_DBG_ERROR, "%s: ACPUDoolBellPoolCreate failed",
			 __func__));
		goto err3;
	}
	cap = acpu_doorbell_get_timeout(acpu->acpu_db_pool, 1000000);
	ctrl = acpu_doorbell_get_timeout(acpu->acpu_db_pool, 1000000);
	if ((cap == INVALID_DB_INDEX) || (ctrl == INVALID_DB_INDEX)) {
		dev_err(xdev->dev, "%s get doorbell from pool timeout",
			__func__);
		ret = -ETIME;
		goto err4;
	}

	ddkhandshake_cmd.all = 0;
	ddkhandshake_cmd.cmd = acpu_dbcmd_ddkmisc;
	ddkhandshake_cmd.subcmd = acpu_dbcmd_ddkmisc_handshake_getcap;

	fwcap = acpu_doorbell_rreg64((uint64_t)acpu->rmmio, DOORBELL_OFFSET(cap));
	dev_info(xdev->dev, "default runtime ACPU  Capability fwcap: 0x%llx.\n",
		 fwcap);
	acpu_doorbell_submit(acpu, cap, ddkhandshake_cmd.all);
	if (acpu_waitfor_value
	    ((volatile uint64_t
	      __iomem)((acpu->rmmio + DOORBELL_OFFSET(cap))), 0x1,
	     0x1) != PVRSRV_OK) {
		dev_err(xdev->dev, "%s DoorBell cap timeout", __func__);
		ret = -ETIME;
		goto err4;
	}

	fwcap = acpu_doorbell_rreg64((uint64_t)acpu->rmmio, DOORBELL_OFFSET(cap));
	dev_info(xdev->dev, "runtime ACPU  Capability fwcap: 0x%llx.\n", fwcap);

	/*ACPU don't clear GPU int by default
	 *Use MSIX int type by default
	 */
	ddkhandshake_cmd.all = 0;
	ddkhandshake_cmd.cmd = acpu_dbcmd_ddkmisc;
	ddkhandshake_cmd.subcmd = acpu_dbcmd_ddkmisc_handshake_cntl;
	ddkhandshake_cmd.fwcntl.clr_gpu_int = 0;
	ddkhandshake_cmd.fwcntl.pcie_int_type = MSIX;
	acpu_doorbell_submit(acpu, ctrl, ddkhandshake_cmd.all);

	if (acpu_waitfor_value
	    ((uint64_t __iomem) ((resource_size_t)acpu->rmmio + DOORBELL_OFFSET(ctrl)),
	     0x1, 0x1) != PVRSRV_OK) {
		dev_err(xdev->dev, "%s DoorBell ctrl timeout", __func__);
		ret = -ETIME;
		goto err4;
	}
	fwcap = acpu_doorbell_rreg64((uint64_t)acpu->rmmio, DOORBELL_OFFSET(ctrl));
	dev_info(xdev->dev, "runtime ACPU fwcntl: 0x%llx.\n", fwcap);
	dev_info(xdev->dev, "acpu init success\n");
	return PVRSRV_OK;

err4:
	acpu_doorbell_pool_destroy(acpu->acpu_db_pool);
err3:
	xdxacpu_isr_uninstall(xdev);
err2:
	iounmap(acpu->rmmio);
err1:
	return ret;
}

static int xdx_acpu_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	int err = xdxacpu_isr_uninstall(xdev);

	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to uninstall xdx acpu isr !!!",
			__func__);
		return err;
	}
	acpu_doorbell_pool_destroy(xdev->acpu.acpu_db_pool);
	iounmap(xdev->acpu.rmmio);
	return 0;
}

static int xdx_acpu_hw_init(void *handle)
{
	return 0;
}

static int xdx_acpu_hw_fini(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version acpu_ip_block_version = {
	.name = "acpu",
	.type = XDX_IP_BLOCK_TYPE_ACPU,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.suspend = xdx_acpu_suspend,
		.resume = xdx_acpu_resume,
		.early_init = xdx_acpu_early_init,
		.early_fini = xdx_acpu_early_fini,
		.sw_init = xdx_acpu_sw_init,
		.sw_fini = xdx_acpu_sw_fini,
		.hw_init = xdx_acpu_hw_init,
		.hw_fini = xdx_acpu_hw_fini,
	},
};
