/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 1, 0)
#include <linux/idr.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)) || (defined(REDHAT_83))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
#include <drm/drm_drv.h>
#endif
#include <drm/xdxgpu_drm.h>

#include "pangu.h"
#include "xdx_device.h"
#include "xdx_sysfs.h"
#include "xdx_drv.h"
#include "xdx_gfx.h"
#include "dm.h"
#include "xdx_vpu.h"
#include "xdx_audio.h"
#include "audio/xdxaudio_hal.h"
#include "xdx_vkms.h"
#include "xdx_acpu.h"
#include "pm/xdx_hwmon.h"
#include "pm/xdx_thm.h"
#include "pm/xdx_cmcu.h"
#include "pm/xdx_pm.h"
#include "dm_fb.h"
#include "osfunc.h"
#include "pm/fwshare/FirmwareInfo.h"

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
#define XDX_PCI_REBAR_CAP            PCI_REBAR_CAP
#define XDX_PCI_REBAR_CAP_SIZES      PCI_REBAR_CAP_SIZES
#define XDX_PCI_REBAR_CTRL_BAR_IDX   PCI_REBAR_CTRL_BAR_IDX
#define XDX_PCI_REBAR_CTRL_BAR_SIZE  PCI_REBAR_CTRL_BAR_SIZE  /* BAR size */
#else
#define XDX_PCI_REBAR_CAP            4
#define XDX_PCI_REBAR_CAP_SIZES      0x00FFFFF0
#define XDX_PCI_REBAR_CTRL_BAR_IDX   0x00000007
#define XDX_PCI_REBAR_CTRL_BAR_IDX   0x00000007
#define XDX_PCI_REBAR_CTRL_BAR_SIZE  0x00001F00  /* BAR size */
#endif
#define XDX_PCI_RES_TYPE_MASK \
  	(IORESOURCE_IO | IORESOURCE_MEM | IORESOURCE_PREFETCH |\
  	 IORESOURCE_MEM_64)

#define DEFAULT_CARDBUS_IO_SIZE		(256)
#define DEFAULT_CARDBUS_MEM_SIZE	(64*1024*1024)
/* pci=cbmemsize=nnM,cbiosize=nn can override this */
unsigned long pci_cardbus_io_size = DEFAULT_CARDBUS_IO_SIZE;
unsigned long pci_cardbus_mem_size = DEFAULT_CARDBUS_MEM_SIZE;

#define DEFAULT_HOTPLUG_IO_SIZE		(256)
#define DEFAULT_HOTPLUG_MEM_SIZE	(2*1024*1024)
/* pci=hpmemsize=nnM,hpiosize=nn can override this */
unsigned long pci_hotplug_io_size  = DEFAULT_HOTPLUG_IO_SIZE;
unsigned long pci_hotplug_mem_size = DEFAULT_HOTPLUG_MEM_SIZE;

#define PCI_P2P_DEFAULT_MEM_ALIGN	0x100000	/* 1MiB */
#define PCI_P2P_DEFAULT_IO_ALIGN	0x1000		/* 4KiB */
#define PCI_P2P_DEFAULT_IO_ALIGN_1K	0x400		/* 1KiB */
#define	to_pci_host_bridge(n) container_of(n, struct pci_host_bridge, dev)

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static DEFINE_RWLOCK(resource_lock);
#endif

struct xdx_pci_dev_resource {
  	struct list_head list;
  	struct resource *res;
  	struct pci_dev *dev;
  	resource_size_t start;
  	resource_size_t end;
  	resource_size_t add_size;
  	resource_size_t min_align;
  	unsigned long flags;
};

/* mmio register read/write functions */
uint64_t xdx_mmio_readq(struct xdx_device *xdev, uint32_t offset)
{
	if (offset < xdev->rmmio_size)
		return (readq(xdev->rmmio + offset));
	BUG();
	return 0;
}

void xdx_mmio_writeq(struct xdx_device *xdev, uint32_t offset,
		     uint64_t v)
{
	if (offset < xdev->rmmio_size)
		writeq(v, ((void __iomem *)xdev->rmmio) + offset);
	else
		BUG();
}

uint32_t xdx_mmio_readl(struct xdx_device *xdev, uint32_t offset)
{
	if (offset < xdev->rmmio_size)
		return (readl(xdev->rmmio + offset));
	BUG();
	return 0;
}

void xdx_mmio_writel(struct xdx_device *xdev, uint32_t offset,
		     uint32_t v)
{
	if (offset < xdev->rmmio_size)
		writel(v, ((void __iomem *)xdev->rmmio) + offset);
	else
		BUG();
}

uint16_t xdx_mmio_readw(struct xdx_device *xdev, uint32_t offset)
{
        if (offset < xdev->rmmio_size)
                return (readw(xdev->rmmio + offset));
        BUG();
        return 0;
}

void xdx_mmio_writew(struct xdx_device *xdev, uint32_t offset,
                     uint16_t v)
{
        if (offset < xdev->rmmio_size)
                writew(v, ((void __iomem *)xdev->rmmio) + offset);
        else
                BUG();
}

uint8_t xdx_mmio_readb(struct xdx_device *xdev, uint32_t offset)
{
	if (offset < xdev->rmmio_size)
		return (readb(xdev->rmmio + offset));
	BUG();
}

void xdx_mmio_writeb(struct xdx_device *xdev, uint32_t offset, uint8_t value)
{
	if (offset < xdev->rmmio_size)
		writeb(value, xdev->rmmio + offset);
	else
		BUG();
}

uint32_t xdx_device_rreg(struct xdx_device *xdev, uint32_t reg,
			 uint32_t acc_flags)
{
	if ((reg * 4) < xdev->rmmio_size)
		return readl(((void __iomem *)xdev->rmmio) + (reg * 4));
	BUG();
}

void xdx_device_wreg(struct xdx_device *xdev, uint32_t reg, uint32_t v)
{
	if ((reg * 4) < xdev->rmmio_size)
		writel(v, ((void __iomem *)xdev->rmmio) + (reg * 4));
	else
		BUG();
}

bool xdx_device_is_dc_supported(enum xdx_asic_type asic_type)
{
	switch (asic_type) {
	case CHIP_PANGU_A0:
	case CHIP_PANGU_B0:
		return true;
	default:
		DRM_WARN_ONCE("unsupported asic type: 0x%x\n", asic_type);
		return false;
	}
}

/**
 * @brief Get clock info from BIOS or firmware
 *
 * @param xdev
 * @return int
 */
static int xdx_device_get_clock_info(struct xdx_device *xdev)
{
	xdev->pm.default_gfxclk = xdev->fw_st_info.default_gpu_freq;
	xdev->pm.current_gfxclk = get_gpu_freq(xdev);

	return 0;
}

static void xdx_device_enable_virtual_display(struct xdx_device *xdev)
{
#ifndef NO_HARDWARE
	xdev->flags.virtual_display = false;

	if (xdx_virtual_display) {
		const char *pci_address_name = pci_name(xdev->pdev);
		char *pciaddstr, *pciaddstr_tmp, *pciaddname_tmp, *pciaddname;

		pciaddstr = kstrdup(xdx_virtual_display, GFP_KERNEL);
		pciaddstr_tmp = pciaddstr;
		while ((pciaddname_tmp = strsep(&pciaddstr_tmp, ";"))) {
			pciaddname = strsep(&pciaddname_tmp, ",");
			if (!strcmp("all", pciaddname)
			    || !strcmp(pci_address_name, pciaddname)) {
				long num_crtc;
				int res = -1;

				xdev->flags.virtual_display = true;

				if (pciaddname_tmp)
					res = kstrtol(pciaddname_tmp, 10,
						      &num_crtc);

				if (!res) {
					if (num_crtc < 1)
						num_crtc = 1;
					if (num_crtc > 6)
						num_crtc = 6;
					xdev->vkmsdev.num_crtc = num_crtc;
				} else {
					xdev->vkmsdev.num_crtc = 1;
				}
				break;
			}
		}

		DRM_INFO("virtual display string:%s, %s:virtual_display:%d, num_crtc:%d\n",
			 xdx_virtual_display, pci_address_name,
			 xdev->flags.virtual_display, xdev->vkmsdev.num_crtc);

		kfree(pciaddstr);
	}
#else
	xdev->flags.virtual_display = true;
#endif /* NO_HARDWARE */
}

static void xdx_device_module_param_handle(struct xdx_device *xdev)
{
	if (xdx_virtual_vblank)
		xdev->flags.virtual_vblank = true;

	if (xdx_virtual_output)
		xdev->flags.virtual_output = true;

	if (xdx_fake_mode)
		xdev->flags.fake_mode = true;

	xdev->dc_heap_size = xdx_dc_heap_size;

	xdx_device_enable_virtual_display(xdev);
}

static int xdx_device_ip_early_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		xdev->ip_blocks[i]->status.early_init = 1;

		if (!xdev->ip_blocks[i]->version.funcs.early_init)
			continue;

		ret = xdev->ip_blocks[i]->version.funcs.early_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.early_init = 0;
			dev_err(xdev->dev,
				"early_init of IP block <%s> failed %d\n",
				xdev->ip_blocks[i]->version.name, ret);
		}
	}

	return 0;
}

static int xdx_device_ip_early_fini(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.early_init)
			continue;

		xdev->ip_blocks[i]->status.early_init = 0;

		if (!xdev->ip_blocks[i]->version.funcs.early_fini)
			continue;

		ret = xdev->ip_blocks[i]->version.funcs.early_fini(xdev);
		if (ret) {
			dev_err(xdev->dev,
				"early_fini of IP block <%s> failed %d\n",
				xdev->ip_blocks[i]->version.name, ret);
		}
	}

	return 0;
}

static int xdx_device_ip_sw_fini(struct xdx_device *xdev)
{
	int i, ret;

	for (i = xdev->num_ip_blocks - 1; i >= 0; i--) {
		if (!xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.sw_fini) {
			dev_warn(xdev->dev, "sw_fini of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			continue;
		}

		ret = xdev->ip_blocks[i]->version.funcs.sw_fini(xdev);
		if (ret) {
			dev_err(xdev->dev, "sw_fini of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.sw_init = 0;
	}

	return ret;
}

static int xdx_device_ip_sw_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.early_init)
			continue;

		if (xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.sw_init) {
			dev_err(xdev->dev, "sw_init of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			break;
		}

		ret = xdev->ip_blocks[i]->version.funcs.sw_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.sw_init = 0;
			dev_err(xdev->dev, "sw_init of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.sw_init = 1;
	}

	if (ret)
		xdx_device_ip_sw_fini(xdev);

	return ret;
}

static int xdx_device_ip_hw_fini(struct xdx_device *xdev)
{
	int i, ret;

	for (i = xdev->num_ip_blocks - 1; i >= 0; i--) {
		if (!xdev->ip_blocks[i]->status.hw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.hw_fini) {
			dev_warn(xdev->dev, "hw_fini of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			continue;
		}

		ret = xdev->ip_blocks[i]->version.funcs.hw_fini(xdev);
		if (ret) {
			dev_err(xdev->dev, "hw_fini of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.hw_init = 0;
	}

	return ret;
}

static int xdx_device_ip_hw_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (xdev->ip_blocks[i]->status.hw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.hw_init) {
			dev_err(xdev->dev, "hw_init of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			break;
		}

		ret = xdev->ip_blocks[i]->version.funcs.hw_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.hw_init = 0;
			dev_err(xdev->dev, "hw_init of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.hw_init = 1;
	}

	if (ret)
		xdx_device_ip_hw_fini(xdev);

	return ret;
}

#if 1
static int xdx_device_ip_init(struct xdx_device *xdev)
{
	int ret;

	ret = xdx_device_ip_early_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to early init IPs\n");
		return ret;
	}

	ret = xdx_device_ip_sw_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to sw init IPs\n");
		goto failed_sw_init;
	}

	xdx_vmm_init(&xdev->vmm);

	xdev->soc_funcs.hw_init(xdev);

	ret = xdx_device_ip_hw_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to hw init IPs\n");
		goto failed_hw_init;
	}

#ifdef 	SUPPORT_VPU
	ret = xdx_vpu_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init vpu\n");
		goto failed_hw_init;
	}
#endif

#if	defined(SUPPORT_DMA_TRANSFER) || defined(SUPPORT_DMA_TRANSFER_B0)
	ret = xdx_dma_dev_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init dma\n");
		goto failed_hw_init;
	}
#endif

#ifdef 	SUPPORT_AUDIO
	ret = xdx_audio_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init audio\n");
		if (ret != -ENXIO)
			goto failed_hw_init;
	}
#endif

#ifdef 	SUPPORT_PANGU_PM
	ret = xdx_pm_init(xdev);
	if( ret != 0 ) {
    	dev_err(xdev->dev, "%s : failed to init pm!!!\n", __func__);
		goto failed_hw_init;
    }
#endif

return 0;

failed_hw_init:
	xdx_device_ip_sw_fini(xdev);
failed_sw_init:
	xdx_device_ip_early_fini(xdev);
	return ret;
}

static int xdx_device_ip_fini(struct xdx_device *xdev)
{

#ifdef SUPPORT_AUDIO
	xdx_audio_fini(xdev);
#endif

#if defined(SUPPORT_DMA_TRANSFER) ||  defined(SUPPORT_DMA_TRANSFER_B0)
	xdx_dma_dev_finit(xdev);
#endif

#ifdef SUPPORT_VPU
	xdx_vpu_fini(xdev);
#endif

#ifdef 	SUPPORT_PANGU_PM
	xdx_pm_fini(xdev);
#endif

	xdx_device_ip_hw_fini(xdev);

	xdx_vmm_fini(&xdev->vmm);

	xdx_device_ip_sw_fini(xdev);

	xdx_device_ip_early_fini(xdev);

	return 0;
}

#else
static int xdx_device_ip_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	xdx_device_module_param_handle(xdev);

#ifdef 	SUPPORT_PANGU_ACPU
	ret = xdx_acpu_init(xdev);
	if (ret)
		return ret;
#endif

	ret = xdx_gfx_init(xdev);
	if (ret)
		return ret;

	ret = xdx_dm_init(xdev);
	if (ret)
		return ret;

#ifdef 	SUPPORT_VPU
	ret = xdx_vpu_init(xdev);
	if (ret)
		return ret;
#endif

#ifdef 	SUPPORT_DMA_TRANSFER
	xdx_dma_init(xdev);
	if (ret)
		return ret;
#endif

#ifdef 	SUPPORT_AUDIO
	xdx_audio_init(xdev);
	if (ret)
		return ret;
#endif
	return ret;
}

static int xdx_device_ip_fini(struct xdx_device *xdev)
{
	int ret;

#ifdef 	SUPPORT_VPU
	xdx_vpu_fini(xdev);
#endif

#ifdef 	SUPPORT_PANGU_ACPU
	ret = xdx_acpu_fini(xdev);
	if (ret)
		return ret;
#endif

#if defined SUPPORT_PANGU_DC_HDMI
    ret = xdx_hdmi_deinit(xdev->ddev);
	if (ret)
		return ret;
#endif

#if defined SUPPORT_PANGU_DC_DP
	ret = xdx_dp_deinit(xdev->ddev);
	if (ret)
		return ret;
#endif

	ret = xdx_dm_fini(xdev);
	if (ret)
		return ret;
    //TODO
    //xdx_fbdev_fini((struct xdx_fbdev*)xdev->fbdev);

#ifdef 	SUPPORT_DMA_TRANSFER
	xdx_dma_fini(xdev);
#endif
#ifdef 	SUPPORT_AUDIO
	xdx_audio_fini(xdev);
	if (ret)
		return ret;
#endif

	ret = xdx_gfx_fini(xdev);

	xdx_irq_fini(xdev);

	return ret;
}

#endif

static void xdx_device_vram_init(struct xdx_device *xdev)
{
	uint64_t limit = (uint64_t)xdx_vram_limit << 20;
	uint64_t vis_vram_limit = (uint64_t)xdx_vis_vram_limit << 20;
	uint32_t regvalue = 0;
	uint8_t gddr_mask = 0;
	uint8_t count = 4, gddr_mask_tmp = 0;
	uint8_t channel_no = 0;
	uint8_t interleave = 0;
	uint8_t tmp = 0;
	uint64_t gddr_die_sizeGB = 1;
	uint64_t gddr_total_sizeGB = 0;

	xdev->vram_start = xdev->hw_info.gddr_dev_phyaddr_base;
	/**
	 * TODO:
	 * should be queried from firmware or read from a scratch
	 * register
	 * gddr_mask(bit0-3)--0:no gddr,1:with 1 gddr
	 * interleave(bit4-7)--0:no interleave 1:gddr01 or gddr23 2:gddr0123 3:dggr03
	 * gddr_die_sizeGB(bit8-11)--0:1GB 1:2GB 2:4GB 3:8GB
	 * gddr_total_sizeGB(bit12-19)
	 * resv(bit20-31)
	 */
	regvalue = xdx_mmio_readl(xdev, PANGU_REG_GDDR_INFO);
	gddr_mask = regvalue & 0xf;
	gddr_mask_tmp = gddr_mask;
	interleave = (regvalue & 0xf0) >> 4;
	tmp = (regvalue & 0xf00) >> 8;
	while (tmp != 0) {
		gddr_die_sizeGB = gddr_die_sizeGB*2;
		tmp--;
	}

	gddr_total_sizeGB = (regvalue & 0xff000) >> 12;
	while (count--) {
		if (gddr_mask_tmp & 1)
			channel_no++;
		gddr_mask_tmp = gddr_mask_tmp >> 1;
		dev_info(xdev->dev, "gddr_mask_tmp=%d, channel_no=%d\n", gddr_mask_tmp, channel_no);
	}

	xdev->fw_st_info.gddr_count = channel_no;

	if (interleave) {//interleave enabled
#if defined XDX_SUPPORT_VERFW
		if (gddr_total_sizeGB != gddr_die_sizeGB*channel_no) {
			dev_err(xdev->dev, "undefined GDDR total size, pls check fw version\n");
			return;
		}
#endif
		xdev->vram_size = gddr_total_sizeGB << 30;
	} else {
		xdev->vram_size = gddr_die_sizeGB << 30;
	}

	dev_info(xdev->dev, "gddr_mask=%d, interleave=%d, gddr_die_size=%llxGB\n",
			gddr_mask, interleave, gddr_die_sizeGB);
	dev_info(xdev->dev, "gddr_total_size=%llxGB, channel_no=%d, xdev->vram_size=%llx\n",
			gddr_total_sizeGB, channel_no, xdev->vram_size);

	xdev->real_vram_size = xdev->vram_size - xdev->hw_info.reserved_vram_size;
	xdev->vram_end = xdev->vram_start + xdev->real_vram_size - 1;
	if (limit && limit < xdev->real_vram_size)
		xdev->real_vram_size = limit;

	dev_info(xdev->dev, "VRAM: %lluM 0x%016llX - 0x%016llX (%lluM used)\n",
			xdev->vram_size >> 20,
			xdev->vram_start,
			xdev->vram_end,
			xdev->real_vram_size >> 20);

	/* Reduce size of CPU-visible VRAM if requested */
	if (vis_vram_limit &&
	    vis_vram_limit < xdev->visible_vram_size)
		xdev->visible_vram_size = vis_vram_limit;

	if (xdev->visible_vram_size == xdev->vram_size)
		xdev->visible_vram_size = xdev->real_vram_size;

	dev_info(xdev->dev, "Visable VRAM: %lluM 0x%016llX - 0x%016llX (%lluM used)\n",
			xdev->aper_size >> 20,
			xdev->aper_base,
			xdev->aper_base + xdev->aper_size - 1,
			xdev->visible_vram_size >> 20);
}

static void xdx_dev_msi_config(struct xdx_device *xdev)
{
	struct pci_dev *pdev = xdev->pdev;
	int pos = pci_find_capability(pdev, PCI_CAP_ID_MSI);
	u32 addrHI, addrLO, data, msiEn;
	if (xdev->irq.irq_type != XDX_IRQ_MSI) return;
	pci_read_config_dword(pdev, pos + PCI_MSI_ADDRESS_HI, &addrHI);
	pci_read_config_dword(pdev, pos + PCI_MSI_ADDRESS_LO, &addrLO);
	pci_read_config_dword(pdev, pos + PCI_MSI_DATA_64, &data);
	xdx_mmio_writel(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_ADDR_H_PF0, addrHI);
	xdx_mmio_writel(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_ADDR_L_PF0, addrLO);
	xdx_mmio_writel(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_DATA_PF0, data);
	msiEn = xdx_mmio_readl(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_EN);
	dev_info(xdev->dev, "msiEn=%08X\n", msiEn);
	xdx_mmio_writel(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_EN, 0);
	xdx_mmio_writel(xdev, PANGU_REG_HOST_MISC_BASE + mmGEN_MSI_EN, msiEn);
}

int xdx_device_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
#ifndef NO_HARDWARE
	struct pci_dev *pdev = xdev->pdev;
#else
	struct platform_device *pdev = xdev->pdev;
#endif
	int ret;
        int i;

	xdx_device_module_param_handle(xdev);

	switch(xdev->asic_type) {
	case CHIP_PANGU_A0:
		pangu_a0_soc_setup(xdev);
		break;
	case CHIP_PANGU_B0:
		pangu_b0_soc_setup(xdev);
		break;
	default:
		dev_err(&pdev->dev, "wrong asic type\n");
		return -ENODEV;
	}

	xdev->shutdown = false;
	xdev->gfx.gfx_timeout = msecs_to_jiffies(10000);
	xdev->vdma_timeout = xdev->gfx.gfx_timeout;
	xdev->gfx.compute_timeout = msecs_to_jiffies(60000);
        for (i = 0; i < XDX_SNAPSHOT_HEAP_MAX; ++i) {
                xdev->gfx.pvr_heap_iter[i] = NULL;
	        xdev->gfx.pvr_pwr_buffer[i] = NULL;
        }
	atomic_set(&xdev->in_s3, 0);
	atomic_set(&xdev->in_s4, 0);

#ifndef NO_HARDWARE
	dev_info(xdev->dev,
		 "initializing device (0x%04X:0x%04X 0x%04X:0x%04X 0x%02X).\n",
		 pdev->vendor, pdev->device, pdev->subsystem_vendor,
		 pdev->subsystem_device, pdev->revision);

	xdev->fw_st_info.device_id = pdev->device;
	xdx_irq_init(xdev);

	/* Registers mapping */
	xdev->rmmio_base = pci_resource_start(xdev->pdev, 0);
	xdev->rmmio_size = pci_resource_len(xdev->pdev, 0);
	xdev->rmmio = ioremap(xdev->rmmio_base, xdev->rmmio_size);
    printk("Debug: xdev->rmmio address:0x%llx  value:0x%llx\n", (long long int)(&xdev->rmmio), (long long int)(xdev->rmmio));
	if (xdev->rmmio == NULL) {
		dev_err(xdev->dev, "failed to map mmio registers\n");
		return -ENOMEM;
	}
	dev_info(xdev->dev, "register mmio base: 0x%016llx\n",
		 (uint64_t)xdev->rmmio_base);
	dev_info(xdev->dev, "register mmio size: %uM\n",
		 (uint32_t)(xdev->rmmio_size >> 20));

	/* Get CPU visable VRAM */
	xdev->aper_base = pci_resource_start(xdev->pdev, 2);
	xdev->aper_size = pci_resource_len(xdev->pdev, 2);
	xdev->visible_vram_size = xdev->aper_size;
	/* reserve PAT memory space to WC for VRAM */
	arch_io_reserve_memtype_wc(xdev->aper_base, xdev->aper_size);
	/* Add an MTRR for the VRAM */
	xdev->vram_mtrr = arch_phys_wc_add(xdev->aper_base, xdev->aper_size);

	dev_info(xdev->dev, "visable vram base: 0x%016llx\n",
		 (uint64_t)xdev->aper_base);
	dev_info(xdev->dev, "visable vram size: %dM\n",
		 (uint32_t)(xdev->aper_size >> 20));

	/* must call query soc info firstly */
	xdev->soc_funcs.get_device_info(xdev);

	/* enable msi config when IRQ type is MSI */
	if(XDX_IRQ_MSI == xdev->irq.irq_type)
		xdx_dev_msi_config(xdev);

	xdx_device_vram_init(xdev);

#ifdef SUPPORT_PANGU_ACPU
	xdev->shareram_base = pci_resource_start(xdev->pdev, 4);
	xdev->shareram_size = pci_resource_len(xdev->pdev, 4);
	xdev->shareram_useful_size = PANGU_SHARERAM_SIZE;//only 256K used for driver, fw reserved other for tool
	xdev->share_ram = ioremap_wc(xdev->shareram_base, xdev->shareram_useful_size);

	xdx_fwinfo_init(xdev);

	ret = cache_fw_static_info(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to read from fw share ram\n");
		goto err_clock_info;
	}
	xdxgpu_pm_sysfs_init(xdev);
#endif

	xdev->gtt_start = xdev->hw_info.host_dev_phyaddr_base;
	xdev->shareram_start = xdev->hw_info.shareram_dev_phyaddr_base;


	ret = xdx_device_get_clock_info(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to get clock info\n");
		goto err_clock_info;
	}
#else
	xdev->flags.no_hardware = 1;
	xdev->visible_vram_size = 256 * 1024 * 1024;
#endif

	/* early init functions */
	ret = xdx_device_ip_init(xdev);
	if (ret)
		goto err_clock_info;

	ret = xdx_fbdev_init(ddev);
	if (ret)
		goto faild_reg_fbdev;

	xdev->power_state = XDX_PS0;
	return 0;

faild_reg_fbdev:
	xdx_fbdev_fini(xdev->fbdev);

err_clock_info:
#ifndef NO_HARDWARE
	iounmap(xdev->rmmio);
	iounmap(xdev->share_ram);
#endif
	return ret;
}

int xdx_device_fini(struct xdx_device *xdev)
{
	struct xdx_fbdev* fbdev = xdev->fbdev;

	xdx_irq_fini(xdev);
	xdx_fbdev_fini(fbdev);
	xdx_device_ip_fini(xdev);
#ifndef NO_HARDWARE
	iounmap(xdev->rmmio);
	iounmap(xdev->share_ram);
	arch_phys_wc_del(xdev->vram_mtrr);
	arch_io_free_memtype_wc(xdev->aper_base, xdev->aper_size);
#endif
	return 0;
}

/**
 * xdx_device_ip_suspend - run suspend for hardware IPs
 *
 * @xdev: xdxgpu_device pointer
 *
 * Main suspend function for hardware IPs.  The list of all the hardware
 * IPs that make up the asic is walked, clockgating is disabled and the
 * suspend callbacks are run.  suspend puts the hardware and software state
 * in each IP into a state suitable for suspend.
 * Returns 0 on success, negative error code on failure.
 */
int xdx_device_ip_suspend(struct xdx_device *xdev)
{
	int r, i;

	for (i = xdev->num_ip_blocks -1; i >= 0; i--) {
		/* display is handled separately */
		if (xdev->ip_blocks[i]->version.type == XDX_IP_BLOCK_TYPE_DC)
			continue;
		DRM_DEBUG("suspending IP block <%s>\n",
			 xdev->ip_blocks[i]->version.name);

		r = xdev->ip_blocks[i]->version.funcs.suspend(xdev);

		if (r) {
			DRM_ERROR("suspend of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, r);
			return r;
		}

		xdev->ip_blocks[i]->status.hw_init = false;
	}

	return 0;
}

static int xdx_device_ip_resume(struct xdx_device *xdev)
{
	int r;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		/* display is handled separately */
		if (xdev->ip_blocks[i]->version.type == XDX_IP_BLOCK_TYPE_DC)
			continue;
		DRM_DEBUG("resuming IP block <%s>\n",
			 xdev->ip_blocks[i]->version.name);

		r = xdev->ip_blocks[i]->version.funcs.resume(xdev);

		if (r) {
			DRM_ERROR("resume of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, r);
			return r;
		}

		xdev->ip_blocks[i]->status.hw_init = true;
	}
	return 0;
}

static int xdx_device_early_suspend(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int r;

	drm_kms_helper_poll_disable(ddev);
        if (xdev->fbdev)
                drm_fb_helper_set_suspend_unlocked(&((struct xdx_fbdev*)xdev->fbdev)->helper, 1);
	r = xdev->dm.base.version.funcs.suspend(xdev);
	if (r) {
		dev_err(xdev->dev, "display suspend failed (%d).\n", r);
		return r;
	}
	xdev->dm.base.status.hw_init = false;
	return 0;
}

static int xdx_device_early_resume(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	struct drm_crtc *crtc;
	struct drm_pending_vblank_event *event;

	/*
	 * During S3/S4, PCIe would lose power supply and SOC FW
	 * would be reloaded. Which means display HW would be
	 * powered on here. So turn off the display crtc to
	 * avoid abnormal display before gpu core and dc ip block
	 * have been resumed
	 */
	drm_modeset_lock_all(ddev);
	list_for_each_entry(crtc, &ddev->mode_config.crtc_list, head) {
		event = crtc->state->event;
		crtc->state->event = NULL;
		crtc->helper_private->atomic_disable(crtc, NULL);
		crtc->state->event = event;
	}
	drm_modeset_unlock_all(ddev);

	return 0;
}

static int xdx_device_late_resume(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int r;

	/* don't light-up the screen when pm thaw */
	if (atomic_read(&xdev->in_s4))
		return 0;

	r= xdev->dm.base.version.funcs.resume(xdev);
	if (r) {
		dev_err(xdev->dev, "display resume failed (%d).\n", r);
		return r;
	}
	xdev->dm.base.status.hw_init = true;
        if (xdev->fbdev)
                drm_fb_helper_set_suspend_unlocked(&((struct xdx_fbdev*)xdev->fbdev)->helper, 0);
	drm_kms_helper_poll_enable(ddev);

	return 0;
}

/**
 * xdx_device_suspend - initiate device suspend
 *
 * @dev: drm dev pointer
 * @fbcon : notify the fbdev of suspend
 *
 * Puts the hw in the suspend state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver suspend.
 */
int xdx_device_suspend(struct xdx_device *xdev, bool fbcon)
{
	int r;
	r = xdx_device_early_suspend(xdev);
	if (r)
		return r;

	r = xdx_device_ip_suspend(xdev);

	if (r) {
		dev_err(xdev->dev, "xdx_device_ip_suspend failed (%d).\n", r);
		return r;
	}
#ifdef SUPPORT_PANGU_ACPU
	r =  xdx_switch_power_state(xdev, XDX_PS2);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS2);
		return r;
	}
#endif
	/* TODO: release host resources */
	/* TODO: evict memory */
	return 0;
}

/**
 * xdx_device_resume - initiate device resume
 *
 * @dev: drm dev pointer
 * @fbcon : notify the fbdev of resume
 *
 * Bring the hw back to operating state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver resume.
 */
int xdx_device_resume(struct xdx_device *xdev, bool fbcon)
{
	int r;
	struct drm_device *ddev = xdev_to_drm(xdev);

	PublicFirmwareInfo_T *fwinfo;
	u32 timeout = 30;
	fwinfo =(PublicFirmwareInfo_T *)xdev->fwinfo.fwinfo;
#ifdef SUPPORT_PANGU_ACPU
	r =  xdx_switch_power_state(xdev, XDX_PS0);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS0);
		return r;
	}
#endif

	/* FIXME
	 * workarround solution, system hang for s3/s4 abort */
	do
	{
		//printk(KERN_ERR"%s %d sync with fw lpmSync =%d\n",__func__,__LINE__,fwinfo->lpmSync);
		if(timeout-- == 0)
		{
			printk(KERN_ERR"%s %d sync with fw timeout lpmSync =%d\n",__func__,__LINE__,fwinfo->lpmSync);
			break;
		}
		mdelay(100);
	}while(fwinfo->lpmSync ==0);
	fwinfo->lpmSync =0;

	r = xdx_device_early_resume(xdev);
	if (r)
		return r;

	/* TODO: reallocate host resources */
	r = xdx_device_ip_resume(xdev);
	if (r) {
		dev_err(xdev->dev, "xdx_device_ip_resume failed (%d).\n", r);
		return r;
	}

	r = xdx_device_late_resume(xdev);
	if (r)
		return r;
#ifdef SUPPORT_AUDIO
       xdx_audio_resume(xdev);
#endif

	/*
	 * monitors' PHY link-status will change(e.g. new cable in or one cable out)
	 * or may fail to change while in s3/s4;
	 * just send an event and let userspace tell us what to do.
	 *
	 * has no side effect for normal use.
	 * */
	drm_kms_helper_hotplug_event(ddev);

	return 0;
}

int xdx_device_reset(struct xdx_device *xdev)
{
	dev_warn(xdev->dev, "%s: no implemented\n", __func__);
	return -1;
}

int xdx_device_ip_block_add(struct xdx_device *xdev,
			    const struct xdx_ip_block *ip_block)
{
	if (!ip_block)
		return -EINVAL;

	dev_info(xdev->dev, "add ip block number %d <%s>\n", xdev->num_ip_blocks,
		ip_block->version.name);

	xdev->ip_blocks[xdev->num_ip_blocks++] = (struct xdx_ip_block *)ip_block;
	return 0;
}

int xdx_device_info_ioctl(struct drm_device *dev, void *data,
			  struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct drm_xdxgpu_info *info = data;
	void __user *out = (void __user *)(uintptr_t)info->return_pointer;
	uint32_t size = info->return_size;

	if (!info->return_size || !info->return_pointer)
		return -EINVAL;

	switch (info->query) {
	case XDXGPU_INFO_DEVICE_INFO: {
		struct drm_xdxgpu_info_device gpu_info = {};
		struct xdx_fw_static_info *info;
		info = &xdev->fw_st_info;

		gpu_info.asic_id = xdev->asic_type;
		gpu_info.core_num = xdev->hw_info.core_num;
		gpu_info.gddr_count = info->gddr_count;
		gpu_info.gddr_size = info->gddr_size;
		gpu_info.gddr_speed = info->gddr_speed;
		gpu_info.gpu_primary_id = info->gpu_primary_id;
		gpu_info.vendor = info->vendor;
		gpu_info.chip_id = info->chip_id;
		gpu_info.device_id = info->device_id;
		gpu_info.card_type = info->card_type;
		memcpy(gpu_info.serial_number, info->serial_number, sizeof(gpu_info.serial_number));
		gpu_info.fw_version = info->fw_ver;
		gpu_info.vbios_version = info->vbios_ver;
		gpu_info.fan_ctrl_mode = info->fan_ctrl_mode;
		gpu_info.support_fan_info = info->support_fan_info;
		gpu_info.support_temp_info = info->support_temp_info;
		gpu_info.support_power_info = info->support_power_info;
		gpu_info.gpu_default_freq = info->default_gpu_freq;

		return copy_to_user(out, &gpu_info,
					min((size_t)size, sizeof(gpu_info))) ? -EFAULT : 0;
	}
	default:
		dev_err(xdev->dev, "Invalid request %d\n", info->query);
		return -EINVAL;
	}

	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static struct xdx_pci_dev_resource *res_to_dev_res(struct list_head *head,
					       struct resource *res)
{
	struct xdx_pci_dev_resource *dev_res;

	list_for_each_entry(dev_res, head, list) {
		if (dev_res->res == res)
			return dev_res;
	}

	return NULL;
}

static resource_size_t get_res_add_size(struct list_head *head,
					struct resource *res)
{
	struct xdx_pci_dev_resource *dev_res;

	dev_res = res_to_dev_res(head, res);
	return dev_res ? dev_res->add_size : 0;
}

static resource_size_t get_res_add_align(struct list_head *head,
					 struct resource *res)
{
	struct xdx_pci_dev_resource *dev_res;

	dev_res = res_to_dev_res(head, res);
	return dev_res ? dev_res->min_align : 0;
}

static void free_list(struct list_head *head)
{
	struct xdx_pci_dev_resource *dev_res, *tmp;

	list_for_each_entry_safe(dev_res, tmp, head, list) {
		list_del(&dev_res->list);
		kfree(dev_res);
	}
}
static int add_to_list(struct list_head *head,
  		 struct pci_dev *dev, struct resource *res,
 		 resource_size_t add_size, resource_size_t min_align)
{
  	struct xdx_pci_dev_resource *tmp;
  
  	tmp = kzalloc(sizeof(*tmp), GFP_KERNEL);
  	if (!tmp) {
  		pr_warn("add_to_list: kmalloc() failed!\n");
  		return -ENOMEM;
  	}
  
  	tmp->res = res;
  	tmp->dev = dev;
  	tmp->start = res->start;
  	tmp->end = res->end;
  	tmp->flags = res->flags;
  	tmp->add_size = add_size;
  	tmp->min_align = min_align;
  
  	list_add(&tmp->list, head);
  
  	return 0;
}

static void remove_from_list(struct list_head *head,
				 struct resource *res)
{
	struct xdx_pci_dev_resource *dev_res, *tmp;

	list_for_each_entry_safe(dev_res, tmp, head, list) {
		if (dev_res->res == res) {
			list_del(&dev_res->list);
			kfree(dev_res);
			break;
		}
	}
}
#endif

static inline void reset_resource(struct resource *res)
{
	res->start = 0;
	res->end = 0;
	res->flags = 0;
}

static inline u64 xdx_pci_rebar_size_to_bytes(int size)
{
 	return 1ULL << (size + 20);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static struct resource *find_free_bus_resource(struct pci_bus *bus,
			 unsigned long type_mask, unsigned long type)
{
	int i;
	struct resource *r;

	pci_bus_for_each_resource(bus, r, i) {
		if (r == &ioport_resource || r == &iomem_resource)
			continue;
		if (r && (r->flags & type_mask) == type && !r->parent)
			return r;
	}
	return NULL;
}
#endif

resource_size_t resource_alignment(struct resource *res)
{
	switch (res->flags & (IORESOURCE_SIZEALIGN | IORESOURCE_STARTALIGN)) {
	case IORESOURCE_SIZEALIGN:
		return resource_size(res);
	case IORESOURCE_STARTALIGN:
		return res->start;
	default:
		return 0;
	}
}

static unsigned long pci_cardbus_resource_alignment(struct resource *res)
{
	if (res->flags & IORESOURCE_IO)
		return pci_cardbus_io_size;
	if (res->flags & IORESOURCE_MEM)
		return pci_cardbus_mem_size;
	return 0;
}

#if 0//TODO
resource_size_t pci_iov_resource_size(struct pci_dev *dev, int resno)
{
	if (!dev->is_physfn)
		return 0;

	return dev->sriov->barsz[resno - PCI_IOV_RESOURCES];
}


resource_size_t __weak pcibios_iov_resource_alignment(struct pci_dev *dev,
						      int resno)
{
	return pci_iov_resource_size(dev, resno);
}

resource_size_t pci_sriov_resource_alignment(struct pci_dev *dev, int resno)
{
  	return pcibios_iov_resource_alignment(dev, resno);
}
#endif  

static inline resource_size_t pci_resource_alignment(struct pci_dev *dev,
						     struct resource *res)
{
#ifdef CONFIG_PCI_IOV
#if 0//TODO
	int resno = res - dev->resource;
	if (resno >= PCI_IOV_RESOURCES && resno <= PCI_IOV_RESOURCE_END)
		return pci_sriov_resource_alignment(dev, resno);
#endif
#endif
	if (dev->class >> 8  == PCI_CLASS_BRIDGE_CARDBUS)
		return pci_cardbus_resource_alignment(res);
	return resource_alignment(res);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static resource_size_t calculate_iosize(resource_size_t size,
		resource_size_t min_size,
		resource_size_t size1,
		resource_size_t old_size,
		resource_size_t align)
{
	if (size < min_size)
		size = min_size;
	if (old_size == 1)
		old_size = 0;
	/* To be fixed in 2.5: we should have sort of HAVE_ISA
	   flag in the struct pci_bus. */
#if defined(CONFIG_ISA) || defined(CONFIG_EISA)
	size = (size & 0xff) + ((size & ~0xffUL) << 2);
#endif
	size = ALIGN(size + size1, align);
	if (size < old_size)
		size = old_size;
	return size;
}

static resource_size_t calculate_memsize(resource_size_t size,
		resource_size_t min_size,
		resource_size_t size1,
		resource_size_t old_size,
		resource_size_t align)
{
	if (size < min_size)
		size = min_size;
	if (old_size == 1)
		old_size = 0;
	if (size < old_size)
		size = old_size;
	size = ALIGN(size + size1, align);
	return size;
}

static inline resource_size_t calculate_mem_align(resource_size_t *aligns,
						  int max_order)
{
	resource_size_t align = 0;
	resource_size_t min_align = 0;
	int order;

	for (order = 0; order <= max_order; order++) {
		resource_size_t align1 = 1;

		align1 <<= (order + 20);

		if (!align)
			min_align = align1;
		else if (ALIGN(align + min_align, min_align) < align1)
			min_align = align1 >> 1;
		align += aligns[order];
	}

	return min_align;
}

resource_size_t __weak pcibios_window_alignment(struct pci_bus *bus,
						unsigned long type)
{
	return 1;
}

static resource_size_t window_alignment(struct pci_bus *bus,
					unsigned long type)
{
	resource_size_t align = 1, arch_align;

	if (type & IORESOURCE_MEM)
		align = PCI_P2P_DEFAULT_MEM_ALIGN;
	else if (type & IORESOURCE_IO) {
		/*
		 * Per spec, I/O windows are 4K-aligned, but some
		 * bridges have an extension to support 1K alignment.
		 */
		if (bus->self->io_window_1k)
			align = PCI_P2P_DEFAULT_IO_ALIGN_1K;
		else
			align = PCI_P2P_DEFAULT_IO_ALIGN;
	}

	arch_align = pcibios_window_alignment(bus, type);
	return max(align, arch_align);
}

static int pbus_size_mem(struct pci_bus *bus, unsigned long mask,
			 unsigned long type, unsigned long type2,
			 unsigned long type3,
			 resource_size_t min_size, resource_size_t add_size,
			 struct list_head *realloc_head)
{
	struct pci_dev *dev;
	resource_size_t min_align, align, size, size0, size1;
	resource_size_t aligns[18];	/* Alignments from 1Mb to 128Gb */
	int order, max_order;
	struct resource *b_res = find_free_bus_resource(bus,
					mask | IORESOURCE_PREFETCH, type);
	resource_size_t children_add_size = 0;
	resource_size_t children_add_align = 0;
	resource_size_t add_align = 0;

	if (!b_res)
		return -ENOSPC;

	memset(aligns, 0, sizeof(aligns));
	max_order = 0;
	size = 0;

	list_for_each_entry(dev, &bus->devices, bus_list) {
		int i;

		for (i = 0; i < PCI_NUM_RESOURCES; i++) {
			struct resource *r = &dev->resource[i];
			resource_size_t r_size;

			if (r->parent || (r->flags & IORESOURCE_PCI_FIXED) ||
			    ((r->flags & mask) != type &&
			     (r->flags & mask) != type2 &&
			     (r->flags & mask) != type3))
				continue;
			r_size = resource_size(r);
#ifdef CONFIG_PCI_IOV
			/* put SRIOV requested res to the optional list */
			if (realloc_head && i >= PCI_IOV_RESOURCES &&
					i <= PCI_IOV_RESOURCE_END) {
				add_align = max(pci_resource_alignment(dev, r), add_align);
				r->end = r->start - 1;
				add_to_list(realloc_head, dev, r, r_size, 0/* don't care */);
				children_add_size += r_size;
				continue;
			}
#endif
			/*
			 * aligns[0] is for 1MB (since bridge memory
			 * windows are always at least 1MB aligned), so
			 * keep "order" from being negative for smaller
			 * resources.
			 */
			align = pci_resource_alignment(dev, r);
			order = __ffs(align) - 20;
			if (order < 0)
				order = 0;
			if (order >= ARRAY_SIZE(aligns)) {
				dev_warn(&dev->dev, "disabling BAR %d: %pR (bad alignment %#llx)\n",
					 i, r, (unsigned long long) align);
				r->flags = 0;
				continue;
			}
			size += max(r_size, align);
			/* Exclude ranges with size > align from
			   calculation of the alignment. */
			if (r_size <= align)
				aligns[order] += align;
			if (order > max_order)
				max_order = order;

			if (realloc_head) {
				children_add_size += get_res_add_size(realloc_head, r);
				children_add_align = get_res_add_align(realloc_head, r);
				add_align = max(add_align, children_add_align);
			}
		}
	}

	min_align = calculate_mem_align(aligns, max_order);
	min_align = max(min_align, window_alignment(bus, b_res->flags));
	size0 = calculate_memsize(size, min_size, 0, resource_size(b_res), min_align);
	add_align = max(min_align, add_align);
	if (children_add_size > add_size)
		add_size = children_add_size;
	size1 = (!realloc_head || (realloc_head && !add_size)) ? size0 :
		calculate_memsize(size, min_size, add_size,
				resource_size(b_res), add_align);
	if (!size0 && !size1) {
		if (b_res->start || b_res->end)
			dev_info(&bus->self->dev, "disabling bridge window %pR to %pR (unused)\n",
				 b_res, &bus->busn_res);
		b_res->flags = 0;
		return 0;
	}
	b_res->start = min_align;
	b_res->end = size0 + min_align - 1;
	b_res->flags |= IORESOURCE_STARTALIGN;
	if (size1 > size0 && realloc_head) {
		add_to_list(realloc_head, bus->self, b_res, size1-size0, add_align);
		dev_printk(KERN_DEBUG, &bus->self->dev, "bridge window %pR to %pR add_size %llx add_align %llx\n",
			   b_res, &bus->busn_res,
			   (unsigned long long) (size1 - size0),
			   (unsigned long long) add_align);
	}
	return 0;
}

static void pbus_size_io(struct pci_bus *bus, resource_size_t min_size,
		resource_size_t add_size, struct list_head *realloc_head)
{
	struct pci_dev *dev;
	struct resource *b_res = find_free_bus_resource(bus, IORESOURCE_IO,
							IORESOURCE_IO);
	resource_size_t size = 0, size0 = 0, size1 = 0;
	resource_size_t children_add_size = 0;
	resource_size_t min_align, align;

	if (!b_res)
		return;

	min_align = window_alignment(bus, IORESOURCE_IO);
	list_for_each_entry(dev, &bus->devices, bus_list) {
		int i;

		for (i = 0; i < PCI_NUM_RESOURCES; i++) {
			struct resource *r = &dev->resource[i];
			unsigned long r_size;

			if (r->parent || !(r->flags & IORESOURCE_IO))
				continue;
			r_size = resource_size(r);

			if (r_size < 0x400)
				/* Might be re-aligned for ISA */
				size += r_size;
			else
				size1 += r_size;

			align = pci_resource_alignment(dev, r);
			if (align > min_align)
				min_align = align;

			if (realloc_head)
				children_add_size += get_res_add_size(realloc_head, r);
		}
	}

	size0 = calculate_iosize(size, min_size, size1,
			resource_size(b_res), min_align);
	if (children_add_size > add_size)
		add_size = children_add_size;
	size1 = (!realloc_head || (realloc_head && !add_size)) ? size0 :
		calculate_iosize(size, min_size, add_size + size1,
			resource_size(b_res), min_align);
	if (!size0 && !size1) {
		if (b_res->start || b_res->end)
			dev_info(&bus->self->dev, "disabling bridge window %pR to %pR (unused)\n",
				 b_res, &bus->busn_res);
		b_res->flags = 0;
		return;
	}

	b_res->start = min_align;
	b_res->end = b_res->start + size0 - 1;
	b_res->flags |= IORESOURCE_STARTALIGN;
	if (size1 > size0 && realloc_head) {
		add_to_list(realloc_head, bus->self, b_res, size1-size0,
			    min_align);
		dev_printk(KERN_DEBUG, &bus->self->dev, "bridge window %pR to %pR add_size %llx\n",
			   b_res, &bus->busn_res,
			   (unsigned long long)size1-size0);
	}
}
#endif

static int xdx_pci_rebar_find_pos(struct pci_dev *pdev, int bar)
{
	unsigned int pos, nbars, i;
	u32 ctrl;

	pos = pci_find_ext_capability(pdev, PCI_EXT_CAP_ID_REBAR);
	if (!pos)
		return -ENOTSUPP;

	pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
	nbars = (ctrl & PCI_REBAR_CTRL_NBAR_MASK) >> PCI_REBAR_CTRL_NBAR_SHIFT;

	for (i = 0; i < nbars; i++, pos += 8) {
		int bar_idx;

		pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
		bar_idx = ctrl & XDX_PCI_REBAR_CTRL_BAR_IDX;
		if (bar_idx == bar)
			return pos;
	}

	return -ENOENT;
}

u32 xdx_pci_rebar_get_possible_sizes(struct pci_dev *pdev, int bar)
{
	int pos;
	u32 cap;

	pos = xdx_pci_rebar_find_pos(pdev, bar);
	if (pos < 0)
		return 0;

	pci_read_config_dword(pdev, pos + XDX_PCI_REBAR_CAP, &cap);
	cap &= XDX_PCI_REBAR_CAP_SIZES;
	return (fls(cap >> 4) - 1);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static int xdx_pci_rebar_set_size(struct pci_dev *pdev, int bar, int size)
{
  	int pos;
  	u32 ctrl;
  
  	pos = xdx_pci_rebar_find_pos(pdev, bar);
  	if (pos < 0)
  		return pos;
  
  	pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
  	ctrl &= ~XDX_PCI_REBAR_CTRL_BAR_SIZE;
  	ctrl |= size << 8;
  	pci_write_config_dword(pdev, pos + PCI_REBAR_CTRL, ctrl);
  	return 0;
}

static int xdx_pci_rebar_get_current_size(struct pci_dev *pdev, int bar)
{
  	int pos;
  	u32 ctrl;
  
  	pos = xdx_pci_rebar_find_pos(pdev, bar);
  	if (pos < 0)
  		return pos;
  
  	pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
  	return ((ctrl & XDX_PCI_REBAR_CTRL_BAR_SIZE) >> 8);
}
#endif


static void xdx_pci_release_resource(struct pci_dev *dev, int resno)
{
  	struct resource *res = dev->resource + resno;
  
  	dev_info(&dev->dev, "BAR %d: releasing %pR\n", resno, res);
  	release_resource(res);
  	res->end = resource_size(res) - 1;
  	res->start = 0;
  	res->flags |= IORESOURCE_UNSET;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static void assign_requested_resources_sorted(struct list_head *head,
				 struct list_head *fail_head)
{
	struct resource *res;
	struct xdx_pci_dev_resource *dev_res;
	int idx;

	list_for_each_entry(dev_res, head, list) {
		res = dev_res->res;
		idx = res - &dev_res->dev->resource[0];
		if (resource_size(res) &&
		    pci_assign_resource(dev_res->dev, idx)) {
			if (fail_head) {
				/*
				 * if the failed res is for ROM BAR, and it will
				 * be enabled later, don't add it to the list
				 */
				if (!((idx == PCI_ROM_RESOURCE) &&
				      (!(res->flags & IORESOURCE_ROM_ENABLE))))
					add_to_list(fail_head,
						    dev_res->dev, res,
						    0 /* don't care */,
						    0 /* don't care */);
			}
			reset_resource(res);
		}
	}
}

static unsigned long pci_fail_res_type_mask(struct list_head *fail_head)
{
	struct xdx_pci_dev_resource *fail_res;
	unsigned long mask = 0;

	/* check failed type */
	list_for_each_entry(fail_res, fail_head, list)
		mask |= fail_res->flags;

	/*
	 * one pref failed resource will set IORESOURCE_MEM,
	 * as we can allocate pref in non-pref range.
	 * Will release all assigned non-pref sibling resources
	 * according to that bit.
	 */
	return mask & (IORESOURCE_IO | IORESOURCE_MEM | IORESOURCE_PREFETCH);
}

static bool pci_need_to_release(unsigned long mask, struct resource *res)
{
	if (res->flags & IORESOURCE_IO)
		return !!(mask & IORESOURCE_IO);

	/* check pref at first */
	if (res->flags & IORESOURCE_PREFETCH) {
		if (mask & IORESOURCE_PREFETCH)
			return true;
		/* count pref if its parent is non-pref */
		else if ((mask & IORESOURCE_MEM) &&
			 !(res->parent->flags & IORESOURCE_PREFETCH))
			return true;
		else
			return false;
	}

	if (res->flags & IORESOURCE_MEM)
		return !!(mask & IORESOURCE_MEM);

	return false;	/* should not get here */
}



static void pdev_sort_resources(struct pci_dev *dev, struct list_head *head)
{
	int i;

	for (i = 0; i < PCI_NUM_RESOURCES; i++) {
		struct resource *r;
		struct xdx_pci_dev_resource *dev_res, *tmp;
		resource_size_t r_align;
		struct list_head *n;

		r = &dev->resource[i];

		if (r->flags & IORESOURCE_PCI_FIXED)
			continue;

		if (!(r->flags) || r->parent)
			continue;

		r_align = pci_resource_alignment(dev, r);
		if (!r_align) {
			dev_warn(&dev->dev, "BAR %d: %pR has bogus alignment\n",
				 i, r);
			continue;
		}

		tmp = kzalloc(sizeof(*tmp), GFP_KERNEL);
		if (!tmp)
			panic("pdev_sort_resources(): kmalloc() failed!\n");
		tmp->res = r;
		tmp->dev = dev;

		/* fallback is smallest one or list is empty*/
		n = head;
		list_for_each_entry(dev_res, head, list) {
			resource_size_t align;

			align = pci_resource_alignment(dev_res->dev,
							 dev_res->res);

			if (r_align > align) {
				n = &dev_res->list;
				break;
			}
		}
		/* Insert it just before n*/
		list_add_tail(&tmp->list, n);
	}
}

static void __dev_sort_resources(struct pci_dev *dev,
				 struct list_head *head)
{
	u16 class = dev->class >> 8;

	/* Don't touch classless devices or host bridges or ioapics.  */
	if (class == PCI_CLASS_NOT_DEFINED || class == PCI_CLASS_BRIDGE_HOST)
		return;

	/* Don't touch ioapic devices already enabled by firmware */
	if (class == PCI_CLASS_SYSTEM_PIC) {
		u16 command;
		pci_read_config_word(dev, PCI_COMMAND, &command);
		if (command & (PCI_COMMAND_IO | PCI_COMMAND_MEMORY))
			return;
	}

	pdev_sort_resources(dev, head);
}


resource_size_t pcibios_align_resource(void *data, const struct resource *res,
	resource_size_t size, resource_size_t align)
{
	resource_size_t start = res->start;

	if ((res->flags & IORESOURCE_IO) && (start & 0x300))
		start = (start + 0x3ff) & ~0x3ff;

	start = (start + align - 1) & ~(align - 1);

	return start;
}


static int __pci_assign_resource(struct pci_bus *bus, struct pci_dev *dev,
		int resno, resource_size_t size, resource_size_t align)
{
	struct resource *res = dev->resource + resno;
	resource_size_t min;
	int ret;

	min = (res->flags & IORESOURCE_IO) ? PCIBIOS_MIN_IO : PCIBIOS_MIN_MEM;

	/*
	 * First, try exact prefetching match.  Even if a 64-bit
	 * prefetchable bridge window is below 4GB, we can't put a 32-bit
	 * prefetchable resource in it because pbus_size_mem() assumes a
	 * 64-bit window will contain no 32-bit resources.  If we assign
	 * things differently than they were sized, not everything will fit.
	 */
	ret = pci_bus_alloc_resource(bus, res, size, align, min,
				     IORESOURCE_PREFETCH | IORESOURCE_MEM_64,
				     pcibios_align_resource, dev);
	if (ret == 0)
		return 0;

	/*
	 * If the prefetchable window is only 32 bits wide, we can put
	 * 64-bit prefetchable resources in it.
	 */
	if ((res->flags & (IORESOURCE_PREFETCH | IORESOURCE_MEM_64)) ==
	     (IORESOURCE_PREFETCH | IORESOURCE_MEM_64)) {
		ret = pci_bus_alloc_resource(bus, res, size, align, min,
					     IORESOURCE_PREFETCH,
					     pcibios_align_resource, dev);
		if (ret == 0)
			return 0;
	}

	/*
	 * If we didn't find a better match, we can put any memory resource
	 * in a non-prefetchable window.  If this resource is 32 bits and
	 * non-prefetchable, the first call already tried the only possibility
	 * so we don't need to try again.
	 */
	if (res->flags & (IORESOURCE_PREFETCH | IORESOURCE_MEM_64))
		ret = pci_bus_alloc_resource(bus, res, size, align, min, 0,
					     pcibios_align_resource, dev);

	return ret;
}

static int _pci_assign_resource(struct pci_dev *dev, int resno,
				resource_size_t size, resource_size_t min_align)
{
	struct pci_bus *bus;
	int ret;

	bus = dev->bus;
	while ((ret = __pci_assign_resource(bus, dev, resno, size, min_align))) {
		if (!bus->parent || !bus->self->transparent)
			break;
		bus = bus->parent;
	}

	return ret;
}

static struct resource * __request_resource(struct resource *root, struct resource *new)
{
	resource_size_t start = new->start;
	resource_size_t end = new->end;
	struct resource *tmp, **p;

	if (end < start)
		return root;
	if (start < root->start)
		return root;
	if (end > root->end)
		return root;
	p = &root->child;
	for (;;) {
		tmp = *p;
		if (!tmp || tmp->start > end) {
			new->sibling = tmp;
			*p = new;
			new->parent = root;
			return NULL;
		}
		p = &tmp->sibling;
		if (tmp->end < start)
			continue;
		return tmp;
	}
}

struct resource *request_resource_conflict(struct resource *root, struct resource *new)
{
	struct resource *conflict;

	write_lock(&resource_lock);
	conflict = __request_resource(root, new);
	write_unlock(&resource_lock);
	return conflict;
}

resource_size_t __weak pcibios_retrieve_fw_addr(struct pci_dev *dev, int idx)
{
	return 0;
}

static int pci_revert_fw_address(struct resource *res, struct pci_dev *dev,
		int resno, resource_size_t size)
{
	struct resource *root, *conflict;
	resource_size_t fw_addr, start, end;

	fw_addr = pcibios_retrieve_fw_addr(dev, resno);
	if (!fw_addr)
		return -ENOMEM;

	start = res->start;
	end = res->end;
	res->start = fw_addr;
	res->end = res->start + size - 1;
	res->flags &= ~IORESOURCE_UNSET;

	root = pci_find_parent_resource(dev, res);
	if (!root) {
		if (res->flags & IORESOURCE_IO)
			root = &ioport_resource;
		else
			root = &iomem_resource;
	}

	dev_info(&dev->dev, "BAR %d: trying firmware assignment %pR\n",
		 resno, res);
	conflict = request_resource_conflict(root, res);
	if (conflict) {
		dev_info(&dev->dev, "BAR %d: %pR conflicts with %s %pR\n",
			 resno, res, conflict->name, conflict);
		res->start = start;
		res->end = end;
		res->flags |= IORESOURCE_UNSET;
		return -EBUSY;
	}
	return 0;
}

static void pci_std_update_resource(struct pci_dev *dev, int resno)
{
	struct pci_bus_region region;
	bool disable;
	u16 cmd;
	u32 new, check, mask;
	int reg;
	struct resource *res = dev->resource + resno;

	/* Per SR-IOV spec 3.4.1.11, VF BARs are RO zero */
	if (dev->is_virtfn)
		return;

	/*
	 * Ignore resources for unimplemented BARs and unused resource slots
	 * for 64 bit BARs.
	 */
	if (!res->flags)
		return;

	if (res->flags & IORESOURCE_UNSET)
		return;

	/*
	 * Ignore non-moveable resources.  This might be legacy resources for
	 * which no functional BAR register exists or another important
	 * system resource we shouldn't move around.
	 */
	if (res->flags & IORESOURCE_PCI_FIXED)
		return;

	pcibios_resource_to_bus(dev->bus, &region, res);
	new = region.start;

	if (res->flags & IORESOURCE_IO) {
		mask = (u32)PCI_BASE_ADDRESS_IO_MASK;
		new |= res->flags & ~PCI_BASE_ADDRESS_IO_MASK;
	} else if (resno == PCI_ROM_RESOURCE) {
		mask = PCI_ROM_ADDRESS_MASK;
	} else {
		mask = (u32)PCI_BASE_ADDRESS_MEM_MASK;
		new |= res->flags & ~PCI_BASE_ADDRESS_MEM_MASK;
	}

	if (resno < PCI_ROM_RESOURCE) {
		reg = PCI_BASE_ADDRESS_0 + 4 * resno;
	} else if (resno == PCI_ROM_RESOURCE) {

		/*
		 * Apparently some Matrox devices have ROM BARs that read
		 * as zero when disabled, so don't update ROM BARs unless
		 * they're enabled.  See https://lkml.org/lkml/2005/8/30/138.
		 */
		if (!(res->flags & IORESOURCE_ROM_ENABLE))
			return;

		reg = dev->rom_base_reg;
		new |= PCI_ROM_ADDRESS_ENABLE;
	} else
		return;

	/*
	 * We can't update a 64-bit BAR atomically, so when possible,
	 * disable decoding so that a half-updated BAR won't conflict
	 * with another device.
	 */
	disable = (res->flags & IORESOURCE_MEM_64) && !dev->mmio_always_on;
	if (disable) {
		pci_read_config_word(dev, PCI_COMMAND, &cmd);
		pci_write_config_word(dev, PCI_COMMAND,
				      cmd & ~PCI_COMMAND_MEMORY);
	}

	pci_write_config_dword(dev, reg, new);
	pci_read_config_dword(dev, reg, &check);

	if ((new ^ check) & mask) {
		dev_err(&dev->dev, "BAR %d: error updating (%#08x != %#08x)\n",
			resno, new, check);
	}

	if (res->flags & IORESOURCE_MEM_64) {
		new = region.start >> 16 >> 16;
		pci_write_config_dword(dev, reg + 4, new);
		pci_read_config_dword(dev, reg + 4, &check);
		if (check != new) {
			dev_err(&dev->dev, "BAR %d: error updating (high %#08x != %#08x)\n",
				resno, new, check);
		}
	}

	if (disable)
		pci_write_config_word(dev, PCI_COMMAND, cmd);
}

#if 0//TODO
void pci_iov_update_resource(struct pci_dev *dev, int resno)
{
	struct pci_sriov *iov = dev->is_physfn ? dev->sriov : NULL;
	struct resource *res = dev->resource + resno;
	int vf_bar = resno - PCI_IOV_RESOURCES;
	struct pci_bus_region region;
	u16 cmd;
	u32 new;
	int reg;

	/*
	 * The generic pci_restore_bars() path calls this for all devices,
	 * including VFs and non-SR-IOV devices.  If this is not a PF, we
	 * have nothing to do.
	 */
	if (!iov)
		return;

	pci_read_config_word(dev, iov->pos + PCI_SRIOV_CTRL, &cmd);
	if ((cmd & PCI_SRIOV_CTRL_VFE) && (cmd & PCI_SRIOV_CTRL_MSE)) {
		dev_WARN(&dev->dev, "can't update enabled VF BAR%d %pR\n",
			 vf_bar, res);
		return;
	}

	/*
	 * Ignore unimplemented BARs, unused resource slots for 64-bit
	 * BARs, and non-movable resources, e.g., those described via
	 * Enhanced Allocation.
	 */
	if (!res->flags)
		return;

	if (res->flags & IORESOURCE_UNSET)
		return;

	if (res->flags & IORESOURCE_PCI_FIXED)
		return;

	pcibios_resource_to_bus(dev->bus, &region, res);
	new = region.start;
	new |= res->flags & ~PCI_BASE_ADDRESS_MEM_MASK;

	reg = iov->pos + PCI_SRIOV_BAR + 4 * vf_bar;
	pci_write_config_dword(dev, reg, new);
	if (res->flags & IORESOURCE_MEM_64) {
		new = region.start >> 16 >> 16;
		pci_write_config_dword(dev, reg + 4, new);
	}
}
#endif

void pci_update_resource(struct pci_dev *dev, int resno)
{
	if (resno <= PCI_ROM_RESOURCE)
		pci_std_update_resource(dev, resno);
#if 0//TODO
#ifdef CONFIG_PCI_IOV
	else if (resno >= PCI_IOV_RESOURCES && resno <= PCI_IOV_RESOURCE_END)
		pci_iov_update_resource(dev, resno);
#endif
#endif
}

int pci_assign_resource(struct pci_dev *dev, int resno)
{
	struct resource *res = dev->resource + resno;
	resource_size_t align, size;
	int ret;

	if (res->flags & IORESOURCE_PCI_FIXED)
		return 0;

	res->flags |= IORESOURCE_UNSET;
	align = pci_resource_alignment(dev, res);
	if (!align) {
		dev_info(&dev->dev, "BAR %d: can't assign %pR (bogus alignment)\n",
			 resno, res);
		return -EINVAL;
	}

	size = resource_size(res);
	ret = _pci_assign_resource(dev, resno, size, align);

	/*
	 * If we failed to assign anything, let's try the address
	 * where firmware left it.  That at least has a chance of
	 * working, which is better than just leaving it disabled.
	 */
	if (ret < 0) {
		dev_info(&dev->dev, "BAR %d: no space for %pR\n", resno, res);
		ret = pci_revert_fw_address(res, dev, resno, size);
	}

	if (ret < 0) {
		dev_info(&dev->dev, "BAR %d: failed to assign %pR\n", resno,
			 res);
		return ret;
	}

	res->flags &= ~IORESOURCE_UNSET;
	res->flags &= ~IORESOURCE_STARTALIGN;
	dev_info(&dev->dev, "BAR %d: assigned %pR\n", resno, res);
	if (resno < PCI_BRIDGE_RESOURCES)
		pci_update_resource(dev, resno);

	return 0;
}

int pci_reassign_resource(struct pci_dev *dev, int resno, resource_size_t addsize,
			resource_size_t min_align)
{
	struct resource *res = dev->resource + resno;
	unsigned long flags;
	resource_size_t new_size;
	int ret;

	if (res->flags & IORESOURCE_PCI_FIXED)
		return 0;

	flags = res->flags;
	res->flags |= IORESOURCE_UNSET;
	if (!res->parent) {
		dev_info(&dev->dev, "BAR %d: can't reassign an unassigned resource %pR\n",
			 resno, res);
		return -EINVAL;
	}

	/* already aligned with min_align */
	new_size = resource_size(res) + addsize;
	ret = _pci_assign_resource(dev, resno, new_size, min_align);
	if (ret) {
		res->flags = flags;
		dev_info(&dev->dev, "BAR %d: %pR (failed to expand by %#llx)\n",
			 resno, res, (unsigned long long) addsize);
		return ret;
	}

	res->flags &= ~IORESOURCE_UNSET;
	res->flags &= ~IORESOURCE_STARTALIGN;
	dev_info(&dev->dev, "BAR %d: reassigned %pR (expanded by %#llx)\n",
		 resno, res, (unsigned long long) addsize);
	if (resno < PCI_BRIDGE_RESOURCES)
		pci_update_resource(dev, resno);

	return 0;
}

static void reassign_resources_sorted(struct list_head *realloc_head,
		struct list_head *head)
{
	struct resource *res;
	struct xdx_pci_dev_resource *add_res, *tmp;
	struct xdx_pci_dev_resource *dev_res;
	resource_size_t add_size, align;
	int idx;

	list_for_each_entry_safe(add_res, tmp, realloc_head, list) {
		bool found_match = false;

		res = add_res->res;
		/* skip resource that has been reset */
		if (!res->flags)
			goto out;

		/* skip this resource if not found in head list */
		list_for_each_entry(dev_res, head, list) {
			if (dev_res->res == res) {
				found_match = true;
				break;
			}
		}
		if (!found_match)/* just skip */
			continue;

		idx = res - &add_res->dev->resource[0];
		add_size = add_res->add_size;
		align = add_res->min_align;
		if (!resource_size(res)) {
			res->start = align;
			res->end = res->start + add_size - 1;
			if (pci_assign_resource(add_res->dev, idx))
				reset_resource(res);
		} else {
			res->flags |= add_res->flags &
				 (IORESOURCE_STARTALIGN|IORESOURCE_SIZEALIGN);
			if (pci_reassign_resource(add_res->dev, idx,
						  add_size, align))
				dev_printk(KERN_DEBUG, &add_res->dev->dev,
					   "failed to add %llx res[%d]=%pR\n",
					   (unsigned long long)add_size,
					   idx, res);
		}
out:
		list_del(&add_res->list);
		kfree(add_res);
	}
}

static void __assign_resources_sorted(struct list_head *head,
				 struct list_head *realloc_head,
				 struct list_head *fail_head)
{
	/*
	 * Should not assign requested resources at first.
	 *   they could be adjacent, so later reassign can not reallocate
	 *   them one by one in parent resource window.
	 * Try to assign requested + add_size at beginning
	 *  if could do that, could get out early.
	 *  if could not do that, we still try to assign requested at first,
	 *    then try to reassign add_size for some resources.
	 *
	 * Separate three resource type checking if we need to release
	 * assigned resource after requested + add_size try.
	 *	1. if there is io port assign fail, will release assigned
	 *	   io port.
	 *	2. if there is pref mmio assign fail, release assigned
	 *	   pref mmio.
	 *	   if assigned pref mmio's parent is non-pref mmio and there
	 *	   is non-pref mmio assign fail, will release that assigned
	 *	   pref mmio.
	 *	3. if there is non-pref mmio assign fail or pref mmio
	 *	   assigned fail, will release assigned non-pref mmio.
	 */
	LIST_HEAD(save_head);
	LIST_HEAD(local_fail_head);
	struct xdx_pci_dev_resource *save_res;
	struct xdx_pci_dev_resource *dev_res, *tmp_res, *dev_res2;
	unsigned long fail_type;
	resource_size_t add_align, align;

	/* Check if optional add_size is there */
	if (!realloc_head || list_empty(realloc_head))
		goto requested_and_reassign;

	/* Save original start, end, flags etc at first */
	list_for_each_entry(dev_res, head, list) {
		if (add_to_list(&save_head, dev_res->dev, dev_res->res, 0, 0)) {
			free_list(&save_head);
			goto requested_and_reassign;
		}
	}

	/* Update res in head list with add_size in realloc_head list */
	list_for_each_entry_safe(dev_res, tmp_res, head, list) {
		dev_res->res->end += get_res_add_size(realloc_head,
							dev_res->res);

		/*
		 * There are two kinds of additional resources in the list:
		 * 1. bridge resource  -- IORESOURCE_STARTALIGN
		 * 2. SR-IOV resource   -- IORESOURCE_SIZEALIGN
		 * Here just fix the additional alignment for bridge
		 */
		if (!(dev_res->res->flags & IORESOURCE_STARTALIGN))
			continue;

		add_align = get_res_add_align(realloc_head, dev_res->res);

		/*
		 * The "head" list is sorted by the alignment to make sure
		 * resources with bigger alignment will be assigned first.
		 * After we change the alignment of a dev_res in "head" list,
		 * we need to reorder the list by alignment to make it
		 * consistent.
		 */
		if (add_align > dev_res->res->start) {
			resource_size_t r_size = resource_size(dev_res->res);

			dev_res->res->start = add_align;
			dev_res->res->end = add_align + r_size - 1;

			list_for_each_entry(dev_res2, head, list) {
				align = pci_resource_alignment(dev_res2->dev,
							       dev_res2->res);
				if (add_align > align) {
					list_move_tail(&dev_res->list,
						       &dev_res2->list);
					break;
				}
			}
		}

	}

	/* Try updated head list with add_size added */
	assign_requested_resources_sorted(head, &local_fail_head);

	/* all assigned with add_size ? */
	if (list_empty(&local_fail_head)) {
		/* Remove head list from realloc_head list */
		list_for_each_entry(dev_res, head, list)
			remove_from_list(realloc_head, dev_res->res);
		free_list(&save_head);
		free_list(head);
		return;
	}

	/* check failed type */
	fail_type = pci_fail_res_type_mask(&local_fail_head);
	/* remove not need to be released assigned res from head list etc */
	list_for_each_entry_safe(dev_res, tmp_res, head, list)
		if (dev_res->res->parent &&
		    !pci_need_to_release(fail_type, dev_res->res)) {
			/* remove it from realloc_head list */
			remove_from_list(realloc_head, dev_res->res);
			remove_from_list(&save_head, dev_res->res);
			list_del(&dev_res->list);
			kfree(dev_res);
		}

	free_list(&local_fail_head);
	/* Release assigned resource */
	list_for_each_entry(dev_res, head, list)
		if (dev_res->res->parent)
			release_resource(dev_res->res);
	/* Restore start/end/flags from saved list */
	list_for_each_entry(save_res, &save_head, list) {
		struct resource *res = save_res->res;

		res->start = save_res->start;
		res->end = save_res->end;
		res->flags = save_res->flags;
	}
	free_list(&save_head);

requested_and_reassign:
	/* Satisfy the must-have resource requests */
	assign_requested_resources_sorted(head, fail_head);

	/* Try to satisfy any additional optional resource
		requests */
	if (realloc_head)
		reassign_resources_sorted(realloc_head, head);
	free_list(head);
}

static void pdev_assign_resources_sorted(struct pci_dev *dev,
				 struct list_head *add_head,
				 struct list_head *fail_head)
{
	LIST_HEAD(head);

	__dev_sort_resources(dev, &head);
	__assign_resources_sorted(&head, add_head, fail_head);

}

static void pbus_assign_resources_sorted(const struct pci_bus *bus,
					 struct list_head *realloc_head,
					 struct list_head *fail_head)
{
	struct pci_dev *dev;
	LIST_HEAD(head);

	list_for_each_entry(dev, &bus->devices, bus_list)
		__dev_sort_resources(dev, &head);

	__assign_resources_sorted(&head, realloc_head, fail_head);
}

static void assign_fixed_resource_on_bus(struct pci_bus *b, struct resource *r)
{
	int i;
	struct resource *parent_r;
	unsigned long mask = IORESOURCE_IO | IORESOURCE_MEM |
			     IORESOURCE_PREFETCH;

	pci_bus_for_each_resource(b, parent_r, i) {
		if (!parent_r)
			continue;

		if ((r->flags & mask) == (parent_r->flags & mask) &&
		    resource_contains(parent_r, r))
			request_resource(parent_r, r);
	}
}

static void pdev_assign_fixed_resources(struct pci_dev *dev)
{
	int i;

	for (i = 0; i <  PCI_NUM_RESOURCES; i++) {
		struct pci_bus *b;
		struct resource *r = &dev->resource[i];

		if (r->parent || !(r->flags & IORESOURCE_PCI_FIXED) ||
		    !(r->flags & (IORESOURCE_IO | IORESOURCE_MEM)))
			continue;

		b = dev->bus;
		while (b && !r->parent) {
			assign_fixed_resource_on_bus(b, r);
			b = b->parent;
		}
	}
}

static struct pci_bus *find_pci_root_bus(struct pci_bus *bus)
{
	while (bus->parent)
		bus = bus->parent;

	return bus;
}

struct pci_host_bridge *pci_find_host_bridge(struct pci_bus *bus)
{
	struct pci_bus *root_bus = find_pci_root_bus(bus);

	return to_pci_host_bridge(root_bus->bridge);
}

void pcibios_resource_to_bus(struct pci_bus *bus, struct pci_bus_region *region,
			     struct resource *res)
{
	struct pci_host_bridge *bridge = pci_find_host_bridge(bus);
	struct resource_entry *window;
	resource_size_t offset = 0;

	resource_list_for_each_entry(window, &bridge->windows) {
		if (resource_contains(window->res, res)) {
			offset = window->offset;
			break;
		}
	}

	region->start = res->start - offset;
	region->end = res->end - offset;
}

static void pci_setup_bridge_mmio(struct pci_dev *bridge)
{
	struct resource *res;
	struct pci_bus_region region;
	u32 l;

	/* Set up the top and bottom of the PCI Memory segment for this bus. */
	res = &bridge->resource[PCI_BRIDGE_RESOURCES + 1];
	pcibios_resource_to_bus(bridge->bus, &region, res);
	if (res->flags & IORESOURCE_MEM) {
		l = (region.start >> 16) & 0xfff0;
		l |= region.end & 0xfff00000;
		dev_info(&bridge->dev, "  bridge window %pR\n", res);
	} else {
		l = 0x0000fff0;
	}
	pci_write_config_dword(bridge, PCI_MEMORY_BASE, l);
}

static void pci_setup_bridge_mmio_pref(struct pci_dev *bridge)
{
	struct resource *res;
	struct pci_bus_region region;
	u32 l, bu, lu;

	/* Clear out the upper 32 bits of PREF limit.
	   If PCI_PREF_BASE_UPPER32 was non-zero, this temporarily
	   disables PREF range, which is ok. */
	pci_write_config_dword(bridge, PCI_PREF_LIMIT_UPPER32, 0);

	/* Set up PREF base/limit. */
	bu = lu = 0;
	res = &bridge->resource[PCI_BRIDGE_RESOURCES + 2];
	pcibios_resource_to_bus(bridge->bus, &region, res);
	if (res->flags & IORESOURCE_PREFETCH) {
		l = (region.start >> 16) & 0xfff0;
		l |= region.end & 0xfff00000;
		if (res->flags & IORESOURCE_MEM_64) {
			bu = upper_32_bits(region.start);
			lu = upper_32_bits(region.end);
		}
		dev_info(&bridge->dev, "  bridge window %pR\n", res);
	} else {
		l = 0x0000fff0;
	}
	pci_write_config_dword(bridge, PCI_PREF_MEMORY_BASE, l);

	/* Set the upper 32 bits of PREF base & limit. */
	pci_write_config_dword(bridge, PCI_PREF_BASE_UPPER32, bu);
	pci_write_config_dword(bridge, PCI_PREF_LIMIT_UPPER32, lu);
}

static void pci_setup_bridge_io(struct pci_dev *bridge)
{
	struct resource *res;
	struct pci_bus_region region;
	unsigned long io_mask;
	u8 io_base_lo, io_limit_lo;
	u16 l;
	u32 io_upper16;

	io_mask = PCI_IO_RANGE_MASK;
	if (bridge->io_window_1k)
		io_mask = PCI_IO_1K_RANGE_MASK;

	/* Set up the top and bottom of the PCI I/O segment for this bus. */
	res = &bridge->resource[PCI_BRIDGE_RESOURCES + 0];
	pcibios_resource_to_bus(bridge->bus, &region, res);
	if (res->flags & IORESOURCE_IO) {
		pci_read_config_word(bridge, PCI_IO_BASE, &l);
		io_base_lo = (region.start >> 8) & io_mask;
		io_limit_lo = (region.end >> 8) & io_mask;
		l = ((u16) io_limit_lo << 8) | io_base_lo;
		/* Set up upper 16 bits of I/O base/limit. */
		io_upper16 = (region.end & 0xffff0000) | (region.start >> 16);
		dev_info(&bridge->dev, "  bridge window %pR\n", res);
	} else {
		/* Clear upper 16 bits of I/O base/limit. */
		io_upper16 = 0;
		l = 0x00f0;
	}
	/* Temporarily disable the I/O range before updating PCI_IO_BASE. */
	pci_write_config_dword(bridge, PCI_IO_BASE_UPPER16, 0x0000ffff);
	/* Update lower 16 bits of I/O base/limit. */
	pci_write_config_word(bridge, PCI_IO_BASE, l);
	/* Update upper 16 bits of I/O base/limit. */
	pci_write_config_dword(bridge, PCI_IO_BASE_UPPER16, io_upper16);
}

static void __pci_setup_bridge(struct pci_bus *bus, unsigned long type)
{
	struct pci_dev *bridge = bus->self;

	dev_info(&bridge->dev, "PCI bridge to %pR\n",
		 &bus->busn_res);

	if (type & IORESOURCE_IO)
		pci_setup_bridge_io(bridge);

	if (type & IORESOURCE_MEM)
		pci_setup_bridge_mmio(bridge);

	if (type & IORESOURCE_PREFETCH)
		pci_setup_bridge_mmio_pref(bridge);

	pci_write_config_word(bridge, PCI_BRIDGE_CONTROL, bus->bridge_ctl);
}

void __weak pcibios_setup_bridge(struct pci_bus *bus, unsigned long type)
{
}

void pci_setup_bridge(struct pci_bus *bus)
{
	unsigned long type = IORESOURCE_IO | IORESOURCE_MEM |
				  IORESOURCE_PREFETCH;

	pcibios_setup_bridge(bus, type);
	__pci_setup_bridge(bus, type);
}


static void __pci_bus_assign_resources(const struct pci_bus *bus,
				struct list_head *realloc_head,
				struct list_head *fail_head)
{
	struct pci_bus *b;
	struct pci_dev *dev;

	pbus_assign_resources_sorted(bus, realloc_head, fail_head);

	list_for_each_entry(dev, &bus->devices, bus_list) {
		pdev_assign_fixed_resources(dev);

		b = dev->subordinate;
		if (!b)
			continue;

		__pci_bus_assign_resources(b, realloc_head, fail_head);

		switch (dev->class >> 8) {
		case PCI_CLASS_BRIDGE_PCI:
			if (!pci_is_enabled(dev))
				pci_setup_bridge(b);
			break;

		case PCI_CLASS_BRIDGE_CARDBUS:
			pci_setup_cardbus(b);
			break;

		default:
			dev_info(&dev->dev, "not setting up bridge for bus %04x:%02x\n",
				 pci_domain_nr(b), b->number);
			break;
		}
	}
}

static void __pci_bridge_assign_resources(const struct pci_dev *bridge,
					  struct list_head *add_head,
					  struct list_head *fail_head)
{
	struct pci_bus *b;

	pdev_assign_resources_sorted((struct pci_dev *)bridge,
					 add_head, fail_head);

	b = bridge->subordinate;
	if (!b)
		return;

	__pci_bus_assign_resources(b, add_head, fail_head);

	switch (bridge->class >> 8) {
	case PCI_CLASS_BRIDGE_PCI:
		pci_setup_bridge(b);
		break;

	case PCI_CLASS_BRIDGE_CARDBUS:
		pci_setup_cardbus(b);
		break;

	default:
		dev_info(&bridge->dev, "not setting up bridge for bus %04x:%02x\n",
			 pci_domain_nr(b), b->number);
		break;
	}
}

static void pci_bus_size_cardbus(struct pci_bus *bus,
			struct list_head *realloc_head)
{
	struct pci_dev *bridge = bus->self;
	struct resource *b_res = &bridge->resource[PCI_BRIDGE_RESOURCES];
	resource_size_t b_res_3_size = pci_cardbus_mem_size * 2;
	u16 ctrl;

	if (b_res[0].parent)
		goto handle_b_res_1;
	/*
	 * Reserve some resources for CardBus.  We reserve
	 * a fixed amount of bus space for CardBus bridges.
	 */
	b_res[0].start = pci_cardbus_io_size;
	b_res[0].end = b_res[0].start + pci_cardbus_io_size - 1;
	b_res[0].flags |= IORESOURCE_IO | IORESOURCE_STARTALIGN;
	if (realloc_head) {
		b_res[0].end -= pci_cardbus_io_size;
		add_to_list(realloc_head, bridge, b_res, pci_cardbus_io_size,
				pci_cardbus_io_size);
	}

handle_b_res_1:
	if (b_res[1].parent)
		goto handle_b_res_2;
	b_res[1].start = pci_cardbus_io_size;
	b_res[1].end = b_res[1].start + pci_cardbus_io_size - 1;
	b_res[1].flags |= IORESOURCE_IO | IORESOURCE_STARTALIGN;
	if (realloc_head) {
		b_res[1].end -= pci_cardbus_io_size;
		add_to_list(realloc_head, bridge, b_res+1, pci_cardbus_io_size,
				 pci_cardbus_io_size);
	}

handle_b_res_2:
	/* MEM1 must not be pref mmio */
	pci_read_config_word(bridge, PCI_CB_BRIDGE_CONTROL, &ctrl);
	if (ctrl & PCI_CB_BRIDGE_CTL_PREFETCH_MEM1) {
		ctrl &= ~PCI_CB_BRIDGE_CTL_PREFETCH_MEM1;
		pci_write_config_word(bridge, PCI_CB_BRIDGE_CONTROL, ctrl);
		pci_read_config_word(bridge, PCI_CB_BRIDGE_CONTROL, &ctrl);
	}

	/*
	 * Check whether prefetchable memory is supported
	 * by this bridge.
	 */
	pci_read_config_word(bridge, PCI_CB_BRIDGE_CONTROL, &ctrl);
	if (!(ctrl & PCI_CB_BRIDGE_CTL_PREFETCH_MEM0)) {
		ctrl |= PCI_CB_BRIDGE_CTL_PREFETCH_MEM0;
		pci_write_config_word(bridge, PCI_CB_BRIDGE_CONTROL, ctrl);
		pci_read_config_word(bridge, PCI_CB_BRIDGE_CONTROL, &ctrl);
	}

	if (b_res[2].parent)
		goto handle_b_res_3;
	/*
	 * If we have prefetchable memory support, allocate
	 * two regions.  Otherwise, allocate one region of
	 * twice the size.
	 */
	if (ctrl & PCI_CB_BRIDGE_CTL_PREFETCH_MEM0) {
		b_res[2].start = pci_cardbus_mem_size;
		b_res[2].end = b_res[2].start + pci_cardbus_mem_size - 1;
		b_res[2].flags |= IORESOURCE_MEM | IORESOURCE_PREFETCH |
				  IORESOURCE_STARTALIGN;
		if (realloc_head) {
			b_res[2].end -= pci_cardbus_mem_size;
			add_to_list(realloc_head, bridge, b_res+2,
				 pci_cardbus_mem_size, pci_cardbus_mem_size);
		}

		/* reduce that to half */
		b_res_3_size = pci_cardbus_mem_size;
	}

handle_b_res_3:
	if (b_res[3].parent)
		goto handle_done;
	b_res[3].start = pci_cardbus_mem_size;
	b_res[3].end = b_res[3].start + b_res_3_size - 1;
	b_res[3].flags |= IORESOURCE_MEM | IORESOURCE_STARTALIGN;
	if (realloc_head) {
		b_res[3].end -= b_res_3_size;
		add_to_list(realloc_head, bridge, b_res+3, b_res_3_size,
				 pci_cardbus_mem_size);
	}

handle_done:
	;
}

static void pci_bridge_check_ranges(struct pci_bus *bus)
{
	u16 io;
	u32 pmem;
	struct pci_dev *bridge = bus->self;
	struct resource *b_res;

	b_res = &bridge->resource[PCI_BRIDGE_RESOURCES];
	b_res[1].flags |= IORESOURCE_MEM;

	pci_read_config_word(bridge, PCI_IO_BASE, &io);
	if (!io) {
		pci_write_config_word(bridge, PCI_IO_BASE, 0xe0f0);
		pci_read_config_word(bridge, PCI_IO_BASE, &io);
		pci_write_config_word(bridge, PCI_IO_BASE, 0x0);
	}
	if (io)
		b_res[0].flags |= IORESOURCE_IO;

	/*  DECchip 21050 pass 2 errata: the bridge may miss an address
	    disconnect boundary by one PCI data phase.
	    Workaround: do not use prefetching on this device. */
	if (bridge->vendor == PCI_VENDOR_ID_DEC && bridge->device == 0x0001)
		return;

	pci_read_config_dword(bridge, PCI_PREF_MEMORY_BASE, &pmem);
	if (!pmem) {
		pci_write_config_dword(bridge, PCI_PREF_MEMORY_BASE,
					       0xffe0fff0);
		pci_read_config_dword(bridge, PCI_PREF_MEMORY_BASE, &pmem);
		pci_write_config_dword(bridge, PCI_PREF_MEMORY_BASE, 0x0);
	}
	if (pmem) {
		b_res[2].flags |= IORESOURCE_MEM | IORESOURCE_PREFETCH;
		if ((pmem & PCI_PREF_RANGE_TYPE_MASK) ==
		    PCI_PREF_RANGE_TYPE_64) {
			b_res[2].flags |= IORESOURCE_MEM_64;
			b_res[2].flags |= PCI_PREF_RANGE_TYPE_64;
		}
	}

	/* double check if bridge does support 64 bit pref */
	if (b_res[2].flags & IORESOURCE_MEM_64) {
		u32 mem_base_hi, tmp;
		pci_read_config_dword(bridge, PCI_PREF_BASE_UPPER32,
					 &mem_base_hi);
		pci_write_config_dword(bridge, PCI_PREF_BASE_UPPER32,
					       0xffffffff);
		pci_read_config_dword(bridge, PCI_PREF_BASE_UPPER32, &tmp);
		if (!tmp)
			b_res[2].flags &= ~IORESOURCE_MEM_64;
		pci_write_config_dword(bridge, PCI_PREF_BASE_UPPER32,
				       mem_base_hi);
	}
}

static void __pci_bus_size_bridges(struct pci_bus *bus, struct list_head *realloc_head)
{
	struct pci_dev *dev;
	unsigned long mask, prefmask, type2 = 0, type3 = 0;
	resource_size_t additional_mem_size = 0, additional_io_size = 0;
	struct resource *b_res;
	int ret;

	list_for_each_entry(dev, &bus->devices, bus_list) {
		struct pci_bus *b = dev->subordinate;
		if (!b)
			continue;

		switch (dev->class >> 8) {
		case PCI_CLASS_BRIDGE_CARDBUS:
			pci_bus_size_cardbus(b, realloc_head);
			break;

		case PCI_CLASS_BRIDGE_PCI:
		default:
			__pci_bus_size_bridges(b, realloc_head);
			break;
		}
	}

	/* The root bus? */
	if (pci_is_root_bus(bus))
		return;

	switch (bus->self->class >> 8) {
	case PCI_CLASS_BRIDGE_CARDBUS:
		/* don't size cardbuses yet. */
		break;

	case PCI_CLASS_BRIDGE_PCI:
		pci_bridge_check_ranges(bus);
		if (bus->self->is_hotplug_bridge) {
			additional_io_size  = pci_hotplug_io_size;
			additional_mem_size = pci_hotplug_mem_size;
		}
		/* Fall through */
	default:
		pbus_size_io(bus, realloc_head ? 0 : additional_io_size,
			     additional_io_size, realloc_head);

		/*
		 * If there's a 64-bit prefetchable MMIO window, compute
		 * the size required to put all 64-bit prefetchable
		 * resources in it.
		 */
		b_res = &bus->self->resource[PCI_BRIDGE_RESOURCES];
		mask = IORESOURCE_MEM;
		prefmask = IORESOURCE_MEM | IORESOURCE_PREFETCH;
		if (b_res[2].flags & IORESOURCE_MEM_64) {
			prefmask |= IORESOURCE_MEM_64;
			ret = pbus_size_mem(bus, prefmask, prefmask,
				  prefmask, prefmask,
				  realloc_head ? 0 : additional_mem_size,
				  additional_mem_size, realloc_head);

			/*
			 * If successful, all non-prefetchable resources
			 * and any 32-bit prefetchable resources will go in
			 * the non-prefetchable window.
			 */
			if (ret == 0) {
				mask = prefmask;
				type2 = prefmask & ~IORESOURCE_MEM_64;
				type3 = prefmask & ~IORESOURCE_PREFETCH;
			}
		}

		/*
		 * If there is no 64-bit prefetchable window, compute the
		 * size required to put all prefetchable resources in the
		 * 32-bit prefetchable window (if there is one).
		 */
		if (!type2) {
			prefmask &= ~IORESOURCE_MEM_64;
			ret = pbus_size_mem(bus, prefmask, prefmask,
					 prefmask, prefmask,
					 realloc_head ? 0 : additional_mem_size,
					 additional_mem_size, realloc_head);

			/*
			 * If successful, only non-prefetchable resources
			 * will go in the non-prefetchable window.
			 */
			if (ret == 0)
				mask = prefmask;
			else
				additional_mem_size += additional_mem_size;

			type2 = type3 = IORESOURCE_MEM;
		}

		/*
		 * Compute the size required to put everything else in the
		 * non-prefetchable window.  This includes:
		 *
		 *   - all non-prefetchable resources
		 *   - 32-bit prefetchable resources if there's a 64-bit
		 *     prefetchable window or no prefetchable window at all
		 *   - 64-bit prefetchable resources if there's no
		 *     prefetchable window at all
		 *
		 * Note that the strategy in __pci_assign_resource() must
		 * match that used here.  Specifically, we cannot put a
		 * 32-bit prefetchable resource in a 64-bit prefetchable
		 * window.
		 */
		pbus_size_mem(bus, mask, IORESOURCE_MEM, type2, type3,
				realloc_head ? 0 : additional_mem_size,
				additional_mem_size, realloc_head);
		break;
	}
}
 
static int xdx_pci_reassign_bridge_resources(struct pci_dev *bridge, unsigned long type)
{
  	struct xdx_pci_dev_resource *dev_res;
  	struct pci_dev *next;
  	LIST_HEAD(saved);
  	LIST_HEAD(added);
  	LIST_HEAD(failed);
  	unsigned int i;
  	int ret;
  
  	/* Walk to the root hub, releasing bridge BARs when possible */
  	next = bridge;
  	do {
  		bridge = next;
  		for (i = PCI_BRIDGE_RESOURCES; i < PCI_BRIDGE_RESOURCE_END;
  		     i++) {
  			struct resource *res = &bridge->resource[i];
  
  			if ((res->flags ^ type) & XDX_PCI_RES_TYPE_MASK)
  				continue;
  
  			/* Ignore BARs which are still in use */
  			if (res->child)
  				continue;
  
  			ret = add_to_list(&saved, bridge, res, 0, 0);
  			if (ret)
  				goto cleanup;
  
  			dev_info(&bridge->dev, "BAR %d: releasing %pR\n",
  				 i, res);
  
  			if (res->parent)
  				release_resource(res);
  			res->start = 0;
  			res->end = 0;
  			break;
  		}
  		if (i == PCI_BRIDGE_RESOURCE_END)
  			break;
  
  		next = bridge->bus ? bridge->bus->self : NULL;
  	} while (next);
  
  	if (list_empty(&saved))
  		return -ENOENT;
  
  	__pci_bus_size_bridges(bridge->subordinate, &added);
  	__pci_bridge_assign_resources(bridge, &added, &failed);
  	BUG_ON(!list_empty(&added));
  
  	if (!list_empty(&failed)) {
  		ret = -ENOSPC;
  		goto cleanup;
  	}
  
  	list_for_each_entry(dev_res, &saved, list) {
  		/* Skip the bridge we just assigned resources for. */
  		if (bridge == dev_res->dev)
  			continue;
  
  		bridge = dev_res->dev;
  		pci_setup_bridge(bridge->subordinate);
  	}
  
  	free_list(&saved);
  	return 0;
cleanup:
  	/* restore size and flags */
  	list_for_each_entry(dev_res, &failed, list) {
  		struct resource *res = dev_res->res;
  
  		res->start = dev_res->start;
  		res->end = dev_res->end;
  		res->flags = dev_res->flags;
  	}
  	free_list(&failed);
  
  	/* Revert to the old configuration */
  	list_for_each_entry(dev_res, &saved, list) {
  		struct resource *res = dev_res->res;
  
  		bridge = dev_res->dev;
  		i = res - bridge->resource;
  
  		res->start = dev_res->start;
  		res->end = dev_res->end;
  		res->flags = dev_res->flags;
  
  		pci_claim_resource(bridge, i);
  		pci_setup_bridge(bridge->subordinate);
  	}
  	free_list(&saved);
  	return ret;
}

static int xdx_pci_resize_resource(struct pci_dev *dev, int resno, int size)
{
        struct resource *res = dev->resource + resno;
        int old, ret;
        u32 sizes;
        u16 cmd;

        /* Make sure the resource isn't assigned before resizing it. */
        if (!(res->flags & IORESOURCE_UNSET))
                return -EBUSY;

        pci_read_config_word(dev, PCI_COMMAND, &cmd);
        if (cmd & PCI_COMMAND_MEMORY)
                return -EBUSY;

        sizes = xdx_pci_rebar_get_possible_sizes(dev, resno);
        if (!sizes)
                return -ENOTSUPP;

        //if (!(sizes & BIT(size)))
        //        return -EINVAL;

        old = xdx_pci_rebar_get_current_size(dev, resno);
        if (old < 0)
                return old;

        ret = xdx_pci_rebar_set_size(dev, resno, size);
        if (ret)
                return ret;

        res->end = res->start + xdx_pci_rebar_size_to_bytes(size) - 1;

        /* Check if the new config works by trying to assign everything. */
        ret = xdx_pci_reassign_bridge_resources(dev->bus->self, res->flags);
        if (ret)
                goto error_resize;

        return 0;

error_resize:
        xdx_pci_rebar_set_size(dev, resno, old);
        res->end = res->start + xdx_pci_rebar_size_to_bytes(old) - 1;
        return ret;
}
#endif

int xdx_device_resize_fb_bar(struct pci_dev *dev, int resno, u64 rbar_size)
{
	struct pci_bus *root;
	struct resource *res;
	unsigned i;
	u16 cmd;
	int result;

	/* Check if the root BUS has 64bit memory resources */
	root = dev->bus;
	while (root->parent)
		root = root->parent;

	pci_bus_for_each_resource(root, res, i) {
		if (res && res->flags & (IORESOURCE_MEM | IORESOURCE_MEM_64) &&
		    res->start > 0x100000000ull)
			break;
	}

	/* Trying to resize is pointless without a root hub window above 4GB */
	if (!res)
		return 0;

	/* Disable memory decoding while we change the BAR addresses and size */
	pci_read_config_word(dev, PCI_COMMAND, &cmd);
	pci_write_config_word(dev, PCI_COMMAND, cmd & ~PCI_COMMAND_MEMORY);

	xdx_pci_release_resource(dev, resno);
	xdx_pci_release_resource(dev, resno + 2);

	res = dev->resource + resno;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	result = pci_resize_resource(dev, resno, rbar_size);
#else
	result = xdx_pci_resize_resource(dev, resno, rbar_size);
#endif	
	if (result == -ENOSPC)
		dev_info(&dev->dev,
			 "Not enough PCI address space for a large BAR.");
	else if (result == -ENOENT)
		dev_info(&dev->dev,
			 "A PCI device is busy and cannot reallocte resource,Manually modify the BIOS if necessary!!!!");
	else if (result && result != -ENOTSUPP)
		dev_info(&dev->dev, "Problem resizing BAR%d (%d).", resno,
			 result);

	pci_assign_unassigned_bus_resources(dev->bus);

	pci_write_config_word(dev, PCI_COMMAND, cmd);
	return 0;
}
