/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <linux/device.h>
#include <linux/stat.h>
#include <linux/sysfs.h>
#include <linux/pm_runtime.h>
#include <drm/xdxgpu_drm.h>

#include "xdx_sysfs.h"
#include "xdx_device.h"
#include "physheap.h"
#include "pvrsrv.h"
#include "pm/xdx_pm.h"
#include "pvrversion.h" 
#include "process_stats.h"
#include "rgxdevice.h"

static inline struct xdx_device *kdev_minor_to_xdev(struct device *kdev)
{
	struct drm_minor *minor = dev_get_drvdata(kdev);

	return drm_to_xdev(minor->dev);
}

static ssize_t xdx_dbdf_show(struct device *dev,
			     struct device_attribute *attr,
			     char *buf)
{
	uint32_t bus, did, fn, domain;
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	bus = xdev->pdev->bus->number;
	domain = pci_domain_nr(xdev->pdev->bus);
	did = PCI_SLOT(xdev->pdev->devfn);
	fn = PCI_FUNC(xdev->pdev->devfn);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%04x:%02x:%02x.%d\n", domain, bus, did, fn);
#else	
	return sprintf(buf, "%04x:%02x:%02x.%d\n", domain, bus, did, fn);
#endif
}

static ssize_t xdx_product_name_show(struct device *dev,
			     struct device_attribute *attr,
			     char *buf)
{
	char name[20];
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	switch (xdev->asic_type) {
	case CHIP_PANGU_A0:
	case CHIP_PANGU_B0:
		strcpy(name, "PANGU");
		break;
	default:
		strcpy(name, "Unknown_Xdxgpu");
		break;
	}
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%s\n", name);
#else	
	return sprintf(buf, "%s\n", name);
#endif
}

static ssize_t xdx_enable_runpm_show(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	char enable[2];
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	if (xdev->enable_runpm)
		strcpy(enable, "1");
	else
		strcpy(enable, "0");
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%s\n", enable);
#else
	return sprintf(buf, "%s\n", enable);
#endif
}

static ssize_t xdx_enable_runpm_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf,
				      size_t count)
{
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);
	struct drm_device *ddev = xdev_to_drm(xdev);
	bool enable;
	int err;

	err = kstrtobool(buf, &enable);
	if (err)
		return err;
	
	if (enable == xdev->enable_runpm)
		return count;

	/* dynamically turn on/off runtime pm feature */
	xdev->enable_runpm = enable;
	if (enable) {
		pm_runtime_use_autosuspend(ddev->dev);
		pm_runtime_set_autosuspend_delay(ddev->dev, 5000);
		pm_runtime_allow(ddev->dev);
		pm_runtime_mark_last_busy(ddev->dev);
		pm_runtime_put_autosuspend(ddev->dev);
	} else {
		pm_runtime_get_sync(ddev->dev);
		pm_runtime_forbid(ddev->dev);
		pm_runtime_dont_use_autosuspend(ddev->dev);
	}

	return count;
}

static ssize_t version_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	PVRSRV_DATA *pvrsrv_data = PVRSRVGetPVRSRVData();
	int i;
	int count  = 0;  

	uint32_t version[7] = {
		PVRVERSION_UNPACK_MAJ(pvrsrv_data->sDriverInfo.sUMBuildInfo.ui32BuildVersion),
		PVRVERSION_UNPACK_MIN(pvrsrv_data->sDriverInfo.sUMBuildInfo.ui32BuildVersion),
		PVRVERSION_UNPACK_MAJ(pvrsrv_data->sDriverInfo.sKMBuildInfo.ui32BuildVersion),
		PVRVERSION_UNPACK_MIN(pvrsrv_data->sDriverInfo.sKMBuildInfo.ui32BuildVersion),
		PVRVERSION_MAJ, 
		PVRVERSION_MIN, 
		PVRVERSION_BUILD
	};

	for (i = 0; i < 7; i++) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
		count += sysfs_emit_at(buf, count, "%d ", version[i]);
		if (i == 6)
			count += sysfs_emit_at(buf, count, "\n");
#else
		count += snprintf(buf + count, PAGE_SIZE - count, "%d ", version[i]);
		if (i == 6)
			count += snprintf(buf + count, PAGE_SIZE - count, "\n");
#endif
	}

	return count;
}

static ssize_t gpu_static_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{	
	struct xdx_fw_static_info *info;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	u32 gddr_type = get_gddr_type(xdev);
	info = &xdev->fw_st_info;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%d %d %d %d %llu %d \n%d %d %d %d %d \n%d %d %d %d %d %d %d %u\n",
			xdev->asic_type, xdev->hw_info.core_num, info->gddr_count,
			info->gddr_size, info->gddr_speed, info->gpu_primary_id, info->vendor, 
			info->chip_id, info->device_id, info->card_type, info->fw_ver,
			info->vbios_ver, info->fan_ctrl_mode, info->support_fan_info,
			info->support_temp_info, info->support_power_info, info->default_gpu_freq, info->gddr_vendor, gddr_type);

#else
	return sprintf(buf, "%d %d %d %d %llu %d \n%d %d %d %d %d \n%d %d %d %d %d %d %d %u\n",
			xdev->asic_type, xdev->hw_info.core_num, info->gddr_count,
			info->gddr_size, info->gddr_speed, info->gpu_primary_id, info->vendor, 
			info->chip_id, info->device_id, info->card_type, info->fw_ver,
			info->vbios_ver, info->fan_ctrl_mode, info->support_fan_info,
			info->support_temp_info, info->support_power_info, info->default_gpu_freq, info->gddr_vendor, gddr_type);
#endif
}

static ssize_t gpu_mem_util_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	uint32_t real_size;
	int i,j;
	IMG_UINT32 ui32LivePids;
	PVRSRV_PER_PROCESS_DATA *psPerProcessData;
	PVRSRV_PER_PROCESS_DATA *psPerProcessDatatemp;
	PVRSRV_ERROR err;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	int count  = 0;

	PVRSRV_DEVICE_NODE *psDeviceNode;
	PHYS_HEAP_MEM_STATS *mem_stats;
	PVRSRV_PHYS_HEAP ids[PVRSRV_PHYS_HEAP_LAST];
	int idx = 0;

	ids[idx++] = PVRSRV_PHYS_HEAP_GPU_LOCAL;
	ids[idx++] = PVRSRV_PHYS_HEAP_CPU_LOCAL;
#ifdef PVRSRV_PHYS_HEAP_MMU
	ids[idx++] = PVRSRV_PHYS_HEAP_MMU;
#endif
#ifdef PVRSRV_PHYS_HEAP_DC
	ids[idx++] = PVRSRV_PHYS_HEAP_DC;
#endif
	if(xdev->real_vram_size > xdev->aper_size)
		ids[idx++] = PVRSRV_PHYS_HEAP_GPU_PRIVATE;

	psDeviceNode = PVRSRVGetPVRSRVData()->psDeviceNodeList;
	mem_stats = kzalloc(sizeof(PHYS_HEAP_MEM_STATS)*idx, GFP_KERNEL);
	if (!mem_stats)
		return 0;
	PhysHeapGetMemInfo(psDeviceNode, idx, ids, mem_stats);

	/*
	 * 1. enable resizebar: GPU_LOCAL + DC + MMU is the total size, put DC in usedSize
	 * 2. disable resizebar: GPU_LOCAL + GPU_PRIVATE + MMU + DC is the total size, put DC in usedSize
	 * 3. put the reserved 64M in TotalSize and usedSize
	 */
	for(i = 2; i < idx; i++){
		if(ids[i] == PVRSRV_PHYS_HEAP_DC){
			mem_stats[0].ui64TotalSize += mem_stats[i].ui64TotalSize;
		} else {
			mem_stats[0].ui64TotalSize += mem_stats[i].ui64TotalSize;
			mem_stats[0].ui64FreeSize += mem_stats[i].ui64FreeSize;
		}
	}
	mem_stats[0].ui64TotalSize += xdev->hw_info.reserved_vram_size;

	for (i = 0; i < 2; i++) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
		count += sysfs_emit_at(buf, count, "%llu %llu ",
				mem_stats[i].ui64FreeSize,mem_stats[i].ui64TotalSize);
#else
		count += snprintf(buf + count , PAGE_SIZE - count , "%llu %llu ",
				mem_stats[i].ui64FreeSize,mem_stats[i].ui64TotalSize);
#endif
	}

	err = PVRSRVQueryProcessDATA(&ui32LivePids, &psPerProcessData);
	if (err) {
		dev_err(xdev->dev, "Fail to query process memory\n");
		return err;
	}
	real_size = ui32LivePids * sizeof(PVRSRV_PER_PROCESS_DATA);
	psPerProcessDatatemp = psPerProcessData;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	count += sysfs_emit_at(buf, count,
		"\n%u", ui32LivePids);
#else
	count += snprintf(buf + count , PAGE_SIZE - count , "\n%u", ui32LivePids);
#endif

	for (i = 0; i < ui32LivePids; i++) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
		count += sysfs_emit_at(buf, count, "\n%u ", psPerProcessDatatemp->ui32Pid);
#else
		count += snprintf(buf + count , PAGE_SIZE - count, "\n%u ", psPerProcessDatatemp->ui32Pid);
#endif

		for (j = 0; j < 22; j++) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
			count += sysfs_emit_at(buf, count , "%llu ",psPerProcessDatatemp->i64StatValue[j]);
#else
			count += snprintf(buf + count ,  PAGE_SIZE - count ,
				"%llu ",psPerProcessDatatemp->i64StatValue[j]);
#endif
		}
		psPerProcessDatatemp = psPerProcessDatatemp + 1;
		if (psPerProcessDatatemp->ui32Pid == 0)
			break;
	}

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	count += sysfs_emit_at(buf, count,"\n");
#else
	count  += snprintf(buf + count , PAGE_SIZE - count ,"\n");
#endif

	kfree(mem_stats);
	return count;
}

static ssize_t gpu_util_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	PVRSRV_DEVICE_DEBUG_INFO *psDebugInfo = &psDeviceNode->sDebugInfo;
	RGXFWIF_GPU_UTIL_STATS psGpu;
	int i;

	PVRSRV_RGXDEV_INFO *psDevInfo = psDeviceNode->pvDevice;
	PVRSRV_ERROR eError = PVRSRV_OK;

	int count  = 0; 

	eError = psDevInfo->pfnGetGpuUtilStats(psDeviceNode,
				psDebugInfo->hGpuUtilUserDebugFS,
				&psGpu);


#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	count += sysfs_emit_at(buf, count, "%15d %15llu %15llu %15llu %15llu\n",
				psGpu.bValid, psGpu.ui64GpuStatActive, psGpu.ui64GpuStatBlocked, 
				psGpu.ui64GpuStatIdle, psGpu.ui64GpuStatCumulative);

	for(i=0;i<9;i++)
	{
		count += sysfs_emit_at(buf, count, "%15llu %15llu %15llu %15llu\n",
				psGpu.aaui64DMOSStatActive[i][0], psGpu.aaui64DMOSStatBlocked[i][0],
				psGpu.aaui64DMOSStatIdle[i][0], psGpu.aaui64DMOSStatCumulative[i][0]);
	}
	count += sysfs_emit_at(buf, count, "  %llu \n", psGpu.ui64TimeStamp);
#else
	count += snprintf(buf + count,PAGE_SIZE - count ,"%15d %15llu %15llu %15llu %15llu\n",
				psGpu.bValid, psGpu.ui64GpuStatActive, psGpu.ui64GpuStatBlocked,
				psGpu.ui64GpuStatIdle, psGpu.ui64GpuStatCumulative);

	for(i=0;i<9;i++)
	{
		count += snprintf(buf + count,PAGE_SIZE - count ,"%15llu %15llu %15llu %15llu\n",
				psGpu.aaui64DMOSStatActive[i][0], psGpu.aaui64DMOSStatBlocked[i][0],
				psGpu.aaui64DMOSStatIdle[i][0], psGpu.aaui64DMOSStatCumulative[i][0]);
	}
	count += snprintf(buf + count,PAGE_SIZE - count ,"  %llu \n", psGpu.ui64TimeStamp);
#endif

	return count;
}

static ssize_t gddr_crc_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	GDDRStat_T gddrstate = get_gddr_crc(xdev);
	int i;
	int count = 0;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	/**
	 * old fw: FwInfo version is smaller than 5
	 * 1. display total crc in the first line
	*/
	if(xdev->fw_st_info.fw_info_ver < 5)
	{
		count += sysfs_emit_at(buf, count, "%d %d %d %d\n",
				gddrstate.crcWriteCount[0], gddrstate.crcReadCount[0],
				gddrstate.crcWriteRetryCount[0], gddrstate.crcReadRetryCount[0]);
		for(i = 1; i < GDDR_PHY_COUNT; i++)
		{
			count += sysfs_emit_at(buf, count, "%d %d %d %d\n",
				-1, -1, -1, -1);
		}
	} else {
		for(i = 0; i < GDDR_PHY_COUNT; i++)
		{
			count += sysfs_emit_at(buf, count, "%d %d %d %d\n",
					gddrstate.crcWriteCount[i], gddrstate.crcReadCount[i],
					gddrstate.crcWriteRetryCount[i], gddrstate.crcReadRetryCount[i]);
		}
	}
	
#else
	if(xdev->fw_st_info.fw_info_ver < 5)
	{
		count += snprintf(buf + count, PAGE_SIZE - count, "%d %d %d %d\n",
				gddrstate.crcWriteCount[0], gddrstate.crcReadCount[0],
				gddrstate.crcWriteRetryCount[0], gddrstate.crcReadRetryCount[0]);
		for(i = 1; i < GDDR_PHY_COUNT; i++)
		{
			count += snprintf(buf + count, PAGE_SIZE - count, "%d %d %d %d\n",
				-1, -1, -1, -1);
		}
	} else {
		for(i = 0; i < GDDR_PHY_COUNT; i++)
		{
			count += snprintf(buf + count, PAGE_SIZE - count, "%d %d %d %d\n",
					gddrstate.crcWriteCount[i], gddrstate.crcReadCount[i],
					gddrstate.crcWriteRetryCount[i], gddrstate.crcReadRetryCount[i]);
		}
	}
#endif

	return count;
}

#if defined(SUPPORT_LINUX_DVFS)
static ssize_t xdx_enable_dvfs_show(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;
	bool enabled;

	if (!psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg.bSupportDVFS) {
		return sprintf(buf, "NO DVFS Support\n");
	}

	enabled = xdx_dvfs_enabled(xdev->dev);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%d\n", enabled);
#else
	return sprintf(buf, "%d\n", enabled);
#endif
}

static ssize_t xdx_enable_dvfs_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf,
				      size_t count)
{
	int err;
	bool enabled;
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);
	PVRSRV_DEVICE_NODE *psDeviceNode = (PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev;

	if (!psDeviceNode->psDevConfig->sDVFS.sDVFSDeviceCfg.bSupportDVFS) {
		printk("NO DVFS Support!\n");
		return count;
	}

	err = kstrtobool(buf, &enabled);
	if (err)
		return err;

	if (enabled)
		xdx_dvfs_enable(xdev->dev);
	else
		xdx_dvfs_disable(xdev->dev);

	return count;
}
#endif /* SUPPORT_LINUX_DVFS */

static ssize_t xdx_gpu_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	PVRSRV_DEVICE_NODE *psDeviceNode;
	PHYS_HEAP_MEM_STATS *mem_stats;
	PVRSRV_PHYS_HEAP ids[1] = { PVRSRV_PHYS_HEAP_GPU_LOCAL };
	int count  = 0; 

	psDeviceNode = PVRSRVGetPVRSRVData()->psDeviceNodeList;
	mem_stats = kzalloc(sizeof(PHYS_HEAP_MEM_STATS), GFP_KERNEL);
	if (!mem_stats)
		return 0;
	PhysHeapGetMemInfo(psDeviceNode, 1, ids, mem_stats);

#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	count += sysfs_emit_at(buf, count, "VRAM total size:0x%llx\n",
			mem_stats[0].ui64TotalSize);
#else
	count  += snprintf(buf + count , PAGE_SIZE - count , "VRAM total size:0x%llx\n",
			mem_stats[0].ui64TotalSize);
#endif

	kfree(mem_stats);
	return count;
}

static DEVICE_ATTR_RO(xdx_dbdf);
static DEVICE_ATTR_RO(xdx_product_name);
static DEVICE_ATTR_RO(version);

static DEVICE_ATTR_RO(gpu_static_info);
static DEVICE_ATTR_RO(gpu_mem_util_info);
static DEVICE_ATTR_RO(gpu_util_info);
static DEVICE_ATTR_RO(gddr_crc);

static DEVICE_ATTR_RW(xdx_enable_runpm);
#if defined(SUPPORT_LINUX_DVFS)
static DEVICE_ATTR_RW(xdx_enable_dvfs);
#endif /* SUPPORT_LINUX_DVFS */

/* Due to a "-" sign in filename, generate the attr without macro. */
static struct device_attribute dev_attr_xdx_gpu_info = {
	.attr = { .name = "gpu-info", .mode = 0444 },
	.show = xdx_gpu_info_show,
};

static const struct attribute * const xdx_attrs[] = {
	&dev_attr_xdx_dbdf.attr,
	&dev_attr_xdx_product_name.attr,
	&dev_attr_version.attr,
	&dev_attr_xdx_enable_runpm.attr,
#if defined(SUPPORT_LINUX_DVFS)
	&dev_attr_xdx_enable_dvfs.attr,
#endif /* SUPPORT_LINUX_DVFS */
	NULL
};

static const struct attribute * const xdx_device_attrs[] = {
	&dev_attr_xdx_gpu_info.attr,
	&dev_attr_gpu_static_info.attr,
	&dev_attr_gpu_mem_util_info.attr,
	&dev_attr_gpu_util_info.attr,
	&dev_attr_gddr_crc.attr,
	NULL
};

void xdx_sysfs_init(struct xdx_device *xdev)
{
	struct device *kdev = xdev_to_drm(xdev)->primary->kdev;
	int r;

	/* sysfs under drm folder */
	r = sysfs_create_files(&kdev->kobj, (const struct attribute **)xdx_attrs);
	if (r)
		DRM_ERROR("xdxgpu sysfs init failed under drm folder\n");
	/* sysfs under device folder */
	r = sysfs_create_files(&xdev->pdev->dev.kobj, (const struct attribute **)xdx_device_attrs);
	if (r)
		DRM_ERROR("xdxgpu sysfs init failed under device folder\n");
}

void xdx_sysfs_fini(struct xdx_device *xdev)
{
	struct device *kdev = xdev_to_drm(xdev)->primary->kdev;

	sysfs_remove_files(&kdev->kobj, (const struct attribute **)xdx_attrs);
	sysfs_remove_files(&xdev->pdev->dev.kobj, (const struct attribute **)xdx_device_attrs);
}
