/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */


#include <linux/io.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/shmem_fs.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/timer.h>
#include <drm/drm_modeset_helper.h>
#ifdef __amd64__
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
#include <asm/set_memory.h>
#endif
#endif
/* Our header */
#include "hantro_priv.h"
#include "hx280enc.h"
#include "hantrodec.h"
#include "hantrocache.h"
#include "hantrodec400.h"
#include "hantro_axife.h"
#include "hantro_mmu.h"
#include <linux/of_reserved_mem.h>
#include <linux/of_irq.h>
#include "hantro_vcmd.h"

#ifdef VSI_FPGA
#include "hantro_fpga.h"
#endif
#include "pangu.h"
#include "xdx_device.h"
#include "xdx_drv.h"
#include "xdxvpu_irq.h"
#include "vpu_xmem.h"

struct hantro_device_handle hantro_dev;
int useirq = 1;
int dbg_mmap = 0;

module_param(useirq, int, 0444);
module_param(dbg_mmap, int, 0644);

int xdx_vpu_drm_open(struct xdx_device *xdev, struct xdx_drm_fpriv *fpriv)
{
	int ret = 0;
	struct vpu_fpriv *vpu_fpriv_data = NULL;

	if (!fpriv || fpriv->vpu_fpriv) {
		dev_err(xdev->dev, "vpu_fpriv is exist.\n");
		return -EEXIST;
	}

	vpu_fpriv_data = kzalloc(sizeof(struct vpu_fpriv), GFP_KERNEL);
	if (!vpu_fpriv_data)
		return -ENOMEM;

	INIT_LIST_HEAD(&vpu_fpriv_data->mm.bo_list);
	mutex_init(&vpu_fpriv_data->mm.mutex);

	fpriv->vpu_fpriv = vpu_fpriv_data;

	return ret;
}

void xdx_vpu_drm_release(struct xdx_device *xdev, struct xdx_drm_fpriv *fpriv)
{

	if (fpriv->vpu_fpriv)
		kfree(fpriv->vpu_fpriv);

	fpriv->vpu_fpriv = NULL;
	return;
}


//#if KERNEL_VERSION(4, 13, 0) > LINUX_VERSION_CODE
#if 0
void debug_dma_alloc_coherent(struct device *dev, size_t size,
			      dma_addr_t dma_addr, void *virt)
{
}

void debug_dma_free_coherent(struct device *dev, size_t size, void *virt,
			     dma_addr_t addr)
{
}
#endif

#if !defined(CONFIG_SW64)
#ifndef virt_to_bus
static inline unsigned long virt_to_bus(void *address)
{
	return (unsigned long)address;
}
#endif
#endif

struct drm_gem_object *
hantro_get_gem_from_dmabuf(struct dma_buf *dma_buf)
{
	struct drm_gem_hantro_object *cma_obj =
		(struct drm_gem_hantro_object
			 *)(((struct dmapriv *)dma_buf->priv)->self);

	if (cma_obj)
		return &cma_obj->base;

	return NULL;
}

int hantro_memalloc(struct hantro_mem_handle *phandle)
{
	struct slice_info *pslice = getslicenode(phandle->sliceidx);
	unsigned int config = pslice->config;

	if (!pslice)
		return -EINVAL;
	/*mmu part is not really finished yet since it needs op with MMU HW*/
	phandle->size = PAGE_ALIGN(phandle->size);
	if (config & CONFIG_HANTROMMU) {
		phandle->vaddr = vmalloc(phandle->size);
		phandle->paddr = page_to_phys(vmalloc_to_page(phandle->vaddr));
	} else {
		phandle->vaddr = dma_alloc_coherent(pslice->dev, phandle->size,
						    &phandle->paddr,
						    GFP_KERNEL | GFP_DMA);
	}
	if (!phandle->vaddr)
		return -ENOMEM;
	else
		return 0;
}

int hantro_memfree(struct hantro_mem_handle *phandle)
{
	struct slice_info *pslice = getslicenode(phandle->sliceidx);
	unsigned int config = pslice->config;

	if (!pslice)
		return -EINVAL;

	if (config & CONFIG_HANTROMMU)
		vfree(phandle->vaddr);
	else
		dma_free_coherent(pslice->dev, phandle->size, phandle->vaddr,
				  phandle->paddr);
	return 0;
}

#ifdef USE_DTB_PROBE

int getnodetype(const char *name)
{
	if (strstr(name, NODENAME_DECODER) == name)
		return HANTRO_CORE_DEC;
	if (strstr(name, NODENAME_ENCODER) == name)
		return HANTRO_CORE_ENC;
	if (strstr(name, NODENAME_CACHE) == name)
		return HANTRO_CORE_CACHE;
	if (strstr(name, NODENAME_DEC400) == name)
		return HANTRO_CORE_DEC400;
	if (strstr(name, NODENAME_AXIFE) == name)
		return HANTRO_CORE_AXIFE;
	if (strstr(name, NODENAME_MMU) == name)
		return HANTRO_CORE_MMU;
	if (strstr(name, NODENAME_VCMD) == name)
		return HANTRO_CORE_VCMD;
	if (strstr(name, NODENAME_IM) == name)
		return HANTRO_CORE_IM;
	if (strstr(name, NODENAME_DECJPG) == name)
		return HANTRO_CORE_DECJPG;
	if (strstr(name, NODENAME_ENCJPG) == name)
		return HANTRO_CORE_ENCJPG;

	return HANTRO_CORE_UNKNOWN;
}

static dtbnode *trycreatenode(struct platform_device *pdev,
			      struct device_node *ofnode, int sliceidx,
			      int parenttype, phys_addr_t parentaddr)
{
	struct fwnode_handle *fwnode;
	struct resource r;
	int i, na, ns, ret;
	int endian = of_device_is_big_endian(ofnode);
	u32 reg_u32[4];
	u64 ioaddress, iosize;

	dtbnode *pnode = kzalloc(sizeof(dtbnode), GFP_KERNEL);

	if (!pnode)
		return NULL;

	XDX_DEV_INFO("try create node %s", ofnode->name);
	pnode->type = getnodetype(ofnode->name);
	pnode->parentaddr = parentaddr;
	pnode->parenttype = parenttype;
	pnode->sliceidx = sliceidx;
	pnode->ofnode = ofnode;
	fwnode = &ofnode->fwnode;

	na = of_n_addr_cells(ofnode);
	ns = of_n_size_cells(ofnode);
	if (na > 2 || ns > 2) {
		pr_err("cell size too big");
		kfree(pnode);
		return NULL;
	}

	fwnode_property_read_u32_array(fwnode, "reg", reg_u32, na + ns);
	if (na == 2) {
		if (!endian) {
			ioaddress = reg_u32[0];
			ioaddress <<= 32;
			ioaddress |= reg_u32[1];
		} else {
			ioaddress = reg_u32[1];
			ioaddress <<= 32;
			ioaddress |= reg_u32[0];
		}
	} else {
		ioaddress = reg_u32[0];
	}
	if (ns == 2) {
		if (!endian) {
			iosize = reg_u32[na];
			iosize <<= 32;
			iosize |= reg_u32[na + 1];
		} else {
			iosize = reg_u32[na + 1];
			iosize <<= 32;
			iosize |= reg_u32[na];
		}
	} else {
		iosize = reg_u32[na];
	}
	pnode->ioaddr = ioaddress;
	pnode->iosize = iosize;
	XDX_DEV_INFO("node regio =%llx:%llx", ioaddress, iosize);

	for (i = 0; i < 4; i++) {
		pnode->irq[i] = -1;
		if (of_irq_to_resource(ofnode, i, &r) > 0) {
			int irq = of_irq_get(ofnode, i);

			XDX_DEV_INFO("irq %d:%s = %lld:%d", i, r.name, r.start, irq);
			if (irq > 0)
				pnode->irq[i] = irq;
		}
	}

	switch (pnode->type) {
	case HANTRO_CORE_DEC:
		ret = hantrodec_probe(pnode, useirq, 0, NULL);
		break;
	case HANTRO_CORE_ENC:
		ret = hantroenc_probe(pnode, useirq, 0, NULL);
		break;
	case HANTRO_CORE_CACHE:
#ifdef HAS_CACHECORE
		ret = cache_probe(pnode, useirq, 0, NULL);
#endif
		break;
	case HANTRO_CORE_DEC400:
#ifdef HAS_DEC400
		ret = hantro_dec400_probe(pnode, 0, NULL);
#endif
		break;
	case HANTRO_CORE_AXIFE:
#ifdef HAS_AXIFE
		ret = hantro_axife_probe(pnode, 0, NULL);
#endif
		break;
	case HANTRO_CORE_MMU:
#ifdef HAS_MMU
		int tmp = 0;
		if(hantro_dev.asic_type == CHIP_PANGU_B0){
#ifdef SUPPORT_VPU_SMMU
			tmp = 1;
#endif
		}
		if(tmp == 0)
			ret = hantroMMUprobe(pnode, 0, NULL, pdev, 0, 0);
#endif
		break;
	case HANTRO_CORE_VCMD:
		ret = 0;
		//ret;	//add VCMD node data analyze here
		break;
	default:
		ret = -EINVAL;
		break;
	}
	if (ret < 0) {
		kfree(pnode);
		pnode = NULL;
	}
	return pnode;
}

static int hantro_analyze_subnode(struct platform_device *pdev,
				  struct device_node *slice, int sliceidx)
{
	dtbnode *head, *nhead, *newtail, *node;

	XDX_DEV_INFO("dev node %s", slice->name);

	head = kzalloc(sizeof(dtbnode), GFP_KERNEL);
	if (!head)
		return -ENOMEM;
	head->type = HANTRO_CORE_SLICE;
	head->parenttype = HANTRO_CORE_SLICE;
	head->ofnode = slice;
	head->ioaddr = -1;
	head->iosize = 0;
	head->next = NULL;

	/*
	 *this is a wide first tree structure iteration,
	 *result is stored in slice info
	 */
	while (head) {
		newtail = NULL;
		nhead = newtail;
		while (head) {
			struct device_node *child, *ofnode = head->ofnode;

			for_each_child_of_node(ofnode, child) {
				node = trycreatenode(pdev, child, sliceidx,
						     head->type, head->ioaddr);
				if (node) {
					if (!nhead) {
						newtail = node;
						nhead = newtail;
					} else {
						newtail->next = node;
					}
					node->next = NULL;
					newtail = node;
				}
			}
			node = head->next;
			kfree(head);
			head = node;
		}
		head = nhead;
	}
	return 0;
}
#endif //USE_DTB_PROBE

static struct hantro_slice hantro_slice_head[MAX_SLICE_NUM];

int hantro_get_vcmdsup(struct drm_device *dev, void *data,
		       struct drm_file *file_priv)
{
	struct hantro_slice *hantro_slice_node;
	hantro_ioctl_id ioctl_id_par;
	u32 slice;

	ioctl_id_par.data = *(unsigned int *)data;
	slice = ioctl_id_par.ID_PAR.node_idx;

	if (slice >= MAX_SLICE_NUM)
		return -1;

	hantro_slice_node = &hantro_slice_head[slice];
	return hantro_slice_node->vcmd_en;
}

static int hantro_set_vcmdsup(u32 sliceidx, u32 vcmd_en)
{
	struct hantro_slice *hantro_slice_node;

	if (sliceidx >= MAX_SLICE_NUM)
		return -1;

	hantro_slice_node = &hantro_slice_head[sliceidx];
	hantro_slice_node->vcmd_en = vcmd_en;
	return 1;
}

#ifdef USE_DTB_PROBE
static int hantro_check_vcmd(struct platform_device *pdev,
			     struct device_node *slice)
{
	struct device_node *child_level_0, *child_level_1;
	int type;

	for_each_child_of_node(slice, child_level_0) {
		for_each_child_of_node(child_level_0, child_level_1) {
			type = getnodetype(child_level_0->name);
			if (type == HANTRO_CORE_VCMD)
				return 1;
		}
	}
	return 0;
}
#endif //USE_DTB_PROBE

static void transfer_vcmdslice_to_norslice(int sliceidx)
{
	vcmd_slice_str *cur_vcmd_slice;
	struct device *dev;
	struct slice_info *cur_nor_slice;
	struct vcmd_core *vcmd_core_tmp;
	struct hantrodec_t *dec_pcore;
	struct hantroenc_t *enc_pcore;

	cur_vcmd_slice = get_slice_by_sliceidx(sliceidx);
	if (!cur_vcmd_slice) {
		pr_debug("error %s vcmd slice is null\n", __func__);
		return;
	}

	if (cur_vcmd_slice->dec_vcmd.subsys_num != 0) {
		dev = cur_vcmd_slice->dec_vcmd.dev;
	} else if (cur_vcmd_slice->enc_vcmd.subsys_num != 0) {
		dev = cur_vcmd_slice->enc_vcmd.dev;
	} else {
		pr_debug("error %s vcmd slice is null\n", __func__);
		return;
	}

	sliceidx = addslice(dev, (phys_addr_t)-1, 0);

	cur_nor_slice = getslicenode_ininit(sliceidx);

	cur_nor_slice->config = get_vcmd_slice_config(sliceidx);

	if (cur_vcmd_slice->dec_vcmd.subsys_num != 0) {
		vcmd_core_tmp = cur_vcmd_slice->dec_vcmd.vcmd_core;

		while (vcmd_core_tmp) {
			dec_pcore = vmalloc(sizeof(*dec_pcore));

			if (!dec_pcore)
				break;

			memset(dec_pcore, 0, sizeof(*dec_pcore));
			dec_pcore->core_id = vcmd_core_tmp->core_id;
			dec_pcore->sliceidx = sliceidx;
			add_decnode(sliceidx, dec_pcore);
			vcmd_core_tmp = vcmd_core_tmp->core_next;
		}
	}

#ifndef HAS_VC8000E_CLOSE
	if (cur_vcmd_slice->enc_vcmd.subsys_num != 0) {
		vcmd_core_tmp = cur_vcmd_slice->enc_vcmd.vcmd_core;

		while (vcmd_core_tmp) {
			enc_pcore = vmalloc(sizeof(*enc_pcore));
			if (!enc_pcore)
				break;
			memset(enc_pcore, 0, sizeof(*enc_pcore));
			enc_pcore->core_id = vcmd_core_tmp->core_id;
			enc_pcore->core_cfg.sliceidx = sliceidx;
			add_encnode(sliceidx, enc_pcore);
			vcmd_core_tmp = vcmd_core_tmp->core_next;
		}
	}
#endif
	slice_init_finish();
}
/*
static int map_register(struct vm_area_struct *vma, unsigned long busaddr, unsigned long size)
{
	size_t mapsize = (vma->vm_end - vma->vm_start);
	int ret;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 3, 0))
	vma->vm_flags |= VM_IO;
#else
	vm_flags_set(vma, VM_IO);
#endif
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	ret = remap_pfn_range(vma, vma->vm_start, busaddr >> PAGE_SHIFT, mapsize,
		vma->vm_page_prot) ? -EAGAIN : 0;

	hantro_mmaplog("%s:%lx:%ld=%d", __func__, busaddr, size, ret);
	return ret;
}
*/
static int map_memblock(struct vm_area_struct *vma, unsigned long busaddr, unsigned long blocksize)
{
	size_t mapsize = vma->vm_end - vma->vm_start;
	int ret = 0;

	if (!(vma->vm_flags & VM_MAYSHARE))
		ret = -EPERM;
	else {
		vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
		ret = remap_pfn_range(vma, vma->vm_start, busaddr >> PAGE_SHIFT,
					mapsize, vma->vm_page_prot) ? -EAGAIN : 0;
	}
	hantro_mmaplog("%s:%lx:%lu=%d", __func__, busaddr, blocksize, ret);
	return ret;
}


int hantro_map_internal_address(struct file *filp, struct vm_area_struct *vma)
{
	hantro_ioctl_id coreid;
	u32 maincoretype;

	coreid.data = (u32)vma->vm_pgoff;
	hantro_mmaplog("%s:nodeid=%d,maincore=%d:subtype=%d:coreidx=%d", __func__,
		coreid.ID_PAR.node_idx, coreid.ID_PAR.group_idx, coreid.ID_PAR.sub_mod_idx, coreid.ID_PAR.codec_idx);
	if (coreid.ID_PAR.group_idx == CODEC_DEC_FLGA)
		maincoretype = NODE_TYPE_DEC;
	else if (coreid.ID_PAR.group_idx == CODEC_ENC_FLGA)
		maincoretype = NODE_TYPE_ENC;
	else
		return -ENODEV;
	switch (coreid.ID_PAR.sub_mod_idx) {
#ifndef HAS_VCMD
	case CORE_FLAG:
		if (maincoretype == NODE_TYPE_DEC) {
			struct hantrodec_t *pdec = get_decnodes(coreid.ID_PAR.node_idx, coreid.ID_PAR.codec_idx);
			if (pdec)
				return map_register(vma, pdec->multicorebase_actual, pdec->iosize);
			else
				return -ENODEV;
		}
		if (maincoretype == NODE_TYPE_ENC) {
			struct hantroenc_t *penc = get_encnodes(coreid.ID_PAR.node_idx, coreid.ID_PAR.codec_idx);
			if (penc)
				return map_register(vma, penc->core_cfg.base_addr, penc->core_cfg.iosize);
			else
				return -ENODEV;
		}
	break;
	case DEC400_FLAG:
		{
			struct dec400_t *pdec400 = get_dec400nodebytype(coreid.ID_PAR.node_idx, maincoretype, coreid.ID_PAR.codec_idx);
			if (pdec400)
				return map_register(vma, pdec400->core_cfg.dec400corebase, pdec400->core_cfg.iosize);
			else
				return -ENODEV;
		}
	case CACHE_FLAG:
		{
			struct cache_dev_t *pcache = get_cachenodebytype(coreid.ID_PAR.node_idx, maincoretype, coreid.ID_PAR.codec_idx);
			if (pcache)
				return map_register(vma, pcache->core_cfg.base_addr, pcache->core_cfg.iosize);
			else
				return -ENODEV;
		}
	case MMU0_FLAG:
	case MMU1_FLAG:
		{
			struct mmu_t *pmmu = get_mmunodebytype(coreid.ID_PAR.node_idx, maincoretype);
			if (pmmu)
				return map_register(vma, pmmu->core_cfg.mmucorebase, pmmu->core_cfg.iosize);
			else
				return -ENODEV;
		}
#else
	case VCMD_FLAG:
	case VCMD_BUF_FLAG:
		{
			vcmd_dev_str *subsys_dev = get_dev_by_sliceidx(coreid.ID_PAR.node_idx, coreid.ID_PAR.group_idx);
			if (subsys_dev) {
				if (coreid.ID_PAR.sub_mod_idx == VCMD_FLAG)
					return map_memblock(vma, subsys_dev->vcmd_buf_mem_pool->busAddress, subsys_dev->vcmd_buf_mem_pool->size);
				else
					return map_memblock(vma, subsys_dev->vcmd_status_buf_mem_pool->busAddress, subsys_dev->vcmd_status_buf_mem_pool->size);
			} else
				return -ENODEV;
		}
#endif
	default:
		return -ENXIO;
	}
	return 0;
}
/*
static int hantro_drm_probe(struct platform_device *pdev)
{
#ifdef USE_DTB_PROBE
	struct device *dev = &pdev->dev;
	int result = 0;
	int sliceidx = -1;

	if (dev->of_node) {
		if (hantro_check_vcmd(pdev, dev->of_node)) {
			sliceidx = hantro_vcmd_probe(pdev, dev->of_node);
			hantro_set_vcmdsup(sliceidx, 1);
			hantro_dev.config |= get_vcmd_slice_config(sliceidx);
			transfer_vcmdslice_to_norslice(sliceidx);
			hantro_vcmd_init(sliceidx);

		} else {
			//probe from system DTB
			//try to attach 1st rsv mem to dtb node
			result = of_reserved_mem_device_init(dev);
			pr_info("try reserve mem =%d", result);

			if (result == 0)
				sliceidx = addslice(dev, -1, 0);
			else
				 //ileave to end of init,
				 //set to default drm platform dev
				 //and default cma area
				 //
				sliceidx = addslice(NULL, -1, 0);

#if USE_HW == 1
			//go throug all sub dtb node' resources
			if (sliceidx >= 0 && dev->of_node)
				hantro_analyze_subnode(pdev, dev->of_node,
						       sliceidx);
#endif
			hantro_set_vcmdsup(sliceidx, 0);
		}
	}
#endif
	pr_info("dev %s probe", pdev->name);

	return 0;
}

static int hantro_drm_remove(struct platform_device *pdev)
{
	return 0;
}
*/
static const struct platform_device_id hantro_drm_platform_ids[] = {
	{
		.name = DRIVER_NAME,
	},
	{},
};
MODULE_DEVICE_TABLE(platform, hantro_drm_platform_ids);

static const struct of_device_id hantro_of_match[] = {
	/*to match dtb, else reg io will fail*/
	{
		.compatible = "thunderbay,hantro",
	},
	{ /* sentinel */ }
};

static int hantro_pm_suspend(struct device *kdev)
{
	//what should we do on HW? Disable IRQ, slow down clock,
	// disable DMA, etc?
	//printk("\nhantro suspend");
	return 0;
}

static int hantro_pm_resume(struct device *kdev)
{
	//printk("\nhantro resume");
	return 0;
}

static const struct dev_pm_ops hantro_pm_ops = {
	/* since we only support S3, only several interfaces should be supported
	 * echo -n "freeze" (or sth else) > /sys/power/state will trigger them
	 * current suspend and resume seem to be enough
	 * maybe suspend_noirq and resume_noirq will be inserted in future
	 */
	//.prepare
	.suspend = hantro_pm_suspend,
	//.suspend_late
	//.suspend_noirq

	//.resume_noirq
	//.resume_early
	.resume = hantro_pm_resume,
	//.complete
};
/*
static struct platform_driver hantro_drm_platform_driver = {
	.probe = hantro_drm_probe,
	.remove = hantro_drm_remove,
	.driver = {
			.name = DRIVER_NAME,
			.owner = THIS_MODULE,
			.of_match_table = hantro_of_match,
			.pm = &hantro_pm_ops,
		},
	.id_table = hantro_drm_platform_ids,
};
*/
static const struct platform_device_info hantro_platform_info = {
	.name = DRIVER_NAME,
	.id = -1,
	.dma_mask = DMA_BIT_MASK(64),
};

static void release_norslice_node(void)
{
	{
		int i, slicen = get_slicenumber();
		struct hantroenc_t *pcore, *pnext;

		XDX_DEV_DEBUG("%s slicen %d\n", __func__, slicen);
		for (i = 0; i < slicen; i++) {
			pcore = get_encnodes(i, 0);
			while (pcore) {
				pnext = pcore->next;
				XDX_DEV_DEBUG("%s vfree\n", __func__);
				vfree(pcore);
				pcore = pnext;
			}
		}
	}

	{
		struct slice_info *post, *prev;

		prev = getslicenode_ininit(0);
		post = prev;
		while (prev) {
			post = prev->next;
			pr_debug("%s kfree\n", __func__);
			kfree(prev);
			prev = post;
		}
	}
}

//#if KERNEL_VERSION(4, 10, 0) > LINUX_VERSION_CODE
#if 0
static int hantro_major = 1; /* dynamic */
#endif
static void hantro_cleanup(struct xdx_device *xdev)
{
	int vcmd_en;
#ifdef HAS_MMU
	int tmp = 0;
#endif
	hantro_ioctl_id ioctl_id_par;

	xdev->vpu.config = 0;

	ioctl_id_par.ID_PAR.node_idx = 0;
	vcmd_en = hantro_get_vcmdsup(NULL, &ioctl_id_par.data, NULL);

	if (vcmd_en == 0) {
		hantro_unlinksysfsAPI(); //this must be before slice clean up
#if USE_HW == 1 /*hw node release*/
#ifdef HAS_VC8000D
		hantrodec_cleanup();
#endif
#ifdef HAS_VC8000E
		hantroenc_cleanup();
#endif
#ifdef HAS_CACHECORE
		cache_cleanup();
#endif
#ifdef HAS_DEC400
		hantro_dec400_cleanup();
#endif
#ifdef HAS_MMU
		tmp = 0;
		if(xdev->asic_type == CHIP_PANGU_B0){
#ifdef SUPPORT_VPU_SMMU
			tmp = 1;
#endif
		}
		if(tmp == 0)
			hantroMMUCleanup();
#endif

#ifdef HAS_AXIFE
		hantro_axife_cleanup();
#endif

#endif
		/*this one must be after above ones to maintain list*/
		slice_remove();
#if 0
#ifdef VSI_FPGA
		hantro_fpga_release();
#endif
#endif

	} else {
		release_norslice_node();
		hantro_vcmd_cleanup();
#ifdef HAS_MMU
		tmp = 0;
		if(xdev->asic_type == CHIP_PANGU_B0){
#ifdef SUPPORT_VPU_SMMU
			tmp = 1;
#endif
		}
		if(tmp == 0)
			hantroMMUCleanup();
#endif
	}
	releaseFenceData();
}
/*
static void probe_hantroHW(struct xdx_device *xdev, unsigned long reg_base,
				  unsigned long ddr_base)
{
	int i, k, coren;
	int ret;
	// struct device *dev=xdev->dev;

	for (i = 0; i < get_slicenumber(); i++) {
		struct slice_info *pslice = getslicenode_ininit(i);

		coren = get_slicecorenum(i, HANTRO_CORE_DEC);
		for (k = 0; k < coren; k++) {
			struct hantrodec_t *decnode = get_decnodes(i, k);

			if (!decnode)
				break;
			decnode->multicorebase += reg_base;
			decnode->multicorebase_actual += reg_base;

			ret = hantrodec_probe(NULL, useirq, 1, decnode);
			if (ret < 0) {
				remove_node(decnode, HANTRO_CORE_DEC);
				return;
			}
		}
		coren = get_slicecorenum(i, HANTRO_CORE_ENC);
		for (k = 0; k < coren; k++) {
			struct hantroenc_t *encnode = get_encnodes(i, k);

			if (!encnode)
				break;
			encnode->core_cfg.base_addr += reg_base;

			ret = hantroenc_probe(NULL, useirq, 1, encnode);
			if (ret < 0)
				remove_node(encnode, HANTRO_CORE_ENC);
		}
		coren = get_slicecorenum(i, HANTRO_CORE_CACHE);
		for (k = 0; k < coren; k++) {
			struct cache_dev_t *cache = get_cachenodes(i, k);

			if (!cache)
				break;
#ifdef HAS_CACHECORE
			cache->com_base_addr += reg_base;
			cache->core_cfg.base_addr += reg_base;

			ret = cache_probe(NULL, useirq, 1, cache);
			if (ret < 0)
				remove_node(cache, HANTRO_CORE_CACHE);
#endif
		}
		coren = get_slicecorenum(i, HANTRO_CORE_DEC400);
		for (k = 0; k < coren; k++) {
			struct dec400_t *dec400 = get_dec400nodes(i, k);

			if (!dec400)
				break;
#ifdef HAS_DEC400
			dec400->core_cfg.dec400corebase += reg_base;
			ret = hantro_dec400_probe(NULL, 1, dec400);
			if (ret < 0)
				remove_node(dec400, HANTRO_CORE_DEC400);
#endif
		}
		coren = get_slicecorenum(i, HANTRO_CORE_AXIFE);
		for (k = 0; k < coren; k++) {
			struct axife_t *axife = get_axifenodes(i, k);

			if (!axife)
				break;
#ifdef HAS_AXIFE
			axife->core_cfg.axifecorebase += reg_base;
			ret = hantro_axife_probe(NULL, 1, axife);
			if (ret < 0)
				remove_node(axife, HANTRO_CORE_AXIFE);
#endif
		}
		coren = get_slicecorenum(i, HANTRO_CORE_MMU);
		for (k = 0; k < coren; k++) {
			struct mmu_t *mmu = get_mmunode(i, k);

			if (!mmu)
				break;
#ifdef HAS_MMU
			ret = hantroMMUprobe(NULL, 1, mmu,
					     xdev, ddr_base, reg_base);
			if (ret < 0)
				remove_node(mmu, HANTRO_CORE_MMU);
#endif
		}
		// Add VCMD HW probe here
		hantro_dev.config |= pslice->config;
	}
}
*/
static int xdx_vpu_hw_reinit(struct xdx_device *xdev)
{
	int i, k, coren;
#ifdef HAS_MMU
        int ret;
		int tmp = 0;
#endif

#ifdef HAS_VCMD
        xdx_vcmd_hw_reset(xdev);
#endif

	for (i = 0; i < get_slicenumber(); i++) {
		getslicenode_ininit(i);

#ifdef HAS_AXIFE
		coren = get_slicecorenum(i, HANTRO_CORE_AXIFE);
		for (k = 0; k < coren; k++) {
			struct axife_t *axife = get_axifenodes(i, k);

			if (!axife)
				break;

			AXIFEEnable(axife->hwregs);
		}
#endif

#ifdef HAS_MMU
#if 0
		coren = get_slicecorenum(i, HANTRO_CORE_MMU);
		for (k = 0; k < coren; k++) {
			struct mmu_t *mmu = get_mmunode(i, k);

			if (!mmu)
				break;

			MMUEnableStep1(mmu->hwregs, mmu->g_mmu);
		}
#endif
#endif
	}

#ifdef HAS_MMU
	tmp = 0;
	if(xdev->asic_type == CHIP_PANGU_B0){
#ifdef SUPPORT_VPU_SMMU
		tmp = 1;
#endif
	}
		if(tmp == 0){
#ifdef HAS_VC8000D
        ret = hantroMMUHwInit(xdev, HANTRO_CORE_DEC);
        if (ret != MMU_STATUS_OK) {
	        XDX_DEV_ERROR("%s,%d hantroMMUHwInit failed for dec\n", __func__, __LINE__);
		}
#endif
#ifdef HAS_VC8000E
        ret = hantroMMUHwInit(xdev, HANTRO_CORE_ENC);
        if (ret != MMU_STATUS_OK) {
		XDX_DEV_ERROR("%s,%d hantroMMUHwInit failed for enc\n", __func__, __LINE__);
		}
#endif
		}
#endif

	return 0;
}

int xdx_vpu_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);

	int i;
	int slice_num = 0;
#ifdef HAS_MMU
	int tmp;
	tmp = 0;
#endif
	hantro_dev.drm_dev = ddev;

	hantro_dev.asic_type = xdev->asic_type;

	//drm_mode_config_init(ddev);

	vpu_debubfs_init(xdev);

	/*
	 *_init functions will init static vairables,
	 *while probe will init dynamic emelemts from DTB
	 */
	/*slice init must be in first to clear list*/
	slice_init();
#ifdef HAS_VC8000E
	hantroenc_init();
#endif
#ifdef HAS_VC8000D
	hantrodec_init();
#endif
#ifdef HAS_CACHECORE
	cache_init();
#endif
#ifdef HAS_DEC400
	hantrodec400_init();
#endif
#ifdef HAS_AXIFE
	hantroaxife_init();
#endif

	xdev->vpu.config = 0;

	if (xdev->aper_size <= NO_RESIZEBAR_SIZE)
		xdev->vpu.config |= CONFIG_INVAS;

	initFenceData();
	xdx_vpu_isr_install(xdev);

#ifndef USE_DTB_PROBE //static table analyze, dec and enc must be in the front
#if USE_HW == 1
    vcmd_core_array_init();
#ifdef HAS_VCMD
	slice_num = hantro_vcmd_probe(xdev, NULL);
	for (i = 0; i < slice_num; i++) {
		hantro_set_vcmdsup(i, 1);
		transfer_vcmdslice_to_norslice(i);
		xdev->vpu.config |= get_vcmd_slice_config(i);
	}
	hantro_vcmd_init(xdev);
#else

#ifdef VSI_FPGA
#if 0
	struct hantro_pci_t normal_pci_par;

	if (pcie_init(&normal_pci_par) < 0)
		pr_debug("%s,%d err pci init failed\n", __func__, __LINE__);
#endif
	reg_base = (uint8_t *)xdev->rmmio_base;
	ddr_base = (uint8_t *)xdev->aper_base; // + 0x80000000;//temporarily set to GDRR 2G as start

	xdxvpu_dev_reg_map(xdev);
	set_mmu_bypass(xdev);

	if (useirq)
		irq_vcmd_init(xdxvpu_dev);

	hantro_fpga_init(ddr_base, 0x1100000);
//	reg_base = normal_pci_par.pci_base_reg_hw;
//	ddr_base = normal_pci_par.pci_base_ddr_hw;
#endif

#ifdef HAS_VC8000D
	result = hantrodec_probe(NULL, useirq, 0, NULL);
	if(result){
		XDX_DEV_ERROR("hantrodec_probe failed");
		return result;
	}
#endif
#ifdef HAS_VC8000E
	result = hantroenc_probe(NULL, useirq, 0, NULL);
	if(result){
		XDX_DEV_ERROR("hantroenc_probe failed");
		return result;
	}
#endif
#ifdef HAS_CACHECORE
	result = cache_probe(NULL, useirq, 0, NULL);
	if(result){
		XDX_DEV_ERROR("cache_probe failed");
		return result;
	}
#endif
#ifdef HAS_DEC400
	result = hantro_dec400_probe(NULL, 0, NULL);
	if(result){
		XDX_DEV_ERROR("hantro_dec400_probe failed");
		return result;
	}
#endif
#ifdef HAS_AXIFE
	result = hantro_axife_probe(NULL, 0, NULL);
	if(result){
		XDX_DEV_ERROR("hantro_axife_probe failed");
		return result;
	}
#endif
#ifdef HAS_MMU
	tmp = 0;
	if(xdev->asic_type == CHIP_PANGU_B0){
#ifdef SUPPORT_VPU_SMMU
		tmp = 1;
#endif
	}
	if(tmp == 0){
		result =
				hantroMMUprobe(NULL, 0, NULL, xdev, ddr_base, reg_base);
			if(result){
				XDX_DEV_ERROR("hantroMMUprobe failed");
				return result;
			}
	}
#endif
#ifdef HAS_VCMD
	//result; //add static VCMD array analyze here
#endif
#endif
#endif //USE_HW==1
#endif //USE_DTB_PROBE

#ifdef HAS_AXIFE
	hantro_axife_probe(xdev);
#endif

#ifndef HAS_VCMD
	if (get_slicenumber() == 0)
		addslice(hantro_dev.drm_dev->dev, -1,
			 0); //for PC, no HW, create a default dev
	for (i = 0; i < get_slicenumber(); i++) {
		struct slice_info *pslice = getslicenode_ininit(i);

		if (!pslice->dev)
			pslice->dev = hantro_dev.drm_dev->dev;
		result = hantro_createsysfsAPI(i, pslice->dev);
		if (result != 0)
			XDX_DEV_ERROR("create sysfs %d fail", i);
	}
	slice_printdebug();
	probe_hantroHW(xdev, reg_base, ddr_base);
	slice_init_finish();
	slice_printdebug();
#endif
	XDX_DEV_INFO("VPU device created");
	return 0;
}

int xdx_vpu_fini(struct xdx_device *xdev)
{

	vpu_isr_uninstall(xdev);
	//vs_dc_fini(&xdev_to_dc(xdev));
	hantro_cleanup(xdev);
	vpu_debugfs_exit();

//	drm_mode_config_cleanup(ddev);

	return 0;
}

static int xdx_vpu_resume(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	xdx_vpu_hw_reinit(xdev);
	xdxgpu_irq_set(xdev, xdev->vpu.irq_src, 0, XDX_IRQ_STATE_ENABLE);
	return 0;
}

static int xdx_vpu_suspend(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
#ifdef HAS_VCMD
	int i, slice_num;
	vcmd_slice_str *cur_slice;

	slice_num = get_vcmd_slice_num();
	for (i = 0; i < slice_num; i++) {
		cur_slice = get_slice_by_sliceidx(i);
		if (!cur_slice) {
			pr_debug("%s vcmd slice num %d is null\n", __func__, i);
			continue;
		}
		vcmd_dev_suspend(&cur_slice->dec_vcmd);
		vcmd_dev_suspend(&cur_slice->enc_vcmd);
	}
#endif
	xdxgpu_irq_set(xdev, xdev->vpu.irq_src, 0, XDX_IRQ_STATE_DISABLE);
	flush_workqueue(xdev->vpu.vpu_workq);

	return 0;
}

static int xdx_vpu_early_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	INIT_LIST_HEAD(&xdev->vpu.vcmd_slice);
	return 0;
}
static int xdx_vpu_early_fini(void *handle)
{
	return 0;
}

static int xdx_vpu_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	xdev->vpu.vpu_workq = create_singlethread_workqueue("xdxgpu-vpu-irq");
	INIT_WORK(&xdev->vpu.vpuirq_dec_work, vpuirq_dec_work_handle);
	INIT_WORK(&xdev->vpu.vpuirq_enc_work, vpuirq_enc_work_handle);
	return 0;
}

static int xdx_vpu_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;

	cancel_work_sync(&xdev->vpu.vpuirq_dec_work);
	cancel_work_sync(&xdev->vpu.vpuirq_enc_work);
	destroy_workqueue(xdev->vpu.vpu_workq);
	xdev->vpu.vpu_workq = NULL;
	return 0;
}

static int xdx_vpu_hw_init(void *handle)
{
	return 0;
}
static int xdx_vpu_hw_fini(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version vpu_ip_block_version = {
	.name = "vpu",
	.type = XDX_IP_BLOCK_TYPE_VPU,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.suspend = xdx_vpu_suspend,
		.resume = xdx_vpu_resume,
		.early_init = xdx_vpu_early_init,
		.early_fini = xdx_vpu_early_fini,
		.sw_init = xdx_vpu_sw_init,
		.sw_fini = xdx_vpu_sw_fini,
		.hw_init = xdx_vpu_hw_init,
		.hw_fini = xdx_vpu_hw_fini,
	},
};
